<?php
	/**
	*	@brief A Payable Invoice (from a Supplier)
	*
	*	@author Nathan Gray
	*	@ingroup perp_ap
	*	
	*	Invoice encapsulates an invoice from a Supplier.
	*	An invoice contains various information such as relevant dates, reference numbers
	*	and financial information, as well as one or more InvoiceLines.
	*
	*	Invoices can be automatically generated based off existing PurchaseOrders or Shipments, or
	*	they can be entered starting from a blank and adding the lines manually.  Automatically
	*	generated Invoices can have their lines overriden to match the invoice the Supplier sent.
	*/
	require_once(EGW_INCLUDE_ROOT . '/perp_api/inc/class.Invoice.inc.php');
	require_once('class.PayableInvoiceLine.inc.php');

	class PayableInvoice extends Invoice {
		/**
		*	Hold payment on this invoice?
		*/
		protected $hold;

		/**
		*	Method by which this invoice was paid
		*/
		protected $payment_method;

		/**
		*	Post this invoice to the General Ledger
		*/
		public function post() {
			parent::post();

			if($this->invoice_transaction_id) {
				return false;
			}

			$config = config::read('perp_ap');
			$ap_account = $config['ap_account'];

			if(!$ap_account) {
				throw new egw_exception(lang('Accounts Payable general ledger account is not set.  Set it in Admin -> Accounts Payable -> Site Configuration'));
			}
			$trans_type = $config['invoice_transaction_type'];
			$invoice_narrative = $config['invoice_narrative'];

			$supplier_data = ExecMethod('perp_ap.bo_perp_supplier.get_supplier', $this->supplier_id);

			$invoice_narrative = sprintf($invoice_narrative,
				$supplier_data['supplier_number'],
				$this->get_our_reference_number()
			);
			
			// This needs to be atomic
			$transaction = CreateObject('perp_ledger.GLTransaction');
			$transaction->set_transaction_type($trans_type);
			$transaction->set_date($this->issue_date);

			// Transaction needs one line for each line, one line to balance with associated BankTransactionLine and SupplierTransactionLine
			$ap_total = 0;
			$gl_accounts = array();

			foreach($this->invoice_lines as $line_no => $line) {
				$ap_total = bcsub($ap_total, $line->post($transaction), 2);
			}
			$ap_line = $transaction->add_line($ap_account, $invoice_narrative, $ap_total);

			// Create transaction for supplier
			$supplier_line = CreateObject('perp_ap.SupplierTransactionLine', $ap_line);
			$supplier_line->set_supplier($this->supplier_id);
			$supplier_line->set_due_date($this->get_due_date());
			$supplier_line->set_exchange_rate($this->get_exchange_rate());
			$supplier_line->set_link('perp_ap', 'INVOICE-' . $this->get_invoice_id());
			$supplier_line->set_supplier_reference($this->get_their_reference_number);
			$supplier_line->set_comment($this->comments);


			$GLOBALS['egw']->db->transaction_begin();
			if($GLOBALS['egw']->db->Link_ID->transCnt == 0) {
				/*
				* Database doesn't support transactions.
				* Reconnect to do transactions, and shove it in GLOBALS[egw]->db so everywhere
				* uses the same connection
				*/
				$old_db =& $GLOBALS['egw']->db;
				$GLOBALS['egw']->db =& ExecMethod('perp_api.bo_perp_api.reconnect_transactional');
				$GLOBALS['egw']->db->transaction_begin();
			}
			$result = $transaction->write();
			$this->invoice_transaction_id = $transaction->get_transaction_number();
			$this->write();

			if(!is_numeric($result) || $this->invoice_transaction_id !== $transaction->get_transaction_number()) {
				$GLOBALS['egw']->db->transaction_abort();
				$GLOBALS['egw']->db->transaction_commit();
				throw new egw_exception(lang('Critical failure!  Invoice %1 does not know it was posted as transaction #%2!',
					$this->get_our_reference_number(),
					$transaction->get_transaction_number()
				));
			}

			$result = $GLOBALS['egw']->db->transaction_commit();

			if($old_db) {
				// Swap the original database back, so it doesn't cause problems elsewhere
				$GLOBALS['egw']->db = $old_db;
			}
			return $result;
		}

		/**
		*	Get the tax group for this invoice
		*/
		public function get_tax_group() {
			if($this->supplier_id) {
				$supplier = ExecMethod('perp_ap.bo_perp_supplier.get_supplier', $this->supplier_id);
				return $supplier['tax_group'];
			}
		}

		/**
		*	Calculate the applicable taxes for an InvoiceLine and add it to the taxes array
		*
		*	@param line 	InvoiceLine to calculate taxes for
		*	@param amount	float	Amount the line is for (why do I need this, it should be in the InvoiceLine?)
		*	@param source_location	int	Location the invoice came from, some taxes depend on this.
		*	@param tax_group	int	ID of the tax group applicable to this line
		*/
		protected function calc_taxes($line, $amount, $source_location, $tax_group = 0) {
			// Taxes!
			$taxes = array();
			if($line && !$line->get_stock_id()) {
				// Try and find the stock id
				if($line->get_item_code()) {
					$item = ExecMethod2('perp_ap.bo_perp_catalog.get_catalog_item', $this->supplier, $line->get_item_code());
					if($item['stock_id']) {
						$line->set_stock_id($item['stock_id']);
					}
				}
			}
			if($line && !$line->get_stock_id()) {
				$this->set_comment($this->get_comment() . "\n" . lang('Unable to calculate taxes for %1', $line->get_description()));
			}

			$sql = 'SELECT 
					perp_tax_authority.tax_authority_id,
					perp_tax_authority.tax_authority_name,
					perp_tax_authority.tax_authority_in_account,
					perp_tax_authority.registration_number,
					perp_tax_group_tax.tax_on_tax,
					perp_tax_group_tax.calc_order,
					perp_tax_auth_rate.tax_category,
					perp_tax_auth_rate.tax_rate
				FROM perp_tax_group_tax
				JOIN perp_tax_auth_rate ON 
					perp_tax_group_tax.tax_authority_id = perp_tax_auth_rate.tax_authority ';
			if($source_location) {
				$sql .= 'AND perp_tax_auth_rate.tax_source_location = ' . $source_location . ' ';
			} 
			$sql .='JOIN perp_tax_authority ON
					perp_tax_group_tax.tax_authority_id = perp_tax_authority.tax_authority_id
				JOIN perp_stock_tax_category ON
					perp_tax_auth_rate.tax_category = perp_stock_tax_category.tax_category_id AND
					perp_tax_group_tax.tax_authority_id = perp_stock_tax_category.tax_authority_id ';
			if($line && $line->get_stock_id()) {
				$sql .= 'AND perp_stock_tax_category.stock_id = ' . $line->get_stock_id();
			}
			$sql .= ' WHERE perp_tax_group_tax.tax_group_id = ' . $tax_group .'
				ORDER BY calc_order';

			$GLOBALS['egw']->db->query($sql, __LINE__, __FILE__);

			while($row = $GLOBALS['egw']->db->row(true)) {
				$taxes[$row['tax_authority_id']][$row['tax_category']]['description'] = ($row['tax_rate'] * 100) . '% ' . $row['tax_authority_name'] . 
					($this->supplier == 0 ? ' (' . $row['registration_number'] . ')' : '');
				$taxes[$row['tax_authority_id']][$row['tax_category']]['gl_account'] = $row['tax_authority_in_account'];
				$taxes[$row['tax_authority_id']][$row['tax_category']]['tax_on_tax'] = $row['tax_on_tax'];
				$taxes[$row['tax_authority_id']][$row['tax_category']]['calc_order'] = $row['calc_order'];
				$taxes[$row['tax_authority_id']][$row['tax_category']]['tax_rate'] = $row['tax_rate'];
			}

			$this->tax_on_tax_total = array();

			foreach($taxes as $authority => &$tax_category) {
				foreach($tax_category as $cat => &$tax) {
					$tax_amount = $amount;
					if($tax['tax_on_tax']) {
						foreach($tax_on_tax_total[$cat] as $calc_order => $row) {
							if($calc_order < $tax['calc_order']) {
								$tax_amount += $row;
							}
						}
					}
					$tax_amount = round($tax_amount * $tax['tax_rate'], 2);
					if(!is_array($this->taxes[$authority][$cat])) {
						$this->tax_lines[$authority][$cat] = $tax;
					} else {
						$this->tax_lines[$authority][$cat] = array_merge($this->taxes[$authority][$cat], $tax);
					}
					$this->tax_lines[$authority][$cat]['quantity'] = 1;
					$this->tax_lines[$authority][$cat]['unit_price'] += $tax_amount;
					$tax_on_tax_total[$cat][$tax['calc_order']] += $tax_amount;
				}
			}
		}

		/**
		*	Suggest a reference number for this invoice using configured parameters
		*
		*	@param invoice_id	ID of the invoice to suggest the reference number for (0 for a new Invoice)
		*	@param supplier_id	The Supplier who issued the invoice
		*	@return string Our reference number
		*/
		public function suggest_invoice_number() {
			if($this->get_our_reference_number()) {
				return $this->get_our_reference_number();
			}

			$config = config::read('perp_ap');
			$invoice_number_format = $config['invoice_number'];

			// Get supplier_id & next_supplier_num
			if($this->supplier_id) {
				$GLOBALS['egw']->db->select('perp_invoice',
					array(
						'ifnull(count(invoice_id),1) AS next_supplier_num',
						'supplier_num',
					),
						array('perp_invoice.supplier_id = ' . $this->supplier_id),
					__LINE__, __FILE__, False,
					'GROUP BY perp_invoice.supplier_id', 'perp_ap', 0,
					'JOIN perp_supplier ON
						perp_invoice.supplier_id = perp_supplier.supplier_id'
				);
				$row = $GLOBALS['egw']->db->row(True);
			}
			if($row) {
				$supplier_num = $row['supplier_num'];
				$next_supplier_num = $row['next_supplier_num'] + 1;
			} else {
				$supplier = ExecMethod('perp_ap.bo_perp_supplier.get_supplier', $this->supplier_id);
				$supplier_num = $supplier['supplier_num'];
				$next_supplier_num = 1;
			}

			// get next_po_id
			$GLOBALS['egw']->db->select('perp_purchase_order',
				array('max(purchase_order_id) + 1 AS next_po_id'),
				array(),
				__LINE__, __FILE__, False, False, 'perp_ap');
			$row = $GLOBALS['egw']->db->row(True);
			$next_po_id = $row['next_po_id'];

                        ereg('date\([dDjlNSwzWFmMntYycU.]+\)', $invoice_number_format, $regs);
                        if(count($regs)) {
                                eval('$date = ' . $regs[0] . ';');
                                $po_code = str_replace($regs[0], '%5$s', $invoice_number_format);
                        }
			$invoice_number = sprintf($invoice_number_format,
				$supplier_num,
				'<Not Implemented>',
				$next_supplier_num,
				date('H:i'),
				$date
			);

			return $invoice_number;
		}

		/**
		*	Get the bank account this Invoice was paid from
		*	@return int bank account number
		*/
		public function get_bank_account() {
			return $this->bank_account;
		}
		/**
		*	Set the bank account this Invoice was paid from
		*	@param account	int	Bank account number
		*/
		public function set_bank_account($account) {
			$accounts = array_keys(ExecMethod('perp_ledger.bo_perp_ledger.get_bank_accounts'));

			if($account && in_array($account, $accounts)) {
				$this->bank_account = $account;
			}
		}


		/** 
		*	?
		*/
		public function get_hold() {
			return $this->hold;
		}
		/**
		*	?
		*/
		public function set_hold($hold) {
			$this->hold = $hold;
		}

		/**
		*	Get the method of payment used to pay this invoice
		*	@return int payment_method_id
		*/
		public function get_payment_method() {
			return $this->payment_method;
		}
		/**
		*	Set the method of payment used to pay this invoice
		*	@param pm int payment_method_id
		*/
		public function set_payment_method($pm) {
			$pms = array_keys(ExecMethod('perp_api.bo_perp_api.get_payment_methods'));

			if($pm && in_array($pm, $pms)) {
				$this->payment_method = $pm;
			}
		}

		/**
		*	Get a list of payments on this invoice.
		*/
		public function get_payments() {
			$GLOBALS['egw']->db->select('perp_supplier_payment',
				array(
					'payment_id',
					'perp_supplier_payment.trans_id',
					'perp_supplier_payment.amount',
					'allocation_date',
					'allocated_from',
					'payment_number',
					'payment_method_name'
				),
				array(
					'invoice_id'	=>	$this->invoice_id
				),
				__LINE__, __FILE__, 0, '', 'perp_ap', -1,
				'JOIN perp_bank_ledger ON
					perp_supplier_payment.trans_id = perp_bank_ledger.trans_id
				JOIN perp_payment_method ON
					perp_bank_ledger.payment_method = perp_payment_method.payment_method_id'
			);
			$payment_list = array();
			if(!is_object($GLOBALS['perp_api']['bo'])) {
				$GLOBALS['perp_api']['bo'] =& CreateObject('perp_api.bo_perp_api');
			}
			while($row = $GLOBALS['egw']->db->row(true)) {
				$row['allocation_date'] = $GLOBALS['egw']->db->from_timestamp($row['allocation_date']);
				$row['formatted_amount'] = $GLOBALS['perp_api']['bo']->get_formatted_price($this->invoice_currency, $row['amount']);
				$payment_list[] = $row;
			}
			return $payment_list;
		}

		/**
		*	Prepare this object for serialization
		*	@return array of parameters to preserve
		*/
		public function __sleep() {
			return parent::__sleep() + array(
				'paid', 'bank_account', 
				'hold', 
				'purchase_order_id', 'invoice_trans_id', 
				'paid_trans_id',
			);
		}
	}
?>
