<?php
	/**
	*	@brief An order from a Supplier
	*
	*	PurchaseOrder encapsulates an order from a Supplier off a catalog.
	*	The general process for creating a PurchaseOrder is an employee (usually in the 
	*	Purchasing department) will put together an order.  One of their superiors
	*	will approve the purchase and send it to the Supplier.\n
	*	When the Supplier ships the order (fully or partially) they send notification
	*	of what was sent, and a Shipment is created to track it.
	*	As goods from the PurchaseOrder are received, they are marked as such. As 
	*	the Supplier invoices for the order, the lines are marked as invoiced.  When the entire
	*	PurchaseOrder is received and invoiced, it is closed.
	*
	*	@author Nathan Gray
	*	@ingroup perp_ap
	*/
	
	require_once(EGW_INCLUDE_ROOT . '/perp_api/inc/class.Order.inc.php');
	require_once('class.PurchaseOrderLine.inc.php');

	class PurchaseOrder extends Order {
		
		/**
		*	Database table for purchase orders
		*/
		protected static $table = 'perp_purchase_order';

		/**
		*	Database table for purchase order detail lines
		*/
		protected static $detail_table = 'perp_purchase_order_detail';

		/**
		*	perp_ledger::GLTransaction 
		*	@todo Check if this is used
		*/
		public $transaction;

		/**
		*	General Ledger Account for goods paid for but not yet received
		*/
		public $gnr_account;

		/**
		*	Date this PurchaseOrder is fully invoiced
		*/
		protected $invoiced_date;

		/**
		*	Location for delivery and receiving
		*/
		protected $into_stock_location;

		/**
		*	If each request needs to be uniquely numbered.  Not implemented
		*/
		protected $requisition_number;

		/**
		*	Constructor
		*	@param po_id Optional.  ID of an existing PurchaseOrder to read
		*/
		public function PurchaseOrder($po_id = null) {
			parent::__construct($po_id);

			$config = config::read('perp_ap');
			$this->gnr_account = $config['gnr_account'];

			if(is_null($po_id)) {
				$this->po_id = null;
				$this->supplier_id = null;
				
				$supervisors = $GLOBALS['phpgw_info']['user']['preferences']['perp_ap']['supervisors'];
				$this->approver = is_array($supervisors) ? current($supervisors) : $supervisors;
			}
		}

		/**
		*	Read the given purchase order from the database
		*	@param po_id int ID number of the PurchaseOrder to read.
		public function read($po_id) {
			parent::read($po_id);
			$GLOBALS['egw']->db->select(self::$table, array(
					'purchase_order_number',
					'supplier',
					'comments',
					'created_date',
					'order_date',
					'arrival_date',	//Requested arrival
					'initiator',
					'approver',
					'approved_date',
					'received_date',
					'invoiced_date',
					'into_stock_location',
					'contact',
					'requisition_number',
					'currency',
					'terms'),
				array( 'purchase_order_id' => $po_id), 
				__LINE__, __FILE__, 'perp_ap');
			$row = $GLOBALS['egw']->db->row(true);
			$this->po_number = $row['purchase_order_number'];
			$this->into_stock_location = $row['into_stock_location'];
			$this->requisition_number = $row['requisition_number'];

			// Get lines
			$GLOBALS['egw']->db->select(self::$detail_table, array(
					'line_no',
					'received_date',
					'invoiced_date',
					'item_code',
					'item_description AS description',
					'gl_account',
					'quantity_invoiced',
					'unit_price',
					'line_price',
					'standard_cost_per_unit',
					'quantity_ordered',
					'quantity_received',
					'completed'),
				array('purchase_order_id' => $po_id),
				__LINE__, __FILE__, False,
				'ORDER BY line_no', 'perp_ap'
			);

			$received = True;
			while($row = $GLOBALS['egw']->db->row(true)) {
				$received = $received && $row['completed'];
				$line =& $this->order_lines[$row['line_no']];
				if($line) {
					$row += $line->toArray();
				}
				$this->order_lines[$row['line_no']] = new PurchaseOrderLine($this, $row);
			}
			
		}
		*/

		/**
		*	Writes the given purchase order to the database
		*	@return int the purchase order ID (not the PO number)
		*/
		public function write() {
			return parent::write();
			$result = $this->check();
			if($result !== True) {
				throw new Exception('Purchase Order can not be saved.  (' . $result . ')');
			}

			// Write PO
			$GLOBALS['egw']->db->insert(self::$table, array(
				'purchase_order_id'	=>	$this->order_id,
				'purchase_order_number'	=>	$this->our_reference_number,
				'supplier'		=>	$this->supplier_id,
				'comments'		=>	$this->comments,
				'created_date'		=>	$this->get_originating_date(),
				'order_date'		=>	$this->issue_date,
				'arrival_date'		=>	$this->requested_delivery_date,
				'initiator'		=>	$this->get_originating_user(),
				'approver'		=>	$this->approving_user,
				'approved_date'		=>	$this->approval_date,
				'received_date'		=>	$this->received_date,
				'invoiced_date'		=>	$this->invoiced_date,
				'into_stock_location'	=>	$this->into_stock_location,
				'contact'		=>	$this->their_contact_person,
				'requisition_number'	=>	$this->requisition_number,
				'currency'		=>	$this->currency,
				'terms'			=>	$this->terms),
				array('purchase_order_id'	=>	$this->po_id),
				__LINE__, __FILE__, 'perp_ap');

			$this->order_id = $this->order_id ? $this->order_id : $GLOBALS['egw']->db->get_last_insert_id(self::$table, 'purchase_order_id');

			// Clear Lines
			$GLOBALS['egw']->db->delete(self::$detail_table, 
				array('purchase_order_id' => $this->order_id), 
				__LINE__, __FILE__, 'perp_ap');

			// Write lines
			$count = 0;
			foreach($this->order_lines as $key => $line) {
				$line->write($count++);
			}
	
			if($this->transaction) {
				$difference = $this->transaction->check();
				if($difference) {
					$this->transaction->add_line($this->gnr_account, 'Purchase Order', -$difference);
				}
				try{
					// Er... Can't nest database transactions. Oh well...
					$this->transaction->write();
				} catch (Exception $e) {
					$GLOBALS['egw']->db->transaction_abort();
					return $e->getMessage();
				}
			}

			$GLOBALS['egw']->db->transaction_commit();

			return $this->order_id;
		}

		/**
		*	Check that this purchase order is valid.
		*	@return boolean True or string error message
		*/
		public function check() {
			$result = parent::check();
			if($result !== true) {
				return $result;
			}

			// Check all lines for completeness
			if($this->approval_date) {
				$invoiced_date = null;
				$received_date = null;
			
				foreach($this->order_lines as $line) {
					if($line instanceof PurchaseOrderLine && $line->get_received_date()) {
						$received_date = max($received_date, $line->get_received_date());
					} else {
						$received_date = null;
					}
					if($line instanceof PurchaseOrderLine && $line->get_invoiced_date()) {
						$invoiced_date = max($invoiced_date, $line->get_invoiced_date());
					} else {
						$invoiced_date = null;
					}
	
				}

				// If all the lines are received || invoiced, this PO is too
				if(!$this->received_date) {
					$this->received_date = $received_date;
				}
				if(!$this->invoiced_date) {
					$this->invoiced_date = $invoiced_date;
				}
			}

			return true;
		}

		/**
		*	Checks to see if the provided user is allowed to approve this PO
		*	This depends on both permissions and the status of the PurchaseOrder.
		*	@param id eGroupWare user ID
		*	@return boolean Approval allowed or error message.
		*/
		public function is_approver($id) {
			return $this->is_admin($id);
		}

		/**
		*	Checks to see if the provided user is allowed to deny this PO
		*	This depends on both permissions and the status of the PurchaseOrder.
		*	@param id eGroupWare user ID
		*	@return boolean Deny allowed or error message.
		*/
		public function is_denier($id) {
			$err = $this->is_approver($id);
			return ($err === True && !$this->get_approval_date());
		}

		/**
		*	Deny this PurchaseOrder.
		*	This will cancel this PurchaseOrder
		*	@return true or error message
		*/
		public function deny() {
			$denier = $GLOBALS['egw_info']['user']['account_id'];
			$check = $this->is_denier($denier);
			if($check === True) {
				return $this->cancel();
			} else {
				return $check;
			}
		}

		/**
		*	Checks to see if the provided user is allowed to cancel this PO
		*	This depends on both permissions and the status of the PurchaseOrder.
		*	@param id eGroupWare user ID
		*	@return boolean Cancel allowed or error message.
		*/
		public function is_canceller($id) {
			$perm = $GLOBALS['egw']->acl->check('order', EGW_ACL_DELETE, 'perp_ap');
			if(!$this->approval_date) {
				$perm = ($id == $this->get_originating_user()) || $perm;
			}
			return $perm;
		}

		/**
		*	Cancel this PurchaseOrder.  
		*	PurchaseOrders cannot be cancelled if we've already received some goods or an Invoice.
		*	@return boolean PurchaseOrder cancelled
		*/
		public function cancel($cancelling_user_id) {
			if($this->is_canceller($id)) {
				if($this->approval_date) {
					// Check for partial approval
					foreach($this->order_lines as $line) {
						if($line->get_received_date()) {
							return False;
						}
						if($line->get_invoiced_date()) {
							return False;
						}	
					}
					
				// Cancel a PO Request
				} elseif(!$this->order_id) {
					return False;
				}
				parent::cancel($cancelling_user_id);
				$GLOBALS['egw']->db->delete(self::$table, 
					array('purchase_order_id' => $this->get_order_id()),
					__LINE__, __FILE__, 'perp_ap'
				);
				$GLOBALS['egw']->db->delete(self::$detail_table,
					array('purchase_order_id' => $this->get_order_id()),
					__LINE__, __FILE__, 'perp_ap'
				);
				return True;
			} else {
				return lang('Access is not allowed');
			}
		}

		/**
		*	Can this PurchaseOrder be shipped?
		*/
		public function can_ship() {
			$test = parent::can_ship();

			if($test !== true) {
				return $test;
			}
			$test = $test && $GLOBALS['egw']->acl->check('shipment', EGW_ACL_ADD, $GLOBALS['egw_info']['flags']['currentapp']);
			return $test;
		}

		/**
		*	Mark a PurchaseOrderLine as (partially) received
		*
		*	@param shipment_id	int	ID of the shipment the goods came in on
		*	@param line_no		int	ID number of the PurchaseOrderLine received against
		*	@param quantity		float	Quantity received, in Supplier's unit of measure
		*	@return PurchaseOrderLine	Line received against.
		*/
		public function receive($shipment_id, $line_no, $quantity) {
			if(!$quantity) {
				return;
			}

			$receive_line = $this->get_line($line_no);

			if($receive_line) {
				$receive_line->receive($shipment_id, $quantity);
			}
			$this->write();
			return $receive_line;
		}

		/** 
		*	Marks a PurchaseOrderLine as (partially) invoiced.
		*	Only updates purchase_order_details, checks & updates invoice date
		*	@param line_no Either a single line_no, or an array of line_no => quantity
		*	@param invoice_qty float Optional.  Quantity being invoiced.
		*/
		public function invoice($line_no, $invoice_qty = '') {
			if(is_array($line_no)) {
				foreach($line_no as $item => $qty) {
					$this->_invoice($item, $qty);
				}
			} elseif($invoice_qty) {
				$this->_invoice($line_no, $invoice_qty);
			}
			$this->write();
		}
		/**
		*	Actually invoice one PurchaseOrderLine
		*	@param line_no Either a single line_no, or an array of line_no => quantity
		*	@param invoice_qty float Optional.  Quantity being invoiced.
		*/
		private function _invoice($line_no, $invoice_qty) {
			$invoice_line = $this->get_line($line_no);
			if($invoice_line) {
				$invoice_line->invoice($invoice_qty);
			} else {
				throw new Exception(lang('Could not find line to invoice'));
			}
		}

		// Accessors

		/**
		*	Get the unique system identifier for this PurchaseOrder
		*	@return int purchase_order_id
		*/
		public function get_purchase_order_id() {
			return $this->po_id;
		}
		/**
		*	Get the human reference number for this PurchaseOrder
		*	@return string purchase order number
		*/
		public function get_purchase_order_number() {
			return $this->po_number;
		}

		/**
		*	Set the human reference number for this PurchaseOrder.
		*	PO number only set once, can't be changed.
		*	Actually, you should really use approve($num) instead.
		*	@param num	string Reference to use for this PO
		*/
		public function set_purchase_order_number($num) {
			if(!$this->po_number) {
				$this->po_number = $num;
				return True;
			} else {
				return False;
			}
		}

		/**
		*	Get the number of days required for the supplier to fully deliver
		*	the order.
		*	@return int Number of days required for delivery
		*/
		private function get_lead_time() {
			$lead_time = 0;

			foreach($this->order_lines as $key => $line) {
				$stock = ExecMethod2('perp_ap.bo_perp_catalog.get_catalog_item', $this->supplier_id, $line->get_supplier_stock_code());
				$lead_time = max($lead_time, $stock['lead_time']);
			}
			return $lead_time;
		}

		/**
		*	Get the requested date for delivery.  By default, this is the fastest
		*	the Supplier can fully deliver the goods, taking into account lead time required.
		*	@return Requested / expected arrival date (U)
		*/
		public function get_requested_delivery_date() {
			if($this->requested_delivery_date) {
				return $this->requested_delivery_date;
			} else {
				$d = $this->get_issue_date();
				$date = date('U', mktime(0,0,0,
					date('m', $d), 
					date('d', $d) + $this->get_lead_time(), 
					date('Y', $d)));
				return $date;
			}
		}

		/**
		*	Set the requested / expected date of delivery.
		*	If the Supplier can't deliver (according to lead times in the catalog),
		*	it will warn that the Supplier can't meet the date, and the date will be corrected, 
		*	until the PurchaseOrder is saved.  After the PurchaseOrder is written, it will only warn.
		*
		*	@param date Requested / expected arrival date (U)
		*	@return null or error message
		*/
		public function set_requested_delivery_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for arrival date');
			}
			$this->requested_delivery_date = $date;
			$lead_time = $this->get_lead_time();
			$order_date = $this->get_issue_date();
			$lead_date = mktime(0,0,0, date('m', $order_date), date('d', $order_date) + $lead_time, date('Y', $order_date));
			if($this->requested_delivery_date < $lead_date) {
				$lead_time = date('d', $lead_date - $this->issue_date);
				$msg = lang('Arrival date less than lead time needed');
				if(!$this->requested_delivery_date) {
					$this->requested_delivery_date = $lead_date;
					$msg .= ' - ' . lang('Adjusted by %1 days', sprintf("%+d", $lead_time));
				} else {
					$msg .= ' - ' . lang('Should be %1', date($GLOBALS['egw_info']['user']['preferences']['common']['dateformat'], $lead_date));
				}
				return $msg;
			} 
		}

		/**
		*	Get the location we want the goods to go to.
		*	@return int location ID
		*/
		public function get_into_stock_location() {
			return $this->into_stock_location;
		}
		/**
		*	Set the location we want the goods to go to.
		*	@param location	 int location ID
		*/
		public function set_into_stock_location($location) {
			$this->into_stock_location = $location;
		}

		/**
		*	Get the date this PO was completely received
		*	@return date (Y-m-d)
		*/
		public function get_received_date() {
			return $this->received_date;
		}
		/**
		*	Set the date this PO was completely received
		*	@param date (U)
		*/
		public function set_received_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for received date');
			}
			$this->received_date = $date;
		}

		/**
		*	Get the date this PO was completely invoiced
		*	@return date (U)
		*/
		public function get_invoiced_date() {
			return $this->invoiced_date;
		}
		/**
		*	Set the date this PO was completely invoiced
		*	@param date (U)
		*/
		public function set_invoiced_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for invoiced_date');
			}
			$this->invoiced_date = $date;
		}

		public function toArray($get_lines) {
			$data = parent::toArray($get_lines);
			return $data;
		}

		/**
		*       Prepare this object for serialization
		*
		*       @return array Properties to preserve across serialization.
		*/
		public function __sleep() {
			return parent::__sleep() + array('po_id', 'po_number', 'supplier_id', 'comments', 'created_date',
				'order_date', 'into_stock_location', 'contact', 
				'approver', 'approved_date', 'requisition_number', 'arrival_date', 
				'currency', 'gnr_account', 'transaction', 'invoiced_date', 'received_date',
				'lines');
		}

		public function suggest_order_number() {
			return $this->suggest_purchase_order_number($this->get_order_id());
		}

		/**
		*	Suggest a purchase order number (human reference) for the given po_id
		*	Uses config var
		*	@param po_id	int	ID of the PurchaseOrder we're suggesting a number for
		*	@return string suggested number based on site wide settings
		*/
		public function suggest_purchase_order_number($po_id) {
			// Get po number format
			$config = config::read('perp_ap');
			$po_code = $config['po_number'];
		
			// Get supplier_id & next_supplier_num
			if($po_id) {
				$where = array('supplier_id = (
					SELECT supplier
					FROM ' . self::order_table . '
					WHERE order_id = ' . $po_id . ')'
				);
			} elseif ($this->get_supplier_id()) {
				$supplier = ExecMethod('perp_ap.bo_perp_supplier.get_supplier', $this->get_supplier_id());
				$where[] = 'perp_order.supplier_id = ' .  $this->get_supplier_id();
				$supplier_num = $supplier['supplier_num'];
			} else {
				return false;
			}

			$GLOBALS['egw']->db->select(self::order_table,
				array('count(order_id) + 1 AS next_supplier_num',
					self::order_table . '.supplier_id',
					'supplier_num'),
				$where,
				__LINE__, __FILE__, False, 
				'GROUP BY supplier_id', 'perp_ap', 0,
				'JOIN perp_supplier ON 
					' . self::order_table . '.supplier_id = perp_supplier.supplier_id'
			);
			$row = $GLOBALS['egw']->db->row(True);
			if(!$supplier_num) {
				$supplier_num = $row['supplier_num'];
			}
			$next_supplier_num = $row['next_supplier_num'];

			// get next_po_id
			$GLOBALS['egw']->db->select(
				self::order_table,
				array('max(order_id) + 1 AS next_po_id'),
				array(),
				__LINE__, __FILE__, False, False, 'perp_ap');
			$row = $GLOBALS['egw']->db->row(True);
			$next_po_id = $row['next_po_id'];

			// get next_by_date
			$where['supplier_id'] = $this->supplier_id;
			$where[] = 'our_reference_number IS NOT NULL';
			$where[] = 'issue_date BETWEEN \'' . $GLOBALS['egw']->db->to_timestamp(date('U', mktime(0,0,0,date('m'), date('d'), date('Y')))) . 
					'\' AND \'' . $GLOBALS['egw']->db->to_timestamp(date('U', mktime(0,0,0,date('m'), date('d') + 1, date('Y')))) .'\'';
			$GLOBALS['egw']->db->select(self::order_table,
				array('count(order_id) + 1 AS next_by_date'),
				$where,
				__LINE__, __FILE__, False
			);
			$row = $GLOBALS['egw']->db->row(True);
			$next_by_date = $row['next_by_date'];
				
			ereg('date\([dDjlNSwzWFmMntYycU.]+\)', $po_code, $regs);
			if(count($regs)) {
				eval('$date = ' . $regs[0] . ';');
				$po_code = str_replace($regs[0], '%5$s', $po_code);
			}
			
			$po_num = sprintf($po_code, 
				$supplier_num,
				$next_po_id,
				$next_supplier_num,
				sprintf("%02d", $next_by_date),
				$date
			);
			return $po_num;
		}
	}
?>
