<?php
	/**
	*	@brief One item on a PurchaseOrder
	*
	*	PurchaseOrderLine encapsulates one item being ordered from a Supplier on a PurchaseOrder.
	*	Each PurchaseOrderLine must have a quantity, and must reference an item in a Supplier's 
	*	catalog.  
	*
	*	@author Nathan Gray
	*	@ingroup perp_ap
	*/

	require_once(EGW_INCLUDE_ROOT . '/perp_api/inc/class.OrderLine.inc.php');

	class PurchaseOrderLine extends OrderLine {

		/**
		*	Date this PurchaseOrderLine was completely received
		*/
		protected $received_date;

		/**
		*	Date this PurchaseOrderLine was completely invoiced
		*/
		protected $invoiced_date;

		/**
		*	This part of the order is fulfilled
		*/
		protected $completed = False;

		/**
		*	Received this time (used for receiving)
		*/
		protected $received;

		/**
		*	ID of the shipment the goods came in on.
		*	Used for receiving.
		*/
		protected $shipment_id = null;

		/**
		*	Constructor.  Will try and find any information not provided.
		*	@param po	PurchaseOrder this line is part of
		*	@param info	Optional.  Array of information for this line
		*/
		public function PurchaseOrderLine(PurchaseOrder $po, $info) {
			parent::__construct($po, $info);
			$this->received_date = $info['received_date'];
			$this->invoiced_date = $info['invoiced_date'];
			$this->qty_invoiced = $info['qty_invoiced'];
			$this->quantity_received = $info['quantity_received'];
			$this->completed = $info['completed'];

			// Try for missing info
			if($this->get_supplier_stock_code() && !(
				$this->description && 
				$this->get_gl_account() && 
				$this->unit_price)) {
				$item = ExecMethod2('perp_ap.bo_perp_catalog.get_catalog_item', $po->get_supplier_id(), $this->get_supplier_stock_code());
				if(!$this->description) {
					$this->description = $item['description'];
				}
				if(!$this->unit_price) {
					$this->unit_price = $item['price'];
				}
				if(!$this->stock_id) {
					$this->stock_id = $item['stock_id'];
				}
				if($item['stock_id'] && !$this->gl_account) {
					$item = ExecMethod2('perp_inventory.bo_perp_stock.get_stock_item', $item['stock_id']);
					if(count($item)) {
						if(!$this->get_gl_account()) {
							$this->set_gl_account(bo_perp_stock::get_gl_accounts($item['stock_id'], bo_perp_inventory::ON_HAND));
						}
					}
				}
			}
		}

		// Accessors
		// TODO: add the rest of the accessors

		/**
		*	Get our system reference (not human reference) for the stock item being ordered.
		*	This will be null if the item being ordered is not one of our stock items.
		*	@return int stock_id
		*/
		public function get_stock_id() {
			return $this->stock_id;
		}

		/**
		*	Get the Supplier's reference for the item being ordered
		*/
		public function get_item_code() {
			return $this->get_supplier_code();
		}
		/**
		*	Set the Supplier's reference for the item being ordered
		*	@param code	string	Supplier's reference for the item
		*	@return	null or error message if the code is not a catalog item
		*/
		public function set_item_code($code) {
			$check = ExecMethod2('perp_ap.bo_perp_catalog.get_catalog_item', $this->purchase_order->get_supplier(), $code);
			if(is_array($check)) {
				$this->set_supplier_code($code);
			} else {
				return lang('Invalid supplier stock code');
			}
		}

		/**
		*	Get the description of this line
		*	@return string Line description
		*/
		public function get_description() {
			return $this->description;
		}

		/**
		*	Add some links to status
		*/
		public function get_status() {
			$list = parent::get_status();
			foreach($list as &$item) {
				$item += array(
					'history_name'  =>      ucfirst(strtolower($item['type'])),
					'reference'     =>      array(
						'app'   =>      'perp_ap',
						'id'    =>      $item['type'] . '-' . $item['reference_id'],
						'title' =>      $item['our_reference_number']
					)
				);
			}
			return $list;
		}

		/**
		*	Is this line completely delivered and paid for?
		*	@return boolean Line completed
		*/
		public function is_complete() {
			return parent::is_complete() && 
				($this->get_quantity_ordered() == $this->get_quantity_received());
		}

		/**
		*	Get the date this PurchaseOrderLine was completely invoiced
		*	@return Invoiced date (Y-m-d)
		*/
		public function get_invoiced_date() {
			return $this->invoiced_date;
		}
		
		/**
		*	Get the date this PurchaseOrderLine was completely delivered
		*	@return Received date (Y-m-d)
		*/
		public function get_received_date() {
			return $this->received_date;
		}

		public function get_ship_unit_price() {
			return $this->get_unit_price();
		}

		/**
		*	Mark this ShippableItem as (partially) shipped.
		*	@todo: we don't mark as shipped yet
		*/
		public function ship(ShipmentLine &$shipment_line) {}

		/**
		*	Receive a certain number of this item off the given shipment.
		*	@param receipt_line	ReceiptLine being received
		*	@return string May return an error message if something's wrong.
		*/
		public function receive(ReceiptLine &$receipt_line) {
			if(!$qty) {
				return lang('No quantity specified');
			}

			// Set quantity received.  goods_received record set in boshipment
			$this->quantity_received += $qty;
			$this->received = $qty;
			$this->shipment_id = $shipment_id;
			
			// Close this line if we've received & invoiced all of them
			if($this->get_quantity_received() == $this->get_quantity_ordered()) {
				$this->received_date = date('Y-m-d');
				if($this->get_quantity_invoiced() == $this->get_quantity_ordered()) {
					$this->completed = True;
					$this->write();
				}
			}
		}

		/**
		*	Get the data in this PurchaseOrderLine as an array.
		*	@return the properties of this line as an array
		*/   
		public function toArray() {
			$data = parent::toArray();
			$data += array(
				'item_code'		=>	$this->get_supplier_stock_code(),
				'stock_id'		=>	$this->stock_id,
				'received_date'		=>	$this->received_date,
				'invoiced_date'		=>	$this->invoiced_date,
				'description'		=>	$this->description,
				'gl_account'		=>	$this->gl_account,
				'quantity_invoiced'	=>	$this->qty_invoiced,
				'unit_price'		=>	$this->unit_price,
				'quantity_ordered'	=>	$this->quantity_ordered,
				'quantity_received'	=>	$this->quantity_received,
				'completed'		=>	$this->completed
			);
			return $data;
		}

		/**
		*	Prepare this object for serialization
		*	@return array Properties to preserve across serialization.
		*/
		public function __sleep() {
			$properties = $this->toArray();
			$properties['received'] = $this->received;
			$properties['purchase_order'] = $this->purchase_order;
			return parent::__sleep() + array_keys($properties);
		}

		/**
		*	Restore this object from serialization.
		*/
		public function __wakeup() {
		}
	}
?>
