<?php
	/**
	*	Encapsulating object for receiving a ShipmentLine.
	*	
	*	This is used when a ShipmentLine is received.  
	*	It will handle creating a received goods record, and 
	*	updating stock counts.
	*
	*	@ingroup perp_ap
	*	@author Nathan Gray
	*/

	require_once(EGW_INCLUDE_ROOT . '/perp_api/inc/class.ShipmentLine.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_api/inc/class.ReceiptLine.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_inventory/inc/class.bo_perp_stock_history.inc.php');

	class ReceiveLine implements ReceiptLine {
		
		/**
		*	Quantity received off the shipment, in the supplier's
		*	unit of measure.
		*/
		protected $quantity_received;

		/**
		*	ShipmentLine received
		*/
		protected $shipment_line;

		/**
		*	Who received the goods
		*/
		protected $received_by;

		/**
		*	Date the goods were received
		*/
		protected $receipt_date = null;

		/**
		*	Serial numbers of serialized goods
		*/
		protected $serials;

		/**
		*	Constructor.
		*	Takes the ShipmentLine and quantity received
		*	@param shipment_line ShipmentLine received or int goods_received_id to load
		*	@param quantity Quantity received in supplier's unit of measure
		*	@param serial_numbers Serial numbers of serialized goods
		*/
		public function __construct(&$shipment_line, $received_by = null, $quantity=0, array $serial_numbers = array(), $receipt_date = null) {
			if($shipment_line instanceof ShipmentLine) {
				$this->shipment_line = $shipment_line;
				$this->quantity_received = $quantity;
				if($received_by) {
					$this->received_by = $received_by;
				} else {
					$this->received_by = $GLOBALS['egw_info']['user']['account_id'];
				}
				$this->serials = $serial_numbers;
				if($receipt_date) {
					$this->receipt_date = $receipt_date;
				} else {
					$this->receipt_date = time();
				}
			} elseif (is_numeric($shipment_line)) {
				$this->read($shipment_line);
			}
		}

		/**
		*	Read info from the database
		*
		*	@param id goods_reveived_id
		*/
		public function read($id) {
			$GLOBALS['egw']->db->select(
				'perp_goods_received',
				array('*'),
				array('goods_received_id' => (int)$id),
				__LINE__, __FILE__, 0,
				'', 'perp_ap', 1,
				'LEFT JOIN perp_stock_cost ON
					perp_stock_cost.stock_cost_id = perp_goods_received.stock_cost_id'
			);
			$this->data = $GLOBALS['egw']->db->row(true);
			$this->data['receive_date'] = $GLOBALS['egw']->db->from_timestamp($this->data['receive_date']);
		}

		/**
		*	Receive the goods into inventory, create
		*	database entry.
		*	Calls ShipmentLine::receive()
		*/
		public function receive() {
			// Check to see if it's inventory goods
			$supplier_id = $this->shipment_line->get_owner()->get_supplier_id();
			$shippable = $this->shipment_line->get_shippable_item();
			if($shippable instanceof OrderLine) {
				$supplier_code = $shippable->get_supplier_stock_code();
				$stock_id = $shippable->get_stock_id();
				$location_id = $shippable->get_owner()->get_location_id();
			}

			$cat_item = ExecMethod2('perp_ap.bo_perp_catalog.get_catalog_item', $supplier_id, $supplier_code, $stock_id);
			if($cat_item) {
				$qty_our_units = $this->quantity_received;

			// Check for serial numbers
				if($stock_id) {
					$stock_item = ExecMethod('perp_inventory.bo_perp_stock.get_stock_item', $stock_id);
					if($stock_item['controlled'] && $stock_item['serial_setup'] == -1) {
						// Just check that quantities match for now
						$serial_count = array_sum($this->serials);
						if($serial_count != $qty_our_units) {
							throw new Exception(lang('Problem with serial numbers for %1, %2 needed but %3 were given', ExecMethod('perp_inventory.bo_perp_stock.link_title', $stock_id), $qty_our_units, $serial_count));
						}
					}

				// Add to inventory
					$stock_cost_id = ExecMethod2('perp_inventory.bo_perp_inventory.add_stock_cost',
						$cat_item['stock_id'],
						$qty_our_units,
						$shippable->get_unit_price(),
						$cat_item['average_indirect_cost']
					);
				}

				if($stock_cost_id) {
					ExecMethod2(
						'perp_inventory.bo_perp_inventory.receive', 
						$cat_item['stock_id'], 
						$location_id, 
						$qty_our_units, 
						$stock_cost_id, 
						$this->serials,
						$this->shipment_line->get_owner()->get_our_reference_number()
					);
				}
				$reference = lang('Received on %1', $this->shipment_line->get_owner()->get_our_reference_number());
				foreach($this->serials as $serial => $qty) {
					bo_perp_stock_history::add_history(
						$stock_id, $serial, 
						$reference, 
						array(
							'app'	=>	'perp_ap',
							 'id'	=>	'SHIPMENT-' . $this->shipment_line->get_owner()->get_shipment_id()
						), 
						__LINE__, __FILE__
					);
				}

				// GL Transaction happens on invoice

			} elseif ($stock_id) {
				throw new egw_exception(lang('Could not find catalog entry, but this is supposed to be one of our items:<br />%1<br />
					You should check the supplier catalog for "%2", and try removing and re-adding this line to the Purchase Order.',
					$this->shipment_line->__toString(),
					$supplier_code
				));
			}

			// Create received record
			$GLOBALS['egw']->db->insert('perp_goods_received',
				array(
					'shipment_id'	=>	$this->shipment_line->get_owner()->get_shipment_id(),
					'line_no'	=>	$this->shipment_line->get_line_number(),
					'quantity_units'	=>	$this->quantity_received,
					'receive_date'	=>	$this->receipt_date,
					'stock_cost_id'	=>	$stock_cost_id,
					'received_by'	=>	$this->received_by,
				),
				array(),
				__LINE__, __FILE__, 'perp_ap'
			);

			// Mark shipment as received
			$this->shipment_line->receive($this);
		}

		public function get_receipt_date() {
			return $this->receipt_date;
		}

		public function get_shipment_line() {
			return $this->shipment_line;
		}

		public function toArray() {
			return $this->data;
		}
	}
?>
