<?php
	/**
	*	@brief An entry into the Supplier Ledger
	*
	*	SupplierTransactionLine encapsulates a single entry onto the Supplier Ledger.
	*	It must be linked to a perp_ledger::TransactionLine in a perp_leger::GLTransaction.
	*
	*	@author Nathan Gray
	*	@ingroup perp_ap
	*/
	require_once(EGW_SERVER_ROOT . '/perp_ledger/inc/class.GLTransaction.inc.php');

	class SupplierTransactionLine extends TransactionLine {
		/**
		*	Database table for Supplier Ledger
		*/
		private $table = 'perp_supplier_ledger';

		/**
		*	ID of the supplier
		*/
		private $supplier;

		/**
		*	Due date.  Used for invoices
		*/
		private $due_date;

		/**
		*	@deprecated
		*/
		private $settled;

		/**
		*	Exchange rate at the time the transaction was entered
		*/
		private $exchange_rate;

		/**
		*	@deprecated
		*/
		private $purchase_order_id;

		/**
		*	Link to another aspect of pERP for reference, such as an Invoice, Shipment or Payment.
		*	Similar to eGroupWare's links, <application_name>:<location>:<record_id>.  Generated by
		*	bolink::link()
		*/
		private $link;

		/**
		*	Supplier's reference number for this transaction.
		*/
		private $supplier_reference;

		/**
		*	Free for text.  Not used yet, should probably allow Infologs instead.
		*/
		private $comment;

		/**
		*	Constructor
		*	@param parent	TransactionLine	Transaction to the General Ledger to associate with.  
		*	@param line_no	Transaction line number to read, if the GLTransaction has already been posted
		*/
		public function SupplierTransactionLine($parent, $line_no = '') {
			if(is_a($parent, 'TransactionLine')) {
                                // An associated transaction
                                $this->transaction = $parent->transaction;
                                if(!$this->transaction->get_transaction_number()) {
                                        $this->set_account($parent->get_account());
                                        $this->set_amount($parent->get_amount());
                                        $this->narrative = $parent->get_narrative();
                                }
				$parent->add_associated_line($this);
				if($this->transaction->get_transaction_number()) {
					return $this->read($this->transaction->get_transaction_number(), $line_no);
				}
                        } elseif( is_a($parent, 'GLTransaction')) {
                                // New transaction
                                $this->transaction = $parent;
				if($this->transaction->get_transaction_number()) {
					return $this->read($this->transaction->get_transaction_number(), $line_no);
				}
                        }
		}

		/**
		*	Read this TransactionLine from the Supplier Ledger, sets this SupplierTransactionLine's parameters.
		*	@param trans_id	int	Transaction number to read
		*	@param line_no	int	Line number of the transaction to read
		*/
		public function read($trans_id, $line_no) {
			$GLOBALS['egw']->db->select($this->table, 
				array(
					'trans_id',
					'line_no',
					'supplier_id',
					'due_date',
					'exchange_rate',
					'amount',
					'supplier_reference',
					'link',
					'comment'
				),
				array(
					'trans_id'	=>	$trans_id,
					'line_no'	=>	$line_no
				),
				__LINE__, __FILE__, 'perp_ap'
			);
			$row = $GLOBALS['egw']->db->row(true);
			$this->supplier		=	$row['supplier_id'];
			$this->due_date		=	$row['due_date'];
			$this->exchange_rate	=	$row['exchange_rate'];
			$this->amount		=	$row['amount'];
			$this->purchase_order_id	=	$row['purchase_order_id'];
			$this->supplier_reference	=	$row['supplier_reference'];
			$this->link		=	$row['link'];
			$this->comment		=	$row['comment'];
		}

		/**
		*	Write (or 'post') this TransactionLine to the Supplier Ledger
		*	@param line_no	The line number of the GLTransaction that this TransactioLine is associated with
		*/
		public function write($line_no) {
			$GLOBALS['egw']->db->insert($this->table, array(
				'trans_id'	=>	$this->transaction->get_transaction_number(),
				'line_no'	=>	$line_no,
				'supplier_id'	=>	$this->supplier,
				'due_date'	=>	$this->due_date,
				'settled'	=>	$this->settled,
				'exchange_rate'	=>	$this->exchange_rate,
				'amount'	=>	$this->get_amount(),
				'link'		=>	$this->link,
				'supplier_reference'	=>	$this->supplier_reference,
				'comment'	=>	$this->comment),
				array(
					'trans_id'	=>	$this->transaction->get_transaction_number(),
					'line_no'	=>	$line_no
				),
				__LINE__, __FILE__, 'perp_ap'
			);
		}

		/**
		*	Get the supplier for this transaction
		*/
		public function get_supplier() {
			return $this->supplier;
		}

		/**
		*	Set the supplier for this transaction
		*	@param supplier	int	supplier_id
		*/
		public function set_supplier($supplier) {
			$this->supplier = $supplier;
		}

		/**
		*	Set the date payment is due (for invoice transactions)
		*	@param date	Date payment is due (Y-m-d)
		*/
		public function set_due_date($date) {
			$this->due_date = $date;
		}

		/**
		*	@deprecated
		*/
		public function set_settled($settled) {
			$this->settled = $settled;
		}

		/**
		*	Get the exchange rate at the time of this transaction
		*	@return float exchange rate
		*/
		public function get_exchange_rate() {
			return $this->exchange_rate;
		}
		/**
		*	Set the exchange rate at the time of this transaction
		*	@param rate float Exchange rate
		*/
		public function set_exchange_rate($rate) {
			$this->exchange_rate = $rate;
		}

		/**
		*	@depricated
		*/
		public function set_purchase_order_id($po_id) {
			$this->purchase_order_id = $po_id;
		}

		/**
		*	@deprecated, use set_link instead
		*/
		public function set_reference($ref) {
			$this->reference = $ref;
		}

		/**
		*	Set a link to another part of pERP (or eGroupWare) such as 
		*	an Invoice or Payment for reference
		*/
		public function set_link($link_to_app, $link_to_id) {
			$temp_id = 0;
			$title = egw_link::title($link_to_app, $link_to_id);
			$this->link = egw_link::link(' ', $temp_id, $link_to_app, $link_to_id, $title);
		}

		/**
		*	Set the Supplier's reference for this transaction
		*	@param supplier_ref	string	Supplier's reference number
		*/
		public function set_supplier_reference($supplier_ref) {
			$this->supplier_reference = $supplier_ref;
		}

		/**
		*	Set the comment on this transaction
		*	@param	c	string	Free form comments
		*/
		public function set_comment($c) {
			$this->comment = $c;
		}

		/**
		*	Prepare this object for serialization
		*
		*	@return array Properties to preserve across serialization.
		*/
		public function __sleep() {
			$sleep = parent::sleep();
			$sleep += array('supplier', 'due_date', 'settled', 'exchange_rate', 'purchase_order_id', 'link', 'supplier_reference', 'comment');
			return $sleep;
		}

		/**
		*	Return a string representation of this line. 
		*	Very useful for debugging purposes, not much else.
		*
		*	@return string Representation of this TransactionLine
		*/
		public function __toString() {
			return "SupplierTransactionLine \tSupplier: " . $this->supplier . "\tAmount: " . $this->amount ."\t(" . $this->link .")\n";
		}
	}
?>
