<?php
	/** 
	*	The *catalog classes are for dealing with a supplier's catalog items
	*	(catalogue) and possible links to our stock items.  It is not required
	*	that a supplier's catalog item correspond to one of our stock items.
	*
	*	@author Nathan Gray
	*	@ingroup perp_ap
	*/

	class bo_perp_catalog {
		/**
		*	Reference to storage layer object
		*/
		private $so;
		
		/**
		*	boolean Turn on for debug information
		*/
		private $debug = False;
		
		/**
		*	Constructor
		*/
		public function bo_perp_catalog() {
			$this->so = CreateObject('perp_ap.so_perp_catalog');
		}

		/**
		*	Get a catalog for a supplier given by ID in query['supplier_id']
		*	@nextmatch
		*/
		public function get_supplier_catalog_list($query = array(), &$rows = array(), &$readonlys = array()) {
			$read = $GLOBALS['egw']->acl->check('catalog', EGW_ACL_READ, 'perp_ap');
			if($this->debug) {
				echo 'get_supplier_catalog_list query: ' . ($read ? '' : 'NO ') . 'Read permission<br />';
			}
			if(!$read || !$query['supplier_id']) {
				if(!isset($query['template'])) {
					return array();
				} else {
					$rows = array();
					return 0;
				}
			}

			if($query['filter']) {
				$query['stock_category'] = $query['filter'];
			}
			if(!$query['order']) {
				$query['order'] = 'supplier_code';
			}
			if(!$query['template']) {
				// For a selectbox
				$query['num_rows'] = -1;
			}

			$list = $this->so->get_supplier_catalog_list($query);

			// Selectbox
			if(!isset($query['template'])) {
				foreach($list as $row) {
					$rows[$row['supplier_code']] = $row['supplier_code'] . ' - ' 
					. $row['description'] . ' (' .
					$row['quantity'] . ' ' . $row['supplier_uom'] . 
					($row['stock_id'] ? (' = ' . ($row['quantity'] * $row['conversion_factor']) . ' ' . $row['stock_uom_name']) : '') . 
					')';
				}
				return $rows;
			}

			$add = $GLOBALS['egw']->acl->check('catalog', EGW_ACL_ADD, 'perp_ap');
			$edit = $GLOBALS['egw']->acl->check('catalog', EGW_ACL_EDIT, 'perp_ap');
			$rows = $read ? $list : array();
			$rows['add'] = $add;
			$rows['edit'] = $edit;
			$rows['delete'] = $GLOBALS['egw']->acl->check('catalog', EGW_ACL_DELETE, 'perp_ap');
			$rows['add_to_po'] = $GLOBALS['egw']->acl->check('order', EGW_ACL_EDIT, 'perp_ap');
			if($query['add_prompt']) {
				$rows['add_prompt'] = $query['add_prompt'];
			}

			// Can't delete catalog items on an open PO
			if($rows['delete']) {
				foreach($rows as $key => $row) {
					if(is_array($row)) {
						$readonlys['delete[' . $row['escaped_supplier_code'] . ']'] = $row['in_use'];
					}
				}
			}
			return $query['count'];
		}

		/**
		*	Get a single item from a supplier's catalog
		*	
		*	@param supplier_id	int	Supplier's ID
		*	@param supplier_code	string	Supplier's identifier for the desired catalog item
		*	@param stock_id		int	Optional.  Our stock ID
		*	@return array	Information about the desired catalog item
		*/
		public function get_catalog_item($supplier_id, $supplier_code, $stock_id = '') {
			$read = $GLOBALS['egw']->acl->check('catalog', EGW_ACL_READ, 'perp_ap');
			$list = array();

			if($read) {
				$list = $this->so->get_catalog_item($supplier_id, $supplier_code, $stock_id);
				if($list) {
					$list['packages'] = $this->so->get_shipping_packages($supplier_id, $supplier_code);
				}
			}
			return $list;
		}

		/**
		*	Update an item on a supplier's catalog
		*	
		*	@param supplier_id	int	Supplier's ID
		*	@param supplier_code	string	Supplier's identifier for the desired catalog item
		*	@param item		array	Catalog item information
		*/
		public function edit_catalog($supplier_id, $supplier_code, $item) {
			$edit = $GLOBALS['egw']->acl->check('catalog', EGW_ACL_EDIT, 'perp_ap');
			$add = $GLOBALS['egw']->acl->check('catalog', EGW_ACL_ADD, 'perp_ap');
			$message = true;

			if($edit || $add) {
				if($item['stock_id'] && $item['conversion_factor'] == 0 ) {
					$item['conversion_factor'] = 1;
				}

				// Check for a changed supplier code, and delete / deactivate old one
				if($supplier_code != $item['old_supplier_code']) {
					$old_item = $this->get_catalog_item($supplier_id, $item['old_supplier_code']);
					if(!$this->so->in_use($old_item)) {
						$this->delete($supplier_id, $item['old_supplier_code']);
					} else {
						$message = lang('Supplier Stock Code cannot be edited until all Purchase Orders using it have been received.
							A new catalog item has been created');
					}
				}

				// Get old catalog for history
				if($item['stock_id']) {
					$history_log =& CreateObject('phpgwapi.historylog', 'perp_inventory');
					if(!$old_item['stock_id']) {
						// New supplier
						$history_log->add('sp', $item['stock_id'], $supplier_id, 'added');
					} elseif ($item['stock_id'] != $old_item['stock_id']) {
						// Changed our stock item
						$history_log->add('sp', $old_item['stock_id'], $supplier_id, 'removed');
						$history_log->add('sp', $item['stock_id'], $supplier_id, 'added');
					} elseif ($item['preferred'] != $old_item['preferred']) {
						$history_log->add('ps', $item['stock_id'], $supplier_id, '');
					}
				}

				$this->so->set_catalog_item($supplier_id, $supplier_code, $item);
				if(isset($item['packages'])) {
					$this->so->set_shipping_packages($supplier_id, $supplier_code, $item['packages']);
				}

				return $message;
			}
		}

		/**
		*	Remove an item from a supplier catalog
		*	
		*	@param supplier_id	int	Supplier's ID
		*	@param item_code	string	Supplier's identifier for the desired catalog item
		*/
		public function delete($supplier_id, $item_code) {
			$delete = $GLOBALS['egw']->acl->check('catalog', EGW_ACL_DELETE, 'perp_ap');
			if($delete) {
				$old_item = $this->get_catalog_item($supplier_id, $item_code);
				if($old_item['stock_id']) {
					$history_log =& CreateObject('phpgwapi.historylog', 'perp_inventory');
					$history_log->add('sp', $old_item['stock_id'], $supplier_id, 'removed');
				}
				$this->so->delete($supplier_id, $item_code);
			}
		}

		/**
		*	Link title
		*
		*	@param entry - Key in the form <supplier_id>-<supplier_code>
		*	@return string Title for linking, etc.
		*/
		public function link_title($entry) {
			if(!is_array($entry)) {
				list($supplier_id, $supplier_code) = explode('-', $entry, 2);

				// Extra check needed because the AJAX select widget doesn't return the supplier_id.
				if(is_numeric($supplier_id) && $supplier_code) {
					$item = $this->get_catalog_item($supplier_id, $supplier_code);
				}
				if($item) {
					$title = $item['description'];
					return $title;
				} else {
					// Just use the supplier's code
					return $entry;
				}
			} else {
				return $entry;
			}
		}

		/**
		*	Get the average indirect cost (shipping, taxes, etc) for an item.
		*	Calculated from the indirect cost as actually entered in a stock item's finance tab 
		*	over the past year.
		*
		*	@param supplier_id
		*	@param stock_id
		*	@return double average indirect cost
		*/
		public function calculate_average_indirect_cost($supplier_id, $stock_id) {
			$read = $GLOBALS['egw']->acl->check('catalog', EGW_ACL_READ, 'perp_ap') &&
				$GLOBALS['egw']->acl->check('inventory', EGW_ACL_READ, 'perp_inventory');
			if($read) {
				return $this->so->calculate_average_indirect_cost($supplier_id, $stock_id);
			} else {
				return null;
			}
		}

		/**
		*	Determine if a catalog item is on any open (un-received) Purchase Order
		*/
		public function in_use(array $catalog) {
			return $this->so->in_use($catalog);
		}
	}
?>
