<?php

	/**
	*	Logic layer object for incoming shipments
	*
	*	@author Nathan Gray
	*	@ingroup perp_ap
	*	@todo Shipments should be encapsulated into an object 
	*/

	require_once(EGW_INCLUDE_ROOT . '/perp_api/inc/class.bo_perp_shipment.inc.php');
	require_once('class.IncomingShipment.inc.php');

	class bo_perp_incoming_shipment extends bo_perp_shipment {
	
		/**
		*	boolean Enable for debug messages
		*/
		private $debug = False;

		/**
		*	Constructor
		*/
		public function __construct() {
			parent::__construct();
			$this->so_perp_shipment =& CreateObject('perp_ap.so_perp_incoming_shipment');
		}

		/**
		*	Get a list of shipments
		*
		*	@nextmatch
		*/
		public function get_shipment_list($query = array(), &$rows = array(), &$readonlys = array()) {
			$query['client_id'] = 0;

			// Check if the current user is a supplier
			$supplier_ids = ExecMethod('perp_ap.bo_perp_supplier.get_supplier_by_user_account');
			if(is_array($supplier_ids) && count($supplier_ids) > 0 && !in_array($query['supplier_id'], $supplier_ids)) {
				$query['supplier_id'] = $supplier_ids;
			}

			$count = parent::get_shipment_list($query, $rows, $readonlys);

			if($query['hide_supplier']) {
				// All the same supplier, hide supplier column
				$rows['supplier_id'] = True;
			}
			return $count;
		}

		/**
		*	Get a list of items on order from a supplier that can be shipped
		*
		*	@nextmatch
		*/
		public function get_shippable_items($query = array(), &$rows = array(), &$readonlys = array()) {
			// Check if the current user is a supplier
			$supplier_ids = ExecMethod('perp_ap.bo_perp_supplier.get_supplier_by_user_account');
			if(is_array($supplier_ids) && count($supplier_ids) > 0 && !in_array($query['supplier_id'], $supplier_ids)) {
				$query['supplier_id'] = $supplier_ids;
			}

			parent::get_shippable_items($query, $list, $readonlys);

			$items = array();
			$count = -0;
			$order_id = 0;
			foreach($list as $row) {
				if($row['order_id'] != $po_id) {
					$order_id = $row['order_id'];
					$count++;
					$rows['order'][$count]['lines'][] = 'Title';
				}
				$row['shippable_class'] = 'PurchaseOrder';
				$rows['order'][$count]['order_id'] = $row['purchase_order_id'];
				$rows['order'][$count]['our_reference_number'] = $row['our_reference_number'];
				$rows['order'][$count]['into_stock_location'] = $row['into_stock_location'];
				$rows['order'][$count]['lines'][] = $row;
			}
			return count($rows);
		}

		/**
		*	Get a list of possible statuses for shipments
		*	@return array of statuses
		*/
		public function get_status_list() {
			return array(
				''		=>	'All',
				'pending'	=>	lang('pending'),
				'in-transit'	=>	lang('in-transit'),
				'arrived'	=>	lang('arrived'),
				'overdue'	=>	lang('overdue'),
			);
		}

		/**
		* get title for a perp supplier shipment entry identified by $entry
		*                               
		* Is called as hook to participate in the linking                
		*
		* @param entry int/array key to look up or array with shipment entry
		* @return string  with title/boolean
		*/

		public function link_title($entry) {
			if(is_array($entry)) {
				return $entry['our_reference_number'];
			} elseif(is_numeric($entry)) {
				$shipment = new IncomingShipment($entry);
				return $shipment->get_our_reference_number();
			}
			return false;
		}

		/**
		*	Get an approximation of the indirect costs, such as freight or duty,
		*	associated with getting a shipment here.  The shipment should be invoiced
		*	already, and must be received already.
		*
		*	If the shipment is spread over multiple invoices, with items from other
		*	shipments on one or both invoices, the cost will not be accurate.
		*
		*	If the shipment has not been invoiced yet, the average indirect cost, 
		*	as set in the supplier's catalog, will be used instead of the actual costs.
		*
		*	First, we look for invoices for the associated Purchase Order Lines that
		*	are on the shipment.  Any lines that are not associated with an InvoicableItem
		*	are assumed to be indirect costs.  The total of all indirect costs on any
		*	invoices found is used as the total indirect cost for this shipment.
		*	If no invoices are found, the total average indirect cost for items is used,
		*	multiplied by the quantity received.
		*
		*	@param shipment IncomingShipment to calculate the indirect cost for
		*	@return float Total indirect cost
		*/
		public function get_indirect_cost(IncomingShipment &$shipment, array &$source_lines = array()) {
			if(!$shipment->get_actual_delivery_date()) {
				// Shipment has not been received yet
				return false;
			}


			// Look for invoices
			$invoices = array();
			$lines =& $shipment->get_lines();
			foreach($lines as &$shipment_line) {
				$order_line =& $shipment_line->get_shippable_item();
				if(is_null($order_line)) {
					continue;
				}
				
				$status = $order_line->get_status();
				foreach($status as $status_entry) {
					if($status_entry['type'] == 'INVOICE') {
						$invoices[$status_entry['reference_id']] = $status_entry;
					}
				}
			}

			$total_indirect_cost = 0;
			if(count($invoices) > 0) {
				// Calculate total indirect costs
				foreach($invoices as $invoice_id => $invoice_reference) {
					$invoice =& new PayableInvoice($invoice_id);
					$invoice_lines = $invoice->get_lines();
					foreach($invoice_lines as &$invoice_line) {
						if($invoice_line->get_invoicable_item() == null) {
							$source_lines[] = $invoice_reference +
								$invoice_line->toArray() + array(
								'line_price' => $invoice_line->get_unit_price() * $invoice_line->get_quantity()
							);
							$total_indirect_cost += $invoice_line->get_unit_price() * $invoice_line->get_quantity();
						}
					}
				}
			} else {
				// Look for average indirect cost from supplier
			}

			return $total_indirect_cost;
		}

		/**
		*	Allocate indirect costs across received goods
		*
		*	The shipment must be received, and really should be invoiced.  The total
		*	cost will be allocated across received goods according to the method 
		*	specified.
		*
		*	Method is one of quantity, cost, size, weight
		*
		*	@param shipment Shipment to allocate costs for
		*	@param method Allocation method
		*	@param total_costs Total amount to allocate
		*/
		public function allocate_indirect_costs(IncomingShipment &$shipment, $method, $total_cost) {
			if(!$shipment->get_actual_delivery_date()) {
				// Shipment has not been received yet
				return false;
			}
			if($shipment->get_approval_date()) {
				// Shipment is done
				return lang('Costs have already been allocated for this shipment');
			}
			
			$method = 'allocate_' . $method;
			$line_allocation = $this->$method($shipment, $total_cost);

			foreach($line_allocation as $line) {
				$line['stock_id'] = $line['shipment_line']->get_shippable_item()->get_stock_id();
				$shipment_line = $line['shipment_line'];
				unset($line['shipment_line']);
				$status = $shipment_line->get_shippable_item()->get_status();
				foreach($status as $status_entry) {
					if($status_entry['type'] == 'RECEIVED') {
						$received = new ReceiveLine($status_entry['reference_id']);
						$received_data = $received->toArray();
						if($received_data['shipment_id'] == $shipment->get_shipment_id() && $received_data['line_no'] == $shipment_line->get_line_number()) {
							$line['stock_cost_id'] = $received_data['stock_cost_id'];
							ExecMethod2('perp_inventory.bo_perp_stock.set_cost_details', $line['stock_id'], array($line));
						} else {
							throw new egw_exception(lang('Line %1 (%2) has not been received yet.',
								$shipment_line->get_human_line_number(),
								$shipment_line->get_description()
							));
						}
					}
				}
				
			}

			$shipment->approve($GLOBALS['egw_info']['user']['account_id']);
			return true;
		}

		/**
		*	Calculate the allocation of indirect cost according to quantity
		*/
		protected function allocate_quantity(IncomingShipment &$shipment, $total_cost) {

			$lines =& $shipment->get_lines();

			// Get total quantity
			$total_quantity = 0;
			foreach($lines as &$line) {
				if($line->get_quantity_received() == 0 || $line->get_shippable_item() == null || $line->get_shippable_item()->get_stock_id() == null) {
					continue;
				}
				$total_quantity += $line->get_quantity_received();
			}

			// Allocate cost
			$unit_cost = round($total_cost / $total_quantity, 3);
			$allocation = array();
			foreach($lines as &$line) {
				if($line->get_quantity_received() == 0 || $line->get_shippable_item() == null || $line->get_shippable_item()->get_stock_id() == null) {
					// This line was not received, and will not be included in the allocation.
					continue;
				}
				$allocation[] = array(
					'shipment_line'	=>	&$line,
					'indirect_cost'	=>	$unit_cost
				);
			}
				
			return $allocation;
		}

		/**
		*	Calculate the allocation of indirect cost according to cost
		*/
		protected function allocate_cost(IncomingShipment &$shipment, $total_cost) {

			$lines =& $shipment->get_lines();

			// Allocate cost
			$allocation = array();
			foreach($lines as &$line) {
				if($line->get_quantity_received() == 0 || $line->get_shippable_item() == null || $line->get_shippable_item()->get_stock_id() == null) {
					// This line was not received, and will not be included in the allocation.
					continue;
				}
				$allocation[] = array(
					'shipment_line'	=>	&$line,
					'indirect_cost'	=>	round($total_cost / ($line->get_shippable_item()->get_unit_price() * $line->get_quantity_received()), 3)
				);
			}
				
			return $allocation;
		}

		/**
		*	Calculate the allocation of indirect cost according to size
		*/
		protected function allocate_size(IncomingShipment &$shipment, $total_cost) {

			$lines =& $shipment->get_lines();

			// Get total size
			$total_size = 0;

			// Keep line sizes so we don't have to look them up again
			$line_size = array();
			foreach($lines as &$line) {
				if($line->get_quantity_received() == 0 || $line->get_shippable_item() == null || $line->get_shippable_item()->get_stock_id() == null) {
					continue;
				}
				$catalog = ExecMethod2('perp_ap.bo_perp_catalog.get_catalog_item', $shipment->get_supplier_id(), $line->get_shippable_item()->get_supplier_stock_code());
				foreach($catalog['packages'] as $package) {
					$line_size[$line->get_line_number()] += $package['length'] * $package['width'];
				}
				$total_size += $line_size[$line->get_line_number()];
			}

			// Allocate cost
			if($total_size != 0) {
				$unit_cost = round($total_cost / $total_size, 3);
			} else {
				// Probably no shipping sizes set
				$unit_cost = $total_cost;
			}

			// Allocate cost
			$allocation = array();
			foreach($lines as &$line) {
				if($line->get_quantity_received() == 0 || $line->get_shippable_item() == null || $line->get_shippable_item()->get_stock_id() == null) {
					// This line was not received, and will not be included in the allocation.
					continue;
				}
				$allocation[] = array(
					'shipment_line'	=>	&$line,
					'indirect_cost'	=>	round($unit_cost * $line_size[$line->get_line_number()], 3)
				);
			}
				
			return $allocation;
		}

		protected function allocate_mass(IncomingShipment &$shipment, $total_cost) {

			$lines =& $shipment->get_lines();

			// Get total mass
			$total_mass = 0;

			// Keep line masses so we don't have to look them up again
			$line_mass = array();
			foreach($lines as &$line) {
				if($line->get_quantity_received() == 0 || $line->get_shippable_item() == null || $line->get_shippable_item()->get_stock_id() == null) {
					continue;
				}
				$catalog = ExecMethod2('perp_ap.bo_perp_catalog.get_catalog_item', $shipment->get_supplier_id(), $line->get_shippable_item()->get_supplier_stock_code());
				foreach($catalog['packages'] as $package) {
					$line_mass[$line->get_line_number()] += $package['mass'];
				}
				$total_mass += $line_mass[$line->get_line_number()];
			}

			// Allocate cost
			if($total_mass != 0) {
				$unit_cost = round($total_cost / $total_mass, 3);
			} else {
				// Probably no shipping sizes set
				$unit_cost = $total_mass;
			}

			// Allocate cost
			$allocation = array();
			foreach($lines as &$line) {
				if($line->get_quantity_received() == 0 || $line->get_shippable_item() == null || $line->get_shippable_item()->get_stock_id() == null) {
					// This line was not received, and will not be included in the allocation.
					continue;
				}
				$allocation[] = array(
					'shipment_line'	=>	&$line,
					'indirect_cost'	=>	round($unit_cost * $line_mass[$line->get_line_number()], 3)
				);
			}
				
			return $allocation;
		}
	}
?>
