<?php
	/**
	*	Payment of a supplier invoice 
	*
	*	@author Nathan Gray
	*	@ingroup perp_ap
	*/

	require_once('class.PayableInvoice.inc.php');

	class bo_perp_payment {

		/**
		*	Reference to storage layer object
		*/
		private $so;

		/**
		*	boolean Enable for debug messages
		*/
		private $debug = false;

		/**
		*	Constructor
		*/
		public function bo_perp_payment() {
			$this->so = CreateObject('perp_ap.so_perp_payment');
		}

		/**
		*	Get a list of payments that have been made
		*
		*	@nextmatch
		*/
		public function get_payments($query = array(), &$rows = array(), &$readonlys = array()) {
			if(!$GLOBALS['egw']->acl->check('payment', EGW_ACL_READ, 'perp_ap')) {
				if(isset($query['template'])) {
					$rows = array();
					return 0;
				} else {
					return array();
				}
			}

			$query['supplier_id'] = $query['supplier_id'] ? $query['supplier_id'] : $query['filter'];
			$query['status'] = $query['status'] ? $query['status'] : $query['filter2'];
			$query['order'] = $query['order'] ? $query['order'] : 'payment_number';

			// Check if the current user is a supplier
			$supplier_ids = ExecMethod('perp_ap.bo_perp_supplier.get_supplier_by_user_account');
			if(is_array($supplier_ids) && count($supplier_ids) > 0 && !in_array($query['supplier_id'], $supplier_ids)) {
				$query['supplier_id'] = $supplier_ids;
			}

			$list = $this->so->get_payments(&$query);

			if(!isset($query['template'])) {
				$rows = array();
				foreach($list as $row) {
					$rows[$row['payment_id']] = $this->link_title($row);
				}
				return $rows;
			} else {
				$rows = $list;

				if($GLOBALS['egw']->acl->check('payment', EGW_ACL_DELETE, 'perp_ap')) {
					foreach($rows as &$row) {
						$readonlys['reverse[' . $row['payment_id'] . ']'] = ($row['status'] != 1);
					}
				} else {
					$rows['no_reverse'] = true;
				}
				return $query['count'];
			}
		}

		/**
		*	Get the details of a particular payment
		*	@param payment_id ID of the payment to retrieve
		*	@return array Payment information
		*/
		public function get_payment($payment_id) {
			if(!$GLOBALS['egw']->acl->check('payment', EGW_ACL_READ, 'perp_ap')) {
				return lang('access is not allowed');
			}
			if(!$payment_id) {
				return false;
			}
			$payment =  $this->so->get_payment($payment_id);

			// Check if the current user is a supplier
			$supplier_ids = ExecMethod('perp_ap.bo_perp_supplier.get_supplier_by_user_account');
			if(is_array($supplier_ids) && count($supplier_ids) > 0 && !in_array($payment['supplier_id'], $supplier_ids)) {
				return lang('access is not allowed');
			}

			// Check for reversal information
			$reversal = self::get_reversal($payment_id);
			if($reversal) {
				// Copy the payment ID for UI convenience
				$reversal['payment_id'] = $payment_id;
				$payment['reversal'] = $reversal;
			}

			return $payment;
		}

		/**
		*	Get payment amounts that have not been allocated to an invoice.  These
		*	are probably overpayments or prepayments.
		*
		*	@nextmatch
		*/
		public function get_unallocated_amounts($query = array(), &$rows = array(), &$readonlys = array()) {
			if(!$GLOBALS['egw']->acl->check('payment', EGW_ACL_READ, 'perp_ap')) {
				if(isset($query['template'])) {
					$rows = array();
					return 0;
				} else {
					return array();
				}
			}

			// Check if the current user is a supplier
			$supplier_ids = ExecMethod('perp_ap.bo_perp_supplier.get_supplier_by_user_account');
			if(is_array($supplier_ids) && count($supplier_ids) > 0 && !in_array($query['supplier_id'], $supplier_ids)) {
				$query['supplier_id'] = $supplier_ids;
			}


			$list = $this->so->get_unallocated_amounts($query);

			if(!isset($query['template'])) {
				$rows = array();
				foreach($list as $row) {
					$rows[$row['payment_id']] = $this->link_title($row);
				}
				return $rows;
			} else {
				$rows = $list;
				return $query['count'];
			}
		}

		/**
		*	Post payment 
		*	Processes a payment and:
		*	1. Posts it to the General Ledger, Bank Ledger and Supplier Ledger
		*	2. Marks pre-payments as used (if used)
		*	3. Marks invoices as (partially) paid
		*
		*	@param payment_information array in the form:
		*	@code
		*		array(
		*			'supplier_id'		=>	Supplier ID
		*			'payment_id'		=>	Unique Payment ID
		*			'payment_number'	=>	The reference of the payment (Cheque number)
		*			'payment_method'	=>	Payment Method
		*			'bank_account'		=>	Bank account
		*			'allocation_date'	=>	Date this happens on
		*			'pre_payment'		=>	Additional amount to be pre-paid
		*			'payment_total'		=>	Total payment amount
		*			'rows'	=> array(
		*			row# => array(
		*				'reference_id'	=>	either a payment_id or an invoice_number
		*				'trans_id'	=>	If a pre-payment, the transaction_id of the payment
		*				'allocation_date'	=>	The date the payment was allocated
		*				'invoice_id'	=>	If an invoice, the id of the invoice
		*				'original_amount'	=>	Original amount of the payment / invoice
		*				'amount_owing'		=>	Amount of the invoice left to pay, or amount of pre-payment left un-allocated
		*				'discount_taken'	=>	Discount used from the Invoice
		*				'payment_amount'	=>	The amount to be paid
		*			)
		*			)
		*		)
		*	@endcode
		*/
		public function post_payment($payment_information) {
			if(!$GLOBALS['egw']->acl->check('payment', EGW_ACL_ADD, 'perp_ap')) {
				throw new Exception(lang('access is not allowed'));
			}
			if($this->debug) {
				echo 'Post payment:';
				_debug_array($payment_information);
			}
			$supplier_id 	= $payment_information['supplier_id'];
			$payment_id	= $payment_information['payment_id'];
			$payment_number	= $payment_information['payment_number'];
			$payment_method	= $payment_information['payment_method'];
			$bank_account	= $payment_information['bank_account'];
			$bank		= ExecMethod('perp_ledger.bo_perp_ledger.get_bank_account', $bank_account);
			$pre_payment	= $payment_information['pre_payment'];
			$allocation_date= $payment_information['allocation_date'];
			$payment_total	= $payment_information['payment_total'];

			// Get site settings
			$config = config::read('perp_ap');
			$ap_account = $config['ap_account'];
			$discount_account = $config['discount_account'];
			$payment_narrative = $payment_information['payment_narrative'] ? $payment_information['payment_narrative'] : $config['payments'][$payment_method]['payment_narrative_format'];
			$discount_narrative = $config['discount_narrative'];
			$trans_type = $config['payment_transaction_type'];
			if(!$payment_number) {
				$payment_number = $config['payments'][$payment_method]['next_payment_number'];
			} 


			// Begin transaction
			$GLOBALS['egw']->db->transaction_begin();

			// Update next payment (cheque) number
			$config['payments'][$payment_method]['next_payment_number']++;
			config::save_value('payments', $config['payments'], 'perp_ap');

			if(!$payment_id) {
				$payment_id = $this->so->get_next_payment_id();
			}

			// Check for double entry
			$check_payment = $this->get_payment($payment_id);
			if(is_array($check_payment[0])) {
				return lang('Trying to enter payment ID# %1 again', $payment_id);
			}
			$check = array('payment_number' => $payment_number, 'payment_method' => $payment_method);
			$duplicates = $this->get_payments($check);
			if(count($duplicates) > 0) {
				return lang('Trying to enter payment twice.  Payment Method: %1, Payment Number: %2',
					$payment_method, $payment_number
				);
			}

			$api_config = config::read('perp_api');
			$diff_on_exchange_account = $api_config['unrealized_currency_loss_account'];

			// Check bank currency against invoice currencies
			$bo_api = CreateObject('perp_api.bo_perp_api');
			foreach($payment_information['rows'] as $key => $row) {
				if(is_array($row) && $row['invoice_id'] && $row['amount_paid']) {
					$invoice = new PayableInvoice($row['invoice_id']);
					if($invoice->get_invoice_currency() != $bank['currency']) {
						return lang('Invoice currency (%1) doesn\'t match bank currency (%2)', $invoice->get_currency(), $bank['currency']);
					}
				}
			}

			$query = array(
				'currency_id'	=>	$bank['currency'],
				'num_rows'	=>	1,
				'template'	=>	false
			);
			$bo_api->get_currencies( 
				$query,
				$currency
			);
			if($currency[0]) {
				$currency = $currency[0];
			} else {
				return lang('Could not find currency %1', $bank['currency']);
			}
			$exchange = $currency['exchange_rate'];

			if(!$payment_narrative) {
				$payment_narrative = lang('Payment narrative not set') . " %2$s";
			} else {
				$supplier = ExecMethod('perp_ap.bo_perp_supplier.get_supplier', $supplier_id);
			}
			$payment_narrative = sprintf($payment_narrative,
				$supplier['supplier_num'],
				$payment_number,
				($invoice ? $invoice->get_our_reference_number() : '')
			);

			// Set up discount narrative
			if(!$discount_narrative) {
				$discount_narrative = lang('Discount narrative not set') . " %2$s";
			} elseif(!$supplier) {
				$supplier = ExecMethod('perp_ap.bo_perp_supplier.get_supplier', $supplier_id);
			}
			$discount_narrative = sprintf($discount_narrative,
				$supplier['supplier_num'],
				$payment_number,
				($invoice ? $invoice->get_our_reference_number() : '')
			);

			// Set up transaction
			$transaction = CreateObject('perp_ledger.GLTransaction');
			$transaction->set_transaction_type($trans_type);
			$transaction->set_date($allocation_date);

			$gl_currency_total = round($payment_total,2) * round($exchange,2);

			// Bank account line
			$bank_trans = $transaction->add_line($bank_account, $payment_narrative, -$gl_currency_total);
			$bank_trans = $bank_trans->get_associated_line(0);
			$bank_trans->set_payment_method($payment_method);
			$bank_trans->set_currency($bank['currency']);

			$check_total = 0;
			$supplier_ledger = 0;
			$difference_on_exchange = 0;

			foreach($payment_information['rows'] as $key => $row) {
				if(is_numeric($key) && is_array($row)) {
					if($row['invoice_id']) {
						$invoice =& new PayableInvoice($row['invoice_id']);
					}
					if((is_null($row['discount_taken']) || $row['discount_taken'] === '') &&
					(is_null($row['payment_amount']) || $row['payment_amount'] === '')) {
							// User didn't actually enter anything here, skip it.
						continue;
					}

					if($row['discount_taken']) {

						// Discount line
						$discount_amount = round($row['discount_taken'] * $exchange, 2);

						$transaction->add_line($discount_account, $discount_narrative, -$discount_amount);
						$ap_line = $transaction->add_line($ap_account, $discount_narrative, $discount_amount);
						$supplier_trans = CreateObject('perp_ap.SupplierTransactionLine', $ap_line);
						$supplier_trans->set_account($ap_account);
						$supplier_trans->set_amount($discount_amount);
						$supplier_trans->set_narrative($discount_narrative);
						$supplier_trans->set_supplier($supplier_id);
						$supplier_trans->set_settled(true);
						$supplier_trans->set_exchange_rate($exchange);
						$supplier_trans->set_link('perp_ap', 'INVOICE-' . $row['invoice_id']);
					}

					if($invoice && $invoice->get_transaction()) {
						$trans_id = $invoice->get_transaction();
						$invoice_lines =& $invoice->get_lines();
						foreach($invoice_lines as &$line) {
							$line_no = $line->get_transaction_line();	
							$inv_trans = CreateObject('perp_ledger.GLTransaction', $trans_id, $line_no);
							$invoiced_trans_line = CreateObject('perp_ap.SupplierTransactionLine', $inv_trans, $line_no);
							if(!$invoiced_trans_line || !$invoiced_trans_line->get_exchange_rate()) {
								continue;
							}
							
							if($this->debug) {
								echo 'Difference on exchange. ' . $invoiced_trans_line->get_exchange_rate() . ' -> ' . $exchange;
							}
							if($exchange && $invoiced_trans_line->get_exchange_rate() != $exchange && $diff_on_exchange_account) {
							        $difference_on_exchange += round(($invoiced_trans_line->get_exchange_rate() - $exchange) * ($row['payment_amount'] + $row['discount_taken']), 2);
							}
						}
					}
					$payments[] = array(
						'invoice_id'	=>	$row['invoice_id'],
						'payment_amount'	=>	$row['payment_amount'],
						'allocated_from'	=>	$row['payment_id']
					);
					$check_total = round($check_total, 2) + round($row['payment_amount'], 2);
				}
			}

			// Crazy comparison due to floating point imprecision
			if(($check_total + $pre_payment - $payment_total) > 0.001) {
				return lang("Lines don't add up to total (%1 + %2 != %3)", $check_total, $pre_payment, $payment_total);
			}

			if($difference_on_exchange != 0) {
				$transaction->add_line($diff_on_exchange_account, $invoice->get_currency() . ' ' . $invoiced_trans_line->get_exchange_rate() . ' -> ' . $exchange, -$difference_on_exchange);
			}

			// AP line
			$ap_line = $transaction->add_line($ap_account, $payment_narrative, $gl_currency_total + $difference_on_exchange);
			$supplier_trans = CreateObject('perp_ap.SupplierTransactionLine', $ap_line);
			$supplier_trans->set_account($ap_account);
			$supplier_trans->set_amount($gl_currency_total);
			$supplier_trans->set_narrative(lang('Payment %1', $payment_id));
			$supplier_trans->set_supplier($supplier_id);
			$supplier_trans->set_settled(true);
			$supplier_trans->set_exchange_rate($exchange);
			$supplier_trans->set_link('perp_ap', 'PAYMENT-' . $payment_id);

			if($this->debug) {
				echo '<pre>';
				echo $transaction;
				echo '</pre>';
			}

			if($GLOBALS['egw']->db->Link_ID->transCnt == 0) {
				/*
				* Database doesn't support transactions.
				* Reconnect to do transactions, and shove it in GLOBALS[egw]->db so
				* everywhere uses the same connection
				*/
				$old_db =& $GLOBALS['egw']->db;
				$GLOBALS['egw']->db =& ExecMethod('perp_api.bo_perp_api.reconnect_transactional');
				$GLOBALS['egw']->db->transaction_begin();
			}

			$trans_id = $transaction->write();

			// Update Invoices
			foreach($payment_information['rows'] as $key => $row) {
				if(is_numeric($key) && is_array($row)) {
					if($row['invoice_id']) {
						// Paying an Invoice
						$invoice =& new PayableInvoice($row['invoice_id']);
						$invoice->set_bank_account($bank_account);
						$invoice->set_payment_method($payment_method);
						
						// Difference on exchange
						if($invoice->get_exchange_rate() != $exchange) {

						}

						$invoice->take_discount($row['discount_taken']);

						$invoice->set_paid(($row['discount_taken'] + $row['payment_amount'] == $row['amount_owing']));

						$invoice->write();
					}
				}
			}

			foreach($payments as $payment) {
				// Allocate payments
				$result = $this->allocate_payment($payment_id, $payment_number, $trans_id, $payment['invoice_id'], $payment['payment_amount'], $allocation_date, $payment['allocated_from']);
				if($result === false) {
					// Error, rollback transaction
					$GLOBALS['egw']->db->transaction_abort();
					$payment_id = 0;
					break;
				}
			}

			// Pre-payment
			if($pre_payment != 0 && !$this->allocate_payment($payment_id, $payment_number, $trans_id, null, $pre_payment, null, null)) {
				$GLOBALS['egw']->db->transaction_abort();
				$payment_id = 0;
			}
				
			$GLOBALS['egw']->db->transaction_commit();
			if($old_db) {
				// Swap the original database back, so it doesn't cause problems elsewhere
				$GLOBALS['egw']->db = $old_db;
			}

			return $payment_id;
		}

		/**
		*	Allocate a payment to an invoice
		*
		*	@param payment_id	int	ID of the payment to allocate
		*	@param payment_number	int	Reference number of the payment (for humans)
		*	@param trans_id		int	General Ledger transaction number for the payment
		*	@param invoice_id	int	ID of the invoice to allocate the payment to
		*	@param amount		float	Amount of the payment to allocate to the invoice.  Doesn't have to be all of it.
		*	@param date			Date of the allocation (usually transaction date)
		*	@param pre_payment	float	If there is a prepayment, the amount of the prepayment
		*/
		public function allocate_payment($payment_id, $payment_number, $trans_id, $invoice_id, $amount, $date, $pre_payment) {
			if(is_null($pre_payment)) {
				$pre_payment = 0;
			}
			if($GLOBALS['egw']->acl->check('payment', EGW_ACL_ADD, 'perp_ap') && $amount != 0) {
				return $this->so->allocate_payment($payment_id, $payment_number, $trans_id, $invoice_id, $amount, $date, $pre_payment);
			} else {
				return false;
			}
		}

		/**
		* get title for a perp payment entry identified by $entry
		*
		* Is called as hook to participate in the link
		*
		* @param entry int/array key to look for or array with payment entry
		* @return string the title
		*/
		public function link_title($entry) {
			if(is_array($entry)) {
				return 'Payment ' . $entry['payment_id'];
			} elseif(is_numeric($entry)) {
				return 'Payment ' . $entry;
			}
			return false;
		}

		/**
		*	Undo a payment
		*
		*	The payment is still recorded, and all the ledger entries remain, but a reversing transaction is
		*	done to reverse the effects.
		*
		*	@param payment_id The ID number of the payment to reverse
		*	@return int Reversing payment ID
		*	@throws egw_exception_no_permission if the user does not have access to reverse payments
		*	@throws egw_exception_wrong_userinput if you try to reverse a payment that has cleared
		*/
		public function reverse($payment_id, $reason) {
			if(!$GLOBALS['egw']->acl->check('payment', EGW_ACL_DELETE, 'perp_ap')) {
				throw new egw_exception_no_permission(lang('You do not have permission to reverse payments'));
			}
			$payment = $this->get_payment($payment_id);

			if($payment['status'] != 1) {
				throw new egw_exception_wrong_userinput(lang('Payment %1 cannot be reversed', $payment_id));
			}
			
			if(trim($reason) == '') {
				throw new egw_exception_wrong_userinput(lang('You must provide a reason for the reversal'));
			}

			return $this->so->reverse($payment, $reason);
		}

		/**
		*	Read a reversal record
		*
		*	@param payment_id Either the original payment, or the reversal
		*	@return Array of reversal information
		*/
		public static function get_reversal($payment_id) {
			if(!$GLOBALS['egw']->acl->check('payment', EGW_ACL_READ, 'perp_ap')) {
				return array();
			}
			return so_perp_payment::get_reversal($payment_id);
		}
	}
?>
