<?php
	/**
	*	Logic layer for Purchase Orders
	*	@author Nathan Gray
	*	@ingroup perp_ap
	*/

	require_once(EGW_INCLUDE_ROOT . '/perp_api/inc/class.bo_perp_order.inc.php');
	require_once('class.so_perp_purchase_order.inc.php');

	class bo_perp_purchase_order extends bo_perp_order {

		/**
		*	Constructor
		*/
		public function __construct() {
			parent::__construct();
			$this->so_perp_order = new so_perp_purchase_order();
		}

		/**
		*	Get a list of purchase orders
		*	@nextmatch
		*	Query options:
		*		search:		purchase order number or requisition number
		*		supplier_id:	Restrict to supplier
		*		filter:		Restrict to status
		*/
		public function get_order_list(&$query = array(), &$rows = array(), &$readonlys = array()) {
			$query['client_id'] = 0;
			$query['order_class'] = 'PurchaseOrder';

			// Check if the current user is a supplier
			$supplier_ids = ExecMethod('perp_ap.bo_perp_supplier.get_supplier_by_user_account');
			if(is_array($supplier_ids) && count($supplier_ids) > 0 && !in_array($query['supplier_id'], $supplier_ids)) {
				$query['supplier_id'] = $supplier_ids;
			}

			return parent::get_order_list($query, $rows, $readonlys);
		}

		/**
		*	Get an open PurchaseOrder for a given supplier.
		*	In this case, open means not approved, not at all shipped, not at all invoiced.
		*
		*	@param supplier_id int	ID of the supplier to look for an open purchase order
		*	@return int purchase order ID or null if none found
		*/
		public function get_open_order($supplier_id) {

			$query['supplier_id'] = $supplier_id;

			// Check if the current user is a supplier
			$supplier_ids = ExecMethod('perp_ap.bo_perp_supplier.get_supplier_by_user_account');
			if(is_array($supplier_ids) && count($supplier_ids) > 0 && !in_array($query['supplier_id'], $supplier_ids)) {
				$query['supplier_id'] = $supplier_ids;
			}


			// Check config, if approval is required, grab an un-approved order.
			// If approval is not required, grab an un-shipped order.
			$config = config::read('perp_ap');
			if($config['orders_require_approval']) {
				$query['status'] = 'request';
				$this->get_order_list($query, $list);
				if($list) {
					return $list[0]['order_id'];
				}
			} else {
				// Need to check that order is not shipped or invoiced
				return $this->so_perp_order->get_addable_order($query);
			}
			return null;
		}

		/**
		*	Get a list of possible statuses for PurchaseOrders
		*	@return array of statuses
		*/
		public function get_status_list() {
			return array(
				''	=>	'All',
				'request'	=>	'Requests',
				'approved'	=>	'Approved',
				'shipped'	=>	'Shipped',
				'not_shipped'	=>	'Not shipped',
				'received'	=>	'Received',
				'not_received'	=>	'Not Received',
				'invoiced'	=>	'Invoiced',
				'not_invoiced'	=>	'Not invoiced',
				'closed'	=>	'Closed'
			);
		}

		/**
		*	Get a list of backordered PurchaseOrders
		*	@nextmatch
		*/
		public function get_backorders($query = array(), &$rows, &$readonlys) {
			// Check if the current user is a supplier
			$supplier_ids = ExecMethod('perp_ap.bo_perp_supplier.get_supplier_by_user_account');
			if(is_array($supplier_ids) && count($supplier_ids) > 0 && !in_array($query['supplier_id'], $supplier_ids)) {
				$query['supplier_id'] = $supplier_ids;
			}

			$read = $GLOBALS['egw']->acl->check('order', EGW_ACL_READ, 'perp_ap');
			if(!$read) {
				$rows = array();
				return count($query) ? 0 : array();
			}

			if(!$query['start']) {
				$query['start'] = 0;
			}
			if($query['filter']) {
				$query['supplier_id'] = $query['filter'];
			}

			$list = $this->so_perp_order->get_backorders($query);
			$count = $query['count'];

			if(!count($query)) {
				$pos = array();
				foreach($list as $row) {
					$pos[$row['purchase_order_id']] = $row['purchase_order_number'];
				}
				return $pos;
			}

			$rows = $list;
			return $count;
		}

		/**
		*	Get the default amount of the given stock item to order for this location
		*	@todo This should be in perp_inventory, not here
		*	@param location	int	ID of the location 
		*	@param stock_id	int	ID of the stock item to order
		*	@return float Quantity to order according to location settings
		*/
		public function get_order_quantity($location, $stock_id) {
			return $this->so_perp_order->get_order_quantity($location, $stock_id);
		}

		/** 
		*	Get a list of purchase orders that have been
		*	approved but have not been added to a shipment yet
		*
		*	@param	owner	int	Optional.  User id of the purchase order's creator
		*/
		public function get_unshipped($owner = '') {
			if($GLOBALS['egw']->acl->check('order', EGW_ACL_READ, 'perp_ap')) {
				return $this->so_perp_order->get_unshipped($owner);
			} else {
				return false;
			}
		}

		/**
		* get title for an perp purchase order entry identified by $entry
		*                               
		* Is called as hook to participate in the linking                
		*
		* @param entry int/array key to look up or array with purchase order entry
		* @return string  with title/boolean
		*/

		public function link_title($entry) {
			if(is_array($entry)) {
				return $entry['purchase_order_number'];
			} elseif(is_numeric($entry)) {
				$order = new PurchaseOrder($entry);
				$title = $order->get_our_reference_number();
				if($order->get_their_reference_number()) {
					$title .= '(' . $order->get_their_reference_number() . ')';
				}
				return $title;
			}
			return false;
		}

		/**
		*	Participate in linking
		*	
		*	@pattern is a pattern to search for in the Purchase Orders
		*	
		*	@return array of matching entries, with keys like ORDER-<id>
		*/
		public function link_query($pattern) {
			$query = array(
				'search'	=>	$pattern,
				'num_rows'	=>	-1
			);
			$list = array();
			// Need this swap for permission check to pass
			$old_app = $GLOBALS['egw_info']['flags']['currentapp'];
			$GLOBALS['egw_info']['flags']['currentapp'] = 'perp_ap';

			$this->get_order_list($query, $list);

			$GLOBALS['egw_info']['flags']['currentapp'] = $old_app;

			$result = array();
			foreach($list as $order) {
				$result['ORDER-' . $order['order_id']] = $this->link_title($order['order_id']);
			}
			return $result;
		}

		/**
		*	Get a list of purchase order lines that have been at least partially received
		*	where the quantity shipped does not match the quantity ordered.
		*
		*	@nextmatch
		*
		*	Extra parameters:
		*	- supplier_id
		*	- shipment_id
		*	- order_id
		*/
		public function get_under_over_list(&$query = array(), &$rows = array(), &$readonly = array()) {
			if(!$GLOBALS['egw']->acl->check('order', EGW_ACL_READ, 'perp_ap')) {
				$rows = array();
				return 0;
			}

			$rows = $this->so_perp_order->get_under_over_list($query);

			return $query['total'];
		}
	}
?>
