<?php
	/** 
	*	Logic object for receiving shipments
	*
	*	@author Nathan Gray
	*	@ingroup perp_ap
	*/

	require_once('class.so_perp_receive.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_inventory/inc/class.bo_perp_stock.inc.php');

	class bo_perp_receive {

		/**
		*	boolean Enable for debug messages
		*/
		private $debug = False;

		/**
		*	Data access object
		*/
		private $so_receive;

		public function __construct() {
			$this->so_receive =& new so_perp_receive();
		}

		/**
		*	Get a list of receivable goods (shipments not recieved)
		*
		*	@nextmatch
		*	Query options:
		*		- supplier_id	-	Restrict to one supplier
		*		- shipment_id	-	Restrict to one shipment
		*		- location_id	-	Restrict to one location, set to filter
		*	
		*/
		public function get_receivable_list($query = array(), &$rows = array(), &$readonly = array()) {
			if(!$GLOBALS['egw']->acl->check('shipment', EGW_ACL_READ, 'perp_ap')) {
				$rows = array();
				return 0;
			}

			if(!isset($query['order'])) {
				$query['order'] = 'perp_shipment.requested_delivery_date';
			}
			if(isset($query['filter'])) {
				$query['location_id'] = $query['filter'];
			}

			$rows = $this->so_receive->get_receivable_list($query);

			foreach($rows as $key => &$row) {
				$readonly[$key + 1 . '[serials]'] = !$row['controlled'];
			}

			return $query['count'];
		}

		/**
		*	Add an item to a shipment, and receive that item.
		*
		*	@return boolean
		*/
		public function receive_item(&$shipment, $content) {
			if(!($shipment instanceof IncomingShipment)) {
				if(is_numeric($shipment)) {
					$shipment = new IncomingShipment($shipment);
				} else {
					throw new egw_exception(lang('Missing shipment to receive'));
				}
			}

			if(!$content['received_by']) {
				$content['received_by'] = $GLOBALS['egw_info']['user']['account_id'];
			}
			if(!is_array($content['serial_numbers'])) {
				$content['serial_numbers'] = array();
			}

			// Check that required serial numbers were provided
			$catalog_item = ExecMethod2('perp_ap.bo_perp_catalog.get_catalog_item', $content['supplier_id'], $content['stock_code']);
			if($catalog_item['stock_id']) {
				$stock_item = ExecMethod('perp_inventory.bo_perp_stock.get_stock_item', $catalog_item['stock_id']);
				// Require serial numbers
				if($stock_item['serial_setup'] == bo_perp_stock::SERIAL_ALWAYS) {
					$found = 0;
					foreach($content['serial_numbers'] as $serial => $quantity) {
						$found += $quantity;
					}
					if($found != $content['quantity_received']) {
						throw new egw_exception(lang('Missing required serial / batch numbers'));
					}
				}
			}
			
			// Build an index of shipment lines by supplier stock code
			$line_index = array();
			$lines =& $shipment->get_lines();
			foreach($lines as &$line) {
				if($line->get_shippable_item() != null) {
					$line_index[$line->get_shippable_item()->get_supplier_stock_code()][] = $line->get_line_number();
				} else {
					$line_index[''][] = $line->get_line_number();
				}
			}

			/**
			*	Check to make sure the item is on the order, and find the order line
			*/
			$order = new PurchaseOrder($content['order_id']);
			if($order->get_supplier_id() != $shipment->get_supplier_id()) {
				throw new egw_exception(lang('Order %1 is not for the right supplier', $order->get_our_reference_number()));
			}
			$lines = $order->get_lines();
			$matching_lines = array();
			foreach($lines as &$line) {
				if($line->get_supplier_stock_code() == $content['stock_code']) {
					$matching_lines[] =& $line;
				}
			}

			// Chose the line to ship & receive
			$order_line = null;
			if(count($matching_lines) == 1) {
				$order_line =& $matching_lines[0];
			} else {
				// Go through all lines for that item
				foreach($matching_lines as &$line) {
					if($line->get_quantity_ordered() < $line->get_quantity_received()) {
						// Found a matching line, use it
						$order_line =& $line;
						break;
					}
				}
			}

			// Receipt Line
			$receipt = null;

			if($order_line != null) {
				/**
				*	First, check for an existing line already on the shipment that is not
				*	completely received.  It could be the shipment was set up in advance.
				*/
				if(isset($line_index[$content['stock_code']])) {
					foreach($line_index[$content['stock_code']] as $line_no) {
						$_order_line =& $shipment->get_line($line_no)->get_shippable_item();

						if($_order_line->get_owner()->get_order_id() == $content['order_id'] && (
								// We found more of the same item, already on shipment
								$order_line->get_line_number() == $_order_line->get_line_number() ||

								// Found a line for this item on the order
								$shipment->get_line($line_no)->get_quantity_received() !=
								$shipment->get_line($line_no)->get_quantity_shipped())) {
							if($shipment->get_line($line_no)->get_quantity_received() + $content['quantity_received'] > $shipment->get_line($line_no)->get_quantity_shipped()) {
								// About to receive more than was shipped... adjust shipment quantity by the difference
								$shipment->get_line($line_no)->set_quantity_shipped($shipment->get_line($line_no)->get_quantity_received() + $content['quantity_received']);
								$shipment->write();

							} elseif ($content['quantity_received'] < 0) {
								// Returning or typo receiving, take some out
								$shipment->get_line($line_no)->set_quantity_shipped($shipment->get_line($line_no)->get_quantity_received() - $content['quantity_received']);
								$shipment->write();
							}
							
							$receipt = new ReceiveLine($shipment->get_line($line_no), $content['received_by'], $content['quantity_received'], $content['serial_numbers'], $content['received_date']); 
						}
					}
				} else {

					/** 
					*	Item is not on the shipment, so we'll add the line, and receive it
					*/

					// Add line to shipment
					$new_line = new ShipmentLine($shipment, $order_line);
					$new_line->set_quantity_shipped($content['quantity_received']);
					$shipment->add_line($new_line);
					$shipment->write();

					// Add line to line index for future reference
					$line_index[$content['stock_code']] = $new_line->get_line_number();

					// Receive the line
					$receipt = new ReceiveLine($new_line, $content['received_by'], $content['quantity_received'], $content['serial_numbers'], $content['received_date']);
				}
			}

			if($receipt != null) {
				$receipt->receive();

				// Check if the PO is fully received
				$order =& $receipt->get_shipment_line()->get_shippable_item()->get_owner();
				if($order->is_complete()) {
					$order->set_closed_date(time());
					$order->write();
				}
				return true;
			}

			// Not on this order or shipment
			return false;
		}

		/**
		*	If an item received cannot be received because it is not on the specified purchase order,
		*	find a list of other purchase orders it could be on.
		*
		*	@param supplier_id
		*	@param supplier_stock_code
		*/
		public function find_other_orders($supplier_id, $supplier_stock_code) {
			$query = array(
				'supplier_id'		=>	$supplier_id,
				'status'		=>	'not_received',
				'supplier_stock_code'	=>	$supplier_stock_code
			);
			$other_orders = array();
			$bo_po = new bo_perp_purchase_order();
			$bo_po->get_order_list($query, $other_orders);

			return $other_orders;
		}

		/**
		*	Tried to receive an item that was not ordered.
		*	
		*	Deal with it.
		*
		*	@todo Still need to handle creating a new order for this item
		*
		*	@param shipment
		*	@param details - Same as what are required for receive_item()
		*	@param action - Action to take
		*	@param target - Order ID to use instead
		*/
		public function not_ordered(&$shipment, $details, $action, $target) {
			switch ($action) {
				case 'cancel':
					// We've already deferred, the decision is to forget about it
					return $this->delete_not_ordered($target);
					break;
				case 'defer':
					$details['shipment_id'] = $shipment->get_shipment_id();
					$id = $this->set_not_ordered(null, $details);
					return ($id && is_numeric($id)) ? true : $id;
					break;
				case 'other_order':
					$details['order_id'] = $target;
					return $this->receive_item($shipment, $details);
					break;
				case 'add_to_order':
					$order = new PurchaseOrder($target);
					if($order->get_supplier_id() == $shipment->get_supplier_id() && $GLOBALS['egw']->acl->check('order', EGW_ACL_EDIT, 'perp_ap')) {
						// Add line to PO.  Constructor will find the missing data from the catalog
						$order->add_line(new PurchaseOrderLine($order, array(
							'supplier_stock_code'	=>	$details['stock_code'],
							'quantity_ordered'	=>	$details['quantity_received'],
						)));
						$order->write();
						$details['order_id'] = $target;
						return $this->receive_item($shipment, $details);
					} else {
						return false;
					}
					break;
				case 'new_order':
					break;
			}
		}

		/**
		*	Get a list of items received but not ordered so something can be done about them.
		*
		*	@nextmatch
		*/
		public function get_not_ordered_list(&$query, &$rows = array(), &$readonlys = array()) {
			if(!$GLOBALS['egw']->acl->check('receive', EGW_ACL_READ, 'perp_ap')) {
				$rows = array();
				return 0;
			}

			if(!$query['order']) {
				$query['order'] = 'id';
			}

			$rows = $this->so_receive->get_not_ordered_list($query);

			if(!$GLOBALS['egw']->acl->check('receive', EGW_ACL_CUSTOM_1, 'perp_ap')) {
				foreach($rows as $key => $row) {
					$readonlys[$key] = array(
						'action'	=>	true
					);
				}
			}
			return $query['total'];
		}

		/**
		*	Tried to receive something that wasn't ordered, but user
		*	doesn't have the knowledge or authority to know what to do
		*	with it, so we'll record it and receive it later, when someone
		*	else has decided what to do with it.
		*
		*	@param id ID of the record being edited
		*	@param details
		*/
		public function set_not_ordered($id, $details) {
			if($GLOBALS['egw']->acl->check('receive', EGW_ACL_EDIT, 'perp_ap')) {
				if(!($details['stock_code'] || $details['supplier_stock_code']) || !$details['quantity_received']) {
					return lang('Missing required parameter: %1', (!$details['stock_code'] ? 'stock_code' : 'quantity'));
				}
				if(!$details['shipment_id']) {
					return lang('Shipment is not set yet');
				}
				if(!$details['supplier_stock_code'] && $details['supplier_id']) {
					$catalog = ExecMethod2('perp_ap.bo_perp_catalog.get_catalog_item', $details['supplier_id'], $details['stock_code']);
					if($catalog) {
						$details['supplier_stock_code'] = $catalog['supplier_code'];
						$details['stock_id']		= $catalog['stock_id'];
						if(!$details['unit_price']) {
							$details['unit_price']		= round($catalog['price'] / ($catalog['quantity'] * $catalog['conversion_factor']), 2);
						}
						if(!$details['description']) {
							$details['description'] = $catalog['description'];
						}
					}
				}
				if(!$details['received_by']) {
					$details['received_by'] = $GLOBALS['egw_info']['user']['account_id'];
				}
				if(!$details['received_date']) {
					$details['received_date'] = time();
				}
				return $this->so_receive->set_not_ordered($id, $details);
			}
		}

		/**
		*	Get an item that was received but not ordered
		*
		*	@param ID
		*/
		public function get_not_ordered($id) {
			if($GLOBALS['egw']->acl->check('receive', EGW_ACL_READ, 'perp_ap')) {
				return $this->so_receive->get_not_ordered($id);
			} else {
				return array();
			}
		}

		/**
		*	Delete a record of something received but not ordered.  This is 
		*	done when either the original order is changed, or the 
		*	item is rejected.
		*
		*	@param id ID of the record to be removed
		*/
		public function delete_not_ordered($id) {
			if($GLOBALS['egw']->acl->check('receive', EGW_ACL_CUSTOM_1, 'perp_ap')) {
				return $this->so_receive->delete_not_ordered($id);
			} else {
				return false;
			}
		}
		
		/**
		*	Get the data for Goods Received Report
		*
		*	
		*/
		public function get_data_goods_received_report($start_date, $end_date, &$goods_received) {
			
			$goods_received = $this->so_receive->get_data_goods_received_report($start_date, $end_date);
			
		}
	}
?>
