<?php
	/**
	*	Data storage layer object for dealing with suppliers
	*
	*	@author Nathan Gray
	*	@ingroup perp_ap
	*/
	class so_perp_supplier {
		/**
		*	Database access object
		*/
		private $db;
	
		/**
		*	Constructor
		*/
		public function so_perp_supplier() {
			$this->db = clone($GLOBALS['egw']->db);
			$this->db->set_app('perp_ap');
		}

		/**
		*	Get list of suppliers according to query parameters
		*	@nextmatch
		*/
		public function get_suppliers(&$query = array()) {
			$where = array();
			if($query['search']) {
				$where[] = 'supplier_num LIKE \'%' . addslashes($query['search']) . '%\' OR ' .
					'supplier_name LIKE \'%' . addslashes($query['search']) . '%\'';
			}
			if($query['supplier_id']) {
				$where['supplier_id'] = $query['supplier_id'];
			}
			if($query['searchletter']) {
				$where[] = '(supplier_num LIKE \'' . $query['searchletter'] . '%\' OR 
					supplier_name LIKE \'' . $query['searchletter'] . '%\')';
			}
			if(isset($query['filter']) && $query['filter'] != '') {
				$where['supplier_status'] = $query['filter'];
			}

			// Get number of rows
			$this->db->select('perp_supplier',
				array('count(supplier_id) AS count'),
				$where,
				__LINE__, __FILE__
			);
			if($row = $this->db->row(true)) {
				$query['count'] = $row['count'];
			}

			$this->db->select('perp_supplier', 
				array('*'),
				$where,
				__LINE__, __FILE__, $query['start'], 
				"ORDER BY " . $query['order'] . ' ' . $query['sort'], 'perp_ap', $query['num_rows']
			);
			$list = array();
			while($row = $this->db->row(True)) {
				$list[] = $row;
			}
			return $list;
		}

		/**
		*	Get an array of supplier information
		*	@param supplier_id	int	ID of the supplier to read
		*	@return array of supplier properties
		*/
		public function get_supplier($supplier_id) {
			$this->db->select('perp_supplier', 
				array('*'),
				array('supplier_id' => $supplier_id),
				__LINE__, __FILE__);
			$list = array();
			if($row = $this->db->row(True)) {
				$row['supplier_since'] = $GLOBALS['egw']->db->from_timestamp($row['supplier_since']);
				return $row;
			} 
			return $list;
		}

		/**
		*	Set the properties of a supplier
		*	@param supplier_id	int	ID of supplier to set, or null for a new supplier
		*	@param supplier		array	Array of Supplier properties to set
		*/
		public function set_supplier($supplier_id, $supplier) {
			// Check for duplicate supplier number
			$where['supplier_num'] = $supplier['supplier_num'];
			if($supplier_id) {
				$where[] = 'supplier_id != ' . $supplier_id;
			}
			$this->db->select('perp_supplier', 
				array('supplier_name'),
				$where,
				__LINE__, __FILE__
			);
			if($this->db->num_rows()) {
				return lang('Duplicate Supplier Number');
			}

			$this->db->insert('perp_supplier',
				array('supplier_id' => $supplier_id,
					'supplier_num'		=>	$supplier['supplier_num'],
					'supplier_name'		=>	$supplier['supplier_name'],
					'supplier_address'	=>	$supplier['supplier_address'],
					'supplier_currency'	=>	$supplier['supplier_currency'],
					'supplier_since'	=>	$GLOBALS['egw']->db->to_timestamp($supplier['supplier_since']),
					'payment_terms'		=>	$supplier['payment_terms'],
					'credit_limit'		=>	$supplier['credit_limit'],
					'bank_account'		=>	$supplier['bank_account'],
					'bank_reference'	=>	$supplier['bank_reference'],
					'bank_particulars'	=>	$supplier['bank_particulars'],
					'remittance'		=>	$supplier['remittance'],
					'tax_group'		=>	$supplier['tax_group'],
					'user_account'		=>	$supplier['user_account'],
					'supplier_status'		=>	$supplier['supplier_status'],
				),
				array('supplier_id' => $supplier_id),
				__LINE__, __FILE__, 'perp_ap'
			);
			$supplier_id = $supplier_id ? $supplier_id : $this->db->get_last_insert_id('perp_supplier', 'supplier_id');
			return $supplier_id;
		}

		/**
		*	Delete a supplier, their catalog, and any Supplier Ledger entries
		*	@param supplier_id	int	ID of Supplier to delete
		*/
		public function delete_supplier($supplier_id) {
			$this->db->delete('perp_supplier', array('supplier_id' => $supplier_id), __LINE__, __FILE__);
			$this->db->delete('perp_supplier_ledger', array('supplier_id' => $supplier_id), __LINE__, __FILE__);
			$this->db->delete('perp_supplier_stock', array('supplier_id' => $supplier_id), __LINE__, __FILE__);
		}

		/**
		*	Function to participate in eGroupWare linking system
		*	Returns a list of Suppliers that match the given pattern. 
		*	Searches supplier_number and supplier_name.  
		*	@param pattern 	string 	Pattern to search for
		*	@return array of matches
		*/
		public function link_query($pattern) {
			$query_str = 'SELECT supplier_id, supplier_num, supplier_name
				FROM perp_supplier
				WHERE supplier_num LIKE '.$this->db->quote('%'.$pattern.'%').' OR
					supplier_name LIKE '.$this->db->quote('%'.$pattern.'%');

			$this->db->query($query_str);
			$list = array();
			while($row = $this->db->row(True)) {
				$list[] = $row;
			}
			return $list;
		}

		/**
		*	Get a supplier's current account balance.
		*
		*	Actually, our account balance with the given supplier
		*
		*	@param supplier_id
		*/
		public function get_account_balance($supplier_id) {
			$join = 'JOIN perp_general_ledger ON 
			        	perp_general_ledger.trans_no = perp_supplier_ledger.trans_id AND
					perp_general_ledger.line_no = perp_supplier_ledger.line_no
				LEFT JOIN perp_bank_ledger ON
			        	perp_bank_ledger.trans_id = perp_general_ledger.trans_no
					AND perp_general_ledger.line_no = perp_general_ledger.line_no
				JOIN perp_account ON
					perp_general_ledger.account = perp_account.account_id
				JOIN perp_account_group ON
					perp_account.account_group = perp_account_group.group_id';

			// Get balance
			$balance = $GLOBALS['egw']->db->select(
				'perp_supplier_ledger',
				array(
					'sum(round(perp_supplier_ledger.amount / perp_supplier_ledger.exchange_rate * (abs(asset_or_liability) / asset_or_liability) * -1, 2)) AS balance',
				),
				array(
					'supplier_id'	=> $supplier_id
				),
				__LINE__, __FILE__, 0,
				'', 'perp_ap', 1,
				$join
			)->fetchSingle();

			return $balance;
		}

		/**
		*	Get a list of entries in the Supplier Ledger for the given supplier
		*	@param supplier_id	int	ID of supplier to search for
		*	@param start		int	Skip a certain number of records
		*	@return Array of information about matching entries.  Doesn't return any TransactionLine objects.
		*/
		public function get_supplier_ledger(&$query) {
			
			$where['supplier_id'] = $query['supplier_id'];

			$join = 'JOIN perp_general_ledger ON 
			        	perp_general_ledger.trans_no = perp_supplier_ledger.trans_id AND
					perp_general_ledger.line_no = perp_supplier_ledger.line_no
				LEFT JOIN perp_bank_ledger ON
			        	perp_bank_ledger.trans_id = perp_general_ledger.trans_no
					AND perp_bank_ledger.line_no = perp_bank_ledger.line_no
				JOIN perp_account ON
					perp_general_ledger.account = perp_account.account_id
				JOIN perp_account_group ON
					perp_account.account_group = perp_account_group.group_id';

			// Get number of rows and balance
			$row = $GLOBALS['egw']->db->select(
				'perp_supplier_ledger',
				array(
					'count(perp_supplier_ledger.trans_id) AS count',
					'sum(round(perp_supplier_ledger.amount / perp_supplier_ledger.exchange_rate * (abs(asset_or_liability) / asset_or_liability) * -1, 2)) AS balance',
				),
				$where,
				__LINE__, __FILE__, 0,
				'', 'perp_ap', 1,
				$join
			)->fetch();
			$query['count'] = $row['count'];
			$query['balance'] = $row['balance'];

			$sql = 'SELECT DISTINCT 
				perp_supplier_ledger.trans_id,
				perp_supplier_ledger.supplier_id,
				perp_general_ledger.trans_date,
				perp_supplier_ledger.exchange_rate,
				round(perp_supplier_ledger.amount / perp_supplier_ledger.exchange_rate * (abs(asset_or_liability) / asset_or_liability) * -1, 2) AS amount,
				perp_supplier_ledger.link,
				perp_supplier_ledger.supplier_reference,
				perp_supplier_ledger.comment,
				perp_general_ledger.narrative,
				perp_bank_ledger.currency
			FROM perp_supplier_ledger ' .
			$join . '
			WHERE ' . $GLOBALS['egw']->db->column_data_implode(' AND ', $where) . ' 
			ORDER BY ' . $query['order'] . ' ' . $query['sort'];

			$this->db->query($sql, __LINE__, __FILE__, $query['start'], $query['num_rows']);
			$list = array();
			while($row = $this->db->row(True)){
				if($row['amount'] > 0 ) {
					$row['credit'] = $row['amount'];
					$row['formatted_credit'] = ExecMethod2('perp_api.bo_perp_api.get_formatted_price', $row['currency'], $row['credit']);
				} else {
					$row['debit'] = -$row['amount'];
					$row['formatted_debit'] = ExecMethod2('perp_api.bo_perp_api.get_formatted_price', $row['currency'], $row['debit']);
				}
				$link = explode(':', $row['link']);
				$app = $link[0];
				unset($link[0]);
				$id = implode(':', $link);
				$title = egw_link::title($app, $id);
				$row['reference'] = array(
					'app'	=>	$app,
					'id'	=>	$id,
					'title'	=>	$title
				);
				$list[] = $row;
			}
			return $list;
		}

		/**
		*	Get a list of supplier IDs associated with the given user account / groups
		*/
		public function get_supplier_by_user_account($account_id) {
			$GLOBALS['egw']->db->select(
				'perp_supplier',
				array('supplier_id'),
				array('user_account'	=>	$account_id),
				__LINE__, __FILE__, false,
				'', 'perp_ap'
			);
			$supplier_ids = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$supplier_ids[] = $row['supplier_id'];
			}
			return $supplier_ids;
		}
	}
?>
