<?php
	/**
	*	In order for another eGroupWare record (pERP included) to be invoiced, there must
	*	be a class that implements InvoicableItem for that record.  The constructor must be
	*	able to take an array with the keys 'invoicable_id' and 'line_number', which 
	*	correspond to the invoicable_id and invoicable_line_id.  Constructor overloading
	*	is fine, but the array must be implemented.
	*
	*	To get your InvoicableItem onto a new Invoice, it should be done via the Invoicable object.
	*	To add your InvoicableItem to an in-progress Invoice with the 'Add Line From... button', there
	*	is a hook that must be implemented: perp_ar:invoicable.
	*
	*	The hook function must return an array, of the form:
	*	array(
	*		- @c 'search'        =>      'perp_ar.ProjectManagerInvoicable.search',
	*		- @c 'config'        =>      'perp_ar.ProjectManagerInvoicable.config',
	*		- @c 'Invoicable'    =>      'perp_ar.ProjectManagerInvoicable',
	*		- @c 'InvoicableItem'        =>      'perp_ar.ProjectManagerInvoicableItem'
	*	);
	*	where:
	*		- search returns nothing, but populates an array of InvoicableItems that can be added,
	*			in the standard nextmatch style.  
	*			It will be passed a nextmatch query array, with an additional client_id key defined
	*		- config Hooks into perp_ar site configuration to set some configuration.  
	*			It will be passed an array containing new settings.
	*			It must return an array(
	*				'template'      =>      <eTemplate name>
	*				'config'        =>      array(
	*					'enabled'	=>	boolean,
	*					<other settings>
	*				),
	*				'help'		=>	'Message displayed under the settings for this app'
	*			);
	*		- Invoicable is the name of the Invoicable class
	*		- InvoicableItem is the name of the InvoicableItem class
	*
	*	@see perp_orders::hook_perp_orders::invoicable_config()
	*	@see ProjectManagerInvoicableItem
	*	
	*
	*	@author Nathan Gray
	*	@ingroup perp_api
	*/

	require_once('class.InvoiceLine.inc.php');

	interface InvoicableItem {
		/**
		*	Mark this InvoicableItem as invoiced.  The InvoiceLine will take care of
		*	any General Ledger transactions, but this object should take care of
		*	any recordkeeping needed to mark itself as invoiced, and prevent 
		*	changes if needed.
		*
		*	@param invoice_line	InvoiceLine that is part of the Invoice
		*/
		public function invoice(InvoiceLine &$invoice_line);

		/**
		*	Return the Invoicable object that this InvoicableItem is part of
		*
		*	@return Invoicable
		*/
		public function get_owner();

		/**
		*	Return a unique identifier for the owner of this InvoicableItem
		*	This is usually the parent object's database key.
		*
		*	@return int
		*/
		public function get_invoicable_id();

		/**
		*	Return a unique (amongst the owning object's InvoicableItems)
		*
		*	@return int identifier for this InvoicableItem 
		*/
		public function get_invoicable_line_id();

		/**
		*	Check that this InvoicableItem can be invoiced.  It can't be
		*	invoiced if it has already been completely invoiced.
		*
		*	@throws Exception with error message if it can't be invoiced.
		*	@return boolean 
		*/
		public function can_invoice();

		/**
		*	Get the quantity to be invoiced, in the supplier's units of measure.
		*	
		*	@return float quantity to be invoiced
		*/
		public function get_invoice_quantity();

		/**
		*	Get unit price for one unit of the InvoicableItem being invoiced.
		*	This is in the supplier's units of measure.
		*
		*	@return float the price of one unit 
		*/
		public function get_unit_price();

		/**
		*	Get the description for the InvoicableItem
		*
		*	@return string description
		*/
		public function get_description();

		/**
		*	Get the general ledger account to be used on the invoice.
		*	May return null if the gl account is not known, in this case
		*	a note will be added to the invoice's comments section.
		*
		*	Implementing classes really should provide an account number.
		*	Methods of determining the account include a custom field, 
		*	but the best is using the config.
		*
		*	@return int general ledger account number
		*/
		public function get_gl_account();

		/**
		*	Return an array that will work for an eGroupWare style link.
		*	This is for allowing cross linking and jumping around.
		*	The array contain a sub-array, with keys app, id:
		*	return array(
		*		array(
		*			'app'	=>	$GLOBALS['egw_info']['flags']['currentapp'],
		*			'id'	=>	'ORDER-' . $this->get_owner->get__id()
		*		)
		*	);
		*
		*	@return array
		*/
		public function get_link();

		/**
		*	Return a list of extra details to be displayed on the invoice, below
		*	the line.  This list cannot be edited, and is an array of details in
		*	the form ('name' => 'Detail name', 'value' => '123')
		*
		*	@param filter The filter parameter from the display select (usually true)
		*	@return array
		*/
		public function get_invoice_details($filter);
	}
?>
