<?php

	/** 
	*	Base class for Invoices, both payable and receivable.  An invoice
	*	contains various information such as relevant dates, reference
	*	numbers and financial information, as well as one or more InvoiceLines.
	*
	*	Invoices can be automatically generated based of existing Orders or Shipments,
	*	or anything that implements the Invoicable interface.  These invoices can then
	*	be altered, if required.  Additionally, Invoices can
	*	be entered starting from a blank and adding the lines manually, but in this 
	*	case, no links to other documents are made.
	*
	*	perp_ap::PayableInvoice and perp_ar::ReceivableInvoice extend
	*	this class.
	*
	*	@ingroup perp_api
	*	@author Nathan Gray
	*/

	require_once( EGW_INCLUDE_ROOT . '/perp_api/inc/class.bo_perp_api.inc.php');

	require_once('class.InvoiceLine.inc.php');
	require_once('class.Invoicable.inc.php');

	class Invoice {

		/** 
		*	Database table for storing Invoices
		*/
		const invoice_table = 'perp_invoice';

		/**
		*	Unique system identifier for the Invoice (integer)
		*/
		protected $invoice_id;

		/**
		*	Our internal reference number for this Invoice
		*/
		protected $our_reference_number;

		/**
		*	The other party's reference number for this Invoice
		*/
		protected $their_reference_number;

		/**
		*	The system user id of the person who originated this Invoice
		*/
		protected $originating_user;

		/**
		*	The system user id of the person who approved this Invoice.
		*	This would be a manager or someone with the authority to allow the Invoice
		*	to proceed.
		*/
		protected $approving_user;

		/**
		*	The addressbook id of the contact person in our organization for this Invoice
		*/
		protected $our_contact_person;

		/**
		*	The addressbook ID of the contact person in their organization for this Invoice.
		*	This may be a sales rep, a purchaser, or someone in the accounting department.
		*/
		protected $their_contact_person;

		/**
		*	System identifier for the billing organisation.
		*	This will be either a Supplier's ID, or 0 if we are the supplier.
		*/
		protected $supplier_id = 0;

		/**
		*	System identifier for the organisation receiving and paying the Invoice.
		*	This will be either a Clients's ID, or 0 if we are the client.
		*/
		protected $client_id = 0;

		/**
		*	The date (and time) this Invoice was originated (created)
		*/
		protected $originating_date = null;

		/**
		*	The date (and time) this Invoice was approved by the approving_user
		*/
		protected $approval_date = null;

		/**
		*	The date (and time) this Invoice was issued.
		*/
		protected $issue_date = null;

		/**
		*	The date this Invoice needs to be paid by
		*/
		protected $due_date = null;

		/**
		*	If an early payment discount is offered, the date the offer expires.
		*/
		protected $discount_date = null;

		/**
		*	If an early payment discount is offered, the amount of discount 
		*	that has been taken.  This will be 0 if no discount is offered.
		*/
		protected $discount_taken = 0;

		/**
		*	The addressbook id of the sending location.  Also know as pay to.
		*/
		protected $bill_from_address;

		/**
		*	The addressbook id of the paying location.  Also known as bill to.
		*/
		protected $bill_to_address;

		/**
		*	Sometimes Invoices do not require approval before proceeding, depending on the organization.
		*/
		protected $requires_approval = true;

		/**
		*	The currency this Invoice is in.
		*/
		protected $invoice_currency;

		/**
		*	Exchange rate into General Ledger currency at the time of invoicing.
		*	This is required because the exchange rate can change between invoicing
		*	and payment.
		*/
		protected $exchange_rate = 1.00;

		/**
		*	Payment terms on this invoice.
		*/
		protected $payment_terms;

		/**
		*	Flag for this Invoice fully paid.
		*/
		protected $invoice_paid;

		/**
		*	Free form notes on this Invoice
		*/
		protected $comments;

		/**
		*	Transaction number of the General Ledger Transaction (perp_ledger::GLTransacton)
		*	caused by posting this invoice
		*/
		protected $invoice_transaction_id;

		/**
		*	Array of InvoiceLines, one for each item on this Invoice
		*/
		protected $invoice_lines = array();

		/**
		*	Array of taxes on this invoice.  This array is in the form
		*	@code
		*	array(
		*		<tax_authority_id>	=>	array(
		*			<tax_category>	=>	array(
		*				'description'	=>	<description>,
		*				'gl_account'	=>	<general ledger account>,
		*				'tax_on_tax'	=>	<tax aggregation level>,
		*				'calc_order'	=>	<calculation order>,
		*				'tax_rate'	=>	<taxation rate>,
		*				'amount'	=>	<tax_amount>,
		*			),
		*		),
		*	)
		*	@endcode
		*/
		protected $tax_lines = array();

		/**
		*	Constructor
		*	@param invoice_id	Optional.  If an invoice_id for an existing Invoice is provided, the Invoice will be loaded.  
		*	@return If invoice_id is provided, it will return whatever Invoice::read() returns.  Otherwise, true.
		*/
		public function __construct($invoice_id = null) {
			if(!is_null($invoice_id)) {
				return $this->read($invoice_id);
			}

			/**
			*	Set some defaults
			*/
			if(!is_object($GLOBALS['perp_api']['bo'])) {
				$GLOBALS['perp_api']['bo'] =& CreateObject('perp_api.bo_perp_api');
			}

			$this->originating_date = date('U');
			$this->issue_date = date('U');

			// Default currency to General Ledger currency
			$config = config::read('perp_ledger');
			$this->invoice_currency = $config['gl_currency'];

			// Set originator to the current user
			$this->set_originating_user($GLOBALS['egw_info']['user']['account_id']);

			// Default contact person to current user
			$this->set_our_contact_person((int)$GLOBALS['egw_info']['user']['person_id']);

			// Check to see if approval is required
			$config = config::read($GLOBALS['egw_info']['flags']['currentapp']);
			$this->requires_approval = $config['invoices_require_approval'];
			
			return true;
		}

		/**
		*	Read the Invoice identified by invoice_id into this Invoice
		*	@param invoice_id The unique system identifier for the Invoice to read.
		*/
		public function read($invoice_id) {
			$GLOBALS['egw']->db->select(self::invoice_table, 
				array(
					'invoice_id',
					'our_reference_number',
					'their_reference_number',
					'originating_user',
					'approving_user',
					'our_contact_person',
					'their_contact_person',
					'supplier_id',
					'client_id',
					'originating_date',
					'issue_date',
					'approval_date',
					'due_date',
					'discount_date',
					'discount_taken',
					'bill_from_address',
					'bill_to_address',
					'requires_approval',
					'invoice_currency',
					'exchange_rate',
					'payment_terms',
					'invoice_paid',
					'comments',
					'invoice_transaction_id'
				),
				array('invoice_id'	=>	$invoice_id),
				__LINE__, __FILE__
			);
			$row = $GLOBALS['egw']->db->row(true);
			$this->invoice_id		=	$row['invoice_id'];
			$this->our_reference_number	=	$row['our_reference_number'];
			$this->their_reference_number	=	$row['their_reference_number'];
			$this->originating_user		=	$row['originating_user'];
			$this->approving_user		=	$row['approving_user'];
			$this->our_contact_person	=	$row['our_contact_person'];
			$this->their_contact_person	=	$row['their_contact_person'];
			$this->supplier_id		=	$row['supplier_id'];
			$this->client_id		=	$row['client_id'];
			$this->originating_date		=	$GLOBALS['egw']->db->from_timestamp($row['originating_date']);
			$this->issue_date		=	$GLOBALS['egw']->db->from_timestamp($row['issue_date']);
			$this->approval_date		=	$GLOBALS['egw']->db->from_timestamp($row['approval_date']);
			$this->due_date			=	$GLOBALS['egw']->db->from_timestamp($row['due_date']);
			$this->discount_date		=	$GLOBALS['egw']->db->from_timestamp($row['discount_date']);
			$this->discount_taken		=	$row['discount_taken'];
			$this->bill_from_address	=	$row['bill_from_address'];
			$this->bill_to_address		=	$row['bill_to_address'];
			$this->requires_approval	=	$row['requires_approval'];
			$this->invoice_currency		=	$row['invoice_currency'];
			$this->exchange_rate		=	$row['exchange_rate'];
			$this->payment_terms		=	$row['payment_terms'];
			$this->invoice_paid		=	$row['invoice_paid'];
			$this->comments			=	$row['comments'];
			$this->invoice_transaction_id	=	$row['invoice_transaction_id'];

			// Read rows
			$GLOBALS['egw']->db->select(InvoiceLine::invoice_line_table,
				array(
					'line_no AS line_number',
					'quantity',
					'description',
					'unit_price',
					'line_type',
					'general_ledger_account',
					'gl_transaction_line',
					'invoicable_class',
					'invoicable_id',
					'invoicable_item_line',
				),
				array('invoice_id'	=>	$invoice_id),
				__LINE__, __FILE__, false,
				'ORDER BY invoice_id, line_no', 'perp_api'
			);
			$rows = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				// Read all rows first
				$rows[] = $row;
			}
			$line_class = get_class($this) . 'Line';
			foreach($rows as $row) {
				$line = new $line_class($this, $row);
				$line->read();
				$this->invoice_lines[$row['line_number']] = $line;
			}
			return true;
		}

		/**
		*	Check this Invoice and write it into the database.
		*	This is not Posting!
		*	Throws exceptions if the Invoice is not valid.
		*/
		public function write() {
			$result = $this->check();
			if($result !== true) {
				throw new Exception(lang('Invoice can not be saved.  (%1)', $result ));
			}

			$GLOBALS['egw']->db->transaction_begin();
			
			// Write Invoice
			$GLOBALS['egw']->db->insert(self::invoice_table,
				$this->toArray(false) + array(
					'class'	=>	get_class($this)
				),
				array(
					'invoice_id'	=>	$this->get_invoice_id()
				),
				__LINE__, __FILE__, 'perp_api'
			);
			if(!$this->invoice_id) {
				$this->invoice_id = $GLOBALS['egw']->db->get_last_insert_id(self::invoice_table, 'invoice_id');
			}

			// Write lines
			$GLOBALS['egw']->db->delete(InvoiceLine::invoice_line_table,
				array(
					'invoice_id'	=>	$this->get_invoice_id()
				),
				__LINE__, __FILE__, 'perp_api'
			);
			foreach($this->invoice_lines as $line_no => $line) {
				$line->write($line_no);
			}

			$GLOBALS['egw']->db->transaction_commit();
			return $this->invoice_id;
		}

		/**
		*	Post this Invoice into the General Ledger and any associated Ledgers.
		*	Posting can only be done once, and cannot be undone.
		*/
		public function post() {
			if($this->get_total() == 0) {
				$this->set_paid(true);
			}
			$this->write();
			if($this->invoice_transaction_id) {
				return false;
			}
			return true;
		}

		/**
		*	Delete this invoice, if it can be canceled
		*
		*	@return true, or error message
		*/
		public function cancel() {
			if(!$this->can_cancel()) {
				return lang('This invoice cannot be canceled');
			}

			$GLOBALS['egw']->db->transaction_begin();
				
			$GLOBALS['egw']->db->delete(self::invoice_table,
				array(
					'invoice_id'	=>	$this->get_invoice_id()
				),
				__LINE__, __FILE__, 'perp_api'
			);
			try {
				foreach($this->invoice_lines as $line_no => $line) {
					$line->cancel($line_no);
					$GLOBALS['egw']->db->delete(InvoiceLine::invoice_line_table,
						array(
							'invoice_id'	=>	$this->get_invoice_id()
						),
						__LINE__, __FILE__, 'perp_api'
					);
				}
			} catch (Exception $e) {
				$GLOBALS['egw']->db->transaction_abort();
				throw $e;
			}

			$GLOBALS['egw']->db->transaction_commit();
			return true;
		}

		/**
		*	Construct an Invoice from an object.
		*	The object must implement Invoicable interface.
		*/
		public function read_from_object(Invoicable &$object) {
			// Common header information
			$this->set_supplier_id((int)$object->get_supplier_id());
			$this->set_client_id((int)$object->get_client_id());

			$this->set_our_contact_person((int)$object->get_our_contact_person());
			$this->set_their_contact_person((int)$object->get_their_contact_person());
		
			$this->set_bill_from_address((int)$object->get_bill_from_address());
			if(!$object->get_bill_from_address()) {
				$this->set_bill_from_address((int)$object->get_their_contact_person());
			}
			$this->set_bill_to_address((int)$object->get_bill_to_address());

			$this->set_invoice_currency($object->get_invoice_currency());
			$this->set_payment_terms((int)$object->get_payment_terms());

			$this->set_our_reference_number($this->suggest_invoice_number());

			// Get tax information 
			$tax_group = $this->get_tax_group();

			$line_class = get_class($this) . 'Line';

			// Add freight line for shipments
			if($object instanceof Shipment) {
				$config = config::read($GLOBALS['egw_info']['flags']['currentapp']);
				if($config['freight_account']) {
					if(!is_object($GLOBALS['perp_api']['bo'])) {
						$GLOBALS['perp_api']['bo'] =& CreateObject('perp_api.bo_perp_api');
					}
					$GLOBALS['perp_api']['bo']->get_shippers(array('shipper_id' => $object->get_shipper_id()), $shippers);
					$shipper_name = $shippers[$object->get_shipper_id()];
					$item = array(
						'unit_price'	=>	0,
						'quantity'	=>	1,
						'description'	=>	lang('Freight charges (%1)', $shipper_name),
						'general_ledger_account'	=>	$config['freight_account']
					);
					$invoice_line = new $line_class($this, $item);
					$line_number = $this->add_line($invoice_line);
				}
			}

			$invoicable_items =& $object->get_invoicable_items();

			foreach($invoicable_items as $key => &$item) {
				if($item->get_invoice_quantity() != 0) {
					$invoice_line = new $line_class($this, $item);
					$line_number = $this->add_line($invoice_line);

					if($tax_group) {
						$found_taxes &= $invoice_line->calculate_taxes($tax_group, $source_tax_location);
					}
				}
			}

			$this->add_tax_lines();
		}

		/**
		*	Adds information to the Invoice's tax_lines array for every tax authority and tax rate applicable to the
		*	supplier or client.  This will be used when creating a blank invoice, or when the actual
		*	tax amounts cannot be calculated.  This does not actually add InvoiceLines to the Invoice,
		*	but it is called from add_tax_lines() if there is no information in the tax_lines array, which
		*	will cause tax lines to be added.
		*
		*	@param tax_group Tax group to add taxes from
		*/
		public function add_blank_tax_lines($tax_group) {
			if(!$tax_group) {
				return false;
			}

			// Get applicable tax rates in that group
			$sql = 'SELECT 
				perp_tax_authority.tax_authority_id,
				perp_tax_authority.tax_authority_name,
				perp_tax_category.tax_category_id,
				perp_tax_category.tax_category_name,
				perp_tax_authority.tax_authority_out_account,
				perp_tax_authority.tax_authority_in_account,
				perp_tax_authority.registration_number,
				perp_tax_auth_rate.tax_rate
			FROM perp_tax_group_tax
			JOIN perp_tax_auth_rate ON 
				perp_tax_group_tax.tax_authority_id = perp_tax_auth_rate.tax_authority 
			JOIN perp_tax_authority ON
				perp_tax_group_tax.tax_authority_id = perp_tax_authority.tax_authority_id
			JOIN perp_tax_category ON
				perp_tax_auth_rate.tax_category = perp_tax_category.tax_category_id
			WHERE perp_tax_group_tax.tax_group_id = ' . $tax_group . '
			ORDER BY calc_order';

			$GLOBALS['egw']->db->query($sql, __LINE__, __FILE__);

			while($row = $GLOBALS['egw']->db->row(true)) {
				$this->tax_lines[$row['tax_category_id']][] = $row + array(
					'description'	=>	($row['tax_rate'] * 100) . '% ' . $row['tax_authority_name'] . ' ' . $row['tax_category_name'] .
								($this->supplier_id == 0 ? ' (' . $row['registration_number'] . ')' : ''),
					'unit_price'	=>	0,
					'general_ledger_account'	=>	($this->supplier_id) ? $row['tax_authority_out_account'] : $row['tax_authority_in_account']
				);
					
			}
		}

		/**
		*	Take all the calculated tax lines in tax_lines and add one InvoiceLine
		*	for each Tax Authority to the Invoice.
		*/
		public function add_tax_lines() {
			// Double check tax lines.  If none were found, add blanks
			if(count($this->tax_lines) == 0) {
				$this->add_blank_tax_lines($this->get_tax_group());
			}
			
			// Remove any lines flagged as TAX so we don't get duplicates
			foreach($this->invoice_lines as $key => &$line) {
				if($line->get_line_type() === InvoiceLine::TAX) {
					$this->remove_line($key);
				}
			}

			$line_class = get_class($this) . 'Line';
			foreach($this->tax_lines as &$tax_category) {
				foreach($tax_category as $tax) {
					$tax_amount = $amount;
					$tax['quantity'] = 1;
					$tax['line_type'] = InvoiceLine::TAX;

					$new_tax_line = new $line_class($this, $tax);
					$this->add_line($new_tax_line);
				}
			}

			// Clear tax information
			$this->tax_lines = array();
		}

		/**
		*	Get the tax group for calculating required taxes.
		*	This is set per supplier / client branch.
		*	@return int tax_group_id
		*/
		public function get_tax_group() {
			return false;
		}

		/**
		*	Get the array of tax information
		*/
		public function &get_tax_array() {
			return $this->tax_lines;
		}

		public function &get_tax_lines() {
			$tax_lines = array();
			foreach($this->invoice_lines as $key => &$line) {
				if($line->get_line_type() == InvoiceLine::TAX) {
					$tax_lines[] =& $line;
				}
			}
			return $tax_lines;
		}

		/**
		*	Permissions and validation methods
		*	@{
		*/

		/**
		*	Check that this Invoice is valid.  Specifically:
		*	- There is at least one InvoiceLine.
		*	- Bill From and Bill To addresses are set
		*	- Each InvoiceLine passes its check
		*
		*	@return boolean true of the order is valid, otherwise returns error message
		*/
		public function check() {
			$test = true;
			if(is_null($this->supplier_id) && is_null($this->client_id)) {
				$test = false;
				return lang('Supplier or client not set.');
			}
			if(trim($this->get_our_reference_number()) == '') {
				return lang('Our reference number is not set');
			}
			if(count($this->invoice_lines) == 0) {
				$test = false;
				return lang('Empty invoice.  Invoices need at least one InvoiceLine');
			}
			if(!is_numeric($this->bill_from_address) || $this->bill_from_address == 0) {
				$test = false;
				return lang('Bill from address is not set.');
			}
			if(!is_numeric($this->bill_to_address) || $this->bill_to_address == 0) {
				$test = false;
				return lang('Bill to address is not set.');
			}
			if(!is_numeric($this->payment_terms) || $this->payment_terms == 0) {
				$test = false;
				return lang('Payment terms are not set.');
			}

			foreach($this->invoice_lines as $key => $line) {
				$line_test = $line->check();
				if($line_test !== true) {
					return $line_test;
				}
			}
			return (boolean)$test;
		}

		/**
		*	Check to see if this invoice can be cancelled
		*
		*	@return boolean 
		*/
		public function can_cancel() {
			return $GLOBALS['egw']->acl->check('invoice', EGW_ACL_DELETE, $GLOBALS['egw_info']['flags']['currentapp']) &&
				is_null($this->invoice_transaction_id);
		}
		
		/**
		*	End of permissions and validation methods
		*	@}
		*/

		/**
		*       Check if the given user can edit this Invoice
		*       @param user_id  User ID of the user to check
		*/
		public function is_editor($user_id) {
			if($this->invoice_transaction_id) {
				// Can't edit after posting.
				return false;
			}
			if($user_id === $this->get_originating_user()) {
				return !(boolean)$this->get_approval_date();
			}
			return $this->is_admin($user_id);
		}

		/**
		*       Check if the given user can administer this Invoice
		*       This includes approving, denying, editing after
		*       approval.  Nothing can be done after posting.
		*       @param user_id  User ID of the user to check
		*/
		public function is_admin($user_id) {
			return array_key_exists($user_id, $this->get_admin_list());
		}

		/**
		*       Get a list of users allowed to administer this Invoice
		*/
		public function get_admin_list() {
			$users = array();
			$approvers = $GLOBALS['egw']->acl->get_ids_for_location('invoice', EGW_ACL_CUSTOM_1);
			if(!is_array($approvers)) {
				return $users;
			}
			foreach($approvers as $id) {
				if($id != 0) {
					$users[$id] = $GLOBALS['egw']->accounts->id2name($id);
				}
				if($id < 0) {
					$users += $GLOBALS['egw']->accounts->members($id);
				}
			}
			return $users;
		}

		/**
		*	Accessor methods
		*	@{
		*/
		
		
		/**
		*	Get the unique system identifier for this Invoice
		*	@return int invoice_id
		*/
		public function get_invoice_id() {
			return $this->invoice_id;
		}

		/**
		*	Get the organization supplying the items billed on this Invoice.
		*	This will be either a Supplier's ID, or 0 if we are the supplier.
		*	@return int supplier_id
		*/
		public function get_supplier_id() {
			return $this->supplier_id;
		}

		/**
		*	Set the organization supplying the items billed on this Invoice.
		*	This will be either a Supplier's ID, or 0 if we are the supplier.
		*	@param supplier_id ID of the supplier
		*/
		public function set_supplier_id($supplier_id) {
			if(!is_integer($supplier_id)) {
				return lang('Invalid supplier');
			}
			$this->supplier_id = $supplier_id;

			// Default bill_from address
			if(!$this->bill_from_address) {
				if($this->supplier_id != 0) {
					$supplier = ExecMethod('perp_ap.bo_perp_supplier.get_supplier', $this->supplier_id);
					$this->set_bill_from_address((int)$supplier['supplier_address']);
					return true;
				}
			} else {
				return true;
			}
		}

		/**
		*	Get the organization receiving the items billed on this Invoice.
		*	This will be either a Client's ID, or 0 if we are the client.
		*	@return int client_id
		*/
		public function get_client_id() {
			return $this->client_id;
		}

		/**
		*	Set the organization receiving the items billed on this Invoice
		*	This will be either a Client's ID, or 0 if we are the client.
		*	@param client_id ID of the client
		*/
		public function set_client_id($client_id) {
			if(!is_integer($client_id)) {
				return lang('Invalid client');
			}
			if($client_id) {
				$this->client_id = $client_id;
				$client_data = ExecMethod('perp_ar.bo_perp_client.get_client', $client_id);
				$this->set_invoice_currency($client_data['currency']);
			}
			return true;
		}


		/**
		*	Set our human reference number for this Invoice
		*	@param reference string our_reference_number
		*/
		public function set_our_reference_number($reference) {
			$this->our_reference_number = $reference;
		}

		/**
		*	Get our human reference number for this Invoice
		*	@return string our_reference_number
		*/
		public function get_our_reference_number() {
			return $this->our_reference_number;
		}

		/**
		*	Set their human reference number for this Invoice
		*	@param reference string their_reference_number
		*/
		public function set_their_reference_number($reference) {
			$this->their_reference_number = $reference;
		}

		/**
		*	Get their human reference number for this Invoice
		*	@return string their_reference_number
		*/
		public function get_their_reference_number() {
			return $this->their_reference_number;
		}

		/**
		*	Set the eGroupWare user id of the originating user.
		*	@param user_id int System user_id
		*/
		public function set_originating_user($user_id) {
			if(!is_integer($user_id)) {
				return lang('Invalid user id for originating user');
			}
			$this->originating_user = $user_id;
		}

		/**
		*	Get the eGroupWare user id of the originating user
		*	@return int user id
		*/
		public function get_originating_user() {
			return $this->originating_user;
		}

		/**
		*	Set the date this Invoice was created
		*	@param date int Unix time stamp
		*/
		public function set_originating_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for originating date');
			}
			$this->originating_date = $date;
		}
		/**
		*	Get the date this Invoice was created
		*	@return int Unix time stamp
		*/
		public function get_originating_date() {
			return $this->originating_date;
		}

		/**
		*	Set the date this Invoice was issued
		*	@param date int Unix time stamp
		*/
		public function set_issue_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for issue date');
			}
			// A change in the issue date affects the due date, in needs to be re-calculated
			$this->due_date = null;

			$this->issue_date = $date;
		}

		/**
		*	Get the date this Invoice was issued
		*	@return int Unix time stamp
		*/
		public function get_issue_date() {
			return $this->issue_date;
		}

		/**
		*	Get the date this Invoice was approved
		*	@return int Unix time stamp
		*/
		public function get_approval_date() {
			return $this->approval_date;
		}

		/** 
		*	Get whether this Invoice requires approval or not
		*	@return boolean requires approval
		*/
		public function requires_approval() {
			return $this->requires_approval;
		}

		/**
		*	Set the date this Invoice is due
		*	@param date int Unix time stamp
		*/
		public function set_due_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for due date');
			}
			$this->due_date = $date;
		}

		/**
		*	Get the payment due date.  If not set, it will be calculated based on terms.
		*	@return int Unix time stamp
		*/
		public function get_due_date() {
			if(!is_null($this->due_date)) {
				return $this->due_date;
			}

			// Calculate due date based on terms.
			$term_list = array();
			$boerp = CreateObject('perp_api.bo_perp_api');
			$query = array('terms' => $this->payment_terms, 'template' => false);
			$boerp->get_payment_terms($query, $term_list);
			$terms = $term_list[0];
			$bill_date = $this->get_issue_date();
			switch($terms['due_date_type']) {
				case 0:
					// Days from invoice date
					$due_date = date('U', mktime(0,0,0,date('m', $bill_date),date('d', $bill_date) + $terms['due_days'], date('Y', $bill_date)));
					break;
				case 1:
					// Days in following month
					$due_date = date('U', mktime(0,0,0,date('m', $get_bill_date) + 1, $terms['due_days'], date('Y', $bill_date)));
					break;
			}
			$this->due_date = $due_date;
			return $due_date;
		}

		/**
		*	Set the date early payment discount expires
		*	@param date int Unix time stamp
		*/
		public function set_discount_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for discount date');
			}
			$this->discount_date = $date;
		}

		/**
		*	Get the discount date
		*	@return int Unix time stamp
		*/
		public function get_discount_date() {
			if(!is_null($this->discount_date)) {
				return $this->discount_date;
			}
			$term_list = array();
			$boerp = CreateObject('perp_api.bo_perp_api');
			$query = array(
				'terms'         =>      $this->payment_terms,
				'template'      =>      false
			);
			$boerp->get_payment_terms($query, $term_list);
			if(count($term_list) != 1) {
				return lang('Could not find the details for the payment terms');
			}
			$terms = $term_list[0];
			$bill_date = $this->get_issue_date();

			switch($terms['discount_date_type']) {
				case 0:
					// Set number of days
					$discount_date = date('U', mktime(0,0,0,date('m', $bill_date),date('d', $bill_date) + $terms['discount_days']));
					break;
				case 1:
					// Calendar day next month
					$discount_date = date('U', mktime(0,0,0,date('m', $bill_date) + 1, $terms['discount_days']));
				default:
					// No discount
					$discount_date = null;
				break;
			}
			$this->discount_date = $discount_date;

			return true;
		}

		/**
		*	Set the billing date on this Invoice.  This is syntactic sugar
		*	for set_issue_date().
		*	@param date Date of the billing
		*/
		public function set_bill_date($date) {
			$this->set_issue_date($date);
		}

		/**
		*	Get the billing date on this Invoice.  This is syntactic sugar
		*	for get_issue_date().
		*	@return timestamp
		*/
		public function get_bill_date() {
			return $this->get_issue_date();
		}

		/**
		*	Take an offered early payment discount.  This is additive,
		*	and adds to any discounts already taken on this invoice.
		*	@todo Check some errors
		*	@param amount float amount of discount taken
		*/
		public function take_discount($amount) {
			$this->discount_taken += $amount;
		}

		/**
		*	Get the amount of discount taken.
		*	@return float Amount of discount taken.
		*/
		public function get_discount_taken() {
			return $this->discount_taken;
		}

		/**
		*	Get the amount of early payment discount available on this Invoice
		*	according to payment terms and discount already taken.
		*	@param date	Optional As at date - today by default
		*	@return Amount of discount available on the given date
		*/
		public function get_discount_available($date = null) {
			if(!$date) {
				$date = date('U');
			}
			if($date <= $this->discount_date) {
				// Calculate discount based on terms
				$term_list = array();
				if(!$GLOBALS['perp_api']['bo']) {
					$GLOBALS['perp_api']['bo'] = CreateObject('perp_api.bo_perp_api');
				}
				$GLOBALS['perp_api']['bo']->get_payment_terms(
					array(
						'terms'	=>	$this->payment_terms, 
						'template' => 	false
					), 
					$term_list
				);
				$terms = $term_list[0];
				// Discount is the discount percent * pre-tax total
				return round(($terms['discount_percent'] / 100) * ($this->get_total() - $this->get_total(InvoiceLine::TAX)) - $this->discount_taken, 2);
			} else {
				// Too late, after deadline
				return 0;
			}
		}

		/**
		*	Get the person in our organization who is to be contacted
		*	regarding this Invoice.
		*	@return int addressbook_id
		*/
		public function get_our_contact_person() {
			return $this->our_contact_person;
		}

		/**
		*	Set the person in our organization who is to be contacted
		*	regarding this Invoice.
		*	@param contact Addressbook ID of the contact
		*	@return true or error message
		*/
		public function set_our_contact_person($contact) {
			if(!is_integer($contact)) {
				return lang('Invalid contact person.');
			}
			return true;
		}

		/**
		*	Get the person in their organization who is to be contacted
		*	regarding this Invoice.
		*	@return int addressbook_id
		*/
		public function get_their_contact_person() {
			return $this->their_contact_person;
		}

		/**
		*	Set the person in their organization who is to be contacted
		*	regarding this Invoice.
		*	@param contact Addressbook ID of the contact
		*	@return true or error message
		*/
		public function set_their_contact_person($contact) {
			if(!is_integer($contact)) {
				return lang('Invalid contact person.');
			}
			$this->their_contact_person = $contact;
			return true;
		}

		/**
		*	Get the location the Bill from address
		*	@return int addressbook ID of the location
		*/
		public function get_bill_from_address() {
			return $this->bill_from_address;
		}

		/**
		*	Set the bill from address
		*	@param location Addressbook ID of the location
		*/
		public function set_bill_from_address($location) {
			if(!is_integer($location)) {
				return lang('Invalid bill from address.');
			}
			$this->bill_from_address = $location;
		}

		/**
		*	Get the bill to address
		*	@return int addressbook ID of the location
		*/
		public function get_bill_to_address() {
			return $this->bill_to_address;
		}

		/**
		*	Set the bill to address
		*	@param location Addressbook ID of the location
		*/
		public function set_bill_to_address($location) {
			if(!is_integer($location)) {
				return lang('Invalid bill to address.');
			}
			$this->bill_to_address = $location;
		}

		/**
		*	Whether incoming or outgoing, find our location associated with this Invoice.
		*	This will use Invoice's bill_to or bill_from addresses to get the associated location.
		*	It is recommended to use InvoiceLine::get_our_location for a more specific location.
		*	@return location_id
		*/
		public function get_our_location() {
			if($this->supplier_id) {
				$address = $this->bill_to_address;
			} elseif($this->client_id) {
				$address = $this->bill_from_address;
			}
			$GLOBALS['egw']->db->select('perp_location', 
				array('location_id'),
				array('delivery_address' => $address),
				__LINE__, __FILE__, false, '', 'perp_inventory', 1
			);
			return $GLOBALS['egw']->db->f('location_id');
		}

		/**
		*	Get the currency this Invoice is being conducted in
		*	@return string three letter currency code
		*/
		public function get_invoice_currency() {
			return $this->invoice_currency;
		}

		/**
		*	Set the currency this Invoice is being conducted in
		*	@todo Better parameter checking
		*	@param currency Three letter currency code from currency list
		*/
		public function set_invoice_currency($currency) {
			if(!is_string($currency)) {
				return lang('Invalid currency.');
			}
			$this->invoice_currency = $currency;
			$this->exchange_rate = 1;
		}

		/**
		*	Get the payment terms for this Invoice
		*	@return int payment_term_id
		*/
		public function get_payment_terms() {
			return $this->payment_terms;
		}

		/**
		*	Set the payment terms for this Invoice
		*	@todo Better parameter checking
		*	@param payment_term_id
		*/
		public function set_payment_terms($payment_term_id) {
			if(!is_integer($payment_term_id)) {
				return lang('Invalid payment terms.');
			}
			$terms = array_keys(ExecMethod('perp_api.bo_perp_api.get_payment_terms'));
			if($payment_term_id && in_array($payment_term_id, $terms)) {
				if($payment_term_id != $this->payment_terms) {
					// Reset due date
					$this->due_date = null;
				}
			}
			$this->payment_terms = $payment_term_id;
		}

		/**
		*	Get comments on this Invoice
		*	@return string comments
		*/
		public function get_comments() {
			return $this->comments;
		}

		/**
		*	Set comments on this Invoice
		*	@param comments 
		*/
		public function set_comments($comments) {
			$this->comments = $comments;
		}

		/**
		*	Get the ID of the General Ledger transaction (perp_ledger::GLTransaction)
		*	used to post this Invoice to the General Ledger
		*/
		public function get_transaction() {
			return $this->invoice_transaction_id;
		}

		/**
		*	Get whether or not this Invoice has been fully paid
		*	@return boolean fully paid
		*/
		public function is_paid() {
			return $this->invoice_paid;
		}

		/**
		*	Set this Invoice as fully paid
		*	@param paid Optional.  Paid status
		*/
		public function set_paid($paid = true) {
			$this->invoice_paid = $paid;
		}

		/**
		*	Get the exchange rate from the Invoice currency into the General
		*	Ledger currency at the time this invoice was entered.
		*	@return float Exchange rate
		*/
		public function get_exchange_rate() {
			if($this->exchange_rate > 0) {
				return $this->exchange_rate;
			}
			$currencies = array();
			$query = array(
				'currency_id'	=>	$this->invoice_currency, 
				'template'	=>	false,
				'num_rows'	=>	1
			); 
			if(!is_object($GLOBALS['perp_api']['bo'])) {
				$GLOBALS['perp_api']['bo'] =& CreateObject('perp_api.bo_perp_api');
			}
			$GLOBALS['perp_api']['bo']->get_currencies(
				$query,
				$currencies
			);
			if($currencies[0]) {
				$this->exchange_rate = $currencies[0]['exchange_rate'];
				return $currencies[0]['exchange_rate'];
			}
		}

		/**
		*	Set the exchange rate for this Invoice
		*	@param rate float exchange rate into the General Ledger currency
		*/
		public function set_exchange_rate($rate) {
			if((float)$rate == 0) {
				throw new Exception(lang('0 is an invalid exchange rate.'));
			}
			$this->exchange_rate = $rate;
		}

		/**
		*	Add an InvoiceLine to this Invoice
		*	@param line	InvoiceLine to add
		*/
		public function add_line(InvoiceLine &$line) {
			$this->invoice_lines[] = $line;
			$line_number = end(array_keys($this->invoice_lines));
			$line->set_line_number($line_number);
			return $line_number;
		}

		/**
		*	Remove an InvoiceLine from this Invoice
		*	@param line_number Number of InvoiceLine to remove
		*/
		public function remove_line($line_number) {
			unset($this->invoice_lines[$line_number]);
		}

		/**
		*	Return the requested InvoiceLine
		*	@param line_number Number of InvoiceLine to return
		*/
		public function get_line($line_number) {
			return $this->invoice_lines[$line_number];
		}

		/**
		*	Get the InvoiceLines
		*/
		public function &get_lines() {
			return $this->invoice_lines;
		}

		/**********************************
		*	End of accessor methods
		*	@}
		*/

		/**
		*	Suggest our reference number (human reference) based on
		*	system settings stored in eGroupWare config variables for
		*	this Invoice.  Depending on permissions, the user may override
		*	the suggested number.
		*
		*	@return string Suggested number based on site settings.
		*/
		public function suggest_invoice_number() {
			return 'Not set, developer should override suggest_invoice_number()';
		}

		/**
		*	Get a total for this Invoice.
		*	@param type Optional.  Only total the given type of InvoiceLine
		*	@return float Total
		*/
		public function get_total($type = null) {
			$total = 0;
			foreach($this->invoice_lines as &$line) {
				if(is_null($type) || $line->get_line_type() == $type) {
					$total += round($line->get_quantity() * $line->get_unit_price(),2);
				}
			}
			return $total;
		}

		/**
		*	Return an array of this Invoice's properties.
		*	Used for display in eTemplates and suck
		*	@param get_lines	boolean Include invoice lines in the array
		*/
		public function toArray($get_lines = false) {
			$data = get_object_vars($this);
			$data['due_date'] = $this->get_due_date();
			$data['discount_date'] = $this->get_discount_date();
			$data['exchange_rate'] = $this->get_exchange_rate();
			$data['total'] = $this->get_total();
			unset($data['invoice_lines']);
			if($get_lines) {
				$data['invoice_lines'] = array();
				foreach($this->invoice_lines as $key => $line) {
					$line_array = $line->toArray();
					$line_array['line_no'] = $key;
					$data['invoice_lines'][] = $line_array;
				}
			}
			return $data;
		}

		/**
		*	Return a string representation of this object.
		*	Very useful for debugging purposes, not much else.
		*	@return string representation of this Invoice.
		*/
		public function __toString() {
			$result = $GLOBALS['egw']->accounts->get_account_name($this->get_originating_user(), $id, $fname, $lname);
			$originating_user = $fname . ' ' . $lname . ' [' . $id . ']';

			$string = sprintf(get_class($this) . " ID: %d\t Our Ref #: %s started %s by %s\tDue: %s\n",
				$this->get_invoice_id(),
				$this->get_our_reference_number(),
				date($GLOBALS['egw_info']['user']['preferences']['common']['dateformat'], $this->get_originating_date()),
				$originating_user,
				date($GLOBALS['egw_info']['user']['preferences']['common']['dateformat'], $this->get_due_date())
			);

			$string .= lang('supplier') . ': ' . $this->get_supplier_id() . "\n";
			$string .= lang('client') . ': ' . $this->get_client_id() . "\n";

			foreach($this->invoice_lines as $line_no => &$line) {
				if($line instanceof InvoiceLine) {
					$string .= "\t$line_no.\t" . $line->__toString() . "\n";
				}
			}
			return $string;
		}

		/**
		*	Prepare this object for serialization
		*	@return array Properties to preserve accros serializtion.
		*/
		public function __sleep() {
			$properties = array(
				'invoice_id',
				'our_reference_number', 'their_reference_number',
				'originating_user', 'approving_user',
				'our_contact_person', 'their_contact_person',
				'supplier_id', 'client_id',
				'bill_from_address', 'bill_to_address',
				'originating_date', 'approval_date', 'issue_date',
				'due_date', 'discount_date',
				'discount_amount',
				'requires_approval',
				'invoice_currency', 'payment_terms', 'exchange_rate',
				'invoice_transaction_id',
				'comments',
				'invoice_lines', 'tax_lines'
			);
			return $properties;
		}
	}
?>
