<?php
	require_once('class.InvoicableItem.inc.php');

	class InvoiceLine {

		const NORMAL	=	0;
		const TAX	=	1;
		const DISCOUNT	=	2;

		/**
		*	Database table for storage
		*/
		const invoice_line_table = 'perp_invoice_line';

		/**
		*	The Invoice this InvoiceLine is part of
		*/
		protected $invoice;

		/**
		*	Line number of this InvoiceLine (integer)
		*/
		protected $line_number;

		/**
		*	InvoicableItem that this InvoiceLine is invoicing
		*/
		protected $invoicable_item = null;

		/**
		*	Quantity invoiced
		*/
		protected $quantity;

		/**
		*	Price for one unit
		*/
		protected $unit_price;

		/**
		*	Type of line (from constants)
		*/
		protected $line_type = InvoiceLine::NORMAL;

		/**
		*	General Ledger account for this invoice line
		*/
		protected $general_ledger_account = null;

		/**
		*	General Ledger transaction line number after this InvoiceLine is posted
		*/
		protected $gl_transaction_line = null;

		/**
		*	Constructor
		*	@param	invoice	Invoice object this InvoiceLine belongs to
		*	@param 	data	Either an array of information for this InvoiceLine, or an InvoiccableItem object
		*/
		public function __construct(Invoice &$invoice, &$data = false) {
			$this->invoice =& $invoice;

			if($data instanceof InvoicableItem) {
				if(is_null($data->get_invoicable_line_id())) {
					throw new Exception(lang('%1 has no line ID, and cannot be saved', get_class($data)));
				}
				$this->invoicable_item =& $data;
				$this->set_quantity($data->get_invoice_quantity());
				$this->set_unit_price($data->get_unit_price());
				$this->set_description($data->get_description());
				$this->general_ledger_account	=	$data->get_gl_account();
			} elseif( is_array($data) ) {
				$this->line_number		=	$data['line_number'];
				$this->quantity			=	$data['quantity'];
				$this->description		=	$data['description'];
				$this->unit_price		=	$data['unit_price'];
				$this->general_ledger_account	=	$data['general_ledger_account'];
				$this->gl_transaction_line	=	$data['gl_transaction_line'];
				$this->line_type		=	$data['line_type'];

				if(!is_null($data['invoicable_class'])) {
					if(!is_object($GLOBALS['perp_api']['bo'])) {
						$GLOBALS['perp_api']['bo'] =& CreateObject('perp_api.bo_perp_api');
					}
					$this->invoicable_item =& $GLOBALS['perp_api']['bo']->get_invoicable_item($data);
					if($this->invoicable_item instanceof InvoicableItem && !$this->general_ledger_account) {
						$this->set_general_ledger_account($this->invoicable_item->get_gl_account());
					}
				}
			}

			// Look for a general ledger account
			if(!$this->general_ledger_account) {
				$this->find_general_ledger_account();
			}

			if(!$this->quantity) {
				$this->quantity = 1;
			}
		}

		/**
		*	Read this line from the database.  It will already be created & mostly read by
		*	Invoice, but this will give a hook to extending classes.
		*/
		public function read() {
			
		}

		/**
		*	Write this line to the database
		*	@param line_number The Identification number of this line.
		*/
		public function write($line_number) {
			$data = array(
					'invoice_id'		=>	$this->invoice->get_invoice_id(),
					'line_no'		=>	$line_number
				) +
				$this->toArray();
			if(!is_null($this->invoicable_item) && !is_null($this->invoicable_item->get_invoicable_line_id())) {
				if($this->invoiceable_item instanceof ShippableItem && $this->invoicable_item->get_shippable_item() instanceof InvoicableItem) {
					$this->invoicable_item = $this->invoicable_item->get_shippable_item();
				}
				$data += array(
					'invoicable_class'	=>	get_class($this->invoicable_item->get_owner()),
					'invoicable_id'		=>	$this->invoicable_item->get_invoicable_id(),
					'invoicable_item_line'	=>	$this->invoicable_item->get_invoicable_line_id()
				);
			}

			$GLOBALS['egw']->db->insert(self::invoice_line_table,
				$data,
				array(
					'invoice_id'	=>	$this->invoice->get_invoice_id(),
					'line_no'	=>	$line_number
				),
				__LINE__, __FILE__, 'perp_api'
			);

			if(!is_null($this->invoicable_item)) {
				$this->invoicable_item->invoice($this);
			}
		}


		/**
		*	Post this InvoiceLine to the General Ledger
		*	Adds a perp_ledger::GLTransactionLine to a perp_ledger::GLTransaction 
		*	in progress.
		*	@param transaction GLTransaction in progress
		*	@return actual value after GL rounds it to 2 decimal places
		*/
		public function post(GLTransaction &$transaction) {
			$line = $transaction->add_line($this->general_ledger_account, $this->description,
				bcmul($this->unit_price, bcmul($this->quantity, $this->invoice->get_exchange_rate(), 3), 3)
			);
			$this->gl_transaction_line = end(array_keys($transaction->get_lines()));
			return $line->get_amount();
		}

		/**
		*	Any actions to perform before being removed from the Invoice
		*	Does not actually remove the line.
		*	
		*	@note Reserved for future use, if InvoiceLines need details.
		*/
		public function remove() {
			// Remove OrderLineDetails
			foreach($this->details as &$detail) {
				$detail->remove();
			}
			$GLOBALS['egw']->db->delete(InvoiceLineDetail::detail_table, 
				array(
					'invoice_id'	=>	$this->invoice->get_invoice_id(),
					'line_no'	=>	$this->line_number,
				),
				__LINE__, __FILE__
			);
		}

		/**
		*	Any actions to perform before the Invoice is cancelled.
		*	Does not actually remove the line or cancel the invoice.
		*/
		public function cancel() {
			//return $this->remove();
			return true;
		}

		/**
		*	Calculate any applicable taxes on this line.
		*
		*	These are stored in the Invoice::tax_lines array,
		*	until they get added as extra lines on the invoice.
		*
		*	@param tax_group ID of the tax group applicable to this line.  This is set
		*		per stock item (stock_id), according to supplier / client settings.
		*	@param source_location Tax location the goods are coming from.  This is set 
		*		per supplier / client.
		*/
		public function calculate_taxes($tax_group, $source_location = null) {
			if($this->line_type == InvoiceLine::TAX) {
				return true;
			}
			if($this->invoicable_item && method_exists($this->invoicable_item, 'get_stock_id')) {
				$stock_id = $this->invoicable_item->get_stock_id();
			}
			if (!$tax_group) {
				$tax_group = $this->invoice->get_tax_group();
			}
			if(!$tax_group) {
				return false;
			}
			if(!is_numeric($source_location)) {
				// Commented out because it's never set, but I think it needs to be somehow.  I'm just not sure how.
				//return false;
			}

			// Get applicable taxes
			$sql = 'SELECT 
					perp_tax_authority.tax_authority_id,
					perp_tax_authority.tax_authority_name,
					perp_tax_authority.tax_authority_in_account,
					perp_tax_authority.tax_authority_out_account,
					perp_tax_authority.registration_number,
					perp_tax_group_tax.tax_on_tax,
					perp_tax_group_tax.calc_order,
					perp_tax_auth_rate.tax_category,
					perp_tax_auth_rate.tax_rate
				FROM perp_tax_group_tax
				JOIN perp_tax_auth_rate ON 
					perp_tax_group_tax.tax_authority_id = perp_tax_auth_rate.tax_authority ';
			if($source_location) {
				$sql .= "AND perp_tax_auth_rate.tax_source_location = $source_location";
			}
			$sql .=	' JOIN perp_tax_authority ON
					perp_tax_group_tax.tax_authority_id = perp_tax_authority.tax_authority_id ';
			if($stock_id) {
				$sql .= '
				JOIN perp_stock_tax_category ON
					perp_tax_auth_rate.tax_category = perp_stock_tax_category.tax_category_id AND
					perp_tax_group_tax.tax_authority_id = perp_stock_tax_category.tax_authority_id  AND
					perp_stock_tax_category.stock_id = ' . $stock_id;
			} 
			$sql .= '
				WHERE perp_tax_group_tax.tax_group_id = ' . $tax_group . '
				ORDER BY calc_order';

			$GLOBALS['egw']->db->query($sql, __LINE__, __FILE__);

			$tax_list = array();
			$taxes =& $this->invoice->get_tax_array();

			while($row = $GLOBALS['egw']->db->row(true)) {
				$tax_list[$row['tax_authority_id']][$row['tax_category']]['description'] = ($row['tax_rate'] * 100) . '% ' . $row['tax_authority_name'] .
					($this->invoice->get_supplier_id() == 0 ? ' (' . $row['registration_number'] . ')' : '');
				$tax_list[$row['tax_authority_id']][$row['tax_category']]['general_ledger_account'] = $row[($this->invoice->get_client_id() ? 'tax_authority_in_account' : 'tax_authority_out_account')];
				$tax_list[$row['tax_authority_id']][$row['tax_category']]['tax_on_tax'] = $row['tax_on_tax'];
				$tax_list[$row['tax_authority_id']][$row['tax_category']]['calc_order'] = $row['calc_order'];
				$tax_list[$row['tax_authority_id']][$row['tax_category']]['tax_rate'] = $row['tax_rate'];
			}

			$this->tax_on_tax_total = array();

			foreach($tax_list as $authority => &$tax_category) {
				foreach($tax_category as $category => &$tax) {
					$tax_amount = $this->unit_price * $this->quantity;
					if($tax['tax_on_tax'] && is_array($tax_on_tax_total[$category])) {
						foreach($tax_on_tax_total[$category] as $calc_order => $row) {
							if($calc_order < $tax['calc_order']) {
								$tax_amount += $row;
							}
						}
					}
					$tax_amount = round($tax_amount * $tax['tax_rate'], 2);
					if(!is_array($taxes[$authority][$category])) {
						$taxes[$authority][$category] = $tax;
					} else {
						$taxes[$authority][$category] = array_merge($taxes[$authority][$category], $tax);
					}
					$taxes[$authority][$category]['quantity'] = 1;
					$taxes[$authority][$category]['line_type'] = InvoiceLine::TAX;
					$taxes[$authority][$category]['unit_price'] += $tax_amount;
					$tax_on_tax_total[$category][$tax['calc_order']] += $tax_amount;
				}
			}

			if(!$stock_id && count($tax_list) > 1 ) {
				// Need to get the tax group tax row
				return lang('Not enough information to calculate taxes for %1, please check tax lines.', $this->description);
			}
			return true;
		}

		/**
		*	Try to find a general ledger account for this InvoiceLine
		*/
		public function find_general_ledger_account() {
			if($this->general_ledger_account) {
				return false;
			}
			if(!$this->stock_id) {
				if(!is_null($this->invoicable_item) && method_exists($this->invoicable_item, 'get_stock_id')) {
					$this->stock_id = $this->invoicable_item->get_stock_id();
					if(!$this->stock_id) {
						return false;
					}
				}
			}

			$query = array(
				'stock_id'	=>	$this->stock_id
			);
			$general_ledger_accounts = array();
			if($query['location_id'] = $this->get_our_location()) {
				$bo_location = CreateObject('perp_inventory.bo_perp_location');
				$bo_location->get_gl_accounts($query, $general_ledger_accounts);
			}
			if($general_ledger_accounts['stock_gl_account']) {
				$this->set_general_ledger_account($general_ledger_accounts['stock_gl_account']);
				return;
			}
			if(count($general_ledger_accounts) > 0) {
				$this->general_ledger_account = $general_ledger_accounts[0]['stock_gl_account'];
			}
			if(count($general_ledger_accounts) != 1) {
				$this->invoice->set_comments($this->invoice->get_comments() . "\n" . 
				lang('Not sure about the general ledger account for %1', $this->description));
			}
		}

		/** 
		*	Check that this line is valid.
		*	Specifically:
		*	- Unit price != 0 or Quantity != 0
		*	- General Ledger Account != 0
		*	@return boolean true if this line passes, otherwise error message
		*/
		public function check() {
			if($this->unit_price == 0 && $this->quantity == 0) {
				return lang('Unit price or quantity must not be 0');
			}
			if(is_null($this->general_ledger_account) || $this->general_ledger_account == 0) {
				return lang('General Ledger account is not set for line #%1: %2.  If the line is for a stock item, check the Stock Category\'s settings.  
					If you do not know which account to select, contact your accounting department.',
					$this->get_line_number() + 1,
					$this->get_description()
				);
			}
			return true;
		}

		/***********************
		*	Accessor methods
		*	@{
		*/
		
		/**
		*	Get the owner of this InvoiceLine
		*	@return Invoice
		*/
		public function &get_invoice() {
			return $this->invoice;
		}

		/**
		*	Get the line number of this InvoiceLine
		*	@return int line number
		*/
		public function get_line_number() {
			return (int)$this->line_number;
		}

		/**
		*	Set the line number of this InvoiceLine.  This
		*	should really only be called from Invoice, used when 
		*	constructing an InvoiceLine from an InvoicableItem.
		*	@param int line number
		*/
		public function set_line_number($line_number) {
			$this->line_number = $line_number;
		}

		/**
		*	Get the quantity of this InvoiceLine
		*	@return float quantity
		*/
		public function get_quantity() {
			return $this->quantity;
		}

		/**
		*	Set the quantity of this InvoiceLine
		*	@param quantity
		*/
		public function set_quantity($quantity) {
			if(getType($quantity) == 'string') {
				$quantity = (float)trim($quantity);
			}
			$this->quantity = $quantity;
		}

		/**
		*	Get the description of this InvoiceLine
		*	@return string description
		*/
		public function get_description() {
			return $this->description;
		}

		/**
		*	Set the description of this InvoiceLine
		*	@param description
		*/
		public function set_description($description) {
			$this->description = $description;
		}

		/**
		*	Set the unit price for this InvoiceLine
		*	@param price
		*/
		public function set_unit_price($price) {
			if(getType($price) == 'string') {
				$price = (float)trim($price);
			}
			$this->unit_price = $price;
		}

		/**
		*	Get the price for one of the item on this InvoiceLine
		*	@return float unit price
		*/
		public function get_unit_price() {
			return $this->unit_price;
		}

		/**
		*	Set the general ledger account for this InvoiceLine
		*	@param gl_account int General Ledger account number
		*/
		public function set_general_ledger_account($gl_account) {
			$this->general_ledger_account = $gl_account;
		}

		/**
		*	Get the general ledger account for this InvoiceLine
		*	@return int the General Ledger account number
		*/
		public function get_general_ledger_account() {
			return $this->general_ledger_account;
		}

		/**
		*	Get the InvoiceLine type for this InvoiceLine.
		*	One of the constants defined for InvoiceLine.
		*/
		public function get_line_type() {
			return $this->line_type;
		}

		/**
		*	Get the Invoicable object
		*/
		public function &get_invoicable_item() {
			return $this->invoicable_item;
		}

		/**
		*	Change the invoicable object
		*/
		public function set_invoicable_item(&$item) {
			$this->invoicable_item =& $item;
		}
		
		/**
		*	Get the line number of the General Ledger transaction
		*	that came from posting the invoice
		*/
		public function get_transaction_line() {
			return $this->gl_transaction_line;
		}

		/**
		*	Get our location_id that is involved with this invoice line.
		*	Check if the associated invoicable_item can give more detail,
		*	such as a sub-location first, otherwise, return the billing
		*	location.
		*/
		public function get_our_location() {
			if(!$this->invoicable_item) {
				return $this->invoice->get_our_location();
			}
			$owner = $this->invoicable_item->get_owner();
			if($owner->get_supplier_id() && method_exists($owner, 'get_ship_to_address')) {
				$address = $owner->get_ship_to_address();
			} elseif($owner->get_client_id() && method_exists($owner, 'get_ship_from_address')) {
				$address = $owner->get_ship_from_address();
			} else {
				return $this->invoice->get_our_location();
			}
			$GLOBALS['egw']->db->select('perp_location',
				array('location_id'),
				array('delivery_address' => $address),
				__LINE__, __FILE__, false, '', 'perp_inventory', 1
			);
			if($row = $GLOBALS['egw']->db->row(true)) {
				return $row['location_id'];
			} else {
				return null;
			}
		}

		/***********************
		*	End of accessor methods
		*	@}
		*/

		/**
		*	Return an array of this OrderLine's properties
		*	@return array
		*/
		public function toArray() {
			return array(
				'line_number'	=>	$this->get_line_number(),
				// Silly humans, using 1 based indexing
				'human_line_number'	=>	$this->get_line_number() + 1,
				'quantity'	=>	$this->quantity,
				'unit_price'	=>	$this->unit_price,
				'description'	=>	$this->description,
				'line_type'	=>	$this->line_type,
				'general_ledger_account'	=>	$this->general_ledger_account,
				'gl_transaction_line'	=>	$this->gl_transaction_line,
			);
		}

		/**
		*	Return a string representation of this object.
		*       Very useful for debugging purposes, not much else.
		*
		*       @return string Representation of this InvoiceLine
		*/
		public function __toString() {
			return sprintf("%-25s\t(%d)%8.3f\t@%8.2f =%16.2f %s",
				$this->description,
				$this->general_ledger_account,
				$this->quantity,
				$this->unit_price,
				$this->unit_price * $this->quantity,
				(!is_null($this->invoicable_item) && method_exists($this->invoicable_item, '__toString') ? "\n\t" . $this->invoicable_item->__toString() : '')
			);
		}

		/**
		*	Prepare this object for serialization
		*
		*	@return array Properties to preserve across serialization
		*/
		public function __sleep() {
			$properties = $this->toArray();
			$properties['invoice'] = true;
			$properties['invoicable_item'] = true;
			return array_keys($properties);
		}

	}

	/**
	*	InvoicableItems can cause issues with the class not being defined.
	*	If no __autoload() is defined, we'll define one here.  It calls perp_api::bo_perp_api::__autoload.
	*
	*	@see perp_api::bo_perp_api::__autoload()
	*/
	if(!function_exists('__autoload')) {
		function __autoload($classname) {
			if(!is_object($GLOBALS['perp']['bo'])) {
				$GLOBALS['perp']['bo'] = CreateObject('perp_api.bo_perp_api');
			}
			return $GLOBALS['perp']['bo']->__autoload($classname);
		}
	}
?>
