<?php
	require_once( EGW_INCLUDE_ROOT . '/perp_api/inc/class.bo_perp_api.inc.php');

	require_once('class.Shippable.inc.php');
	require_once('class.OrderLine.inc.php');

	class Order implements Shippable {
		/**
		*	Database table
		*/
		const order_table = 'perp_order';

		/**
		*	Unique system identifier for the order (integer)
		*/
		protected $order_id;

		/**
		*	Our internal reference number for the order (string)
		*/
		protected $our_reference_number;

		/**
		*	The other party's reference number for the order (string)
		*/
		protected $their_reference_number;

		/**
		*	The system user id of the person who originated the order.
		*/
		protected $originating_user;

		/**
		*	The system user id of the person who approved the order.
		*	This would be a manager or someone with the authority to allow the order to
		*	proceed.
		*/
		protected $approving_user;

		/**
		*	The addressbook id of the contact person in our organization for this Order.
		*/
		protected $our_contact_person;

		/**
		*	The addressbook identification of the contact person in their organization
		*	for this order.  This will usually be a purchaser or a sales representitive.
		*/
		protected $their_contact_person;

		/**
		*	System identifier for the organisation supplying the items requested on this Order.
		*	This will be either a Supplier's ID, or 0 if we are the supplier.
		*/
		protected $supplier_id;

		/**
		*	System identifier for the organisation receiving the items requested on this Order.
		*	This will be either a Clients's ID, or 0 if we are the client.
		*/
		protected $client_id;

		/**
		*	The date (and time) this Order was originated (created)
		*/
		protected $originating_date = null;

		/**
		*	The date (and time) this Order was approved by the approving_user
		*/
		protected $approval_date = null;

		/**
		*	The date (and time) this Order was issued.  This is when the other parties were notified of 
		*	the order.  That is, when the Order was actually made.
		*/
		protected $issue_date = null;

		/**
		*	Requested delivery date (and time)
		*/
		protected $requested_delivery_date;

		/**
		*	Actual delivery date (and time)
		*/
		protected $actual_delivery_date;

		/**
		*	The addressbook id of the sending location
		*/
		protected $ship_from_address;

		/**
		*	The addressbook id of the destination location
		*/
		protected $ship_to_address;

		/**
		*	The addressbook id of the where to send the invoice
		*/
		protected $bill_to_address;

		/**
		*	Sometimes orders do not require approval before proceeding, depending on the organization.
		*/
		protected $requires_approval = true;

		/**
		*	The currency this Order is being conducted in.
		*/
		protected $order_currency;

		/**
		*	Requested / Expected payment terms.  The supplier, which may be us, will use whatever they want, but
		*	it will start as what we think it should be.
		*/
		protected $payment_terms;

		/**
		*	Free form notes on this Order
		*/
		protected $comments;

		/**
		*	Date this order was closed
		*/
		protected $closed_date = null;

		/**
		*	Array of OrderLines, one for each item on this Order
		*/
		protected $order_lines = array();

	
		/**
		*	Construct an Order
		*	@param order_id	Optional.  If an order_id for an existing Order is provided, the Order will be loaded.
		*	@return If order_id is provided, it will return whatever Order::read() returns.  Otherwise, true.
		*/
		public function __construct($order_id = null) {

			if(!is_null($order_id)) {
				return $this->read($order_id);
			}

			/**
			*	Set some defaults
			*/
			$this->originating_date = date('U');
			$this->issue_date = date('U');
			$this->requested_delivery_date = date('U');

			// Set originator to the current user 
			$this->set_originating_user($GLOBALS['egw_info']['user']['account_id']);
			
			// Default contact person to current user
			$this->set_our_contact_person((int)$GLOBALS['egw_info']['user']['person_id']);

			// Check to see if approval is required
			$config = config::read($GLOBALS['egw_info']['flags']['currentapp']);
			if($config) {
				$this->requires_approval = $config['orders_require_approval'];
			}

			return true;
		}

		/**
		*	Read the Order identified by order_id into this Order
		*	@param order_id The unique system identifier for the Order to read.
		*/
		public function read($order_id) {
			$GLOBALS['egw']->db->select(self::order_table,
				array(
					'order_id',
					'our_reference_number',
					'their_reference_number',
					'originating_user',
					'approving_user',
					'our_contact_person',
					'their_contact_person',
					'supplier_id',
					'client_id',
					'originating_date',
					'issue_date',
					'approval_date',
					'requested_delivery_date',
					'actual_delivery_date',
					'ship_from_address',
					'ship_to_address',
					'bill_to_address',
					'requires_approval',
					'order_currency',
					'payment_terms',
					'comments',
					'closed_date'
				),
				array('order_id'	=>	$order_id),
				__LINE__, __FILE__, false, '', 'perp_api'
			);

			$row = $GLOBALS['egw']->db->row(true);

			$this->order_id			=	$row['order_id'];
			$this->our_reference_number	=	$row['our_reference_number'];
			$this->their_reference_number	=	$row['their_reference_number'];
			$this->originating_user		=	$row['originating_user'];
			$this->approving_user		=	$row['approving_user'];
			$this->our_contact_person	=	$row['our_contact_person'];
			$this->their_contact_person	=	$row['their_contact_person'];
			$this->supplier_id		=	$row['supplier_id'];
			$this->client_id		=	$row['client_id'];
			$this->originating_date		=	$GLOBALS['egw']->db->from_timestamp($row['originating_date']);
			$this->issue_date		=	$GLOBALS['egw']->db->from_timestamp($row['issue_date']);
			$this->approval_date		=	$GLOBALS['egw']->db->from_timestamp($row['approval_date']);
			$this->requested_delivery_date	=	$GLOBALS['egw']->db->from_timestamp($row['requested_delivery_date']);
			$this->actual_delivery_date	=	$GLOBALS['egw']->db->from_timestamp($row['actual_delivery_date']);
			$this->ship_from_address	=	$row['ship_from_address'];
			$this->ship_to_address		=	$row['ship_to_address'];
			$this->bill_to_address		=	$row['bill_to_address'];
			$this->requires_approval	=	$row['requires_approval'];
			$this->order_currency		=	$row['order_currency'];
			$this->payment_terms		=	$row['payment_terms'];
			$this->comments			=	$row['comments'];
			$this->closed_date		=	$GLOBALS['egw']->db->from_timestamp($row['closed_date']);

			// Read rows
			$GLOBALS['egw']->db->select(OrderLine::order_line_table,
				array(
					'line_no AS line_number',
					'stock_id',
					'supplier_stock_code',
					'our_stock_code',
					'description',
					'quantity_ordered',
					'unit_price',
					'ship_location'
				),
				array('order_id'	=>	$order_id),
				__LINE__, __FILE__, false, 
				'ORDER BY order_id, line_no', 'perp_api'
			);

			$line_class = get_class($this) . 'Line';
			$line_data = array();

			while($row = $GLOBALS['egw']->db->row(true)) {
				$line_data[] = $row;
			}
			foreach($line_data as $row) {
				$line = new $line_class($this, $row);
				$this->order_lines[$row['line_number']] = $line;
				$line->read();
			}
		}

		/**
		*	Check this Order and write it into the database
		*	Throws exceptions if the Order is not valid.  
		*/
		public function write() {
			$result = $this->check();
			if($result !== True) {
				throw new Exception('Order can not be saved.  (' . $result . ')');
			}

			$GLOBALS['egw']->db->transaction_begin();

			// Write Order
			$GLOBALS['egw']->db->insert(self::order_table,
				$this->toArray(false) + array(
					'class'	=>	get_class($this)
				),
				array(
					'order_id'	=>	$this->get_order_id()
				),
				__LINE__, __FILE__, 'perp_api'
			);
			$this->order_id = $this->order_id ? $this->order_id : $GLOBALS['egw']->db->get_last_insert_id(self::order_table, 'order_id');
			
			// Write lines
			$GLOBALS['egw']->db->delete(OrderLine::order_line_table,
				array(
					'order_id'	=>	$this->get_order_id()
				),
				__LINE__, __FILE__, 'perp_api'
			);
			foreach($this->order_lines as $line_no => $line) {
				$line->write($line_no);
			}

			$GLOBALS['egw']->db->transaction_commit();
			return $this->order_id;
		}

		/**
		*	Permissions functions
		*	@{
		*/

		/**
		*	Check if the given user can edit this Order
		*	@param user_id	User ID of the user to check
		*/
		public function is_editor($user_id) {
			$edit = false;
			if($user_id == $this->get_originating_user()) {
				$edit = !(boolean)$this->get_approval_date();
			}
			return $edit || $this->is_admin($user_id);
		}

		/**
		*	Check if the given user can administer this Order
		*	This includes approving, denying, editing after 
		*	approval
		*	@param user_id	User ID of the user to check
		*/
		public function is_admin($user_id) {
			return array_key_exists($user_id, $this->get_admin_list());
		}
		
		/**
		*	Get a list of users allowed to administer this Order
		*/
		public function get_admin_list() {
			$users = array();
			if(!$this->requires_approval) {
				$users[$this->get_originating_user()] = $GLOBALS['egw']->accounts->id2name($this->get_originating_user());
			}
			$approvers = $GLOBALS['egw']->acl->get_ids_for_location('order', EGW_ACL_CUSTOM_1);
			if(!is_array($approvers)) {
				return $users;
			}
			foreach($approvers as $id) {
				if($id != 0) {
					$users[$id] = $GLOBALS['egw']->accounts->id2name($id);
				}
				if($id < 0) {
					$users += $GLOBALS['egw']->accounts->members($id);
				}
			}
			return $users;
		}

		/**	@}	*/

		/**
		*	Check that this order is valid.  Specifically:
		*	- There is at least one OrderLine.
		*	- Ship To address is set.
		*	- Each OrderLine passes its check
		*
		*	@return boolean true if the order is valid, otherwise returns error message
		*/
		public function check() {
			$pass_check = true;
			if(is_null($this->supplier_id) && is_null($this->client_id)) {
				$pass_check = false;
				return lang('Supplier or client is not set.');
			}
			if(count($this->order_lines) == 0) {
				$pass_check = false;
				return lang('Empty order.  Orders need at least one OrderLine.');
			}
			if(!is_numeric($this->ship_to_address) || $this->ship_to_address == 0) {
				$pass_check = false;
				return lang('Ship to address is not set.');
			}
			
			foreach($this->order_lines as $key => $line) {
				$reason = $line->check();
				if($reason !== true) {
					return $reason;
				}
				$pass_check = true;
			}

			return (boolean)$pass_check;
		}

		/**
		*	Does this order need to be approved?  The organization
		*	may not require it.
		*/
		public function get_approval_required() {
			if($this->requires_approval) {
				return !(boolean)$this->approval_date;
			} else {
				return false;
			}
		}

		/**
		*	Check that this order can be approved.  This requires that Order::check() passes, as well as:
		*	- This Order has not been approved yet
		*	- Our contact person is set
		*	- Our reference number is set
		*	- Each OrderLine passes its approval_check()
		*	
		*	If it passes, and approval is not required, the approving user is set to the 
		*	originating user
		*
		*	@return boolean true if the order can be approved.
		*/
		public function approval_check() {
			$pass_check = $this->check();
			if($pass_check !== true) {
				return $pass_check;
			}
			
			if($this->approval_date) {
				$pass_check = false;
				return lang('Order has already been approved.');
			}
			if(is_null($this->our_contact_person)) {
				return lang('Contact person is not set.');
			}
			if(is_null($this->our_reference_number)) {
				$pass_check = false;
				return lang('Our reference number is not set.');
			}
				
			foreach($this->order_lines as $key => $line) {
				$line_check = $line->approval_check();
				if($line_check !== true) {
					return $line_check;
				}
			}

			if($pass_check === true && (!$this->requires_approval || ($this->requires_approval && !$this->approving_user))) {
				$this->approving_user = $this->get_originating_user();
			}
			return $pass_check;
		}

		/**
		*	Approve this order
		*	@param approving_user_id The system's user id of the person who approved this Order
		*	@throws Exception If the Order cannot be approved.
		*/
		public function approve($approving_user_id) {
			$check = $this->approval_check();
			if($check !== true) {
				throw new Exception(lang('Order could not be approved (%1)', $check));
			}
			if(!$approving_user_id) {
				throw new Exception(lang('No user id provided to approve'));
			}
			$this->approving_user = $approving_user_id;
			$this->approval_date = date('U');
			$this->write();
		}

		/**
		*	Deny this order.  Orders cannot be denied after being approved.  Denied Orders
		*	are immediately cancelled.
		*	@param denying_user_id The system's user id of the person who denied this Order
		*	
		*/
		public function deny($denying_user_id) {
			if($this->approval_date) {
				throw new Exception(lang('Order has already been approved'));
			}
			return $this->cancel($denying_user_id);
		}

		/**
		*	Cancel this order.  Orders cannot be cancelled if some items or money have already
		*	changed hands.  
		*	@throws Exception If the Order cannot be cancelled
		*	@param cancelling_user_id The system's user id of the person who cancelled this order
		*/
		public function cancel($cancelling_user_id) {
			$test = $this->can_cancel($cancelling_user_id);
			if($test !== true) {
				throw new Exception($test);
			}

			$GLOBALS['egw']->db->transaction_begin();
			$GLOBALS['egw']->db->delete(self::order_table,
				array('order_id' => $this->order_id),
				__LINE__, __FILE__
			);
			try {
				foreach($this->order_lines as $line) {
					$line->cancel();
				}
			} catch (Exception $e) {
				$GLOBALS['egw']->db->transaction_abort();
				throw $e;
			}
			$GLOBALS['egw']->db->transaction_commit();
		}

		/**
		*	Can this order be cancelled?  Orders cannot be cancelled if some items or money have already
		*	changed hands.
		*/
		public function can_cancel($cancelling_user_id) {
			if(!$this->is_admin($cancelling_user_id)) {
				return lang('Access is not allowed');
			}
			if($this->is_complete()) {
				return lang('Order is complete');
			}
			foreach($this->order_lines as $order_line) {
				$test = $order_line->can_cancel();
				if($test !== true) {
					return $test;
				}
			}
			return true;
		}

		/**
		*	Is this Order completed (closed) ?
		*	@return boolean
		*/
		public function is_complete() {
			$test = (count($this->order_lines) > 0);
			foreach($this->order_lines as $line) {
				$test = $test && $line->is_complete();
			}
			return $test;
		}

		/**
		*	Can this Order be shipped?
		*	@return boolean true or reason why not
		*/
		public function can_ship() {
			$test = false;
			$reason = '';
			foreach($this->order_lines as $line) {
				$reason = $line->can_ship();
				$test = $test || (is_bool($reason) ? ($reason === true) : false);
			}
			// if $reason == '', lines are all complete (returned false)
			if($test === false) {
				if($reason != '') {
					return $reason;
				} else {
					return false;
				}
			}
			if($this->get_approval_required()) {
				return lang('Order needs to be approved before shipping.');
			}
			return $test;
		}

		/**
		*	Can this Order be received?
		*	@return boolean
		*/
		public function can_receive() {
			$test = (count($this->order_lines) > 0);
			foreach($this->order_lines as $line) {
				$test = ($test && $line->can_receive());
			}
			$test = $test && (($this->requires_approval && !is_null($this->get_approval_date())) || !$this->requires_approval);
			$test = $test && $GLOBALS['egw']->acl->check('receive', EGW_ACL_EDIT, $GLOBALS['egw_info']['flags']['currentapp']);
			return $test;
		}

		/**
		*	Can this Order be invoiced?
		*	@return boolean
		*/
		public function can_invoice() {
			$test = false;
			foreach($this->order_lines as $line) {
				$test = $test || $line->can_invoice();
			}
			$test = $test && !$this->get_approval_required();
			$test = $test && $GLOBALS['egw']->acl->check('invoice', EGW_ACL_ADD, $GLOBALS['egw_info']['flags']['currentapp']);
			return $test;
		}

		/**
		*	Add an OrderLine to this Order
		*	@todo Add some checks on status before adding
		*	@param orderline OrderLine to add to this Order
		*	@return int Line identification number
		*/
		public function add_line(OrderLine &$orderline) {
			$this->order_lines[] =& $orderline;
			$orderline->set_line_number(end(array_keys($this->order_lines)));
			return $orderline->get_line_number();
		}

		/**
		*	Remove an OrderLine from this Order
		*	@todo Add some checks on status before removing
		*	@param line_no Identification number of the line to remove
		*/
		public function remove_line($line_no) {
			$order_line = $this->order_lines[$line_no];
			if($order_line && $order_line->can_cancel()) {
				// These will be removed by remove(), so get them now
				$associated_lines = $order_line->get_detail('associated_line');

				$order_line->remove();
				unset($this->order_lines[$line_no]);

				if($associated_lines) {
					foreach($associated_lines as $detail) {
						$this->remove_line($detail->get_value());
					}
				}
			}
		}

		/**
		*	Return a reference to an OrderLine specified by the Identification number.
		*
		*	@param line_no Identification number of the line to return
		*/
		public function get_line($line_no) {
			return $this->order_lines[$line_no];
		}

		/**
		*	Returns all OrderLines
		*	@return array of OrderLine
		*/
		public function get_lines() {
			return $this->order_lines;
		}

		/**
		*	Returns the monetary total of the Order
		*	@return float Order total
		*/
		public function get_total() {
			$total = 0;
			foreach ($this->order_lines as $order_line) {
				$total += $order_line->get_unit_price() * $order_line->get_quantity_ordered();
			}
			return $total;
		}

		/**
		*	Accessors
		*	@{
		*/

		/**
		*	Get the unique system identifier for this Order
		*	@return int order_id
		*/
		public function get_order_id() {
			return $this->order_id;
		}

		/**
		*	Get the organization supplying the items requested on this Order.
		*	This will be either a Supplier's ID, or 0 if we are the supplier.
		*	@return int supplier_id
		*/
		public function get_supplier_id() {
			return $this->supplier_id;
		}

		/**
		*	Set the organization supplying the items requested on this Order.
		*	This will be either a Supplier's ID, or 0 if we are the supplier.
		*	@param supplier_id ID of the supplier
		*/
		public function set_supplier_id($supplier_id) {
			if(!is_integer($supplier_id)) {
				return lang('Invalid supplier');
			}
			$this->supplier_id = $supplier_id;

			if($this->supplier_id != 0) {
				$supplier = ExecMethod('perp_ap.bo_perp_supplier.get_supplier', $this->supplier_id);
			}

			// Default currency 
			if(!$this->order_currency) {
				if($this->supplier_id != 0) {
					$this->set_order_currency($supplier['supplier_currency']);
				}
			}
			// Default terms 
			if(!$this->payment_terms) {
				if($this->supplier_id != 0) {
					$this->set_payment_terms((int)$supplier['payment_terms']);
				}
			}
			// Default ship_from address
			if(!$this->ship_from_address) {
				if($this->supplier_id != 0) {
					$this->set_ship_from_address((int)$supplier['supplier_address']);
				}
			}
		}

		/**
		*	Get the organization receiving the items requested on this Order.
		*	This will be either a Client's ID, or 0 if we are the client.
		*	@return int client_id
		*/
		public function get_client_id() {
			return $this->client_id;
		}

		/**
		*	Set the organization receiving the items requested on this Order.
		*	This will be either a Client's ID, or 0 if we are the client.
		*	@param client_id ID of the client
		*/
		public function set_client_id($client_id) {
			if(!is_integer($client_id)) {
				return lang('Invalid client');
			}
			$this->client_id = $client_id;
		}


		/**
		*	Set our human reference number for this Order
		*	@param reference string our_reference_number
		*/
		public function set_our_reference_number($reference) {
			$this->our_reference_number = $reference;
		}

		/**
		*	Get our human reference number for this Order
		*	@return string our_reference_number
		*/
		public function get_our_reference_number() {
			return $this->our_reference_number;
		}

		/**
		*	Set their human reference number for this Order
		*	@param reference string their_reference_number
		*/
		public function set_their_reference_number($reference) {
			$this->their_reference_number = $reference;
		}

		/**
		*	Get their human reference number for this Order
		*	@return string their_reference_number
		*/
		public function get_their_reference_number() {
			return $this->their_reference_number;
		}

		/**
		*	Set the eGroupWare user id of the originating user.
		*	@param user_id int System user_id
		*/
		public function set_originating_user($user_id) {
			if(!is_integer($user_id)) {
				return lang('Invalid user id for originating user');
			}
			$this->originating_user = $user_id;
		}

		/**
		*	Get the eGroupWare user id of the originating user
		*	@return int user id
		*/
		public function get_originating_user() {
			return $this->originating_user;
		}

		/**
		*	Set the eGroupWare user id of the user who has / will approve this order.
		*	@param user_id int System user_id
		*/
		public function set_approving_user($user_id) {
			if(!is_integer($user_id)) {
				return lang('Invalid user id for approving user');
			}
			$this->approving_user = $user_id;
		}

		/**
		*	Get the eGroupWare user id of the originating user
		*	@return int user id
		*/
		public function get_approving_user() {
			return $this->approving_user;
		}

		/**
		*	Set the date this Order was created
		*	@param date int Unix time stamp
		*/
		public function set_originating_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for originating date');
			}
			$this->originating_date = $date;
		}
		/**
		*	Get the date this Order was created
		*	@return int Unix time stamp
		*/
		public function get_originating_date() {
			return $this->originating_date;
		}

		/**
		*	Set the date this Order was issued
		*	@param date int Unix time stamp
		*/
		public function set_issue_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for issue date');
			}
			$this->issue_date = $date;
		}

		/**
		*	Get the date this Order was issued
		*	@return int Unix time stamp
		*/
		public function get_issue_date() {
			return $this->issue_date;
		}

		/**
		*	Get the date this Order was approved
		*	@return int Unix time stamp
		*/
		public function get_approval_date() {
			return $this->approval_date;
		}

		/** 
		*	Get whether this Order requires approval or not.
		*	This is different from if it has been approved or not.  Check
		*	approval date to see if the order has been aproved or not, requires_approval
		*	is only to check that the Order required approval.
		*	@return boolean requires approval
		*/
		public function requires_approval() {
			return $this->requires_approval;
		}

		/**
		*	Set the requested delivery date for this Order
		*	@param date int Unix time stamp
		*/
		public function set_requested_delivery_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for requested delivery date');
			}
			$this->requested_delivery_date = $date;
		}

		/**
		*	Get the requested date for delivery
		*	@return int Unix time stamp
		*/
		public function get_requested_delivery_date() {
			return $this->requested_delivery_date;
		}

		/**
		*	Set the actual delivery date for this Order
		*	@param date int Unix time stamp
		*/
		public function set_actual_delivery_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for actual delivery date');
			}
			$this->actual_delivery_date = $date;
		}

		/**
		*	Get the actual date of delivery
		*	@return int Unix time stamp
		*/
		public function get_actual_delivery_date() {
			return $this->actual_delivery_date;
		}

		/**
		*	Set the close date for this Order
		*	@param date int Unix time stamp
		*/
		public function set_closed_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for closed date');
			}
			$this->closed_date = $date;
		}

		/**
		*	Get the close date of this order
		*	@return int Unix time stamp
		*/
		public function get_closed_date() {
			return $this->closed_date;
		}

		/**
		*	Get the person in our organization who is to be contacted
		*	regarding this Order.
		*	@return int addressbook_id
		*/
		public function get_our_contact_person() {
			return $this->our_contact_person;
		}

		/**
		*	Set the person in our organization who is to be contacted
		*	regarding this Order.
		*	@param contact Addressbook ID of the contact
		*	@return true or error message
		*/
		public function set_our_contact_person($contact) {
			if(!is_integer($contact)) {
				return lang('Invalid contact person.');
			}
			$this->our_contact_person = $contact;
			return true;
		}

		/**
		*	Get the person in their organization who is to be contacted
		*	regarding this Order.
		*	@return int addressbook_id
		*/
		public function get_their_contact_person() {
			return $this->their_contact_person;
		}

		/**
		*	Set the person in their organization who is to be contacted
		*	regarding this Order.
		*	@param contact Addressbook ID of the contact
		*	@return true or error message
		*/
		public function set_their_contact_person($contact) {
			if(!is_integer($contact)) {
				return lang('Invalid contact person.');
			}
			$this->their_contact_person = $contact;
			return true;
		}

		/**
		*	Get the location the OrderLines are coming from for this Order
		*	@return int addressbook ID of the location
		*/
		public function get_ship_from_address() {
			return $this->ship_from_address;
		}

		/**
		*	Set the location the OrderLines are coming from for this Order
		*	@param location Addressbook ID of the location
		*/
		public function set_ship_from_address($location) {
			if(!is_integer($location)) {
				return lang('Invalid ship from address.');
			}
			$this->ship_from_address = $location;
		}

		/**
		*	Get the location the OrderLines are going to for this Order
		*	@return int addressbook ID of the location
		*/
		public function get_ship_to_address() {
			return $this->ship_to_address;
		}

		/**
		*	Set the location the OrderLines are going to for this Order
		*	@param location Addressbook ID of the location
		*/
		public function set_ship_to_address($location) {
			if(!is_integer($location)) {
				return lang('Invalid ship to address.');
			}
			$this->ship_to_address = $location;
		}

		/**
		*	Get where to send the invoice for this  Order
		*	@return int addressbook ID of the location
		*/
		public function get_bill_to_address() {
			return $this->bill_to_address;
		}

		/**
		*	Set where to send the invoice for this  Order
		*	@param location Addressbook ID of the location
		*/
		public function set_bill_to_address($address) {
			if(!is_integer($address)) {
				return lang('Invalid bill to address.');
			}
			$this->bill_to_address = $address;
		}

		/**
		*	Get our location involved in this Order.
		*	This will be either the source or destination,
		*	depending on the order.
		*	@return int Inventory location ID or false if it can't be found.
		*/
		public function get_location_id() {
			if($this->supplier_id == 0) {
				$address = $this->ship_from_address;
			} else {
				$address = $this->ship_to_address;
			}
			$links = egw_link::get_links('addressbook', $address, 'perp_inventory');
			$key = current($links);
			if($key) {
				list($location, $location_id) = explode('-', $key);
				return $location_id;
			}
			return false;
		}

		/**
		*	Get the currency this Order is being conducted in
		*	@return string three letter currency code
		*/
		public function get_order_currency() {
			return $this->order_currency;
		}

		/**
		*	Set the currency this Order is being conducted in
		*	@todo Better parameter checking
		*	@param currency Three letter currency code from currency list
		*/
		public function set_order_currency($currency) {
			if(!is_string($currency)) {
				return lang('Invalid currency.');
			}
			$this->order_currency = $currency;
		}

		/**
		*	Get the payment terms for this Order
		*	@return int payment_term_id
		*/
		public function get_payment_terms() {
			return $this->payment_terms;
		}

		/**
		*	Set the payment terms for this Order
		*	@todo Better parameter checking
		*	@param payment_term_id
		*/
		public function set_payment_terms($payment_term_id) {
			if(!is_integer($payment_term_id)) {
				return lang('Invalid payment terms.');
			}
			$this->payment_terms = $payment_term_id;
		}

		/**
		*	Get comments on this Order
		*	@return string comments
		*/
		public function get_comments() {
			return $this->comments;
		}

		/**
		*	Set comments on this Order
		*	@param comments 
		*/
		public function set_comments($comments) {
			$this->comments = $comments;
		}

		/** @} */


		/**
		*	Methods required to implement Invoicable
		*	@{
		*/

		/**
		*	Address the Invoice comes from
		*	@return  Addressbook ID of the location
		*/
		public function get_bill_from_address() {
			return $this->get_ship_from_address();
		}

		/**
		*	Currency the Invoice is in
		*	@return 3 letter currency code
		*/
		public function get_invoice_currency() {
			return $this->get_order_currency();
		}

		/**
		*	Get items to be invoiced
		*	@return order lines
		*/
		public function &get_invoicable_items() {
			return $this->order_lines;
		}

		/**
		*	Get items to be shipped
		*	@return order lines
		*/
		public function &get_shippable_items() {
			return $this->order_lines;
		}

		/**
		*	Return an array of this Order's properties
		*	Used for display in eTemplates and such
		*	@param get_lines	boolean Include order lines in the array
		*/
		public function toArray($get_lines = false) {
			$data = get_object_vars($this);
			unset($data['order_lines']);
			if($get_lines) {
				$data['order_lines'] = array();
				foreach($this->order_lines as $key => $line) {
					$line_array = $line->toArray();
					$line_array['line_no'] = $key;
					$data['order_lines'][] = $line_array;
				}
			}
			return $data;
		}

		/**     
		*	Return a string representation of this object.
		*	Very useful for debugging purposes, not much else.
		*
		*	@return string Representation of this Order
		*/      
		public function __toString() {
			$result = $GLOBALS['egw']->accounts->get_account_name($this->get_originating_user(), $id, $fname, $lname);
			$originating_user = $fname . ' ' . $lname . ' [' . $id . ']';
			unset($fname);
			unset($lname);
			unset($id);

			$string = sprintf(get_class($this) . " ID: %d\t Our Ref #: %s started %s by %s\n",
			$this->get_order_id(),
			$this->get_our_reference_number(),
			date($GLOBALS['egw_info']['user']['preferences']['common']['dateformat'], $this->get_originating_date()),
			$originating_user);

			$string .= lang('supplier') . ': ' . $this->get_supplier_id() . "\n";
			$string .= lang('client') . ': ' . $this->get_client_id() . "\n";

			foreach($this->order_lines as $line_no => $orderline) {
				$string .= "\t$line_no.\t" . $orderline->__toString();
			}
			return $string;
		}

		/**
		*	Prepare this object for serialization
		*
		*	@return array Properties to preserve across serialization.
		*/
		public function __sleep() {
			$properties =  array('order_id', 'our_reference_number', 'their_reference_number', 
				'originating_user', 'approving_user', 
				'our_contact_person', 'their_contact_person',
				'supplier_id', 'client_id',
				'originating_date', 'approval_date', 'issue_date', 
				'requested_delivery_date', 'actual_delivery_date',
				'ship_to_address', 'ship_from_address', 'bill_to_address',
				'requires_approval',
				'order_currency', 'payment_terms',
				'comments','closed_date',
				'order_lines'
			);
			return $properties;
		}

		/**
		*	Suggest our reference number (human reference) based on 
		*	system settings stored in eGroupWare config variables for
		*	this Order.  Depending on permissions, the user may override
		*	the suggested number.
		*	
		*	@return string Suggested number based on site settings.  
		*/
		public function suggest_order_number() {
			return 'Not set';
		}
	}
?>
