<?php
	require_once('class.Order.inc.php');
	require_once('class.OrderLineDetail.inc.php');
	require_once('class.ShippableItem.inc.php');
	require_once('class.InvoicableItem.inc.php');

	class OrderLine implements ShippableItem, InvoicableItem {
		/**
		*	Database table name
		*/
		const order_line_table = 'perp_order_line';

		/**
		*	Reference to the containing Order
		*/
		protected $order;

		/**
		*	Line number on the containing Order
		*/
		protected $line_number;

		/**
		*	System identifier for a stock item in our stock list
		*/
		protected $stock_id;

		/**
		*	Other party's human reference number
		*/
		protected $supplier_stock_code;

		/**
		*	Our human reference number (stock_num)
		*/
		protected $our_stock_code;

		/**
		*	Description of this line
		*/
		protected $description;

		/**
		*	Quantity ordered in supplier's unit of measure
		*/
		protected $quantity_ordered = 0;

		/**
		*	Price for one unit (quantity 1) of the item ordered
		*/
		protected $unit_price = 0;

		/**
		*	Location to ship this OrderLine from.  
		*	0 for any / unknown, otherwise a location_id
		*/
		protected $ship_location = 0;
		
		/**
		*	Details for this OrderLine, such as color, shape, size, etc.
		*/
		protected $details = array();

		/**
		*	Construct an OrderLine
		*	
		*	@param order Order that this line is part of.
		*	@param line_data Optional.  Array of information for this OrderLine.
		*/
		public function __construct(Order &$order, Array &$line_data = null) {
			$this->order =& $order;

			$this->line_number		=	$line_data['line_number'];
			$this->stock_id			=	$line_data['stock_id'];
			$this->supplier_stock_code	=	$line_data['supplier_stock_code'];
			$this->our_stock_code		=	$line_data['our_stock_code'];
			$this->description		=	$line_data['description'];
			$this->set_quantity_ordered($line_data['quantity_ordered']);
			$this->unit_price		=	$line_data['unit_price'];
			$this->ship_location		=	$line_data['ship_location'];

			// Make sure we have our stock code, if available
			if($this->stock_id && !$this->our_stock_code) {
				$stock = ExecMethod('perp_inventory.bo_perp_stock.get_stock_item', $this->stock_id);
				$this->our_stock_code = $stock['stock_num'];
			}
		}

		/**
		*	Read this line from the database
		*	This OrderLine will already have its properties set from being read
		*	in its containing Order.  This gives the OrderLine a chance to set
		*	other properties
		*/
		public function read() {
			// Check for stock_id
			if(!$this->stock_id && $this->order->get_supplier_id()) {
				$catalog_item = ExecMethod2('perp_ap.bo_perp_catalog.get_catalog_item', $this->order->get_supplier_id(), $this->supplier_stock_number);
				if(is_array($catalog_item) && $catalog['stock_id']) {
					$this->stock_id = $catalog['stock_id'];
				}
			}

			// Get details
			$GLOBALS['egw']->db->select(OrderLineDetail::detail_table,
				array(
					'order_id',
					'line_no',
					'detail_no AS detail_number',
					'detail_class',
					'detail_name',
					'detail_value',
					'display'
				),
				array(
					'order_id'	=>	$this->order->get_order_id(),
					'line_no'	=>	$this->line_number
				),
				__LINE__, __FILE__, false,
				'ORDER BY order_id, line_no, detail_no', 'perp_api'
			);
			$list = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$list[] = $row;
			}

			foreach($list as &$detail_data) {
				$detail =& new $detail_data['detail_class']($this, $detail_data);
				$detail->read();
				$this->details[$detail_data['detail_number']] =& $detail;
			}
		}
		

		/**
		*	Write this line to the database
		*	@param line_number The Identification number of this line.
		*/
		public function write($line_number = null) {
			if(!is_null($line_number)) {
				$this->line_number = $line_number;
			}
			$GLOBALS['egw']->db->insert(self::order_line_table,
				array(
					'order_id'	=>	$this->order->get_order_id(),
					'line_no'	=>	$this->line_number,
				) + $this->toArray(),
				array(
					'order_id'	=>	$this->order->get_order_id(),
					'line_no'	=>	$this->line_number
				),
				__LINE__, __FILE__, 'perp_api'
			);
			$GLOBALS['egw']->db->delete(OrderLineDetail::detail_table,
				array(
					'order_id'	=>	$this->order->get_order_id(),
					'line_no'	=>	$this->line_number
				),
				__LINE__, __FILE__, 'perp_api'
			);
					
			foreach($this->details as $detail) {
				$detail->write();
			}
		}

		/**
		*	Check that this line is valid
		*	@todo Currently always returns true
		*	@return true or error message
		*/
		public function check() {
			return true;
		}

		/**
		*	Check that this line is valid for approval
		*	@todo Currently always returns true
		*	@return true or error message
		*/
		public function approval_check() {
			return true;
		}

		/**
		*	Get the line number of this OrderLine
		*/
		public function get_line_number() {
			return $this->line_number;
		}

		/**
		*	Set the line number of this OrderLine
		*	@param line_number int Line number
		*/
		public function set_line_number($line_number) {
			if($this->order->get_line($line_number) !== $this) {
				throw new Exception('Call $order->add_line, don\'t call set_line_number directly');
			}
			$this->line_number = $line_number;
		}

		/**
		*	Get our stock code for this item
		*
		*	@return string
		*/
		public function get_our_stock_code() {
			return $this->our_stock_code;
		}

		/**
		*	Get the supplier's code for this item
		*
		*	@return string
		*/
		public function get_supplier_stock_code() {
			return $this->supplier_stock_code;
		}

		/**
		*	Get our stock id for this item, if it is one of our stock items
		*/
		public function get_stock_id() {
			return $this->stock_id;
		}

		/**
		*	Set our stock id for this item, if it is one of our stock items
		*	@param stock_id int Stock ID
		*/
		public function set_stock_id($stock_id) {
			$this->stock_id = $stock_id;
		}

		/**
		*	Get the quantity ordered
		*	@return float quantity ordered
		*/
		public function get_quantity_ordered() {
			return $this->quantity_ordered;
		}

		/**
		*	Set the quantity ordered
		*	@param quantity Quantity ordered
		*/
		public function set_quantity_ordered($quantity) {
			$this->quantity_ordered = $quantity;
			return true;
		}

		/**
		*	Get the price for one of the items on this OrderLine
		*	@return float
		*/
		public function get_unit_price() {
			return $this->unit_price;
		}

		/**
		*	Set the price for one of the items on this OrderLine
		*	@param unit_price Unit price
		*/
		public function set_unit_price($unit_price) {
			$this->unit_price = $unit_price;
		}

		/**
		*	Get the description of this OrderLine
		*/
		public function get_description() {
			return $this->description;
		}

		/**
		*	Set the description of this OrderLine
		*	@param description Text description of this line, usually the full item name.
		*/
		public function set_description($description) {
			$this->description = $description;
			return true;
		}

		/**
		*	Get the location involved in this OrderLine
		*	shipping or receiving
		*/
		public function get_ship_location() {
			return $this->ship_location;
		}
		
		/**
		*	Set the location involved in this OrderLine
		*	@param location_id ID of a location.
		*/
		public function set_ship_location($location_id) {
			if($location_id != $this->ship_location) {
				if(!$this->ship_location) {
					$this->ship_location = '0';
				}

				// Un-reserve stock at prior location
				if($this->ship_location != 0) {
					ExecMethod2('perp_inventory.bo_perp_inventory.reserve_stock', $this->ship_location, $this->stock_id, -$this->quantity_ordered, lang('Ship location for %1 changed', $this->order->get_our_reference_number()));
				}

				// Reserve stock at new location
				if($location_id != 0) {
					ExecMethod2('perp_inventory.bo_perp_inventory.reserve_stock', $location_id, $this->stock_id, $this->quantity_ordered, lang('Sales Order %1', $this->order->get_our_reference_number()));
				}
			}
			$this->ship_location = $location_id;
			return true;
		}

		/**
		*	Any actions to perform before being removed from the Order
		*	Does not actually remove the line.
		*/
		public function remove() {
			if(!$this->can_cancel()) {
				return false;
			}

			// Remove OrderLineDetails
			foreach($this->details as &$detail) {
				$detail->remove();
			}
			$GLOBALS['egw']->db->delete(OrderLineDetail::detail_table, 
				array(
					'order_id'	=>	$this->order->get_order_id(),
					'line_no'	=>	$this->line_number,
				),
				__LINE__, __FILE__
			);
		}

		/**
		*	Any actions to perform before the Order is cancelled.
		*	Does not actually remove the line or cancel the order.
		*/
		public function cancel() {
			$this->remove();
		}

		/**
		*	Can this OrderLine be cancelled?
		*/
		public function can_cancel() {
			if(!$GLOBALS['egw']->acl->check('order', EGW_ACL_EDIT | EGW_ACL_CUSTOM_1, $GLOBALS['egw_info']['flags']['currentapp'])) {
				return false;
			}
			if($this->get_quantity_invoiced() != 0 || $this->get_quantity_shipped() != 0) {
				// Already shipped or invoiced
				return false;
			}
			/*
			* Now we check for shipped + returned, or invoice + credit.  We still need to keep
			* the line for reference. 
			*/
			if(count($this->get_status())) {
				// Something happened to this line.  Keep it.
				return false;
			}
			
			return true;
		}

		/**
		*	Get any OrderLineDetails describing this OrderLine.
		*	
		*	@param classname Optional.  Restrict returned OrderLineDetails to a specific extending class.
		*	@return array of OrderLineDetail
		*/
		public function &get_details($classname = null) {
			if($classname == null) {
				return $this->details;
			}
			$details = array();
			foreach($this->details as &$detail) {
				if($detail instanceof $classname) {
					$details[] =& $detail;
				}
			}
			return $details;
		}

		/**
		*	Get the OrderLineDetail identified by $detail_number.
		*	If $detail_number is a string, this function will return an array
		*	of OrderLineDetails that have the string as their name.
		*	@param detail_number The id of the detail to get, or the name of the detail
		*	@return OrderLineDetail identified by the number, or array of details by name
		*/
		public function &get_detail($detail_number) {
			if(is_numeric($detail_number)) {
				return $this->details[$detail_number];
			}
			$results = array();
			foreach($this->details as $key => &$detail) {
				if(!$detail) {
					// How'd this happen?
					unset($this->details[$key]);
					continue;
				}
				if($detail->get_name() == $detail_number) {
					$results[$detail->get_number()] =& $detail;
				}
			}
			return $results;
		}

		/**
		*	Add a detail describing some aspect of this OrderLine
		*/
		public function add_detail(OrderLineDetail &$detail) {
			$this->details[] =&$detail;
			$detail->set_number(end(array_keys($this->details)));
			$detail->add();
			return $detail->get_number();
		}

		/**
		*	Add a detail for the General Ledger account that this line
		*	should be on.  This allows the person approving the PO
		*	to specify the GL account for the invoice instead of accounting
		*	needing to call them.  The purchaser will usually have a better
		*	idea of which account should be used.  The account can be changed
		*	on the invoice, so this is just for a default for the invoice.
		*
		*	@param gl_account The general ledger account number to use for this OrderLine when invoiced
		*	@return null or error message if the account is not a valid General Ledger account
		*/
		public function set_gl_account($gl_account) {
			if($gl_account == '') {
				$gl_account = null;
			}

			// Check for valid account
			$accounts = ExecMethod('perp_ledger.bo_perp_ledger.get_accounts', array('search' => $gl_account));
			if(!isset($accounts[$gl_account]) && !is_null($gl_account)) {
				return lang('Invalid general ledger account %1', $account);
			}

			// Reset or set gl account number
			$details = $this->get_detail("general_ledger_account");
			if(count($details) >= 1) {
				$detail =& $details[end(array_keys($details))];
				$detail->set_value($gl_account);
				return true;
			} elseif(!is_null($gl_account)) {
				$detail = new OrderLineDetail($this, 
					array(
						'detail_name'	=>	'general_ledger_account',
						'detail_value'	=>	$gl_account,
						'display'	=>	false
					)
				);
				$this->add_detail($detail);
				return true;
			}
			if(is_null($gl_account) && is_object($detail)) {
				$this->remove_detail($detail->get_number());
				return true;
			}
			return false;
		}

		/**
		*	Get the General Ledger account that is to be used when invoicing.
		*	This account can be overridden when creating the invoice, but it
		*	makes a good default.
		*/
		public function get_gl_account() {
			$details = $this->get_detail("general_ledger_account");
			if(count($details) >= 1) {
				$detail =& $details[end(array_keys($details))];
				return $detail->get_value();
			} else {
				return null;
			}
		}

		/**
		*	Remove an OrderLineDetail from this OrderLine
		*	@param detail_number The id of the detail to get
		*/
		public function remove_detail($detail_number) {
			$detail =& $this->details[$detail_number];
			$detail->remove();
			unset($this->details[$detail_number]);
		}

		/**
		*	Add an association to another OrderLine
		*
		*	OrderLines that are associated will be removed when this OrderLine is removed.
		*	This is a two way link, as this OrderLine will be removed when any associated line is.
		*
		*	@param line_number Line number of the OrderLine to associate
		*/
		public function add_associated_line($line_number) {
			if(!is_numeric($line_number)) {
				throw new egw_exception_invalid_parameter('Wrong datatype, integer expected');
			}
			if($line_number == $this->get_line_number()) {
				return;
			}

			$this->add_detail(new OrderLineDetail($this, array(
				'detail_name'	=>	'associated_line',
				'detail_value'	=>	(int)$line_number,
				'display'	=>	false
			)));

			$other_line =& $this->get_owner()->get_line($line_number);
			if($other_line) {
				$other_line->add_detail(new OrderLineDetail($other_line, array(
					'detail_name'	=>	'associated_line',
					'detail_value'	=>	$this->get_line_number(),
					'display'	=>	false
				)));
			}
		}

		/**
		*	Remove an association to another OrderLine
		*
		*	OrderLines that are associated will be removed when this OrderLine is removed.
		*	If the association is removed, the other lines will not be removed.
		*/
		public function remove_associated_line($line_number) {
			if($line_number == $this->get_line_number()) {
				return;
			}

			$lines =& $this->get_detail('associated_line');
			foreach($lines as $key => $detail) {
				if($detail->get_value() == $line_number) {
					$this->remove_detail($key);
				}
			}
			$other_line =& $this->get_owner()->get_line($line_number);
			if($other_line) {
				$lines = $other_line->get_detail('associated_line');
				foreach($lines as $key => $detail) {
					if($detail->get_value() == $line_number) {
						$other_line->remove_detail($key);
					}
				}
			}
		}

		/**
		*	@todo Implement these
		*	@{
		*/
		public function ship(ShipmentLine &$shipment_line){}
		public function receive(ReceiptLine &$receipt_line){}

		/**
		*	Can this OrderLine be shipped?
		*	This depends on how many there are left to ship, and
		*	if payment is required first.
		*	@todo Currently, this only checks that this OrderLine has been 
		*		fully invoiced.  It should check to see if the invoice(s) 
		*		have been paid.
		*	@return boolean true or reason why not
		*/
		public function can_ship() {
			if($this->is_complete()) {
				return lang('Already complete');
			}

			// Check terms
			if(!is_object($GLOBALS['perp_api']['bo'])) {
				$GLOBALS['perp_api']['bo'] = CreateObject('perp_api.bo_perp_api');
			}
			$GLOBALS['perp_api']['bo']->get_payment_terms(array('terms' => $this->order->get_payment_terms(), 'template' => false), $terms);
			$terms = $terms[0];
			if($terms['due_days'] <= 1) {
				// Payment required
				$term_check = ($this->get_quantity_invoiced() == $this->get_quantity_ordered());
				if(!$term_check) {
					// TODO: Find and check the invoices for paid
					return lang('Order has payment terms of %1 and line %2 has not been invoiced', $terms['payment_term_name'], $this->line_number);
				}
			}
			return ($this->get_quantity_shipped() != $this->get_quantity_ordered());
		}

		/**
		*	Check to see if this ShippableItem can be received.  Depends on
		*	status and permissions.  It can't be received if it has already been completely
		*	shipped & received, or nothing shipped
		*/
		public function can_receive(){
			return !$this->is_complete() && ($this->get_quantity_received() != $this->get_quantity_shipped());
		}
		public function get_ship_quantity() {
			return $this->quantity_ordered - $this->get_quantity_shipped();
		}
		public function get_ship_unit_price() {
			return $this->get_unit_price();
		}

		public function invoice(InvoiceLine &$invoice_line) {
		}
		public function can_invoice() {
			$test = (boolean)($this->get_quantity_invoiced() != $this->quantity_ordered);
			return $test;
		}
		public function get_invoice_quantity() {
			return $this->quantity_ordered - $this->get_quantity_invoiced();
		}

		public function get_quantity_invoiced() {
			if(!$this->order->get_order_id()) {
				return 0;
			}
			$sql = 'SELECT 
				IF(quantity_invoiced IS NULL, 0, quantity_invoiced) AS quantity_invoiced
			FROM perp_order_line
			LEFT JOIN (
				SELECT 
					invoicable_class,
					invoicable_id AS order_id,
					invoicable_item_line AS line_no,
					sum(quantity) AS quantity_invoiced
				FROM perp_invoice_line
				GROUP BY invoicable_id, invoicable_item_line
			) AS invoiced_total ON
				invoiced_total.order_id = perp_order_line.order_id AND
				invoiced_total.line_no = perp_order_line.line_no
			WHERE  perp_order_line.order_id = ' . $this->order->get_order_id() .' AND
				perp_order_line.line_no = ' . $this->line_number;
			$GLOBALS['egw']->db->query($sql, __LINE__, __FILE__);
			$row = $GLOBALS['egw']->db->row(true);
			return $row['quantity_invoiced'];
		}

		public function get_owner() {
			return $this->order;
		}
		public function get_invoicable_id() {
			return $this->order->get_order_id();
		}
		public function get_invoicable_line_id() {
			return $this->line_number;
		}

		/**
		*	Return a list of extra details to be displayed on the invoice, below
		*	the line.  This list cannot be edited, and is an array of details in
		*	the form ('name' => 'Detail name', 'value' => '123')
		*
		*	@param filter The filter parameter from the display select (usually true)
		*	@return array
		*/
		public function get_invoice_details($filter) {
			return array();
		}

		public function get_shippable_id() {
			return $this->order->get_order_id();
		}
		public function get_shippable_line_id() {
			return $this->line_number;
		}
		public function get_link() {
			return array(array(
				'app'	=>	$GLOBALS['egw_info']['flags']['currentapp'],
				'id'	=>	'ORDER-' . $this->order->get_order_id()
				)
			);
		}

		/**
		*	Get quantity shipped
		*/
		public function get_quantity_shipped() {
			if(!$this->order->get_order_id()) {
				return 0;
			}

			$sql = 'SELECT 
				SUM(IF(quantity_shipped IS NULL, 0, quantity_shipped)) AS quantity_shipped
			FROM perp_order_line
			LEFT JOIN (
				SELECT 
					shippable_class,
					shippable_id AS order_id,
					shippable_item_line AS line_no,
					sum(quantity_shipped) AS quantity_shipped
				FROM perp_shipment_line
				GROUP BY order_id, perp_shipment_line.shippable_item_line
			) AS shipped_total ON
				shipped_total.shippable_class LIKE \'%Order\' AND
				shipped_total.order_id = perp_order_line.order_id AND
				shipped_total.line_no = perp_order_line.line_no
			WHERE  perp_order_line.order_id = ' . $this->order->get_order_id() . ' AND
				perp_order_line.line_no = ' . $this->line_number;
			$GLOBALS['egw']->db->query($sql, __LINE__, __FILE__);
			$row = $GLOBALS['egw']->db->row(true);
			return $row['quantity_shipped'];
		}

		/**
		*	Get the quantity of this OrderLine received to date
		*/
		public function get_quantity_received() {
			if(!$this->order->get_order_id()) {
				return 0;
			}

			$sql = 'SELECT 
				SUM(IF(quantity_units IS NULL, 0, quantity_units)) AS quantity_received
			FROM perp_order_line 
			LEFT JOIN perp_shipment_line ON
				perp_shipment_line.shippable_class LIKE \'%Order\' AND
				perp_shipment_line.shippable_id = perp_order_line.order_id AND
				perp_shipment_line.shippable_item_line = perp_order_line.line_no
			LEFT JOIN (
				SELECT 
					shipment_id,
					line_no,
					sum(quantity_units) AS quantity_units
				FROM perp_goods_received
				GROUP BY shipment_id, line_no
			) AS received_total ON
				received_total.shipment_id = perp_shipment_line.shipment_id AND
				received_total.line_no = perp_shipment_line.line_no
			WHERE  perp_order_line.order_id = ' . $this->order->get_order_id() . ' AND
				perp_order_line.line_no = ' . $this->line_number;
			$GLOBALS['egw']->db->query($sql, __LINE__, __FILE__);
			$row = $GLOBALS['egw']->db->row(true);
			return $row['quantity_received'];
		}

		/**
		*	Is this OrderLine complete?
		*	"Complete" by default means shipped and invoiced.  Subclasses may extend and add more
		*	requirements, such as received.
		*/
		public function is_complete() {
			return ($this->get_quantity_ordered() == $this->get_quantity_shipped()) &&
				($this->get_quantity_ordered() == $this->get_quantity_invoiced());
		}

		/**
		*	Get a list of history / status information for this OrderLine.
		*
		*	This list can include:
		*	- Stock On Hand (in the selected location)
		*	- Shipment(s)
		*	- Invoices
		*/
		public function get_status() {
			if(!$this->order->get_order_id()) {
				return array();
			}

			$sql = 'SELECT
				perp_order_line.order_id,
				supplier_stock_code,
				type,
				our_reference_number,
				reference_id,
				related_id,
				quantity

			FROM perp_order_line
			JOIN (
				SELECT 
					\'INVOICE\' AS type,
					perp_invoice_line.invoice_id AS reference_id,
					invoicable_id AS related_id,
					invoicable_item_line AS related_item_line,
					our_reference_number,
					quantity
				FROM  perp_invoice_line 
				JOIN perp_invoice ON 
					perp_invoice_line.invoice_id = perp_invoice.invoice_id
				UNION
				SELECT
					\'SHIPMENT\' AS type,
					perp_shipment_line.shipment_id AS reference_id,
					shippable_id AS related_id,
					shippable_item_line AS related_item_line,
					our_reference_number,
					quantity_shipped AS quantity
				FROM perp_shipment_line
				JOIN perp_shipment ON
					perp_shipment_line.shipment_id = perp_shipment.shipment_id
				UNION
				SELECT
					\'RECEIVED\' AS type,
					perp_goods_received.goods_received_id AS reference_id,
					shippable_id AS related_id,
					shippable_item_line AS related_item_line,
					receive_date as our_reference_number,
					quantity_units as quantity
				FROM perp_goods_received
				JOIN perp_shipment_line ON
					perp_shipment_line.shipment_id = perp_goods_received.shipment_id AND
					perp_shipment_line.line_no = perp_goods_received.line_no
			) AS perp_related_line ON
				perp_related_line.related_id = perp_order_line.order_id AND
				perp_related_line.related_item_line = perp_order_line.line_no
			WHERE perp_order_line.order_id = ' . $this->order->get_order_id() . ' AND
				perp_order_line.line_no = ' . $this->line_number;

			$GLOBALS['egw']->db->query($sql, __LINE__, __FILE__);
			
			$list = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$list[] = $row;
			}

			return $list;
		}

			
		/**
		*	Return an array of this OrderLine's properties
		*	@return array
		*/
		public function toArray() {
			return array(
				'line_number'		=>	$this->line_number,
				'stock_id'		=>	$this->stock_id,
				'supplier_stock_code'	=>	$this->supplier_stock_code,
				'our_stock_code'	=>	$this->our_stock_code,
				'description'		=>	$this->description,
				'quantity_ordered'	=>	$this->quantity_ordered,
				'unit_price'		=>	$this->unit_price,
				'ship_location'		=>	$this->ship_location,
				'link'			=>	$this->get_link()
			);
		}

		/**
		*	Return a string representation of this object.
		*       Very useful for debugging purposes, not much else.
		*
		*       @return string Representation of this OrderLine
		*/
		public function __toString() {
			$string =  sprintf("%d\t%-10s\t%-10s\t%-50s\t%10d @ %10.2f (%s)\n",
				$this->stock_id,
				$this->our_stock_code,
				$this->supplier_stock_code,
				$this->description,
				$this->quantity_ordered,
				$this->unit_price,
				get_class($this)
			);
			foreach($this->details as $detail) {
				$string .= $detail->__toString();
			}
			return $string;
		}

		/**
		*	Prepare this object for serialization
		*
		*	@return array Properties to preserve across serialization
		*/
		public function __sleep() {
			$properties = $this->toArray();
			$properties['order'] = true;
			$properties['details'] = true;
			return array_keys($properties);
		}
	}
?>
