<?php

	/**
	*	Details on an OrderLine.  Used for specifying various things
	*	on an item, such as size or color.
	*	
	*	@author Nathan Gray
	*	@ingroup perp_api
	*/

	class OrderLineDetail {

		/**
		*	Database table for storage
		*/
		const detail_table = 'perp_order_detail';

		/**
		*	OrderLine this detail describes
		*/
		protected $order_line = null;

		/**
		*	Identifying number for this detail
		*/
		protected $detail_number = null;

		/**
		*	Name of the detail.
		*	For example: color, size
		*/
		protected $detail_name = null;

		/**
		*	Value of the detail.
		*	For example: blue, green, large, super-size
		*/
		protected $detail_value = null;

		/**
		*	Display this detail in the list of details?
		*/
		protected $display = true;
		

		public function __construct(OrderLine &$line, $data) {
			$this->order_line =& $line;
			$this->detail_number = $data['detail_number'];
			$this->detail_name = $data['detail_name'];
			$this->detail_value = $data['detail_value'];
			if(isset($data['display'])) {
				$this->display = $data['display'];
			}
		}

		/**
		*	This function is a stub to allow extending functions
		*	to read their own data.  Actual reading is done in
		*	OrderLine.
		*/
		public function read() {
		}

		/**
		*	Write this OrderLineDetail into the database for long term storage
		*/
		public function write() {
			$order_id = $this->order_line->get_owner()->get_order_id();
			$line_number = $this->order_line->get_line_number();

			if(is_null($this->detail_number)) {
				// Line was not added to the order, it's just there.  Don't keep it.
				return;
			}
			$GLOBALS['egw']->db->insert(self::detail_table,
				array(
					'order_id'      =>      $order_id,
					'line_no'       =>      $line_number,
					'detail_no'	=>	$this->detail_number,
					'detail_class'	=>	get_class($this)
				) + $this->toArray(),
				array(
					'order_id'      =>      $order_id,
					'line_no'       =>      $line_number,
					'detail_no'	=>	$this->detail_number,
				),
				__LINE__, __FILE__, 'perp_api'
			);
		}

		/**
		*	Can this OrderLineDetail be edited?
		*
		*	@return boolean
		*/
		public function can_edit() {
			return ($this->get_order_line()->get_quantity_invoiced() == 0);
		}

		/**
		*	Get the OrderLine this OrderLineDetail describes
		*/
		public function &get_order_line() {
			return $this->order_line;
		}
		/**
		*	Set the OrderLine this OrderLineDetail describes
		*/
		public function set_order_line(OrderLine &$order_line) {
			$this->order_line =& $order_line;
		}

		/**
		*	Get the identifying number for this OrderLineDetail
		*/
		public function get_number() {
			return $this->detail_number;
		}

		/**
		*	Set the identifying number for this OrderLineDetail.
		*	This should only be called from OrderLine.
		*	@param number New identifying number. 
		*/
		public function set_number($number) {
			if($this->order_line->get_detail($number) !== $this) {
				throw new Exception('Call $order_line->add_detail, don\'t call set_number directly');
			}
			$this->detail_number = $number;
		}

		/**
		*	Get the name of this detail
		*/
		public function get_name() {
			return $this->detail_name;
		}

		/**
		*	Get the value of this detail
		*/
		public function get_value() {
			return $this->detail_value;
		}
		/**
		*	Set the value of this detail
		*/
		public function set_value($value) {
			$this->detail_value = $value;
		}

		/**
		*	Get whether or not to display this detail in a list
		*	@return boolean display
		*/
		public function get_display() {
			return $this->display;
		}
		/**
		*	Set whether or not to display this detail in a list
		*/
		public function set_display($display) {
			$this->display = (boolean)$display;
		}

		/**
		*	Function called when this OrderLineDetail is added to the
		*	OrderLine.
		*/
		public function add() {}

		/**
		*	Function called when this OrderLineDetail is removed from the
		*	OrderLine.
		*/
		public function remove() {
			$GLOBALS['egw']->db->delete(self::detail_table,
				array(
					'order_id'		=>	$this->order_line->get_owner()->get_order_id(),
					'line_no'		=>	$this->order_line->get_line_number(),
					'detail_no'		=>	$this->detail_number
				),
				__LINE__, __FILE__
			);
		}

		/** 
		*	Return an array of this OrderLineDetail's properties
		*	@return array
		*/
		public function toArray() {
			return array(
				'detail_number'	=>	$this->detail_number,
				'detail_name'	=>	$this->detail_name,
				'detail_value'	=>	$this->detail_value,
				'display'	=>	$this->display
			);
		}

		/**
		*	Return a string representation of this object.
		*
		*	@return string Representation of this OrderLineDetail
		*/
		public function __toString() {
			return sprintf("%-10s\t%-10s\t%-20s\n",
				get_class($this),
				$this->detail_name,
				$this->detail_value
			);
		}

		/**
		*	Return an array of properties to preserve across
		*	serialization.
		*/
		public function __sleep() {
			return( array_keys( get_object_vars( $this ) ) );
		}
	}
?>
