<?php
	require_once( EGW_INCLUDE_ROOT . '/perp_api/inc/class.bo_perp_api.inc.php');

	require_once('class.Shippable.inc.php');
	require_once('class.Invoicable.inc.php');
	require_once('class.ShipmentLine.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_ledger/inc/class.GLTransaction.inc.php');

	class Shipment implements Invoicable {
		/**
		*	Database table
		*/
		const shipment_table = 'perp_shipment';

		/**
		*	Unique system identifier for the shipment (integer)
		*/
		protected $shipment_id;

		/**
		*	Our internal reference number for the shipment (string)
		*/
		protected $our_reference_number;

		/**
		*	The other party's reference number for the shipment (string)
		*/
		protected $their_reference_number;

		/**
		*	The system user id of the person who originated the shipment.
		*/
		protected $originating_user;

		/**
		*	The system user id of the person who approved the shipment.
		*	This would be a manager or someone with the authority to allow the shipment to
		*	proceed.
		*/
		protected $approving_user;

		/**
		*	The addressbook id of the contact person in our organization for this Shipment.
		*/
		protected $our_contact_person;

		/**
		*	The addressbook identification of the contact person in their organization
		*	for this shipment.  This will usually be a purchaser or a sales representitive.
		*/
		protected $their_contact_person;

		/**
		*	System identifier for the organisation supplying the items requested on this Shipment.
		*	This will be either a Supplier's ID, or 0 if we are the supplier.
		*/
		protected $supplier_id;

		/**
		*	System identifier for the organisation receiving the items requested on this Shipment.
		*	This will be either a Clients's ID, or 0 if we are the client.
		*/
		protected $client_id;

		/**
		*	The date (and time) this Shipment was originated (created)
		*/
		protected $originating_date = null;

		/**
		*	The date (and time) this Shipment was approved by the approving_user
		*/
		protected $approval_date = null;

		/**
		*	The date (and time) this Shipment was issued.  This is when the other parties were notified of 
		*	the shipment.  That is, when the Shipment was actually made.
		*/
		protected $issue_date = null;

		/**
		*	Requested delivery date (and time)
		*/
		protected $requested_delivery_date;

		/**
		*	Actual delivery date (and time)
		*/
		protected $actual_delivery_date;

		/**
		*	The addressbook id of the sending location
		*/
		protected $ship_from_address;

		/**
		*	The addressbook id of the destination location
		*/
		protected $ship_to_address;

		/**
		*	ID of the shipper carrying the load
		*/
		protected $shipper;

		/**
		*	Vessel identifier (Optional)
		*/
		protected $vessel;

		/**
		*	Sometimes shipments do not require approval before proceeding, depending on the organization.
		*/
		protected $requires_approval = true;

		/**
		*	Free form notes on this Shipment
		*/
		protected $comments;

		/**
		*	Array of ShipmentLines, one for each item on this Shipment
		*/
		protected $shipment_lines = array();

	
		/**
		*	Construct a Shipment
		*	@param shipment_id	Optional.  If an shipment_id for an existing Shipment is provided, the Shipment will be loaded.
		*	@return If shipment_id is provided, it will return whatever Shipment::read() returns.  Otherwise, true.
		*/
		public function __construct($shipment_id = null) {

			if(!is_null($shipment_id)) {
				return $this->read($shipment_id);
			}

			/**
			*	Set some defaults
			*/
			$this->originating_date = date('U');
			$this->issue_date = date('U');

			// Set originator to the current user 
			$this->set_originating_user($GLOBALS['egw_info']['user']['account_id']);
			
			// Default contact person to current user
			$this->set_our_contact_person((int)$GLOBALS['egw_info']['user']['person_id']);

			// Check to see if approval is required
			$config = config::read($GLOBALS['egw_info']['flags']['currentapp']);
			$this->requires_approval = $config['shipments_require_approval'];

			return true;
		}

		/**
		*	Read the Shipment identified by shipment_id into this Shipment
		*	@param shipment_id The unique system identifier for the Shipment to read.
		*/
		public function read($shipment_id) {
			$GLOBALS['egw']->db->select(self::shipment_table,
				array(
					'shipment_id',
					'our_reference_number',
					'their_reference_number',
					'originating_user',
					'approving_user',
					'our_contact_person',
					'their_contact_person',
					'supplier_id',
					'client_id',
					'originating_date',
					'issue_date',
					'approval_date',
					'requested_delivery_date',
					'actual_delivery_date',
					'ship_from_address',
					'ship_to_address',
					'shipper',
					'vessel',
					'requires_approval',
					'comments'
				),
				array('shipment_id'	=>	$shipment_id),
				__LINE__, __FILE__, false, '', 'perp_api'
			);

			$row = $GLOBALS['egw']->db->row(true);

			$this->shipment_id			=	$row['shipment_id'];
			$this->our_reference_number	=	$row['our_reference_number'];
			$this->their_reference_number	=	$row['their_reference_number'];
			$this->originating_user		=	$row['originating_user'];
			$this->approving_user		=	$row['approving_user'];
			$this->our_contact_person	=	$row['our_contact_person'];
			$this->their_contact_person	=	$row['their_contact_person'];
			$this->supplier_id		=	$row['supplier_id'];
			$this->client_id		=	$row['client_id'];
			$this->originating_date		=	$GLOBALS['egw']->db->from_timestamp($row['originating_date']);
			$this->issue_date		=	$GLOBALS['egw']->db->from_timestamp($row['issue_date']);
			$this->approval_date		=	$GLOBALS['egw']->db->from_timestamp($row['approval_date']);
			$this->requested_delivery_date	=	$GLOBALS['egw']->db->from_timestamp($row['requested_delivery_date']);
			$this->actual_delivery_date	=	$GLOBALS['egw']->db->from_timestamp($row['actual_delivery_date']);
			$this->ship_from_address	=	$row['ship_from_address'];
			$this->ship_to_address		=	$row['ship_to_address'];
			$this->shipper			=	$row['shipper'];
			$this->vessel			=	$row['vessel'];
			$this->requires_approval	=	$row['requires_approval'];
			$this->comments			=	$row['comments'];

			// Read rows
			$GLOBALS['egw']->db->select(ShipmentLine::shipment_line_table,
				array(
					'line_no AS line_number',
					'quantity_shipped',
					'shippable_class',
					'shippable_id',
					'shippable_item_line'
				),
				array('shipment_id'	=>	$shipment_id),
				__LINE__, __FILE__, false, 
				'ORDER BY shipment_id, line_no', 'perp_api'
			);
			$line_data = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$line_data[] = $row;
			}

			// Make sure we have a good class for the lines
			$line_class = get_class($this) . 'Line';
			if(!class_exists($line_class)) {
				ExecMethod('perp_api.bo_perp_api.__autoload', $line_class);
			}
			if(!class_exists($line_class)) {
				$line_class = get_parent_class($this) . 'Line';
			}
			foreach($line_data as $row) {

				$line = new $line_class($this, $row);
				$line->read();
				$this->shipment_lines[$row['line_number']] = $line;
			}
		}

		/**
		*	Check this Shipment and write it into the database
		*	Throws exceptions if the Shipment is not valid.  
		*/
		public function write() {
			$result = $this->check();
			if($result !== True) {
				throw new Exception('Shipment can not be saved.  (' . $result . ')');
			}

			$GLOBALS['egw']->db->transaction_begin();

			// Write Shipment
			$GLOBALS['egw']->db->insert(self::shipment_table,
				$this->toArray(false) + array(
					'class'	=>	get_class($this)
				),
				array(
					'shipment_id'	=>	$this->get_shipment_id()
				),
				__LINE__, __FILE__, 'perp_api'
			);
			if($GLOBALS['egw']->db->affected_rows() > 2) {
				$GLOBALS['egw']->db->transaction_abort();
				throw new egw_exception(lang('There was an error saving, %1 rows were affected.  Probably a duplicate reference number.', $GLOBALS['egw']->db->affected_rows()));
			}
			$this->shipment_id = $this->shipment_id ? $this->shipment_id : $GLOBALS['egw']->db->get_last_insert_id(self::shipment_table, 'shipment_id');
			
			// Write lines
			$GLOBALS['egw']->db->delete(ShipmentLine::shipment_line_table,
				array(
					'shipment_id'	=>	$this->get_shipment_id()
				),
				__LINE__, __FILE__, 'perp_api'
			);
			foreach($this->shipment_lines as $line_no => $line) {
				$line->write($line_no);
			}

			$GLOBALS['egw']->db->transaction_commit();
			return $this->shipment_id;
		}

		/**
		*	Construct a shipment from an object.  
		*	The object must implement the Shippable interface
		*	@param object Shippable object to ship
		*/
		public function read_from_object(Shippable &$object) {
			// Common header information
			$this->set_supplier_id((int)$object->get_supplier_id());
			$this->set_client_id((int)$object->get_client_id());

			$this->set_our_contact_person((int)$object->get_our_contact_person());
			$this->set_their_contact_person((int)$object->get_their_contact_person());
		
			$this->set_ship_from_address((int)$object->get_ship_from_address());
			if(!$object->get_ship_from_address()) {
				$this->set_ship_from_address((int)$object->get_their_contact_person());
			}
			$this->set_ship_to_address((int)$object->get_ship_to_address());

			$this->set_requested_delivery_date((int)$object->get_requested_delivery_date());

			$shippable_items =& $object->get_shippable_items();
			foreach($shippable_items as $key => &$item) {
				if($item->get_ship_quantity() != 0) {
					if(!$item->get_ship_location() || $item->get_ship_location() == $this->get_location_id()) {
						$shipment_line =& new ShipmentLine($this, $item);
						$this->add_line($shipment_line);
					}
				}
			}

			$this->set_our_reference_number($this->suggest_shipment_number());
		}

		/**
		*	Permissions functions
		*	@{
		*/

		/**
		*	Check if the given user can edit this Shipment
		*	@param user_id	User ID of the user to check
		*/
		public function is_editor($user_id) {
			if($this->is_admin($user_id)) {
				return !(boolean)$this->get_actual_delivery_date();
			}
			if($user_id == $this->get_originating_user() || $GLOBALS['egw']->acl->check('shipment', EGW_ACL_EDIT, $GLOBALS['egw_info']['flags']['currentapp'])) {
				return !(boolean)$this->get_approval_date();
			}
		}

		/**
		*	Check if the given user can administer this Shipment
		*	This includes approving, denying, editing after 
		*	approval
		*	@param user_id	User ID of the user to check
		*/
		public function is_admin($user_id) {
			return array_key_exists($user_id, $this->get_admin_list());
		}
		
		/**
		*	Get a list of users allowed to administer this Shipment
		*/
		public function get_admin_list() {
			$users = array();
			if(!$this->requires_approval) {
				$users[$this->get_originating_user()] = $GLOBALS['egw']->accounts->id2name($this->get_originating_user());
			}
			$approvers = $GLOBALS['egw']->acl->get_ids_for_location('shipment', EGW_ACL_CUSTOM_1);
			if(!is_array($approvers)) {
				$approvers = array();
			}
			foreach($approvers as $id) {
				if($id != 0) {
					$users[$id] = $GLOBALS['egw']->accounts->id2name($id);
				}
				if($id < 0) {
					$users += $GLOBALS['egw']->accounts->members($id);
				}
			}
			return $users;
		}

		/**	@}	*/

		/**
		*	Check that this shipment is valid.  Specifically:
		*	- There is at least one ShipmentLine.
		*	- Ship To address is set.
		*	- Each ShipmentLine passes its check
		*
		*	@return boolean true if the shipment is valid, otherwise returns error message
		*/
		public function check() {
			$pass_check = true;
			if(is_null($this->supplier_id)) {
				$pass_check = false;
				return lang('Supplier is not set.');
			}
			if(count($this->shipment_lines) == 0) {
				$pass_check = false;
				return lang('Empty shipment.  Shipments need at least one ShipmentLine.');
			}
			if(!is_numeric($this->ship_to_address) || $this->ship_to_address == 0) {
				$pass_check = false;
				return lang('Ship to address is not set.');
			}
			if(!is_numeric($this->ship_from_address) || $this->ship_from_address == 0) {
				$pass_check = false;
				return lang('Ship from address is not set.');
			}
			
			foreach($this->shipment_lines as $key => &$line) {
				$pass_check &= $line->check();
			}

			return (boolean)$pass_check;
		}

		/**
		*	Check that this shipment can be approved.  This requires that Shipment::check() passes, as well as:
		*	- This Shipment has not been approved yet
		*	- Our contact person is set
		*	- Our reference number is set
		*	- Each ShipmentLine passes its approval_check()
		*	
		*	If it passes, and approval is not required, the approving user is set to the 
		*	originating user
		*
		*	@return boolean true if the shipment can be approved.
		*/
		public function approval_check() {
			$pass_check = $this->check();
			if($pass_check !== true) {
				return $pass_check;
			}
			
			if($this->approval_date) {
				$pass_check = false;
				return lang('Shipment has already been approved.');
			}
			if(is_null($this->our_contact_person)) {
				return lang('Contact person is not set.');
			}
			if(is_null($this->our_reference_number)) {
				$pass_check = false;
				return lang('Our reference number is not set.');
			}
				
			foreach($this->shipment_lines as $key => $line) {
				$line_check = $line->approval_check();
				if($line_check !== true) {
					return $line_check;
				}
			}

			if($pass_check === true && !$this->requires_approval) {
				$this->approving_user = $this->get_originating_user();
			}
			return $pass_check;
		}

		/**
		*	Approve this shipment
		*	@param approving_user_id The system's user id of the person who approved this Shipment
		*	@throws Exception If the Shipment cannot be approved.
		*/
		public function approve($approving_user_id) {
			$check = $this->approval_check();
			if($check !== true) {
				throw new Exception(lang('Shipment could not be approved (%1)', $check));
			}
			if(!$approving_user_id) {
				throw new Exception(lang('No user id provided to approve'));
			}
			$this->approving_user = $approving_user_id;
			$this->approval_date = date('U');
			$this->approval_required = false;
			$this->write();
		}

		/**
		*	Deny this shipment.  Shipments cannot be denied after being approved.  Denied Shipments
		*	are immediately cancelled.
		*	@param denying_user_id The system's user id of the person who denied this Shipment
		*	
		*/
		public function deny($denying_user_id) {
			if($this->approval_date) {
				throw new Exception(lang('Shipment has already been approved'));
			}
			return $this->cancel($denying_user_id);
		}

		/**
		*	Cancel this shipment.  Shipments cannot be cancelled if some items or money have already
		*	changed hands.  
		*	@throws Exception If the Shipment cannot be cancelled
		*	@param cancelling_user_id The system's user id of the person who cancelled this shipment
		*/
		public function cancel($cancelling_user_id) {
			$test = $this->can_cancel($cancelling_user_id);

			if($test !== true) {
				throw new Exception($test);
			}

			$GLOBALS['egw']->db->transaction_begin();
			$GLOBALS['egw']->db->delete(self::shipment_table,
				array('shipment_id' => $this->shipment_id),
				__LINE__, __FILE__
			);
			try {
				foreach($this->shipment_lines as $line) {
					$line->cancel();
				}
				$GLOBALS['egw']->db->delete(ShipmentLine::shipment_line_table,
					array('shipment_id' => $this->shipment_id),
					__LINE__, __FILE__
				);
			} catch (Exception $e) {
				$GLOBALS['egw']->db->transaction_abort();
				throw $e;
			}
			$GLOBALS['egw']->db->transaction_commit();
		}

		/**
		*	Can this shipment be cancelled?  Shipments cannot be cancelled if some items or money have already
		*	changed hands.
		*/
		public function can_cancel($cancelling_user_id) {
			if(!$GLOBALS['egw']->acl->check('shipment', EGW_ACL_DELETE, $GLOBALS['egw_info']['flags']['currentapp'])) {
				return lang('Access is not allowed');
			}
			if($this->is_complete() === true) {
				return lang('Shipment is complete');
			}
			foreach($this->shipment_lines as $shipment_line) {
				$test = $shipment_line->can_cancel();
				if($test !== true) {
					return $test;
				}
			}
			return true;
		}

		/**
		*	Is this Shipment completed (closed) ?
		*	@return boolean
		*/
		public function is_complete() {
			if((boolean)$this->actual_delivery_date) {
				return true;
			}
			$test = (count($this->shipment_lines) > 0);
			foreach($this->shipment_lines as $line) {
				$test = $test && $line->is_complete();
			}
			return $test;
		}

		/**
		*       Does this shipment need to be approved?  The organization
		*       may not require it.
		*/
		public function get_approval_required() {
			if($this->requires_approval) {
				return !(boolean)$this->approval_date;
			} else {
				return false;
			}
		}

		/**
		*	Can this Shipment be shipped?
		*	@return boolean
		*/
		public function can_ship() {
			$test = !$this->get_approval_required() && !$this->get_actual_delivery_date();
			$test = $test && (count($this->shipment_lines) > 0);
			foreach($this->shipment_lines as $line) {
				$check = $line->can_ship();
				if($check !== true) {
					return $check;
				}
			}
			return $test;
		}

		/**
		*	Ship this shipment.
		*/
		public function ship() {
			if(!$this->can_ship()) {
				return false;
			}

			// Get some defaults
			$config = CreateObject('phpgwapi.config', 'perp_ar');
			$config->read_repository();
			$cogs_account = $config->config_data['cogs_account'];

			// Departure location
			$locations = ExecMethod('perp_inventory.bo_perp_location.get_location_list', array('delivery_address' => $this->ship_from_address));
	
			if(count($locations) > 0) {
				$location_id = key($locations);
			} else {
				return lang('Could not determine departure location.');
			}

			// Set up GLTransaction
			$transaction = new GLTransaction();
			$transaction->set_transaction_type($config->config_data['shipment_transaction_type']);

			// This may cause a nested transaction inside GLTransaction
			$GLOBALS['egw']->db->transaction_begin();

			foreach($this->shipment_lines as $line) {
				$line_result = $line->ship($transaction, $location_id, $cogs_account);
				if($line_result !== true) {
					$GLOBALS['egw']->db->transaction_abort();
					return $line_result;
				}
			}

			if($transaction->check() === true) {
				$this->set_actual_delivery_date((int)date('U'));
				$this->write();
				$transaction->write();
				$GLOBALS['egw']->db->transaction_commit();
			}
			return true;
		}

		/**
		*	Can this Shipment be received?
		*	@return boolean
		*/
		public function can_receive() {
			$test = (count($this->shipment_lines) > 0);
			$line_check = false;
			foreach($this->shipment_lines as $line) {
				$line_check = $line_check || $line->can_receive();
			}
			return $test && $line_check;
		}

		/**
		*	Can this Shipment be invoiced?
		*	@return boolean
		*/
		public function can_invoice() {
			$test = false;
			foreach($this->shipment_lines as $line) {
				$test |= $line->can_invoice();
			}
			$test &= (!$this->requires_approval || ($this->requires_approval && $this->approval_date));
			return $test;
		}

		/**
		*	Add an ShipmentLine to this Shipment
		*	@todo Add some checks on status before adding
		*	@param shipmentline ShipmentLine to add to this Shipment
		*	@return int Line identification number
		*/
		public function add_line(ShipmentLine &$shipmentline) {
			$this->shipment_lines[] =& $shipmentline;
			$shipmentline->set_line_number(end(array_keys($this->shipment_lines)));
			return $shipmentline->get_line_number();
		}

		/**
		*	Remove an ShipmentLine from this Shipment
		*	@todo Add some checks on status before removing
		*	@param line_no Identification number of the line to remove
		*/
		public function remove_line($line_no) {
			$shipment_line = $this->shipment_lines[$line_no];
			if($shipment_line) {
				$shipment_line->remove();
			}
			unset($this->shipment_lines[$line_no]);
		}

		/**
		*	Return a reference to an ShipmentLine specified by the Identification number.
		*
		*	@param line_no Identification number of the line to return
		*/
		public function &get_line($line_no) {
			return $this->shipment_lines[$line_no];
		}

		/**
		*	Returns all ShipmentLines
		*	@return array of ShipmentLine
		*/
		public function get_lines() {
			return $this->shipment_lines;
		}

		/**
		*	Accessors
		*	@{
		*/

		/**
		*	Get the unique system identifier for this Shipment
		*	@return int shipment_id
		*/
		public function get_shipment_id() {
			return $this->shipment_id;
		}

		/**
		*	Get the organization supplying the items requested on this Shipment.
		*	This will be either a Supplier's ID, or 0 if we are the supplier.
		*	@return int supplier_id
		*/
		public function get_supplier_id() {
			return $this->supplier_id;
		}

		/**
		*	Set the organization supplying the items requested on this Shipment.
		*	This will be either a Supplier's ID, or 0 if we are the supplier.
		*	@param supplier_id ID of the supplier
		*/
		public function set_supplier_id($supplier_id) {
			if(!is_integer($supplier_id)) {
				return lang('Invalid supplier');
			}
			$this->supplier_id = $supplier_id;

			// Default ship_from address
			if(!$this->ship_from_address) {
				if($this->supplier_id != 0) {
					$supplier = ExecMethod('perp_ap.bo_perp_supplier.get_supplier', $this->supplier_id);
					$this->set_ship_from_address($supplier['supplier_address']);
				}
			}
			return true;
		}

		/**
		*	Get the organization receiving the items requested on this Shipment.
		*	This will be either a Client's ID, or 0 if we are the client.
		*	@return int client_id
		*/
		public function get_client_id() {
			return $this->client_id;
		}

		/**
		*	Set the organization receiving the items requested on this Shipment.
		*	This will be either a Client's ID, or 0 if we are the client.
		*	@param client_id ID of the client
		*/
		public function set_client_id($client_id) {
			if(!is_integer($client_id)) {
				return lang('Invalid client');
			}
			$this->client_id = $client_id;
			return true;
		}


		/**
		*	Set our human reference number for this Shipment
		*	@param reference string our_reference_number
		*/
		public function set_our_reference_number($reference) {
			$this->our_reference_number = $reference;
		}

		/**
		*	Get our human reference number for this Shipment
		*	@return string our_reference_number
		*/
		public function get_our_reference_number() {
			return $this->our_reference_number;
		}

		/**
		*	Set their human reference number for this Shipment
		*	@param reference string their_reference_number
		*/
		public function set_their_reference_number($reference) {
			$this->their_reference_number = $reference;
		}

		/**
		*	Get their human reference number for this Shipment
		*	@return string their_reference_number
		*/
		public function get_their_reference_number() {
			return $this->their_reference_number;
		}

		/**
		*	Set the eGroupWare user id of the originating user.
		*	@param user_id int System user_id
		*/
		public function set_originating_user($user_id) {
			if(!is_integer($user_id)) {
				return lang('Invalid user id for originating user');
			}
			$this->originating_user = $user_id;
		}

		/**
		*	Get the eGroupWare user id of the originating user
		*	@return int user id
		*/
		public function get_originating_user() {
			return $this->originating_user;
		}

		/**
		*	Set the date this Shipment was created
		*	@param date int Unix time stamp
		*/
		public function set_originating_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for originating date');
			}
			$this->originating_date = $date;
		}
		/**
		*	Get the date this Shipment was created
		*	@return int Unix time stamp
		*/
		public function get_originating_date() {
			return $this->originating_date;
		}

		/**
		*	Set the date this Shipment was issued
		*	@param date int Unix time stamp
		*/
		public function set_issue_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for issue date');
			}
			$this->issue_date = $date;
		}

		/**
		*	Get the date this Shipment was issued
		*	@return int Unix time stamp
		*/
		public function get_issue_date() {
			return $this->issue_date;
		}

		/**
		*	Get the date this Shipment was approved
		*	@return int Unix time stamp
		*/
		public function get_approval_date() {
			return $this->approval_date;
		}

		/** 
		*	Get whether this Shipment requires approval or not
		*	@return boolean requires approval
		*/
		public function requires_approval() {
			return $this->requires_approval;
		}

		/**
		*	Set the requested delivery date for this Shipment
		*	@param date int Unix time stamp
		*/
		public function set_requested_delivery_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for requested delivery date');
			}
			$this->requested_delivery_date = $date;
			return true;
		}

		/**
		*	Get the requested date for delivery
		*	@return int Unix time stamp
		*/
		public function get_requested_delivery_date() {
			return $this->requested_delivery_date;
		}

		/**
		*	Set the actual delivery date for this Shipment
		*	@param date int Unix time stamp
		*/
		public function set_actual_delivery_date($date) {
			if(!is_integer($date)) {
				return lang('Invalid date for actual delivery date');
			}
			$this->actual_delivery_date = $date;
		}

		/**
		*	Get the actual date for delivery
		*	@return int Unix time stamp
		*/
		public function get_actual_delivery_date() {
			return $this->actual_delivery_date;
		}

		/**
		*	Get the person in our organization who is to be contacted
		*	regarding this Shipment.
		*	@return int addressbook_id
		*/
		public function get_our_contact_person() {
			return $this->our_contact_person;
		}

		/**
		*	Set the person in our organization who is to be contacted
		*	regarding this Shipment.
		*	@param contact Addressbook ID of the contact
		*	@return true or error message
		*/
		public function set_our_contact_person($contact) {
			if(!is_integer($contact)) {
				return lang('Invalid contact person.');
			}
			$this->our_contact_person = $contact;
			return true;
		}

		/**
		*	Get the person in their organization who is to be contacted
		*	regarding this Shipment.
		*	@return int addressbook_id
		*/
		public function get_their_contact_person() {
			return $this->their_contact_person;
		}

		/**
		*	Set the person in their organization who is to be contacted
		*	regarding this Shipment.
		*	@param contact Addressbook ID of the contact
		*	@return true or error message
		*/
		public function set_their_contact_person($contact) {
			if(!is_integer($contact)) {
				return lang('Invalid contact person.');
			}
			$this->their_contact_person = $contact;
			return true;
		}

		/**
		*	Get the location the ShipmentLines are coming from for this Shipment
		*	@return int addressbook ID of the location
		*/
		public function get_ship_from_address() {
			return $this->ship_from_address;
		}

		/**
		*	Set the location the ShipmentLines are coming from for this Shipment
		*	@param location Addressbook ID of the location
		*/
		public function set_ship_from_address($location) {
			if(!is_integer($location)) {
				return lang('Invalid ship from address.');
			}
			$this->ship_from_address = $location;
			return true;
		}

		/**
		*	Get the location the ShipmentLines are going to for this Shipment
		*	@return int addressbook ID of the location
		*/
		public function get_ship_to_address() {
			return $this->ship_to_address;
		}

		/**
		*	Set the location the ShipmentLines are going to for this Shipment
		*	@param location Addressbook ID of the location
		*/
		public function set_ship_to_address($location) {
			if(!is_integer($location)) {
				return lang('Invalid ship to address.');
			}
			$this->ship_to_address = $location;
			return true;
		}

		/**
		*       Get our location involved in this Shipment.
		*       This will be either the source or destination,
		*       depending on the order.
		*       @return int Inventory location ID or false if it can't be found.
		*/
		public function get_location_id() {
			if($this->supplier_id == 0) {
				$address = $this->ship_from_address;
			} else {
				$address = $this->ship_to_address;
			}
			$links = ExecMethod2('phpgwapi.bolink.get_links',
				'addressbook', $address,
				'perp_inventory'
			);
			$key = current($links);
			if($key) {
				list($location, $location_id) = explode('-', $key);
				return $location_id;
			}
			return false;
		}

		/**
		*	Set the shipper
		*	@param shipper ID of the shipper carrying the load
		*/
		public function set_shipper_id($shipper) {
			$this->shipper = $shipper;
			return true;
		}

		/**
		*	Get the shipper
		*	@return shipper ID of the shipper carrying the load
		*/
		public function get_shipper_id() {
			return $this->shipper;
		}

		/**
		*	Set the vessel identifier of the truck or ship carrying the load
		*	@param vessel identifer of the vessel carrying the load
		*/
		public function set_vessel($vessel) {
			$this->vessel = $vessel;
			return true;
		}

		/**
		*	Get the vessel identifier
		*	@return vessel identifier of the truck or ship carrying the load
		*/
		public function get_vessel() {
			return $this->vessel;
		}

		/**
		*	Get comments on this Shipment
		*	@return string comments
		*/
		public function get_comments() {
			return $this->comments;
		}

		/**
		*	Set comments on this Shipment
		*	@param comments 
		*/
		public function set_comments($comments) {
			$this->comments = $comments;
			return true;
		}

		/** @} */

		/**
		*	Methods required to implement Invoicable
		*	@{
		*/

		/**
		*	Address the Invoice comes from
		*	@return  Addressbook ID of the location
		*/
		public function get_bill_from_address() {
			return $this->get_ship_from_address();
		}

		/**
		*	Address the Invoice is sent to
		*	@return Addressbook ID of the location
		*/
		public function get_bill_to_address() {
			return $this->get_ship_to_address();
		}

		/**
		*	Currency the Invoice is in
		*	@return 3 letter currency code
		*/
		public function get_invoice_currency() {
			foreach($this->shipment_lines as $line) {
				if(!is_null($line->get_shippable_item())) {
					return $line->get_shippable_item()->get_owner()->get_invoice_currency();
				}
			}
		}

		/**
		*	Get payment terms for invoicing
		*/
		public function get_payment_terms() {
			foreach($this->shipment_lines as $line) {
				if(!is_null($line->get_shippable_item())) {
					return $line->get_shippable_item()->get_owner()->get_payment_terms();
				}
			}
		}

		/**
		*	Get items to be invoiced.
		*
		*	We return an array of all the shipment lines.  Immediately after creation,
		*	we have to swap the InvoiceLine's InvoicableItem to the associated shippable item, 
		*	rather than the ShipmentLines themselves so the Shippable Items
		*	get invoiced directly, and are what are referenced by the InvoiceLine.
		*	@returns shipment lines
		*/
		public function &get_invoicable_items() {
			return $this->shipment_lines;
		}

		/**	@}	*/

		/**
		*	Return an array of this Shipment's properties
		*	Used for display in eTemplates and such
		*	@param get_lines	boolean Include shipment lines in the array
		*/
		public function toArray($get_lines = false) {
			$data = get_object_vars($this);
			unset($data['shipment_lines']);
			if($get_lines) {
				$data['shipment_lines'] = array();
				foreach($this->shipment_lines as $key => $line) {
					$line_array = $line->toArray();
					$line_array['line_no'] = $key;
					$data['shipment_lines'][] = $line_array;
				}
			}
			return $data;
		}

		/**     
		*	Return a string representation of this object.
		*	Very useful for debugging purposes, not much else.
		*
		*	@return string Representation of this Shipment
		*/      
		public function __toString() {
			$result = $GLOBALS['egw']->accounts->get_account_name($this->get_originating_user(), $id, $fname, $lname);
			$originating_user = $fname . ' ' . $lname . ' [' . $id . ']';
			unset($fname);
			unset($lname);
			unset($id);

			$string = sprintf(get_class($this) . " ID: %d\t Our Ref #: %s started %s by %s\n",
			$this->get_shipment_id(),
			$this->get_our_reference_number(),
			date($GLOBALS['egw_info']['user']['preferences']['common']['dateformat'], $this->get_originating_date()),
			$originating_user);

			$string .= lang('supplier') . ': ' . $this->get_supplier_id() . "\n";
			$string .= lang('client') . ': ' . $this->get_client_id() . "\n";

			foreach($this->shipment_lines as $line_no => $shipmentline) {
				$string .= "\t$line_no.\t" . $shipmentline->__toString();
			}
			return $string;
		}

		/**
		*	Prepare this object for serialization
		*
		*	@return array Properties to preserve across serialization.
		*/
		public function __sleep() {
			$properties =  array('shipment_id', 'our_reference_number', 'their_reference_number', 
				'originating_user', 'approving_user', 
				'our_contact_person', 'their_contact_person',
				'supplier_id', 'client_id',
				'originating_date', 'approval_date', 'issue_date', 
				'requested_delivery_date', 'actual_delivery_date',
				'ship_to_address', 'ship_from_address',
				'shipper', 'vessel',
				'requires_approval',
				'comments',
				'shipment_lines'
			);
			return $properties;
		}

		public function __wakeup() {
		}

		/**
		*	Suggest our reference number (human reference) based on 
		*	system settings stored in eGroupWare config variables for
		*	this Shipment.  Depending on permissions, the user may override
		*	the suggested number.
		*
		*	Options for shipment number:
		*	- %1$s - Supplier number
		*	- %2$s - Client number
		*	- %3$d - Shipment ID
		*	- date() - formatted date
		*	
		*	@return string Suggested number based on site settings.  
		*/
		public function suggest_shipment_number() {
			// Get number format
			$config = CreateObject('phpgwapi.config');
			$config->read_repository();
			$shipment_code = $config->config_data['shipment_number'];

			// Get supplier
			if($this->supplier_id) {
				$supplier = ExecMethod('perp_ap.bo_perp_supplier.get_supplier', $this->supplier_id);
			}

			// Get client
			if($this->client_id) {
				$client = ExecMethod('perp_ar.bo_perp_client.get_client', $this->client_id);
			}

			// Shipment number
			if($this->shipment_id) {
				$shipment_number = $this->shipment_id;
			} else {
				$GLOBALS['egw']->db->query("SHOW TABLE STATUS LIKE 'perp_shipment'",
					__LINE__, __FILE__);
				if($row = $GLOBALS['egw']->db->row(true)) {
					$shipment_number = $row['Auto_increment'];
				}
			}

			// Date
			preg_match('/date\(.+\)/', $shipment_code, $regs);
			if(count($regs)) {
				eval('$date = date(\'' . 
					substr($regs[0], 5, strlen($regs[0]) - 6) . 
					'\');');
				$shipment_code = str_replace($regs[0], '%4$s', $shipment_code);
			}

			$our_reference = sprintf($shipment_code,
				$supplier['supplier_num'],
				$client['client_num'],
				($shipment_number ? $shipment_number : 'Not set'),
				$date
			);

			// Check for unique reference number
			if(!$this->shipment_id && $our_reference) {
				$tries = 0;
				$correction = $our_reference;
				do {
					$tries++;
					$count = $GLOBALS['egw']->db->select(
						self::shipment_table,
						array('count(shipment_id)'),
						array(
							'our_reference_number' =>  $correction
						),
						__LINE__, __FILE__, 'perp_api'
					)->fetchSingle();
					if($count) {
						$correction = $our_reference . '.' . $tries;
					}
				} while ($count > 0 && $tries < 5);
				$our_reference = $correction;
			}


			return $our_reference;
		}
	}
?>
