<?php
	require_once('class.Shipment.inc.php');
	require_once('class.ShippableItem.inc.php');
	require_once('class.InvoicableItem.inc.php');
	require_once('class.ShipmentLineDetail.inc.php');

	class ShipmentLine implements InvoicableItem {
		/**
		*	Database table name
		*/
		const shipment_line_table = 'perp_shipment_line';

		/**
		*	Reference to the containing Shipment
		*/
		protected $shipment;

		/**
		*	Line number on the containing Shipment
		*/
		protected $line_number;

		/**
		*	ShippableItem that is shipped
		*/
		protected $shippable_item = null;

		/**
		*	Quantity shipped on this shipment in supplier's unit of measure
		*/
		protected $quantity_shipped = 0;

		/**
		*	Details for this ShipmentLine, such as color, shape, size, etc.
		*/
		protected $details = array();

		/**
		*	Construct a ShipmentLine
		*	
		*	@param shipment Shipment that this line is part of.
		*	@param line_data Optional.  Array of information for this ShipmentLine.
		*/
		public function __construct(Shipment &$shipment, &$line_data = null) {
			$this->shipment =& $shipment;
			if(is_array($line_data)) {
				$this->line_number		=	$line_data['line_number'];
				if(!is_null($line_data['shippable_class'])) {
					if(!is_object($GLOBALS['perp_api']['bo'])) {
						$GLOBALS['perp_api']['bo'] =& CreateObject('perp_api.bo_perp_api');
					}
					$this->shippable_item =& $GLOBALS['perp_api']['bo']->get_shippable_item($line_data);
				}
				$this->set_quantity_shipped($line_data['quantity_shipped']);
			} elseif ($line_data instanceof ShippableItem) {
				$this->shippable_item	=&	$line_data;
				$this->quantity_shipped	=	$line_data->get_ship_quantity();
			}
		}

		/**
		*	Read this line from the database
		*	This ShipmentLine will already have its properties set from being read
		*	in its containing Shipment.  This gives the ShipmentLine a chance to set
		*	other properties
		*/
		public function read() {
			// Get details
			$GLOBALS['egw']->db->select(ShipmentLineDetail::detail_table,
				array(
					'shipment_id',
					'line_no',
					'detail_id AS detail_number',
					'detail_class',
					'detail_name',
					'detail_value',
					'display'
				),
				array(
					'shipment_id'	=>	$this->shipment->get_shipment_id(),
					'line_no'	=>	$this->line_number
				),
				__LINE__, __FILE__, false,
				'ORDER BY shipment_id, line_no, detail_id', 'perp_api'
			);
			$list = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$list[] = $row;
			}

			foreach($list as &$detail_data) {
				if(!class_exists($detail_data['detail_class'])) {
					ExecMethod('perp_api.bo_perp_api.__autoload', $detail_data['detail_class']);
				}
				$detail =& new $detail_data['detail_class']($this, $detail_data);
				$detail->read();
				$this->details[$detail_data['detail_number']] =& $detail;
			}
		}

		/**
		*	Write this line to the database
		*	@param line_number The Identification number of this line.
		*/
		public function write($line_number = null) {
			if(!is_null($line_number)) {
				$this->line_number = $line_number;
			}
			$data = array(
				'shipment_id'	=>	$this->shipment->get_shipment_id(),
				'line_no'	=>	$this->line_number,
			) + $this->toArray();
			if(!is_null($this->shippable_item)) {
				$data += array(
					'shippable_class'	=>	get_class($this->shippable_item->get_owner()),
					'shippable_id'		=>	$this->shippable_item->get_shippable_id(),
					'shippable_item_line'	=>	$this->shippable_item->get_shippable_line_id()
				);
			}
			$GLOBALS['egw']->db->insert(self::shipment_line_table,
				$data,
				array(
					'shipment_id'	=>	$this->shipment->get_shipment_id(),
					'line_no'	=>	$this->line_number
				),
				__LINE__, __FILE__, 'perp_api'
			);
			$GLOBALS['egw']->db->delete(ShipmentLineDetail::detail_table,
				array(
					'shipment_id'	=>	$this->shipment->get_shipment_id(),
					'line_no'	=>	$this->line_number
				),
				__LINE__, __FILE__, 'perp_api'
			);
			
			foreach($this->details as $detail) {
				$detail->write();
			}
		}

		/**
		*	Do anything that needs to be done when cancelled
		*/
		public function cancel() {
			$this->remove();
		}

		/**
		*	Check that this line is valid
		*	@todo Currently always returns true
		*	@return true or error message
		*/
		public function check() {
			return true;
		}

		/**
		*	Check that this line is valid for approval
		*	@todo Currently always returns true
		*	@return true or error message
		*/
		public function approval_check() {
			return true;
		}

		/**
		*	Get the line number of this ShipmentLine
		*/
		public function get_line_number() {
			return $this->line_number;
		}

		/**
		*	Set the line number of this ShipmentLine
		*	@param line_number int Line number
		*/
		public function set_line_number($line_number) {
			if($this->shipment->get_line($line_number) !== $this) {
				throw new Exception('Call $shipment->add_line, don\'t call set_line_number directly');
			}
			$this->line_number = $line_number;
		}

		/**
		*	Get the quantity shipped
		*	@return float quantity shipped
		*/
		public function get_quantity_shipped() {
			return $this->quantity_shipped;
		}

		/**
		*	Set the quantity on this shipment
		*	@param quantity Quantity ship
		*
		*	@return true
		*/
		public function set_quantity_shipped($quantity) {
			if($this->shippable_item && method_exists($this->shippable_item, 'get_stock_id') && $this->shippable_item->get_stock_id()) {
				$stock = ExecMethod('perp_inventory.bo_perp_stock.get_stock_item', $this->shippable_item->get_stock_id());
				$quantity = round($quantity, $stock['decimal_places']);
			}
			$this->quantity_shipped = $quantity;

			return true;
		}

		/**
		*	Get the shippable item being shipped
		*	@return ShippableItem
		*/
		public function &get_shippable_item() {
			return $this->shippable_item;
		}

		/**
		*	Get a list of serial numbers for this line
		*
		*	This will first check ShipmentLineDetails and ShippableItem details
		*
		*	@return array of serial => quantity
		*/
		public function get_serial_numbers() {
			$serial_numbers = array();
			$serial_details = $this->get_details('ShipmentSerial');
			foreach($serial_details as $serial_detail) {
				$serial_numbers[$serial_detail->get_serial_number()] = $serial_detail->get_value();
			}

			// Check shippable item for serials from work orders
			if(is_object($this->shippable_item) && method_exists($this->shippable_item, 'get_details')) {
				$shippable_details = $this->get_shippable_item()->get_details('WorkOrderDetail');
				if(is_array($shippable_details)) {
					$bo_inventory =& CreateObject('perp_inventory.bo_perp_inventory');
					$query = array(
						'stock_id'	=>	$this->shippable_item->get_stock_id(),
					);
					foreach($shippable_details as $detail) {
						// Pull the serial number off the work order
						$query['work_order_id'] = $detail->get_work_order_id();
						$query['detail_id'] = $detail->get_work_order_detail_id();
						$bo_inventory->get_controlled_inventory_list($query, $serials);
						foreach($serials as $info) {
							$serial_numbers[$info['serial_number']] = $info['quantity'];
						}
					}
				}
			}
			return $serial_numbers;
		}

		/**
		*	Any actions to perform before being removed from the Shipment
		*	Does not actually remove the line.
		*/
		public function remove() {
			// Remove ShipmentLineDetails
			foreach($this->details as &$detail) {
				$detail->remove();
			}
			$GLOBALS['egw']->db->delete(ShipmentLineDetail::detail_table, 
				array(
					'shipment_id'	=>	$this->shipment->get_shipment_id(),
					'line_no'	=>	$this->line_number,
				),
				__LINE__, __FILE__
			);
		}

		/*
		*	Get any ShipmentLineDetails describing this ShipmentLine.
		*	
		*	@param classname Optional.  Restrict returned ShipmentLineDetails to a specific extending class.
		*	@return array of ShipmentLineDetail
		*/
		public function &get_details($classname = null) {
			if($classname == null) {
				return $this->details;
			}
			$details = array();
			foreach($this->details as &$detail) {
				if($detail instanceof $classname) {
					$details[] =& $detail;
				}
			}
			return $details;
		}

		/**
		*	Get the ShipmentLineDetail identified by $detail_number.
		*	If $detail_number is a string, this function will return an array
		*	of ShipmentLineDetails that have the string as their name.
		*	@param detail_number The id of the detail to get, or the name of the detail
		*	@return ShipmentLineDetail identified by the number, or array of details by name
		*/
		public function &get_detail($detail_number) {
			if(is_numeric($detail_number)) {
				return $this->details[$detail_number];
			}
			$results = array();
			foreach($this->details as &$detail) {
				if($detail->get_name() == $detail_number) {
					$results[$detail->get_number()] =& $detail;
				}
			}
			return $results;
		}

		/**
		*	Add a detail describing some aspect of this ShipmentLine
		*/
		public function add_detail(ShipmentLineDetail $detail) {
			$this->details[] =&$detail;
			$detail->set_number(end(array_keys($this->details)));
			$detail->add();
			return $detail->get_number();
		}

		/**
		*	Remove a ShipmentLineDetail from this ShipmentLine
		*	@param detail_number The id of the detail to get
		*/
		public function remove_detail($detail_number) {
			$detail =& $this->details[$detail_number];
			$detail->remove();
			unset($this->details[$detail_number]);
		}

		/**
		*	Can this ShipmentLine be cancelled?
		*/
		public function can_cancel() {
			return !($this->shipment->get_actual_delivery_date()) &&
				($this->get_quantity_received() == 0);
		}

		/**
		*	Check to see if this ShipmentLine can be shipped
		*
		*	@return true or reason why not
		*/
		public function can_ship(){
			if(is_null($this->shippable_item) || !($this->shippable_item instanceof ShippableItem)) {
				// A shipment line without an item to ship?
				throw new egw_exception_not_found(lang('Unable to find reference line (from order) for Shipment %1 (ID#%3), line %2',
					$this->shipment->get_our_reference_number(),
					$this->__toString(),
					$this->shipment->get_shipment_id()
				));
			}

			// Get serial / batch numbers
			$stock_id = $this->shippable_item->get_stock_id();
			$stock = ExecMethod('perp_inventory.bo_perp_stock.get_stock_item', $stock_id);

			if($stock['controlled']) {
				$serialized_quantity = 0;
				$serials =& $this->get_serial_numbers();
				foreach($serials as $serial_number => $quantity) {
					$serialized_quantity += $quantity;
				}
				if($serialized_quantity < $this->quantity_shipped) {
					return lang('Line %1 is missing serial numbers.  %2 needed, %3 found ',
						$this->line_number + 1,
						$this->quantity_shipped,
						$serialized_quantity
					);
				}
			}
			// Check inventory levels: on hand + reserved >= quantity needed
			$bo_inventory =& CreateObject('perp_inventory.bo_perp_inventory');
			$query = array(
				'stock_id'	=>	$this->shippable_item->get_stock_id(),
				'location_id'	=>	$this->shipment->get_location_id()
			);
			$bo_inventory->get_inventory($query, $inventory);
			if(count($inventory) == 0 || $inventory[0]['quantity_on_hand'] + $inventory[0]['quantity_reserved'] < $this->quantity_shipped) {
				// Not enough inventory where it should be.  Let's look for it to give a good error message.
				$stock_reference = ExecMethod('perp_inventory.bo_perp_stock.link_title', $stock_id);
				$where = array();
				if(!$stock['controlled']) {
					unset($query['location_id']);
					$bo_inventory->get_inventory($query, $inventory);
					foreach($inventory as $location) {
						if($location['quantity_on_hand'] > 0) {
							$where[$location['location_id']] = ExecMethod('perp_inventory.bo_perp_location.link_title', $location['location_id']);
						}
					}
				} elseif($stock['serial_setup'] == -1) {
					// Only check if we want to track serials all the way
					foreach($serials as $serial_number => $quantity) {
						$query['serial_number'] = $serial_number;
						$bo_inventory->get_controlled_inventory_list($query, $locations);
						$count = 0;
						foreach($locations as $location) {
							if($location['location_id']) {
								$where[$location['location_id']] = ExecMethod('perp_inventory.bo_perp_location.link_title', $location['location_id']) . ": $serial_number<br />\n";
							}
						}
					}
				}
				if(count($where) > 0) {
					return lang('Unable to find enough %1 to ship out from %2.  Some was found in: <br />%3', 
						$stock_reference,
						ExecMethod('perp_inventory.bo_perp_location.link_title', $this->shipment->get_location_id()),
						implode(", ", $where)
					);
				} else {
					return lang('Unable to find %1 to ship out from %2 anywhere.', 
						$stock_reference,
						ExecMethod('perp_inventory.bo_perp_location.link_title', $this->shipment->get_location_id())
					);
				}
			}
			return true;
		}

		/**
		*	Actions to perform when shipping this ShipmentLine
		*	Remove stock from inventory, add to a GLTransaction to move value from inventory to 
		*	Cost of Goods Sold account
		*/
		public function ship(GLTransaction &$transaction, $location_id, $default_cogs) {
			// Get stock_id of shipped item
			if(is_null($this->shippable_item) || !method_exists($this->shippable_item, 'get_stock_id')) {
				return lang('Could not determine stock item to ship');
			}
			$stock_id = $this->shippable_item->get_stock_id();
			$stock = ExecMethod('perp_inventory.bo_perp_stock.get_stock_item', $stock_id);

			// Round quantity to correct number of decimals
			$this->quantity_shipped = round($this->quantity_shipped, $stock['decimal_places']);

			// Check for location_id from shippable item
			$item_location = $this->shippable_item->get_ship_location();

			if($item_location == 0 || $item_location != $location_id) {
				// Set ship from location
				$this->shippable_item->set_ship_location($location_id);
				if(method_exists($this->shippable_item, 'write')) {
					$this->shippable_item->write();
				}

				// Reserve stock now
				ExecMethod2('perp_inventory.bo_perp_inventory.reserve_stock', $location_id, $stock_id, $this->quantity_shipped, $this->get_owner()->get_our_reference_number());
			}

			// Get serial / batch numbers
			if($stock['controlled']) {
				$serial_details = $this->get_details('ShipmentSerial');
				foreach($serial_details as $serial_detail) {
					$serials[$serial_detail->get_serial_number()] = $serial_detail->get_value();
				}
			}

			// Remove from inventory (reserved stock)
			$cost = ExecMethod2('perp_inventory.bo_perp_inventory.ship', $stock_id, $location_id, $this->quantity_shipped, $serials, $this->get_owner()->get_shipment_id());

			// Transaction narrative
			$narrative = lang('Shipment %1 - shipped %2 of %3 for Sales Order %4', 
				$this->shipment->get_our_reference_number(),
				$this->quantity_shipped, 
				$stock['stock_num'],
				$this->shippable_item->get_owner()->get_our_reference_number()
			);

			// Find inventory account
			$inventory_account = bo_perp_stock::get_gl_accounts($stock_id, bo_perp_inventory::ON_HAND);
			if(!$inventory_account) {
				return lang('Could not find stock account for %1', $stock['stock_num']);
			}

			// Find Cost of Goods Sold account
			list($sales_category) = ExecMethod('perp_inventory.bo_perp_stock.get_sales_categories', $stock_id);
			$sales_category = ExecMethod('perp_orders.bo_perp_orders.get_sales_category', $sales_category['sales_category_id']);
			if(!$sales_category['cost_of_sales_gl_account']) {
				return lang('Could not find cost of sales account for %1', $stock['stock_num']);
			}

			// Remove from inventory account
			$transaction->add_line($inventory_account, $narrative, -$cost);
			// Add to Cost of Goods Sold
			$transaction->add_line($sales_category['cost_of_sales_gl_account'], $narrative, $cost);

			// Mark shippable item as shipped
			if(!is_null($this->shippable_item)) {
				$this->shippable_item->ship($this);
			}
			return true;
		}

		/**
		*	Actions to perform when receiving this ShipmentLine
		*/
		public function receive(ReceiptLine &$receipt_line) {
			$lines =& $this->shipment->get_lines();
			$totally_received = true;
			foreach($lines as $line) {
				$totally_received &= ($line->get_quantity_received() == $line->get_quantity_shipped());
			}
			if($totally_received) {
				$this->shipment->set_actual_delivery_date((int)$receipt_line->get_receipt_date());
				$this->shipment->write();
			}
		}

		/**
		*	Check to see if this ShipmentLine can be received
		*/
		public function can_receive(){
			return $this->shipment->get_shipment_id() && 
				($this->quantity_shipped != $this->get_quantity_received() ) &&
				!($this->shipment->get_actual_delivery_date());
		}

		/**
		*	Get the quantity of this ShipmentLine already received
		*/
		public function get_quantity_received() {
			if(!$this->shipment->get_shipment_id()) {
				return 0;
			}

			$sql = 'SELECT 
					IF(quantity_units IS NULL, 0, quantity_units) AS received
				FROM perp_shipment_line
				LEFT JOIN (
					SELECT 
						shipment_id,
						line_no,
						sum(quantity_units) AS quantity_units
					FROM perp_goods_received
					GROUP BY shipment_id, line_no
				) AS received_total ON
					received_total.shipment_id = perp_shipment_line.shipment_id AND
					received_total.line_no = perp_shipment_line.line_no
				WHERE 
					perp_shipment_line.shipment_id = ' . $this->shipment->get_shipment_id() . ' AND
					perp_shipment_line.line_no = ' . $this->line_number;
			$GLOBALS['egw']->db->query($sql, __LINE__, __FILE__);
			$row = $GLOBALS['egw']->db->row(true);
			return $row['received'];
		}

		public function invoice(InvoiceLine &$invoice_line) {
			if(!is_null($this->shippable_item) && $this->invoicable_item instanceof InvoicableItem) {
				$this->shippable_item->invoice($invoice_line);
			}
		}
		public function can_invoice() {
			if(is_null($this->shippable_item)) {
				throw new egw_exception_not_found(lang('Unable to find reference line (from order) for Shipment %1 (ID#%3), line %2',
					$this->shipment->get_our_reference_number(),
					$this->__toString(),
					$this->shipment->get_shipment_id()
				));
			}
			return ($this->get_quantity_invoiced() != $this->quantity_shipped) && $this->shippable_item->can_invoice() && !$this->shipment->get_approval_required();
		}
		public function get_invoice_quantity() {
			return $this->quantity_shipped;
		}

		public function get_quantity_invoiced() {
			if($this->shippable_item) {
				return $this->shippable_item->get_quantity_invoiced();
			} else {
				return 0;
			}
		}

		/**
		*	Return a list of extra details to be displayed on the invoice, below
		*	the line.  This list cannot be edited, and is an array of details in
		*	the form ('name' => 'Detail name', 'value' => '123')
		*
		*	@param filter The filter parameter from the display select (usually true)
		*	@return array
		*/
		public function get_invoice_details($filter) {
			return array(
			);
		}

		public function get_owner() {
			return $this->shipment;
		}
		public function get_invoicable_id() {
			return $this->shipment->get_shipment_id();
		}
		public function get_invoicable_line_id() {
			return $this->line_number;
		}
		public function get_description() {
			if($this->shippable_item) {
				return $this->shippable_item->get_description();
			} else {
				return null;
			}
		}
			
		public function get_unit_price() {
			if($this->shippable_item) {
				return $this->shippable_item->get_ship_unit_price();
			} else {
				return null;
			}
		}

		public function get_gl_account() {
			if($this->shippable_item) {
				return $this->shippable_item->get_gl_account();
			} else {
				return null;
			}
		}

		public function get_link() {
			if($this->shippable_item) {
				return $this->shippable_item->get_link();
			} else {
				return null;
			}
		}

		/**
		*	Is this ShipmentLine complete?
		*
		*/
		public function is_complete() {
			return $this->get_quantity_received() == $this->quantity_shipped;
		}

		/**
		*	Return an array of this ShipmentLine's properties
		*	@return array
		*/
		public function toArray() {
			return array(
				'line_number'		=>	$this->line_number,
				'quantity_shipped'	=>	$this->get_quantity_shipped(),
			);
		}

		/**
		*	Return a string representation of this object.
		*       Very useful for debugging purposes, not much else.
		*
		*       @return string Representation of this ShipmentLine
		*/
		public function __toString() {
			$string = get_class($this) . ' ' . $this->get_line_number(). "\t" . $this->quantity_shipped . "\n";
			if(count($this->get_details()) > 0) {
				foreach($this->details as &$detail) {
					$string .= $detail->__toString() . "\n";
				}
			}
			if(!is_null($this->shippable_item)) {
				$string .= $this->shippable_item->__toString();
			}
			return $string;
		}

		/**
		*	Prepare this object for serialization
		*
		*	@return array Properties to preserve across serialization
		*/
		public function __sleep() {
			$properties = $this->toArray();
			$properties['shipment'] = true;
			$properties['shippable_item'] = true;
			$properties['details'] = true;
			return array_keys($properties);
		}
	}
?>
