<?php
	/**
	*	Automatically update exchange rates 
	*	
	*	exchange_update will pull current exchange rates off the web,
	*	and update the exchange rates in pERP.  You can pick the datasouce
	*	from a list of sources.
	*	This can be set to go automatically using asyncservice if asyncservice
	*	is installed.
	*
	*	@author Nathan Gray
	*	@see exchange_datasource
	*	@ingroup perp_api
	*	@ingroup exchange_update
	*/

	class bo_perp_exchange_update {
		/**
		*	Directory where datasources are kept (should be /perp_api/inc/exchange_datasource/)
		*/
		protected static $datasource_dir;

		/**
		*	Name of the datasource to use to get the exchange rates
		*/
		protected $datasource_name;

		/**
		*	Datasource object to use to get the exchange rates
		*/
		protected $datasource;
		
		/**
		*	Array of configuration options
		*/
		protected $config;

		/**
		*	Constructor\n
		*	Sets the datasoruce_dir and reads the datasource to use from config
		*/
		public function bo_perp_exchange_update() {
			$this->datasource_dir = EGW_INCLUDE_ROOT . '/perp_api/inc/exchange_datasource/';
			$this->config =& config::read('perp_api');
			$this->datasource_name = $this->config['exchange_rate_datasource'];
			if(!$this->datasource_name) {
				$this->datasource_name = 'google';
			}
		}

		/**
		*	Get a list of available datasources 
		*	@return array of datasource => name
		*/
		public function get_datasource_list() {
			$list = array();

			// Get a list of classes that can be datasources
			$files = scandir($this->datasource_dir);
			if($files === false) {
				return array();
			}

			foreach($files as $filename) {
				if(!is_file($this->datasource_dir . $filename) ||
					!is_readable($this->datasource_dir . $filename) || 
					substr($filename, -3, 3) != 'php' ||
					$filename == 'class.exchange_datasource.inc.php') {
					continue;
				}
				list($class, $class_name) = explode('.', $filename);

				// We assume if it's a php file in the directory, it's an exchange_datasource
				
				$list[$class_name] = $class_name;
			}
			return $list;
		}

		/**
		*	Get the currently selected datasource
		*	Mostly here for completeness, it will be set via site configuration
		*	@return string Name of the datasource
		*/
		public function get_datasource() {
			return $this->datasource_name;
		}

		/**
		*	Set the datasource to use to get currency exchange rates
		*	Mostly here for completeness, it will be set via site configuration
		*
		*	@param datasource_name string Name of the datasource class to use
		*	@return boolean Successfully set datasource
		*/
		public function set_datasource($datasource_name) {
			// TODO: verify $datasource_name
			$this->datasource_name = $datasource_name;
			config::save_value('exchange_rate_datasource', $datasource_name, 'perp_api');
		}

		/**
		*	Update the exchange rate on all currencies in pERP using the set datasource
		*	@return array of currency_id => new_rate or false on failure
		*/
		public function update_all() {
			if(! ($this->datasource instanceof exchange_datasource)) {
				$this->datasource = $this->get_datasource_class($this->datasource_name);
			}

			if($this->datasource) {
				return $this->datasource->update_all();
			} else {
				return false;
			}
		}

		/**
		*	Update the exchange rate for a given currency
		*
		*	@param currency_id string The 3 letter currency code
		*	@return The new exchange rate or false if the rate could not be determined.
		*/
		public function update($currency_id) {
			if(! ($this->datasource instanceof exchange_datasource)) {
				$this->datasource = $this->get_datasource_class($this->datasource_name);
			}

			if($this->datasource) {
				return $this->datasource->update($currency_id);
			}
		}

		/**
		*	Creates an object for the given datasource name
		*
		*	@param datasource_class string The name of the datasource to create
		*	@return Object datasource class
		*/
		public function get_datasource_class($datasource_class) {
			// Get selected datasouces
			if(!$this->get_datasource()) {
				return false;
			}
			
			require_once($this->datasource_dir . 'class.' . $datasource_class . '.inc.php');

			$datasource = new $datasource_class();
			if($datasource instanceof exchange_datasource) {
				return $datasource;
			} else {
				return false;
			}
		}
	}
?>
