<?php

	/**
	*	Logic layer for working with pERP reports
	*
	*	@author Nathan Gray
	*	@ingroup perp_api
	*	@ingroup perp_report
	*/
	class bo_perp_report {
		/**
		*	Set to true to display extra debug information
		*	@note Debug information will damage the report, and it will not display correctly
		*/
		private $debug = false;

		// Location of the lock file
		private $lock_file = '/tmp/perp_report.lock';

		/**
		*	Constructor\n
		*	bo_perp_report has no storage layer, the reports themselves get their data
		*/
		public function bo_perp_report() {

		}

		/**
		*	Get a list of reports the current user has access to
		*	The application should be set in $query['appname'].
		*
		*	@nextmatch
		*/
		public function get_report_list(&$query = array(), &$rows = array(), &$readonlys = array()) {
			// Get a list of reports the current user has access to
			if(!isset($query['appname'])) {
				$query['appname'] = $GLOBALS['egw_info']['flags']['currentapp'];
			}

			$locations = $GLOBALS['egw']->acl->get_all_location_rights($GLOBALS['egw_info']['user']['account_id'], $query['appname'], True);

			if(!$locations) {
				$rows = array();
				return false;
			}

			// Read config values
			$config = config::read($query['appname']);

			foreach($locations as $location => $rights) {
				if(strpos($location, 'report_') === false || !($rights & EGW_ACL_READ)) {
					unset($locations[$key]);
					continue;
				}

				$row = array(
					'report'	=>	$query['appname'] . '.' . substr($location, 7),
					'report_name'	=>	$location,
					'average_run_time'	=>	round($config[$location]['average_run_time'], 1),
				);
				
				$async_id		=	$row['report'] . '.%';
				if($GLOBALS['egw']->acl->check('run', EGW_ACL_READ, 'admin')) {
					$row['can_adhoc']	=	$config[$location]['can_adhoc'];
					$row['can_schedule']	=	$config[$location]['can_schedule'];
					$admin = true;
				} else {
					$rows['no_can_adhoc']	=	true;
					$rows['no_can_schedule']	=	true;
				}
				$readonlys['cancel[]'] = true;
				$existing = ExecMethod('phpgwapi.asyncservice.read', $async_id);
				if(is_array($existing)) {
					foreach($existing as $id => $schedule) {
						if(!$admin && $schedule['account_id'] != $GLOBALS['egw_info']['user']['account_id']) {
							unset($schedule['account_id']);
						}
						$schedule['escaped_id'] = htmlspecialchars($id);
						$schedule['formatted_next'] = date($GLOBALS['egw_info']['user']['preferences']['common']['dateformat'] . ' ' . 
							($GLOBALS['egw_info']['user']['preferences']['common']['timeformat'] == '24' ? 'H:i' : 'h:i a'),
							$schedule['next']
						);
						$row['schedules'][] = $schedule;
						$readonlys['cancel[' . $schedule['escaped_id'] . ']'] = !($admin || $schedule['account_id'] == $GLOBALS['egw_info']['user']['account_id']);
					}
				}

				$readonlys['adhoc[' . $row['report_name'] . ']']	=	!$config[$location]['can_adhoc'];
				$readonlys['schedule[' . $row['report_name'] . ']']	=	!$config[$location]['can_schedule'];

				/**
				*	Due to the technical restriction of the report converter crashing if
				*	two of them run at the same time, if the user doesn't get their reports
				*	in ODT format, they can't adhoc.
				*	They should be able to schedule for as soon as possible.
				*/
				if('ooo' != $GLOBALS['egw_info']['user']['preferences']['perp_api']['report_output_format']) {
					$readonlys['adhoc[' . $row['report'] . ']']	=	true;
					$rows['no_adhoc']	=	true;
				}

				$rows[] = $row;
			}

			return count($rows);
		}

		/**
		*	Generate a report
		*	
		*	@param report	string	Report name in the form <appname>.<report_name>
		*	@param parameters	array	Parameters for the report
		*	@param data	array	Provide data for the report, if the report doesn't get its own.  Most do.
		*	@return	perp_report / false	If the report is successfully generated, the report is returned.  False on error.
		*/
		public function get_report($report, &$parameters = array(), $data = array()) {
			
			list($appname, $report_name) = explode('.', $report);
			
			if(!$report_name) {
				$report_name = $appname;
				$appname = $GLOBALS['egw_info']['flags']['currentapp'];
			}

			$GLOBALS['egw']->translation->add_app($appname);

			if($this->debug) {
				echo "Application: $appname<br />Report:$report_name<br />";
			}

			// Check ACL (Just on the report, not on the underlying data)
			if($report_name != 'no_report' && !$GLOBALS['egw']->acl->check('report_' . $report_name, EGW_ACL_READ, $appname)) {
				return lang('Access is not allowed to report %1', $report_name);
			}

			$class_file = EGW_INCLUDE_ROOT . '/' . $appname . '/inc/report/class.' . $report_name . '.inc.php';

			if($this->debug) {
				echo 'Debug is on, you won\'t get a file<br />';
				echo "Class: $class_file<br />Template: $template_file<br />";
				echo "Parameters: ";
				_debug_array($parameters);
			}
			if(file_exists($class_file)) {
				include_once($class_file);
				$report = new $report_name($parameters);
			} else {
				echo lang("Couldn't find either report template %1 or class %2", $template_file, $class_file);
			}

			if(is_object($report) && is_array($data) && count($data)) {
				$report->set_data($data);
			} elseif (!is_object($report)) {
				$report = CreateObject('perp_api.perp_report', $parameters);
			}

			return is_object($report) ? $report : false;
		}

		/**
		*	Get a list of supported output formats for reports.
		*	
		*	@return array
		*/
		public function get_supported_formats() {
			$config = config::read('perp_api');

			$list = array(
				'ooo'	=>	'OpenOffice.org'
			);
			if($config['report_java_binary'] != '') {
				$list += array(
					'pdf'	=>	'Adobe pdf',
				//	'rtf'	=>	'Rich Text rtf',
					'ms'	=>	'Microsoft Office'
				);
			}
			return $list;
		}

		/**
		*	Schedule a report for deferred execution
		*
		*	@return async_id, or error message on failure
		*/
		public function schedule_report($appname, $report_name, $schedule_time, $distribution, $settings = array(), $account_id = false) {
			if(strpos($report_name, 'report_') === 0) {
				$report_name = substr($report_name, 7);
			}

			if($settings['async_id']) {
				$async_id = $settings['async_id'];
			}
			if($async_id && !$schedule_time) {
				// No time set, cancel the report
				ExecMethod2('phpgwapi.asyncservice.cancel_timer', $async_id);
			} elseif ($settings['async_id'] && $schedules = ExecMethod('phpgwapi.asyncservice.read', $settings['async_id'])) {
				// report already scheduled, cancel it an save the updated version
				foreach($schedules as $id => $schedule) {
					ExecMethod('phpgwapi.asyncservice.cancel_timer', $id);
				}
			}

			// ID is appname.report_name.owner.{first 200 characters from the settings array}
			unset($settings['async_id']);
			$async_id = $appname . '.' . $report_name . '.' . (int)$account_id . '.' . substr(serialize($settings), 0, 200);
			
			$method = 'perp_api.bo_perp_report.async_report';
			$data = array(
				'async_id'	=>	$async_id,
				'appname'	=>	$appname,
				'report_name'	=>	$report_name,
				'distribution'	=>	$distribution,
				'settings'	=>	$settings
			);

			ExecMethod2('phpgwapi.asyncservice.set_timer', $schedule_time, $async_id, $method, $data, $account_id);

			return $async_id;
		}

		/**
		*	Check to see if the report system is busy
		*/
		public function is_locked() {
			return file_exists($this->lock_file);
		}

		/**
		*	Remove the lock file
		*/
		public function unlock() {
			unlink($this->lock_file);
		}

		/**
		*	Generate a report that has been previously scheduled
		*
		*	@param data An array with keys async_id, distribution and settings
		*/
		public function async_report($data) {
			// Start buffering output, in case there's an error
			ob_start();

			$schedules = ExecMethod('phpgwapi.asyncservice.read', $data['async_id']);
			$schedule = $schedules[$data['async_id']];

			// Maybe already run?
			if($schedule['next'] > time()) {
				// Was run already and the time reset
				return;
			}

			// Check for lock
			$locked = $this->is_locked();

			// Found lock, already generating reports.  OOo reports don't have to wait, it's the conversion that causes problems
			if($locked && $data['settings']['output_format'] != 'ooo') {
				// Log that we're rescheduling
				$GLOBALS['egw']->log->error(array(
					'text'  =>      'W-Report, Had to reschedule report %1',
					'p_1' => $data['async_id'],
					'file'=>__FILE__,'line'=>__LINE__
				));
				$GLOBALS['egw']->log->commit();
	
				// Reschedule this report for later
				$reschedule_id = $data['appname'] . '.' . $data['report_name'] . '.' . $schedule['account_id'] . '.LOCKED-RESCHEDULE';
				$data['async_id'] = $reschedule_id;
				ExecMethod('phpgwapi.asyncservice.cancel_timer', $reschedule_id);

				// We have to increment the time by 1, because otherwise the asyncservice will not schedule it
				ExecMethod2('phpgwapi.asyncservice.set_timer', time() + 1, $reschedule_id, 'perp_api.bo_perp_report.async_report', $data, $schedule['account_id']);
				return false;
			}

			// No lock, create it
			if($data['settings']['output_format'] != 'ooo') {
				touch($this->lock_file);
			}

			// Check for others who want the same report (this report is included)
			$distribution = array();
			$search = "{$data['appname']}.{$data['report_name']}.%." . substr(serialize($data['settings']), 0, 200);
			$schedules = ExecMethod('phpgwapi.asyncservice.read', $search);
			if(!is_array($schedules)) {
				$schedules = array($data['async_id'] => $schedule);
			}
			foreach($schedules as $key => $other_schedule) {
				if($data['settings'] != $other_schedule['data']['settings']) {
					unset($schedules[$key]);
					continue;
				}
				$distribution[] = $other_schedule['data']['distribution'];
				$other_schedule['next'] = ExecMethod2('phpgwapi.asyncservice.next_run', $other_schedule['times'], true);
				ExecMethod2('phpgwapi.asyncservice.write', $other_schedule, true);
			}

			$GLOBALS['egw_info']['flags']['currentapp'] = $data['appname'];

			// Generate the report
			$data['settings']['header'] = 'file';
			$report =& $this->get_report($data['appname'] . '.' . $data['report_name'], $data['settings']);
			
			if(!is_object($report)) {
				echo $report;
				echo 'Report did not get created';
			} else {
				$filename = $report->display();

				// Distribute the report
				foreach($distribution as $owner => $dist) {
					$method = 'distribute_' . $dist['type'];
					$this->$method($filename, $report, $dist, $schedules[$data['async_id']]);
				}

				// Remove the file
				unlink($filename);
			}
			
			// Release the lock
			if($data['settings']['output_format'] != 'ooo') {
				$this->unlock();

				// Check for any reports that were deferred because of locking
				$deferred = ExecMethod('phpgwapi.asyncservice.read', '%.LOCKED-RESCHEDULE');
				if(is_array($deferred) && count($deferred) > 0) {
					// Not sure if anything needs to be done here.
				}
			}
			ob_end_clean();
		}

		/**
		*	Email a report out
		*/
		public function distribute_email($filename, &$report, $settings, $async) {
			require_once( EGW_API_INC . '/class.phpmailer.inc.php');
			$mail = new PHPMailer();

			$mail->From = $settings['from'];
			$mail->FromName = $settings['from_name'] ? $settings['from_name'] : $GLOBALS['egw_info']['server']['site_title'];

			// Email to requestor
			$contact = ExecMethod('addressbook.addressbook_bo.read', 'account:' . ($async['account_id'] ? $async['account_id'] : $GLOBALS['egw_info']['user']['account_id']));
			$mail->AddAddress($contact['email'], $contact['n_fileas']);

			// Check for others
			if(is_array($settings['to'])) {
				foreach($settings['to'] as $address) {
					$mail->AddAddress($address, $name);
				}
			}
			$mail->Subject = $settings['subject'] ? $settings['subject'] : lang('perp_report_mail_subject Report: %1 Domain: %2', 
				lang('report_' . get_class($report)),
				$GLOBALS['egw_info']['user']['domain']
			);
			$replace = "\n";
			$search = '|<br ?/?>|';

			if($settings['body']) {
				$mail->Body = str_replace('<br />', "\n", $settings['body']);
			} else {
				$mail->Body = lang('perp_report_mail_body Report: %1 Domain: %2', 
					lang('report_' . get_class($report)),
					$GLOBALS['egw_info']['user']['domain']
				);
				$mail->Body = preg_replace($search, $replace, $mail->Body);

				$mail->Body .= "\n\n" . lang('This report was automatically generated at your request.') . "\n";
				if(is_array($async['times'])) {
					$mail->Body .= lang('It will be generated again on %1', 
						date(
							$GLOBALS['egw_info']['user']['preferences']['common']['dateformat'] . ' ' .
							($GLOBALS['egw_info']['user']['preferences']['common']['timeformat'] == '24' ? 'H:i' : 'h:i a'),
							ExecMethod('phpgwapi.asyncservice.next_run', $async['times'])
						)
					) . "\n";
				} else {
					$mail->Body .= lang('It will not be generated again unless you request it.');
				}
			}
			$mail->Body .= "\n\n";

			if(file_exists($filename)) {
				$mail->AddAttachment($filename, $report->get_filename(), 'base64', $report->get_mime());
				if($settings['infolog_id']) {
					$infolog = ExecMethod('infolog.infolog_bo.read', $settings['infolog_id']);
					if($infolog) {
						egw_link::link('infolog',$settings['infolog_id'],'file',$filename);
						$infolog['info_status'] = 'done';
						ExecMethod('infolog.infolog_bo.write', $infolog);
					}
				}
			} else {
				$mail->Body .= lang('There was an problem with your report.  The following may help determine what happened:');
				$mail->Body .= "\n" . ob_get_contents() . "\n";
				$mail->Body .= "\n\nAsync:\n" . print_r($async, true) . "\n\nReport:\n" . print_r($report, true);
			}

			$mail->Send();
		}
	}
?>
