<?php

	/** 
	*	Base class for pERP Reporting 
	*	Reporting uses TinyButStrong OpenOffice.org
	*	http://www.tinybutstrong.com/apps/tbsooo/doc.html
	*
	*	Reports should extend this class and override:
	*		-# $template
	*		-# get_data()
	*
	*	@author Nathan Gray
	*	@ingroup perp_api
	*	@defgroup perp_report pERP Reports
	*/

	include_once(EGW_INCLUDE_ROOT . '/perp_api/inc/report/TBSOOo/tbs_class.php');
	include_once(EGW_INCLUDE_ROOT . '/perp_api/inc/report/ooo.inc.php');

	class perp_report {

		/** 
		* @name TinyButStrong OpenOffice.org variables
		* @{
		*/
		private $header = 'content';
		protected $OOo = null;
		private $work_directory = '/tmp';
		// Need to setup a user perf fo inline or attachment to force
		// pdfs and such to downlaod and not display inline
		private $fileaction = 'attachment';
		// currently supported values are
		// ooo, pdf, ms
		// for all others soffice server must be listenning
		// Xvfb :1 &; soffice -quickstart -headless -accept="socket,port=8100;urp;" -display :1 &
		private $output_format = 'ooo';
		// future use for remote OOo server
		// private $soffice_port = '8100';
		// private $soffice_host = 'localhost';
		/* @} */

		/**
		*	@name Global report variables
		*	@{
		*/

		/**
		*	Data used to generate the report
		*	Most reports will override get_data() to provide this themselves
		*/
		protected $data = array();

		/**
		*	Parameters used to generate the report
		*/
		protected $parameters = array();

		/**
		*	Filename (without extension) for the completed report
		*/
		protected $filename = '';

		/* @} */

		/**
		*	The OpenOffice.org file used as a template to create the file
		*	Report-specific, override this.  
		*	The API will take care of multi-lingualism, and checking for a local customization
		*	in the default directory.
		*/
		protected $template= '';
		protected $output_ext='';
		protected $input_ext='';

		/**
		*	Constructor\n
		*	Creates a report object, also creates the TBSOOo object used to create the report
		*	@param 	parameters	array	Parameters used to get the data desired in the report
		*/
		public function perp_report(array $parameters = array()) {
			// Set paths
			$config = config::read('perp_api');
			$this->work_directory = $config['report_work_directory'];
			$java_binary = $config['report_java_binary'];

			// Massage string dates into timestamps (now, yesterday, etc)
			foreach($parameters as $name => $value) {
				if(!is_numeric($value) && strpos($name, 'date') !== false) {
					$parameters[$name] = $this->convert_date($value);
					$parameters[$name . '-string'] = $value;
				}
			}
			$this->parameters = $parameters;

			// Output format
			if(!$this->parameters['output_format']) {
				$this->output_format = $GLOBALS['egw_info']['user']['preferences']['perp_api']['report_output_format'];
			} else {
				$this->output_format = $this->parameters['output_format'];
			}

			// Set default data
			$this->data[0] = array(
				'template_missing'	=>	lang('template is either not set or missing'),
			);

			if(!$this->parameters['header']) {
				$this->header = (isset($_GET['header']) ? $_GET['header'] : 'content');
			} else {
				$this->header = $this->parameters['header'];
			}

			// instantiate a pERP OOo class
			$this->OOo = new pERP_OOo();

			// set the process directory
			if ($this->OOo->SetProcessDir($this->work_directory) == false) {
				die(lang('Unable to set working directory (%1)', $this->work_directory));
			}
			// set the java binary
			if ($this->OOo->SetJavaBinary($java_binary) == false) {
				die(lang('Unable to set java binary (%1)', $java_binary));
			}
			// Set the charset encoding
			$this->OOo->SetDataCharset('UTF8');
		}

		// Common functions

		/**
		*	Return the OpenOffice.org file used as a template
		*
		*	@return string	the name of the OpenOffice.org file
		*/
		public function get_template() {
			// Check subdirectories in order
			$subdirs = array(
				'default',
				strtolower($GLOBALS['egw_info']['user']['preferences']['common']['lang']),
				'en',
			);
			$class = new ReflectionClass(get_class($this));
			$base = dirname($class->getFileName());

			foreach($subdirs as $dir) {
				if(file_exists($base . SEP . $dir . SEP . $this->template)) {
					return $base . SEP . $dir . SEP . $this->template;
				}
			}

			// Check for template
			if(!$this->template) {
				$this->template = EGW_INCLUDE_ROOT . '/perp_api/inc/report/no_report.odt';
				$this->input_ext = 'odt';
				$this->output_ext = $this->get_output_filext();
			}
			return ($this->template);
		}

		/**
		*	Set the OpenOffice.org file to be used as a template
		*	@param template	string The name of the OpenOffice.org file
		*/
		public function set_template($template) {
			$this->template = $template;
			$input_ext_array = explode(".", basename($this->template));
			$this->input_ext = $input_ext_array[count($input_ext_array) - 1];
			$this->output_ext = $this->get_output_filext();
		}

		/**
		*	Get the parameters for this report
		*	@return array The parameters that are / will be used for this report
		*/
		public function get_parameters() {
			return $this->parameters;
		}


		/**
		*	Return the data for the report.  
		*	This will be array() unless set_data() has been called or get_data() overridden.
		*	Most implementing classes should override this function to provide their own data.
		*
		*	@param query array parameters
		*	@return array Report data
		*/
		public function get_data($query = array()) {
			return $this->data;
		}

		/**
		*	Set the data for the report.  This function is not needed if the report generates its own data.
		*
		*	@param new_data array The data for the report
		*/
		public function set_data(array $new_data) {
			$this->data = $new_data;
		}

		/**
		*	Get the destination file neame for the report.
		*
		*	@return Filename, without extension
		*/
		public function get_filename() {
			return ($this->filename != '' ? $this->filename . '.' : rtrim(basename($this->template), $this->input_ext)).$this->output_ext;
		}

		/**
		*	Get the file mime type based on output_ext
		*
		*	@return mime string
		*/
		public function get_mime() {
			switch($this->output_ext) {
				case 'pdf'; return 'application/pdf'; break;
				case 'doc': return 'application/msword'; break;
				case 'rtf': return 'text/rtf'; break;
				// OpenOffice mimetypes
				case 'sxw': return 'application/vnd.sun.xml.writer'; break;
				case 'stw': return 'application/vnd.sun.xml.writer.template'; break;
				case 'sxg': return 'application/vnd.sun.xml.writer.global'; break;
				case 'sxc': return 'application/vnd.sun.xml.calc'; break;
				case 'stc': return 'application/vnd.sun.xml.calc.template'; break;
				case 'sxi': return 'application/vnd.sun.xml.impress'; break;
				case 'sti': return 'application/vnd.sun.xml.impress.template'; break;
				case 'sxd': return 'application/vnd.sun.xml.draw'; break;
				case 'std': return 'application/vnd.sun.xml.draw.template'; break;
				case 'sxm': return 'application/vnd.sun.xml.math'; break;
				case 'odt': return 'application/vnd.oasis.opendocument.text'; break;
				case 'ott': return 'application/vnd.oasis.opendocument.text-template'; break;
				case 'oth': return 'application/vnd.oasis.opendocument.text-web'; break;
				case 'odm': return 'application/vnd.oasis.opendocument.text-master'; break;
				case 'odg': return 'application/vnd.oasis.opendocument.graphics'; break;
				case 'otg': return 'application/vnd.oasis.opendocument.graphics-template'; break;
				case 'odp': return 'application/vnd.oasis.opendocument.presentation'; break;
				case 'otp': return 'application/vnd.oasis.opendocument.presentation-template'; break;
				case 'ods': return 'application/vnd.oasis.opendocument.spreadsheet'; break;
				case 'ots': return 'application/vnd.oasis.opendocument.spreadsheet-template'; break;
				case 'odc': return 'application/vnd.oasis.opendocument.chart'; break;
				case 'odf': return 'application/vnd.oasis.opendocument.formula'; break;
				case 'odb': return 'application/vnd.oasis.opendocument.database'; break;
				case 'odi': return 'application/vnd.oasis.opendocument.image'; break;
				default:    return ''; break;
			}
		}

		/**
		*	Return the output file extension
		*	Based on the input template and output family,
		*	determine the output file extension and return it
		*
		*	@return file_extension string
		*/
		private function get_output_filext() {
			switch ($this->output_format) {
				case 'ooo':
					switch ($this->input_ext) {
						case 'odt':
							$fileext = 'odt';
							break;
						case 'ods':
							$fileext = 'ods';
							break;
					}
					break;
				case 'ms':
					switch ($this->input_ext) {
						case 'odt':
							$fileext = 'doc';
							break;
						case 'ods':
							$fileext = 'xls';
							break;
					}
					break;
				case 'pdf':
					$fileext = 'pdf';
					break;
			}

			return $fileext;
		}

		/** 
		*	Display the report
		*	Includes appropriate headers.  You shouldn't output
		*	anything before this.
		*/
		public function display() {
			// Start timer
			$start_time = microtime(true);

			// Clear stats before checking file_exists
			 clearstatcache();

			// Check for template
			if(!file_exists($this->get_template())) {
				$data['template_missing']	= lang('template is either not set or missing');
				$data['template_name']		= $this->template;
				$this->template = EGW_INCLUDE_ROOT . '/perp_api/inc/report/no_report.odt';
			}

			if (!$this->input_ext || !$this->output_ext) {
				$this->set_template($this->template);
			}

			$data = (isset($data) ? $data : $this->get_data($this->get_parameters()));

			// Check for empty data
			if($data === false) {
				$data = array(
					'template_missing'	=>	lang('Report is missing a parameter, or no data was found.'),
					'template_name'		=>	lang('report_' . get_class($this)) . "\n\n\n\n" . print_r($this->get_parameters(), true)
				);
				$this->template = EGW_INCLUDE_ROOT . '/perp_api/inc/report/no_report.odt';
			}

			// Create new document
			$this->OOo->NewDocFromTpl($this->get_template());

			$this->OOo->LoadXmlFromDoc('content.xml');
			$this->OOo->MergeBlock('data', 'array', array(0=>$data));

			if(!is_array($data)) {
				$data = array();
			}

			foreach($data as $key => $block) {
				if(is_string($key) && is_array($block) && !isset($block['source_type'])) {
					$this->OOo->MergeBlock($key, 'array', $data[$key]);
				} elseif(is_string($key) && is_array($block) && isset($block['source_type'])) {
					$this->OOo->MergeBlock($key, $block['source_type'], $block['query']);
				}
			}
			$this->OOo->SaveXmlToDoc();

			// Image Replacement
			$this->OOo->Image_Replace('content.xml', $data['images']);

			// ===== xml meta ==============================
			$this->OOo->LoadXmlFromDoc('meta.xml');
			$this->OOo->SaveXmlToDoc();

			// Styles data
			$this->OOo->LoadXmlFromDoc('styles.xml');
			$this->OOo->SaveXmlToDoc();
			$this->OOo->Image_Replace('styles.xml', $data['images']);

			// check if a convertion is required
			$this->OOo->ConvertDoc($this->output_format, $this->output_ext);

			// odt file with full path
			$file = $this->OOo->GetPathnameDoc();

			// Stop timer
			$stop_time = microtime(true);

			// Update timing history
			$config = config::read($GLOBALS['egw_info']['flags']['currentapp']);
			$report_config = $config['report_' . get_class($this)];
			if($report_config) {
				if(!is_array($report_config['run_time'])) {
					$report_config['run_time'] = array();
				}
				$report_config['run_time'][] = $stop_time - $start_time;
				while(count($report_config['run_time']) > 10) {
					array_shift($report_config['run_time']);
				}
				$report_config['average_run_time'] = (
					array_sum($report_config['run_time']) / 
					count($report_config['run_time'] )
				);
				config::save_value('report_' . get_class($this), $report_config, $GLOBALS['egw_info']['flags']['currentapp']);
			}

			// Output
			switch ($this->header) {
				// Keep it around, caller should move it
				case 'file':
					return $file;
				break;

				// method 1 : send to the browser via header location (problem with IE)
				case 'location':
					header('Location:'.$file);
				break;

				// method 2 : send to the browser via header content-type
				case 'content':
				default:
					header('Content-type: '.$this->get_mime());
					header('Content-Length: '.filesize($file));
					header('Content-Disposition: '.$this->fileaction.'; filename="' . $this->get_filename() . '"');
					$this->OOo->FlushDoc();
				break;
			}

			// Done with these files, do a cleanup now
			$this->OOo->RemoveDoc();

			// clean old files in process dir (default files older than 2 hours)
			// Don't use this if you use a system directory like /tmp
			//$this->OOo->ClearProcessDir(2, 0);

		}

		/**
		*	When a report is scheduled for delayed execution, the settings are usually
		*	different from an adhoc report.  Dates in particular will be different for
		*	recurring reports.  
		*
		*	If there are settings already stored in the async services, the dates will 
		*	be recalculated based on the next run date and returned.  If not, defaults will be used.  
		*	Either way, this function should return output suitable for display in an
		*	eTemplate, so options need to be provided for selectboxes.
		*
		*	@param async_id The ID of the asyncronous job.  
		*/
		public function get_schedule_settings($async_id = null) {
			$settings = array();

			$async = ExecMethod('phpgwapi.asyncservice.read', $async_id);
			$async = $async[$async_id];
		
			// Try to be intelligent about this
			foreach($this->parameters as $name => $value) {
				if(strpos($name, 'date') !== false) {
					$settings['options-' . $name] = array(
						'now'		=>	'now',
						'yesterday'	=>	'yesterday',
						'tomorrow'	=>	'tomorrow',
						'last Monday'	=>	'Last Monday',
						'last Friday'	=>	'Last Friday',
						'this_month_start'	=>	'Start of this month',
						'this_month_end'	=>	'End of this month',
						'last_month_start'	=>	'Start of last month',
						'last_month_end'	=>	'End of last month',
						'this_year_start'	=>	'Start of this year',
						'this_year_end'		=>	'End of this year',
						'last_year_start'	=>	'Start of last year',
						'last_year_end'		=>	'End of last year',

					);
				}
			}

			foreach($settings as $name => &$value) {
				if(strpos($name, 'options-') !== false && strpos($name, 'date') !== false) {
					if(is_array($value)) {
						foreach($value as $key => &$option) {
							$option = lang($option);
							$option .= ' (' . date(
								$GLOBALS['egw_info']['user']['preferences']['common']['dateformat'] . ' H:i', 
								$this->convert_date($key, $async['next'])
							) . ')';
						}
					}
				}
			}

			return $settings;
		}

		/**
		*	Convert a text date into a timestamp, with more handling than strtotime
		*/
		public function convert_date($time, $now = null) {
			if(is_null($now)) {
				$now = time();
			}
			switch ($time) {
				case 'this_month_start':
					$time = mktime(0,0,0, date('m', $now), 1, date('Y', $now));
					break;
				case 'this_month_end':
					$time = mktime(23,59,59, date('m', $now), date('t', $now), date('Y', $now));
					break;
				case 'last_month_start':
					$time = mktime(0,0,0, date('m', $now) - 1, 1, date('Y', $now));
					break;
				case 'last_month_end':
					$time = mktime(23,59,59, date('m', $now) -1 , date('t', strtotime('-1 month', $now)), date('Y', $now));
					break;
				case 'next_month_start':
					$time = mktime(0,0,0, date('m', $now) + 1, 1, date('Y', $now));
					break;
				case 'next_month_end':
					$time = mktime(23,59,59, date('m', $now) + 1 , date('t', strtotime('+1 month', $now)), date('Y', $now));
					break;
				case 'this_year_start':
					$time = mktime(0,0,0, 1, 1, date('Y', $now));
					break;
				case 'this_year_end':
					$time = mktime(23,59,59, 12, 31, date('Y', $now));
					break;
				case 'last_year_start':
					$time = mktime(0,0,0, 1, 1, date('Y', $now) - 1);
					break;
				case 'last_year_end':
					$time = mktime(23,59,59, 12, 31, date('Y', $now) - 1);
					break;

				default:
					$time = strtotime($time, $now);
					break;
			}

			return $time;
		}
	}
?>
