<?php

	/**
	*       Base class for pERP Reporting
	*       Reporting based on TinyButStrong OpenOffice.org
	*       http://www.tinybutstrong.com/apps/tbsooo/doc.html
	*
	*       @author Nathan Gray
	*       @ingroup perp_api
	*       @defgroup perp_report pERP Reports
	*/

	class pERP_OOo extends clsTinyButStrong {
		/**
		*	PRIVATE VARs
		*/
		private $_process_path = 'tmp/';
		private $_charset = '';
		private $_tempdir = '';
		private $_template_file_ext = '';
		private $_file_ext = '';
		private $_file_format = '';
		private $_xml_filename = '';
		private $java_binary = '';

		/**
		*	PUBLIC FUNCTIONS
		*/

		/**
		*	Set the java binary (path and file)
		*
		*	@param string full path and filename to the java binary
		*
		*	@return true/false
		*/
		public function SetJavaBinary($java_binary) {
			// Add a test here make sure it's exec and a proper java
			$this->java_binary = $java_binary;

			return true;
		}

		/**
		*	Set the temp dir to build reports in, webserver will
		*	require write permissions.
		*
		*	@param string full path and filename to temp dir
		*
		*	@return true/false
		*/
		public function SetProcessDir($process_path) {
			clearstatcache();

			// set the directory for processing temporary OOo files
			if ($process_path == '') {
				$this->meth_Misc_Alert('SetProcessDir method', 'Parameter is empty');
				return false;
			}

			// add a trailing / at the path
			$this->_process_path = $process_path.(substr($process_path, -1, 1) == '/' ? '' : '/');
    
			// test if 'dir' exists
			if (!is_dir($this->_process_path)) {
				$this->meth_Misc_Alert('SetProcessDir method', 'Directory not found : '.$this->_process_path);
				return false;
			}

			// test if 'dir' is writable
			if (!is_writable($this->_process_path)) {
				$this->meth_Misc_Alert('SetProcessDir method', 'Directory not writable : '.$this->_process_path);
				return false;
			}
			return true;
		}

		/**
		*	Set the default character set for OOo documents
		*
		*	@param string the charset for the report
		*		Current values are:
		*			-UTF8
		*			-ISO 8859-1
		*/
		public function SetDataCharset($charset) {
			$this->_charset = strtoupper($charset);
		}

		/**
		*	Create a temp odt file from the template odt in the
		*	temp dir
		*
		*	@param string odt template full path and filename
		*
		*	@return string/false temp path and filename
		*/
		public function NewDocFromTpl($template_filename) {
			clearstatcache();

			// test if OOo source file exist
			if (!file_exists($template_filename)) {
				$this->meth_Misc_Alert('NewDocFromTpl method', 'File not found : '.$template_filename);
				return false;
			}

			// create unique ID
			$unique = md5(microtime());

			// find path, file and extension
			$a_pathinfo = pathinfo($template_filename);
			$this->_template_file_ext = $a_pathinfo['extension'];
			$this->_tempdir = $this->_process_path.$unique;

			// create unique temporary dir
			if (!mkdir($this->_tempdir, 0700)) {
				$this->meth_Misc_Alert('NewDocFromTpl method', 'Can\'t create directory : '.$this->_tempdir);
				return false;
			}

			// copy the ooo template into the temporary dir
			if (!copy($template_filename, $this->_tempdir.'.'.$this->_template_file_ext)) {
				$this->meth_Misc_Alert('NewDocFromTpl method', 'Can\'t copy file to process dir : '.$template_filename);
				return false;
			}

			return $this->_tempdir.'.'.$this->_template_file_ext;
		}

		/**
		*	Inject values into the report template in the temp
		*	report file
		*
		*	@param string filename for the extracted xml file from
		*	the temp report file
		*
		*	@return true/false
		*/
		public function LoadXmlFromDoc($xml_file) {
			clearstatcache();

			$this->_xml_filename = $xml_file;

			// unzip the XML files
			// exec($this->_unzip_bin.' '.$this->_tempdir.'.'.$this->_template_file_ext.' -d '.$this->_tempdir.' '.$this->_xml_filename);
			if (!$this->_unzip()) {
				$this->meth_Misc_Alert('LoadXmlFromDoc method', 'File not found : '.$this->_tempdir.'/'.$this->_template_file_ext);
				return false;
			}

			// test if XML file exist
			if (!file_exists($this->_tempdir.'/'.$this->_xml_filename)) {
				$this->meth_Misc_Alert('LoadXmlFromDoc method', 'File not found : '.$this->_tempdir.'/'.$this->_xml_filename);
				return false;
			}

			// load the template
			$this->ObjectRef = &$this;
			$this->LoadTemplate($this->_tempdir.'/'.$this->_xml_filename, '=~_CharsetEncode');

			// work around - convert apostrophe in XML file needed for TBS functions
			$this->Source = str_replace('&apos;', '\'', $this->Source);

			return true;
		}

		/**
		*	Rebuild the temp report file in odt format
		*
		*	@return true/false
		*/
		public function SaveXmlToDoc() {
			clearstatcache();

			// get the source result
			$this->Show(TBS_NOTHING);

			// store the merge result in place of the XML source file
			$fdw = fopen($this->_tempdir.'/'.$this->_xml_filename, "w");

			// Multi-byte characters cause issues
			if((extension_loaded('mbstring') ||@dl(PHP_SHLIB_PREFIX.'mbstring.'.PHP_SHLIB_SUFFIX)) && ((int) ini_get('mbstring.func_overload')) & 2) {
				$length = mb_strlen($this->Source, 'latin1');
			} else {
				$length = strlen($this->Source);
			}
			fwrite($fdw, $this->Source, $length);
			fclose ($fdw);

			// test if XML file exist
			if (!file_exists($this->_tempdir.'/'.$this->_xml_filename)) {
				$this->meth_Misc_Alert('SaveXmlToDoc method', 'File not found : '.$this->_tempdir.'/'.$this->_xml_filename);
				return false;
			}

			// zip and remove the file
			// exec($this->_zip_bin.' -j -m '.$this->_tempdir.'.'.$this->_template_file_ext.' '.$this->_tempdir.'/'.$this->_xml_filename);
			if(!$this->_zip()) {
				$this->meth_Misc_Alert('SaveXmlToDoc method', 'File not found : '.$this->_tempdir.'/'.$this->_template_file_ext);
				return false;
			}

			return true;
		}

		/**
		*	Convert from odt to user specified format
		*
		*	@param string report file format
		*		Currently avail formats are:
		*			-pdf (Protable Document Format)
		*			-ooo (Open Office)
		*			-ms (Microsoft Office)
		*		string report file extension
		*/
		public function ConvertDoc($file_format, $file_ext) {
			$this->_file_format = $file_format;
			$this->_file_ext = $file_ext;

			// This should have a better check as TBS can output
			// more then odt, so this should only be for others
			$command = $this->java_binary.' -jar '.EGW_INCLUDE_ROOT.'/perp_api/inc/report/jodconverter-2.2.1/lib/jodconverter-cli-2.2.1.jar "'. $this->_tempdir.'.'.$this->_template_file_ext.'" "'.$this->_tempdir.'.'.$this->_file_ext . '"' . (defined('CONVERT_LOG') ? ' 2>&1' : '');
			if ($this->_file_format != 'ooo') {
				$handle = popen($command, 'r');

				// Wait and check for errors
				$errors = '';
				while (!feof($handle)) {
					set_time_limit(20);
					$results = fgets($handle, 256);
					$errors .= $results;

					// avoid a busy wait
					sleep(1);
				}
				pclose($handle);

				// Log errors
				if(strlen($errors)) {
					$severity = 'I';
					if(strpos($errors, 'Exception') !== false) {
						$severity = 'E';
					}
					$GLOBALS['egw']->log->error(array(
						'text'	=>	$severity . '-Report, %1', 'p_1' => nl2br($errors), 'file'=>__FILE__,'line'=>__LINE__
					));
					$GLOBALS['egw']->log->commit();
				}

				// Remove old file after convert
				unlink($this->_tempdir.'.'.$this->_template_file_ext);
			}
		}

		/**
		*	Get full filename and path of finish report file
		*
		*	@return string full path and filename
		*/
		public function GetPathnameDoc() {
			// remove tmp dir
			$this->_RemoveTmpBasenameDir();

			// return path
			return $this->_tempdir.'.'.$this->_file_ext;
		}

		/**
		*	Replacement for readfile, to send the finish report
		*/
		function FlushDoc() {
			$fp = @fopen($this->GetPathnameDoc(), 'rb');
			fpassthru($fp);
			fclose($fp);
		}

		/**
		*	Remove finished report file from temp dir
		*/
		public function RemoveDoc() {
			// remove file
			unlink($this->GetPathnameDoc());
		}

		/**
		*	Remove all files/dirs in the tmp dir in a given time
		*
		*	@param (int, int) (hours, minutes)
		*/
		public function ClearProcessDir($hour = '2', $minut = '0') {
			clearstatcache();

			$now = mktime(date("H")-abs((int)$hour), date("i")-abs((int)$minut), date("s"), date("m"), date("d"), date("Y"));

			if ($dir = @opendir($this->_process_path)) {
				while (($file = readdir($dir)) !== false)  {
					if ($file != ".." && $file != ".") {
						if (filemtime($this->_process_path.$file) < $now) {
							if (!(is_dir($this->_process_path.$file) ? @rmdir($this->_process_path.'/'.$file) : @unlink($this->_process_path.$file))) {
								$this->meth_Misc_Alert('ClearProcessDir method', 'Can\'t remove directory or file : '.$this->_process_path.$file);
							}
						}
					}
				}

				closedir($dir);
			}
		}

		/**
		*	Encode strings to the proper OOo charset
		*
		*	@param string raw string from the template
		*
		*	@return string encoded string
		*/
		public function _CharsetEncode($string_encode) {
			$string_encode = str_replace('&' ,'&amp;', $string_encode);
			$string_encode = str_replace('<' ,'&lt;', $string_encode);
			$string_encode = str_replace('>' ,'&gt;', $string_encode);
			//$string_encode = str_replace("\n", '</text:p><text:p>', $string_encode); // '\n' by XML tags
			$string_encode = str_replace("\n", '<text:line-break/>', $string_encode); // '\n' by XML tags

			// OOo XML charset is utf8
			switch($this->_charset) {
				// no encode
				case 'UTF8': break;
				// encode ISO 8859-1 to UTF8
				case 'ISO 8859-1': 
				default: $string_encode = utf8_encode($string_encode); break;
			}

			// work-around for EURO caracter
			$string_encode = str_replace(chr(0xC2).chr(0x80) , chr(0xE2).chr(0x82).chr(0xAC),  $string_encode); // 

			return $string_encode;
		}

		/**
		*	PRIVATE FUNCTIONS
		*/

		/**
		*	remove tmp dir
		*/
		private function _RemoveTmpBasenameDir() {
			clearstatcache();

			if (is_dir($this->_tempdir) && !rmdir ($this->_tempdir)) {
				$this->meth_Misc_Alert('_RemoveTmpDir method', 'Can\'t remove directory : '.$this->_tempdir);
			}
		}

		/**
		*	unzip the odt so the xml file can be modified
		*
		*	@return true or false
		*/
		private function _unzip() {
			$zip = new ZipArchive;
			if ($zip->open($this->_tempdir.'.'.$this->_template_file_ext) === TRUE) {
				$zip->extractTo($this->_tempdir, $this->_xml_filename);
				$zip->close();
				return true;
			} else {
				$this->meth_Misc_Alert('_unzip method', 'File not found : '.$this->_tempdir.'/'.$this->_template_file_ext);
				return false;
			}
		}

		/**
		*	zip the xml file to create a proper odt
		*
		*	@return true or false
		*/
		private function _zip() {
			$zip = new ZipArchive;
			if ($zip->open($this->_tempdir.'.'.$this->_template_file_ext) === TRUE) {
				$zip->addFile($this->_tempdir.'/'.$this->_xml_filename, $this->_xml_filename);
				$zip->close();
				unlink($this->_tempdir.'/'.$this->_xml_filename);
				return true;
			} else {
				$this->meth_Misc_Alert('_zip method', 'File not found : '.$this->_tempdir.'/'.$this->_template_file_ext);
				return false;
			}
		}

		/**
		*	An attempt to force OO.o to use the correct data types
		*
		*	Types are one of:
		*		currency
		*		float
		*		percentage
		*	Maybe more.
		*/
		function meth_Locator_Replace(&$Txt, &$Loc, &$Value, $CheckSub) {
			if(!$Loc->PrmLst['type']) {
				return parent::meth_Locator_Replace($Txt, $Loc, $Value, $CheckSub);
			}
			
			// Keep these, we have to reset when we're done
			$value_beginning =	$Loc->PosBeg;
			$value_ending	=	$Loc->PosEnd;
			$value_length	=	$value_ending - $value_beginning;

 			// Get container - cell only
			tbs_Locator_EnlargeToStr($Txt,$Loc, '<table:table-cell', '>');

			// Attributes to set
			$cell_value = $Value;
			foreach($Loc->SubLst as $sub) {
				$cell_value = $cell_value[$sub];
			}
			$attributes = array(
				'office:value'		=>	$cell_value,
				'office:value-type'	=>	$Loc->PrmLst['type']
			);
			if($Loc->PrmLst['type'] == 'currency') {
				if($Loc->PrmLst['currency']) {
					$attributes['office:currency'] = $Loc->PrmLst['currency'];
				} else {
					$config = config::read('perp_ledger');
					$attributes['office:currency'] = $config['gl_currency'];
				}
			}

			// Set / reset attributes
			$hack_text = substr($Txt, $Loc->PosBeg, $value_ending - $Loc->PosBeg);
			$hack_length = strlen($hack_text);
			foreach($attributes as $attribute => $value) {
				$count = 0;
				if (preg_match("/$attribute=/", $hack_text)) {
					if ($addlast) {
						$hack_text =& preg_replace("/$attribute=\"(.*?)\"/", "$attribute=\"$value\" $lastattr=\"$lastvalue\"", $hack_text, 1, $count);
						unset($addlast);
						unset($lastattr);
						unset($lastvalue);
					} else {
						$hack_text =& preg_replace("/$attribute=\"(.*?)\"/", "$attribute=\"$value\"", $hack_text, 1, $count);
					}
				} else {
					$addlast = true;
					$lastattr = $attribute;
					$lastvalue = $value;
				}
			}

			// Replace the original bit with the hacked up version
			$Txt =& substr_replace($Txt, $hack_text, $Loc->PosBeg, $value_ending - $Loc->PosBeg);
			
			// Correct Loc to include changes
			$hack_length = strlen($hack_text) - $hack_length;
			$Loc->PosBeg = $value_beginning + $hack_length;
			$Loc->PosEnd = $value_ending + $hack_length;
			
			// Let the parent insert the value
			$new_end= parent::meth_Locator_Replace($Txt, $Loc, $Value, $CheckSub);

			// Reset Loc to before I messed with it so subsequent lines work
			$Loc->PosBeg = $value_beginning;
			$Loc->PosEnd = $value_ending;

			return $new_end;
		}

		/**
		*	Replace images in the ODF file with those provided
		*
		*	Array should be in the form of Array(
		*		'name'	=>	OO.o image name
		*		'path'	=>	Replacement image
		*	)
		*
		*	Optional keys include x, y, width, height
		*		
		*	$param replacements array of file replacements.  
		*/
		public function Image_Replace($file, $replacements = array()) {
			if(!is_array($replacements)) {
				return;
			}

			// Pull content.xml from the zip
			$this->LoadXmlFromDoc($file);

			// Load it into the XML parser
			$xml = simplexml_load_file($this->_tempdir.'/'.$this->_xml_filename);

			// Handy reference of all namespaces in the document, useful for getting anything out
			$namespaces = $xml->getDocNamespaces();

			foreach($replacements as $replace) {
				$path = $replace['path'];
				if(stripos($replace['path'], 'http') !== false) {
					$url = parse_url($replace['path']);

					// Check to see if this is a local file referenced by a URL
					if (!file_exists(EGW_INCLUDE_ROOT . '/./' . $url['path'])) {
						// Not a local file, get it

						// Can't handle this yet.  Quietly log a warning, so the report still works
						$GLOBALS['egw']->log->error(array(
							'text'	=>	'W-Report, Image_Replace \'%2\'  File does not look local: %3', 
							'p_1' => '',
							'p_2' => $replace['name'],
							'p_3' => $path,
							'file'=>__FILE__,'line'=>__LINE__
						));
					} else {
						$query = explode('&', urldecode($url['query']));
						$path = EGW_INCLUDE_ROOT . '/./' . $url['path'];

						if (file_exists($path) && !file_exists($path . ' ' . implode(' ', $query))) {
							// File is not there.  Maybe it's a PHP script that makes an image? (barcode)
							exec('php ' . $path . ' ' . implode(' ', $query) . ' >> ' . $this->_tempdir . '/' . $replace['name']);
							$path = $this->_tempdir . '/' . $replace['name'];
						}
					}
				}
				
				// Get a list of images with that name in the document
				$images = $xml->xpath("//draw:frame[@draw:name='" . $replace['name'] . "']/draw:image");
				if(!is_array($images)) {
					$images = array();
				}
				foreach($images as $image) {

					// Get parent node, which is the containing frame
					$frame =& $image->xpath('..');
					$frame =& $frame[0];
/*
echo $path . '<br />';
echo htmlspecialchars($frame->asXML()) . '<br />';
//*/
					foreach($frame->attributes($namespaces['svg']) as $key => $value) {
						// simpleXML can't handle it, go to DOM
						if(isset($replace[$key])) {
							$element = dom_import_simplexml($frame);
							$element->setAttribute('svg:' . $key, $replace[$key]);
						}
					}
					
					$attributes =& $image->attributes($namespaces['xlink']);
					if($attributes['show'] == 'embed' && file_exists($path)) {
						// Replace image
						list($base, $extension) = explode('.', basename($attributes['href']));

						// simpleXML can't handle it, go to DOM
						$element =& dom_import_simplexml($image);
						$element->setAttribute('xlink:href', str_replace($base, $replace['name'], $attributes['href']));

						$zip = new ZipArchive;
						if ($zip->open($this->_tempdir.'.'.$this->_template_file_ext) === TRUE) {
							$zip->addFile($path, $attributes['href']);
							$zip->close();
						} else {
							//$this->meth_Misc_Alert('Image_Replace method', 'File not found : '.$this->tempdir . '.' . $this->_template_file_ext);
							// Quietly log the error, so the report still works
							$GLOBALS['egw']->log->error(array(
								'text'	=>	'W-Report, Image_Replace \'%2\'  File not found: %3', 
								'p_1' => '',
								'p_2' => $replace['name'],
								'p_3' => $this->tempdir . '.' . $this->_template_file_ext, 
								'file'=>__FILE__,'line'=>__LINE__
							));
						}
					} elseif (!file_exists($path)) {
						//$this->meth_Misc_Alert('Image_Replace method', 'File not found : '.$path);
						// Quietly log the error, so the report still works
						$GLOBALS['egw']->log->error(array(
							'text'	=>	'W-Report, Image_Replace \'%2\'  File not found: %3', 
							'p_1' => '',
							'p_2' => $replace['name'],
							'p_3' => $path,
							'file'=>__FILE__,'line'=>__LINE__
						));
					}
						
				}
				$GLOBALS['egw']->log->commit();
				if(file_exists($this->_tempdir . '/' . $replace['name'])) {
					// Remove temp file
					unlink($this->_tempdir . '/' . $replace['name']);
				}
/*
echo htmlspecialchars($frame->asXML()) . '<br />';
//*/
			}

			// Save any changes
			if(!$xml->asXML($this->_tempdir.'/'.$this->_xml_filename)) {
				$this->meth_Misc_Alert('Image_Replace method', 'Unable to write: '.$this->_tempdir.'/'.$this->_xml_filename);
			}

//echo htmlspecialchars($xml->asXML());
//die();
			// Zip the file again
			$this->_zip();
		}
	}
?>
