<?php
	/**
	*	@brief A Receivable Invoice (sent to a Client)
	*
	*	@author Nathan Gray
	*	@ingroup perp_ar
	*	
	*	Invoice encapsulates an invoice to a Client.
	*	An invoice contains various information such as relevant dates, reference numbers
	*	and financial information, as well as one or more InvoiceLines.
	*
	*	Invoices can be automatically generated based off existing Orders or Shipments, or
	*	they can be entered starting from a blank and adding the lines manually.  Automatically
	*	generated invoices can be overriden.
	*/

	require_once(EGW_INCLUDE_ROOT . '/perp_api/inc/class.Invoice.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_ledger/inc/class.GLTransaction.inc.php');
	require_once('class.ReceivableInvoiceLine.inc.php');

	class ReceivableInvoice extends Invoice {
		/**
		*	Client account ID
		*/
		protected $client_account_id;

		/**
		*	Client branch id
		*/
		protected $client_branch_id;

		/**
		*	Read this invoice from the database
		*	We read the client_account_id and branch_id from the extra
		*	perp_receivable_invoice table.
		*	@param invoice_id ID of the invoice to read.
		*/
		public function read($invoice_id) {
			parent::read($invoice_id);

			$GLOBALS['egw']->db->select('perp_receivable_invoice',
				array(
					'client_account_id',
					'client_branch_id',
				),
				array(
					'invoice_id'	=>	$invoice_id
				),
				__LINE__, __FILE__
			);
			$row = $GLOBALS['egw']->db->row(true);
			$this->client_account_id = $row['client_account_id'];
			$this->client_branch_id = $row['client_branch_id'];
		}

		/**
		*	Write this invoice to the database.
		*	We write the client_account_id and branch_id to the extra
		*	perp_receivable_invoice table.
		*
		*	@return int Invoice ID number, or string error
		*/
		public function write() {
			$this->supplier_id = 0;
			$GLOBALS['egw']->db->transaction_begin();
			$result = parent::write();

			if(!$result) {
				$GLOBALS['egw']->db->transaction_abort();
			}
			$GLOBALS['egw']->db->insert('perp_receivable_invoice',
				array(
					'invoice_id'	=>	$this->invoice_id,
					'client_account_id'	=>	$this->client_account_id,
					'client_branch_id'	=>	$this->client_branch_id
				),
				array(
					'invoice_id'	=>	$this->invoice_id
				),
				__LINE__, __FILE__, 'perp_ar'
			);

			if($GLOBALS['egw']->db->transaction_commit()) {
				return $result;
			} else {
				return lang('Unable to save');
			}
		}

		/**
		*	Post this invoice to the General Ledger
		*/
		public function post() {
			$result = parent::post();

			if($this->invoice_transaction_id || !$result) {
				return false;
			}

			$config = config::read('perp_ar');
			$ar_account = $config['ar_account'];
			$trans_type = $config['invoice_transaction_type'];
			$invoice_narrative = $config['invoice_narrative'];

			$client_data = ExecMethod('perp_ar.bo_perp_client.get_client', $this->client_id);

			$invoice_narrative = sprintf($invoice_narrative,
				$client_data['client_number'],
				$this->get_our_reference_number()
			);
			
			// This needs to be atomic
			$transaction = CreateObject('perp_ledger.GLTransaction');
			$transaction->set_transaction_type($trans_type);
			$transaction->set_date($this->issue_date);

			// Transaction needs one line for each line, one line to balance with associated BankTransactionLine and SupplierTransactionLine
			$ar_total = 0;
			$gl_accounts = array();

			foreach($this->invoice_lines as $line_no => $line) {
				$line_total = $line->post($transaction);
				$ar_total -= $line_total;
			}
			try {
				$ar_line = $transaction->add_line($ar_account, $invoice_narrative, $ar_total);
			} catch (Exception $e) {
				throw new Exception(
					lang(
						'Unable to add Accounts Receivable transaction line.  Please verify your General Ledger accounts, and Accounts Receivable settings (%1)',
						$e->getMessage()
					)
				);
			}


			// Create transaction for supplier
			$client_line = CreateObject('perp_ar.ClientTransactionLine', $ar_line);
			$client_line->set_client($this->client_id);
			$client_line->set_due_date($this->get_due_date());
			$client_line->set_exchange_rate($this->get_exchange_rate());
			$client_line->set_link('perp_ar', 'INVOICE-' . $this->get_invoice_id());
			$client_line->set_client_reference($this->their_reference_number);
			$client_line->set_comment($this->comments);

			$GLOBALS['egw']->db->transaction_begin();

			if($GLOBALS['egw']->db->Link_ID->transCnt == 0) {
				/*
				* Database doesn't support transactions.
				* Reconnect to do transactions, and shove it in GLOBALS[egw]->db so everywhere
				* uses the same connection
				*/
				$old_db =& $GLOBALS['egw']->db;
				$GLOBALS['egw']->db =& ExecMethod('perp_api.bo_perp_api.reconnect_transactional');
				$GLOBALS['egw']->db->transaction_begin();
			}

			$result = $transaction->write();
			$this->invoice_transaction_id = $transaction->get_transaction_number();
			$this->write();

			if(!is_numeric($result) || $this->invoice_transaction_id !== $transaction->get_transaction_number()) {
				$GLOBALS['egw']->db->transaction_abort();
				$GLOBALS['egw']->db->transaction_commit();
				throw new egw_exception(lang('Critical failure!  Invoice %1 does not know it was posted as transaction #%2!',
					$this->get_our_reference_number(),
					$transaction->get_transaction_number()
				));
			}

			$result = $GLOBALS['egw']->db->transaction_commit();

			if($old_db) {
				// Swap the original database back, so it doesn't cause problems elsewhere
				$GLOBALS['egw']->db = $old_db;
			}
			return $result;
		}

		/**
		*	Check that this is a valid invoice
		*
		*	In addition to the checks in the parent, this checks the client account too.
		*	Client branch is optional.
		*	@return true or error message
		*/
		public function check() {
			if(!(int)$this->get_client_account_id()) {
				return lang('Client account is not set');
			}
			return parent::check();
		}

		/**
		*	Add client_account_id and branch id if they can be found
		*/
		public function read_from_object(&$object) {
			if(method_exists($object, 'get_client_account_id')) {
				$this->set_client_account_id((int)$object->get_client_account_id());
			}
			if(method_exists($object, 'get_client_branch_id')) {
				$this->set_client_branch_id((int)$object->get_client_branch_id());
			}
			parent::read_from_object($object);

			if(method_exists($object, 'get_their_reference_number')) {
				$this->set_their_reference_number($object->get_their_reference_number());
			}
		}

		/**
		*	Get tax_group for this invoice.  Pulled from client branch.
		*/
		public function get_tax_group() {
			if($this->client_branch_id) {
				$client_branch = ExecMethod('perp_ar.bo_perp_client_branch.get_client_branch', $this->client_branch_id);
				return $client_branch['tax_group'];
			}
		}

		/**
		*	Get the client account ID number
		*/
		public function get_client_account_id() {
			return $this->client_account_id;
		}

		/**
		*	Set the client account ID number
		*	@param id the new ID number
		*/
		public function set_client_account_id($id) {
			if(!$this->invoice_id) {
				$this->client_account_id = (int)$id;
				if($this->client_id) {
					$client_account = ExecMethod('perp_ar.bo_perp_client_account.get_client_account', $account);
					$this->set_client_id($client_account['client_id']);
				}
			}
		}

		/**
		*	Get the client branch ID number
		*/
		public function get_client_branch_id() {
			return $this->client_branch_id;
		}

		/**
		*	Set the client branch ID number
		*	@param id the new ID number
		*/
		public function set_client_branch_id($id) {
			$this->client_branch_id = (int)$id;
		}

		/**
		*	Get a list of receipts on this invoice
		*
		*	@todo If someone entered a payment in a currency that does
		*		not match the invoice currency, calculate the
		*		amount in the invoice's currency also
		*
		*	@return an array of receipt information
		*/
		public function get_receipts() {
			$GLOBALS['egw']->db->select('perp_client_receipt',
				array(
					'receipt_id',
					'invoice_id',
					'perp_client_receipt.amount',
					'allocation_date',
					'allocated_from',
					'perp_client_receipt.line_no',
					'payment_method_name',
					'currency'
				),      
				array(
					'invoice_id'    =>      $this->invoice_id
				),
				__LINE__, __FILE__, 0, 'ORDER BY receipt_id DESC', 'perp_ar', -1,
				'JOIN perp_bank_ledger ON
					perp_client_receipt.trans_id = perp_bank_ledger.trans_id
				JOIN perp_payment_method ON
					perp_bank_ledger.payment_method = perp_payment_method.payment_method_id'
			);
			if(!is_object($GLOBALS['perp_api']['bo'])) {
				$GLOBALS['perp_api']['bo'] =& CreateObject('perp_api.bo_perp_api');
			}
			$receipt_list = array();
			$different_currencies = false;
			$currency = $this->currency;

			while($row = $GLOBALS['egw']->db->row(true)) {
				$totalpaid += $row['amount'];

				// Check to see if someone mixed currencies
				$different_currencies = $different_currencies || ($row['currency'] != $this->currency);
				$row['allocation_date'] = $GLOBALS['egw']->db->from_timestamp($row['allocation_date']);
				$receipt_list[] = $row;
			}

			// Formatting has to be done after, otherwise it resets the recordset
			foreach($receipt_list as &$row) {
				$row['formatted_amount'] = $GLOBALS['perp_api']['bo']->get_formatted_price($row['currency'], $row['amount']);

				// Calculate amount in invoice's currency
				if($different_currencies) {
					$currency = $row['currency'];
					// TODO
				}
			}

			unset($row);
			$row['receipt_id'] = '';
			$row['invoice_id'] = $this->invoice_id;
			$row['currency'] = $currency;
			$row['amount'] = $totalpaid;
			$row['formatted_amount'] = $GLOBALS['perp_api']['bo']->get_formatted_price($row['currency'], $row['amount']);
			$receipt_list[] = $row;

			return $receipt_list;
		}


		/**
		*	Suggest a reference number for this invoice using configured parameters
		*
		*	@return string Our reference number
		*/
		public function suggest_invoice_number() {
			if($this->get_our_reference_number()) {
				return $this->get_our_reference_number();
			}

			$config = config::read('perp_ar');
			$invoice_number_format = $config['invoice_number'];

			// Get client_id & next_client_num
			if($this->client_id) {
				$GLOBALS['egw']->db->select(Invoice::invoice_table,
					array(
						'invoice_id',
						'client_num',
						'our_reference_number'
					),
						array('perp_invoice.client_id = ' . $this->client_id),
					__LINE__, __FILE__, False,
					'ORDER BY invoice_id DESC', 'perp_ar', 1,
					'JOIN perp_client ON
						perp_invoice.client_id = perp_client.client_id'
				);
				$row = $GLOBALS['egw']->db->row(true);
				if($row) {
					$client_num = $row['client_num'];
				} else {
					$client = ExecMethod('perp_ar.bo_perp_client.get_client', $this->client_id);
					$client_num = $client['client_num'];
				}
			}

                        ereg('date\([dDjlNSwzWFmMntYycU.]+\)', $invoice_number_format, $regs);
                        if(count($regs)) {
                                eval('$date = ' . $regs[0] . ';');
                                $invoice_number_format = str_replace($regs[0], '%3$s', $invoice_number_format);
                        }

			// Extract variables from the last invoice
			$regex = "/(%([1-9])'?.?[0-9]*(\.[0-9]+)?.)/";
			$extract_regex = '/' . preg_replace($regex, '(?P<_\2>.+)', $invoice_number_format) .'/';
			preg_match_all($extract_regex, $row['our_reference_number'], $values, PREG_PATTERN_ORDER);
			$next_client_num = $values['_2'][0] + 1;

			// Make sure it's unique
			$unique = false;

			while(!$unique) {
				$invoice_number = sprintf($invoice_number_format,
					$client_num,
					$next_client_num,
					$date
				);

				$unique = $GLOBALS['egw']->db->query("SELECT count(*) = 0 
					FROM perp_invoice
					WHERE our_reference_number = '$invoice_number'", __LINE__, __FILE__)->fetchSingle();
				$next_client_num++;
			}
			return $invoice_number;
		}

		/**
		*	Prepare this object for serialization
		*	@return array of parameters to preserve
		*/
		public function __sleep() {
			$preserve =  parent::__sleep();
			$preserve[] = 'client_account_id';
			$preserve[] = 'client_branch_id';
			return $preserve;
		}
	}
?>
