<?php
	require_once('class.Staff.inc.php');

	/**
	*	A skill or certificate that a staff person has.  This can include 
	*	certifications such as first aid, skills or abilities...
	*/

	class StaffSkill {

		/**
		*	Staff object owning this skill
		*/
		private $staff;

		/**
		*	ID number, unique amoung the owning Staff object's skills
		*/
		private $skill_id;

		private $name;
		private $start_date;
		private $expiry_date;

		/**
		*	Skill levels are somewhat arbitrary, because it depends on the skill.
		*	Some skill are binary (you have it or you don't), some skills are variable, where the
		*	level depends on your test scores, and some skills are multi-level, where each level
		*	means something different.
		*
		*	Skill level is an int.  
		*	-	-2 - Failed the test
		*	-	-1 - Expired
		*	-	 0 - The skill is lacking, but you still want a record of it
		*	-	>0 - Various meanings, depending on the skill
		*/
		private $skill_level = 1;

		/**
		*	Create a new StaffSkill
		*
		*	Data keys:
		*	- skill_name
		*	- skill_start_date
		*	- skill_expiry_date
		*	- skill_level
		*
		*	@param staff Staff object
		*	@param data Associative array of data
		*/
		public function __construct(Staff &$staff, Array $data) {
			if(!$staff->get_staff_id()) {
				throw new Exception(lang('Invalid staff object passed to StaffSkill'));
			}
			$this->staff =& $staff;
			$this->name = $data['skill_name'];
			$this->start_date = $data['skill_start_date'];
			if($data['skill_expiry_date'] == 0) {
				$this->expiry_date = null;
			} else {
				$this->expiry_date = (int)$data['skill_expiry_date'];
			}
			$this->skill_level = $data['skill_level'];

			if(!is_null($this->expiry_date) && $this->expiry_date < time()) {
				// Skill has expired
				$this->skill_level = -1;
			}
		}

		/**
		*	Set the ID number for this skill.
		*	This function is to be called by Staff::add_skill(),
		*	not anywhere else.
		*
		*	@param id The ID number for this skill
		*/
		public function set_id($id) {
			$this->skill_id = (int)$id;
		}

		public function &get_staff() {
			return $this->staff;
		}

		public function get_name() {
			return $this->name;
		}
		public function set_name($name) {
			$this->name = $name;
		}

		public function get_start_date() {
			return $this->start_date;
		}
		public function set_start_date($date) {
			$this->start_date = (int)$date;
		}

		public function get_expiry_date() {
			return $this->expiry_date;
		}
		public function set_expiry_date($date) {
			if($date == 0 || $date == '') {
				$this->expiry_date = null;
			} else {
				$this->expiry_date = $date;
			}
		}

		public function get_skill_level() {
			return $this->skill_level;
		}
		public function set_skill_level($level) {
			$this->skill_level = (int)$level;
		}

		/**
		*	Write this skill into the database
		*/
		public function write() {
			$data = array(
				'staff_id'		=>	$this->staff->get_staff_id(),
				'skill_id'		=>	$this->skill_id,
				'skill_name'		=>	$this->name,
				'skill_start_date'	=>	$GLOBALS['egw']->db->to_timestamp($this->start_date),
				'skill_expiry_date'	=>	$GLOBALS['egw']->db->to_timestamp($this->expiry_date),
				'skill_level'		=>	$this->skill_level
			);
			$GLOBALS['egw']->db->insert(
				'perp_staff_skill',
				$data, $data,
				__LINE__, __FILE__,
				'perp_human_resources'
			);
		}

		/**
		*	Deletion is interesting, because there is no key defined.
		*	There's no key defined, because there's not a lot of point.
		*	We'll delete based on a complete match
		*/
		public function delete() {
			$GLOBALS['egw']->db->delete(
				'perp_staff_skill',
				array(
					'staff_id'		=>	$this->staff->get_staff_id(),
					'skill_id'		=>	$this->skill_id
				),
				__LINE__, __FILE__
			);
		}

		public function toArray() {
			$data = array(
				'staff_id'		=>	$this->staff->get_staff_id(),
				'skill_id'		=>	$this->skill_id,
				'skill_name'		=>	$this->name,
				'skill_start_date'	=>	$this->start_date,
				'skill_expiry_date'	=>	$this->expiry_date,
				'skill_level'		=>	$this->skill_level
			);
			return $data;
		}

		public function __sleep() {
			return array(
				'staff', 'skill_id',
				'name', 'start_date', 'expiry_date', 'skill_level'
			);
		}
	}
?>
