<?php

	/**
	*	A unified way to enter serial numbers
	*
	*	There are several different options for entering serial numbers:
	*	- Quantity: How many serial numbers are needed?
	*	- Required: Are the serial numbers required, or optional?  This is related
	*		to the stock item's serial_setup, but this says if the serial is required.
	*	- Valid: Does the serial number need to be a valid serial number?  This is almost always
	*		going to be yes.
	*	- Existing: Does the serial number have to already exist in the system?
	*	- Replacement needed: Are we swapping two serialized components, and need a match for every
	*		serial number?
	*
	*	The value required is an array, of the form:
	*	@verbatim
	*	array(
	*		'stock_id'	=>	#,
	*		'quantity'	=>	1 (optional),
	*		'serials'	=>	array(
	*			0	=>	array(
	*				'serial_number'	=>	'SN01234',
	*				'quantity'	=>	1,
	*				'replacement'	=>	'SN56789' (optional)
	*			)
	*		)
	*	)
	* 	@endverbatim
	*
	* @license http://opensource.org/licenses/gpl-license.php GPL - GNU General Public License
	* @package perp_orders
	* @link http://projecterp.org
	* @author Nathan Gray <nathangray@sourceforge.net>
	*/

	class serial_number_widget {

		public $public_functions = array(
			'pre_process'	=>	true,
			'render'	=>	true,
			'post_process'	=>	true,
			'popup'		=>	true
		);

		// This is the name for the editor
		public $human_name = 'Serial Number Entry';

		private $debug = true;

		public static $default_values = array(
			'stock_id'		=>	null,
			'serials'		=>	array(),
			'quantity'		=>	1,
			'required'		=>	false,
			'valid'			=>	true,
			'existing'		=>	false,
			'replacement_needed'	=>	false,
			'location_id'		=>	0,
			
			'base_id'		=>	'',
		);

		public function __construct() {
		}

		/**
		* pre-processing of the extension
		*
		* This function is called before the extension gets rendered
		*
		* @param string $name form-name of the control
		* @param mixed &$value value / existing content, can be modified
		* @param array &$cell array with the widget, can be modified for ui-independent widgets
		* @param array &$readonlys names of widgets as key, to be made readonly
		* @param mixed &$extension_data data the extension can store persisten between pre- and post-process
		* @param object &$tmpl reference to the template we belong too
		* @return boolean true if extra label is allowed, false otherwise
		*/
		public function pre_process($name,&$value,&$cell,&$readonlys,&$extension_data,&$tmpl) {

			if(!is_array($value)) {
				$value = (Array)$value;
			}

			// Make sure all the values and defaults we need are set
			$value =& array_merge(self::$default_values, $value);

			if(!$value['base_id']) {
				$value['base_id'] = $name;
			}


			$cell['type'] = 'html';
		
			if($value['stock_id']) {
				$extension_data['stock_item'] = ExecMethod('perp_inventory.bo_perp_stock.get_stock_item', $value['stock_id']);
			}

			$extension_data['required'] = $value['required'] || $extension_data['stock_item']['serial_setup'] != bo_perp_stock::SERIAL_WHEN_NEEDED;

			$value = self::render($cell, $name, $value, $readonlys, $extension_data, $tmpl);
			unset($cell['onclick']);
			return false;
		}

		function post_process($name,&$value,&$extension_data,&$loop,&$tmpl,$value_in) {
                        $value = unserialize(urldecode($value_in['serials']));
                        return true;
                }

		/**
		*	Render the widget
		*/
		public static function render(&$cell, $form_name, &$value, &$readonly, &$extension_data, &$tmpl) {
			$div_style = $extension_data['required'] ? 'border: 1px dotted red;' : '';

			if(!is_array($extension_data['stock_item']) || $extension_data['stock_item']['controlled'] == 0) {
				// Not controlled
				$div_style .= 'visibility: collapse;';
			}
			$html = "<div style=\"$div_style\">";
			$html .= html::submit_button(
				$form_name . '[button]', 		// element name
				$cell['label'], 			// Label
				// onClick
				$tmpl->js_pseudo_funcs($cell['onclick'], $tmpl->name_vars) . ' get_serial_numbers(this);',
				false, 
				// Extra settings:
				"id=\"{$form_name}[button]\" style=\"$button_style\"", 
				'', 
				'perp_inventory', 'button'
			);
			
			foreach($value as $key => $data) {
				if(is_array($data)) {
					$data = urlencode(serialize($data));
				}

				// Custom generation to set the ID
				$html .= "<input type=\"hidden\" id=\"{$form_name}[$key]\" name=\"{$form_name}[$key]\" value=\"" . html::htmlspecialchars($data) . "\" />\n";
			}

			$html .= '</div>';
			
			$GLOBALS['egw']->js->validate_file('stock', 'serial_numbers', 'perp_inventory');
			

			return $html;
		}

		/**
		*	Open a dialog for entering serial numbers
		*/
		public static function popup($content = array()) {

			// Get values
			foreach(self::$default_values as $key => $default) {
				if(isset($_GET[$key])) {
					if($key == 'serials') {
						$content[$key] = unserialize(urldecode($_GET[$key]));
					} else {
						$content[$key] = $_GET[$key];
					}
				} elseif (!array_key_exists($key, $content)) {
					 $content[$key] = $default;
				}
			}
			if(!is_array($content['serials'])) {
				$content['serials'] = array();
			}

			$msg = array();
			$stock_item = ExecMethod('perp_inventory.bo_perp_stock.get_stock_item', $content['stock_id']);

			// Fill in ranges
			if(isset($content['range'])) {
				foreach($content['range'] as $line) {
                                        if($line['serial_start'] && $line['serial_end']) {
                                                $range = range($line['serial_start'], $line['serial_end']);
                                                $suffix = "%0" . max(strlen((string)$line['serial_start']), strlen((string)$line['serial_end'])) . "d";
                                                foreach($range as $serial) {
                                                        $serial = sprintf($suffix, $serial);
                                                        $content['serials'][] = array(
								'serial_number'	=>	$line['prefix'] . $serial,
								'quantity'	=>	1
							);
                                                }
                                        }
                                }
                                $count = count($serials);
                                unset($content['range']);
			}

			if($location_id && !is_object(self::$bo_inventory)) {
				self::$bo_inventory = new bo_perp_inventory();
			}

			$good_quantity = 0;

			// Validate serial numbers
			foreach($content['serials'] as $key => &$values) {
				if(!is_array($values)) {
					unset($content['serials'][$key]);
					continue;
				}
				$result = self::check_serial($values, $content);
				if($result !== true) {
					unset($content['serials'][$key]);

					if($result) {
						$msg[] = array($result, 'error');
					}
					continue;
				}
				$good_quantity += $values['quantity'];
			}
			if($good_quantity > $content['quantity']) {
				$msg[] = array(lang('Too many serials - entered %1, needed %2 serials', $good_quantity, $content['quantity']), 'error');
			}

			// Write serials back to parent page
			if(count($msg) == 0 && $good_quantity == $content['quantity'] && $content['save']) {
				$serial_id = $content['base_id'] . '[serials]';
				$js = "var serials = window.opener.document.getElementById('$serial_id'); ";
				$js .= "serials.value = '" . urlencode(serialize($content['serials'])) . "'; ";
				if(!$serial_button_id) {
					$serial_button_id = $content['base_id'] . '[button]';
				}
				$js .= "var serial_button = window.opener.document.getElementById('$serial_button_id'); ";
				$js .= "serial_button.value = '&radic;'; ";
				$GLOBALS['egw']->js->set_onload($js);
				$GLOBALS['egw']->js->set_onload('window.close(); ');
			}

			// Using the range is a preference
			$content['no_range'] = !$GLOBALS['egw_info']['user']['preferences']['perp_inventory']['use_range_for_serials'];
			
			// Pad the form with enough boxes
			if($stock_item['serialized']) {
				while(count($content['serials']) < $content['quantity']) {
					$content['serials'][] = array(
						'serial_number'	=>	'',
						'quantity'	=>	1,
						'replacement'	=>	''
					);
				}
			} else {
				$content['serials'][] = array(
					'serial_number'	=>	'',
					'quantity'	=>	min($quantity - $good_quantity, 0),
					'replacement'	=>	''
				);
			}

			// Shift one to allow room for title
			array_unshift($content['serials'], false);

			$preserve = $content;

			// Hide replacement column if not needed
			$content['serials']['no_replacement'] = !$content['replacement_needed'];

			// Hide quantity column, it's always 1 for serialized
			$content['serials']['no_quantity'] = $stock_item['serialized'];

			if($msg) {
				$content['message'] = ExecMethod('perp_api.ui_perp_api.make_message', $msg);
			}

			// Display
                        $GLOBALS['egw_info']['flags']['app_header'] = lang('Enter Serial Numbers');
			$template = new etemplate();
                        $template->read('perp_inventory.serial_popup');
                        $template->exec('perp_inventory.serial_number_widget.popup', $content, $select_box_options, $readonly, $preserve, 2);
		}

		public static function check_serial(&$values, $parameters) {
			// Get rid of blanks
			if(trim($values['serial_number']) == '') {
				return false;
			}

			// Check format
			if($parameters['valid'] || $parameters['existing']) {
				$result = bo_perp_stock::check_serial($parameters['stock_id'], $values['serial_number']);

				// Check the format
				if($parameters['valid'] && $result == bo_perp_stock::NOT_VALID) {
					return lang('%1 is not a valid number for %2', $values['serial_number'], ExecMethod('perp_inventory.bo_perp_stock.link_title', $parameters['stock_id']));
				}

				// If it has to be existing, and check_serial didn't find it, double check (check_serial() doesn't check if the stock item is WHEN_NEEDED)
				if($parameters['existing'] && $result != bo_perp_stock::SERIAL_OK && !so_perp_stock::match_serial($parameters['stock_id'], $values['serial_number'])) {
					return lang(
						'There is no record of a %2 with serial number %1', $values['serial_number'], 
						ExecMethod('perp_inventory.bo_perp_stock.link_title', $stock_id)
					);
				}
				if($parameters['existing'] && $location_id) {
					$query = array(
						'stock_id'	=>	$parameters['stock_id'],
						'location'	=>	$location_id,
						'serial_number'	=>	$values['serial_number']
					);
					$count = self::$bo_inventory->get_controlled_inventory_list($query, $rows);
					if($count == 0) {
						return lang(
							'There is no %2 with serial number %1 at %3', 
							$values['serial_number'], 
							ExecMethod('perp_inventory.bo_perp_stock.link_title', $stock_id),
							ExecMethod('perp_inventory.bo_perp_location.link_title', $location_id)
						);
					}
				}
			}

			if(!$values['quantity']) {
				$values['quantity'] = 1;
			}
			return true;
		}
	}
?>
