<?php
	/**
	*	BankTransaction
	*
	*	This is mostly a convenience class that automatically balances the transaction with 
	*	one line for the bank account.  This means that you shouldn't add any other TransactionLines
	*	with the bank account's General Ledger account.
	*
	*	This class could probably go away to simplify things, and we'd only use GLTransaction.
	*
	*	@author Nathan Gray
	*/
	class BankTransaction extends GLTransaction {
		
		/**
		*	Database table for the bank ledger
		*/
		private $bank_table = 'perp_bank_ledger';
		
		/**
		*	Bank Transaction variables
		*	@{
		*/
		protected $currency;
		protected $exchange_rate = 1;
		protected $narrative;
		protected $payment_method;
		protected $bank_account;
		/* @} */

		/**
		*	General Ledger account for gains or losses when the exchange rate changes
		*/
		protected $currency_loss_account;

		/**
		*	Constructor
		*	@param trans_number int Optional.  If provided, loads the transaction with that ID
		*/
		public function BankTransaction($trans_number = null) {
			$payment_type = $GLOBALS['egw_info']['user']['preferences']['perp_ledger']['default_payment_type'];
			parent::GLTransaction($trans_number);
		}

		/**
		*	Get the exchange rate for this Transaction
		*	@return float Exchange rate
		*/
		public function get_exchange_rate() {
			return $this->exchange_rate;
		}
		/**
		*	Set the exchange rate for this transaction
		*	@param rate	float	Exchange rate
		*/
		public function set_exchange_rate($rate) {
			if(is_numeric($rate)) {
				$this->exchange_rate = $rate;
			}
		}

		/**
		*	Get the narrative for this transaction
		*	@note Transactions don't have narratives, just TransactionLines
		*	@return string Narrative
		*/
		public function get_narrative() {
			return $this->narrative;
		}
		/**
		*	Set the narrative for this transaction
		*	@note Transactions don't have narratives, just TransactionLines
		*	@param narrative string Narrative for this transaction
		*/
		public function set_narrative($narrative) {
			$this->narrative = $narrative;
		}

		/**
		*	Get the payment method for this transaction
		*	@return int payment_method_id
		*/
		public function get_payment_method() {
			return $this->payment_method;
		}
		/**
		*	Set the payment method for this transaction
		*	@param method	int	Payment method ID for this transaction
		*/
		public function set_payment_method($method) {
			if(is_numeric($method)) {
				$this->payment_method = $method;
			}
		}

		/**
		*	Get the bank account for this transaction
		*	@return int Bank account
		*/
		public function get_bank_account() {
			return $this->bank_account;
		}

		/**
		*	set_bank_ccount
		*	Sets the bank account, currency (exchange rate) 
		*	@param account_number - The account number
		*	@return account number or False if not valid bank account
		*/
		public function set_bank_account($account_number) {
			if(settype($account_number, 'int')) {
				$GLOBALS['egw']->db->select('perp_bank_account', array(
					'bank_account_id', 
					'bank_account_name',
					'bank_account_number',
					'currency'),
					array('bank_account_id'	=> $account_number), 
					__LINE__, __FILE__, 'perp_ledger');
				$row = $GLOBALS['egw']->db->row(True);
				if(!$row) {
					return false;
				}
				$this->bank_account = $row['bank_account_id'];
				$this->set_currency($row['currency']);
				return $account_number;
			}
			return false;
		}

		/**
		*	Get the currency for this BankTransaction
		*	@note General Leder transaction is in General Ledger currency only
		*	@return string 3 letter currency id
		*/
		public function get_currency() {
			return $this->currency;
		}

		/**
		*	setCurrency
		*	@param currency string	Must be a valid three letter currency ID
		*	@return new exchange rate or False on error
		*/
		public function set_currency($currency) {
			$exchange_rate = array();
			$boerp = CreateObject('perp_api.bo_perp_api');
			$query = array(
				'currency_id'	=>	$currency,
				'template'	=>	false
			);
			$boerp->get_currencies( $query, $exchange_rate);
			if(count($exchange_rate) == 1) {
				$this->currency = $currency;
				$this->exchange_rate = $exchange_rate[0]['exchange_rate'];
			} else {
				return False;
			}
		}

		/**
		*	Read a previously posted transaction
		*	@param trans_no	ID number of the transaction to read
		*/
		public function read($trans_no) {
			if(!$GLOBALS['egw']->acl->check('bank_trans', EGW_ACL_READ, 'perp_ledger')) {
				return lang('Access denied');
			}
			parent::read($trans_no);
			$GLOBALS['egw']->db->select($this->bank_table, array(
					'bank_account',
					'currency', 
					'exchange_rate',
					'narrative',
					'payment_method',
					'amount'
				),
				array('trans_id' => $trans_no),
				__LINE__, __FILE__, 'perp_ledger');
			$row = $GLOBALS['egw']->db->row(true);
			$this->currency = $row['currency'];
			$this->exchange_rate = $row['exchange_rate'];
			$this->narrative = $row['narrative'];
			$this->payment_method = $row['payment_method'];
			$this->amount = $row['amount'];
		}

		/**
		*	Write (or 'post') this transaction into the General Ledger and Bank Ledger
		*/
		public function write() {
			if(!$GLOBALS['egw']->acl->check('bank_trans', EGW_ACL_ADD, 'perp_ledger')) {
				return lang('Access denied');
			}
			
			$this->amount = -1 * parent::check();
			$difference = parent::check() - abs($this->amount);
			$gl_line = parent::add_line($this->bank_account, $this->narrative, $this->amount);
			$bank_line =& $gl_line->get_associated_line(0);
			$bank_line->set_payment_method($this->payment_method);

			parent::write();
		}

		/**
		*	Return a string representation of this transaction.
		*	Very useful for debugging.
		*/
		public function __toString() {
			$string = sprintf("Bank Transaction %s to %s (%s @ %2.3f)\n",
				$this->payment_method,
				$this->bank_account,
				$this->currency,
				$this->exchange_rate);
			return '<pre>' . $string . "<br />\n" .parent::__toString() . '</pre>';
		}

		/**
		*	Prepare this object for serialization.
		*	@return array of properties to preserve across serialization
		*/
		public function __sleep() {
			return parent::__sleep() + array(
				'payment_method',
				'bank_account',
				'currency',
				'exchange_rate'
			);
		}
	}
?>
