<?php
	/**
	*	BankTransactionLine is a specialized TransactionLine for transactions on the Bank Ledger
	*	It should be added as an associated transaction to a TransactionLine.
	*	@author Nathan Gray
	*	@ingroup perp_ledger
	*	
	*/
	class BankTransactionLine extends TransactionLine {
		/**
		*	Database table for the bank ledger
		*/
		private $table = 'perp_bank_ledger';

		/**
		*	Exchange rate when this transaction was made.  It will be 1
		*	if the bank account currency is the same as the GL currency.
		*/
		private $exchange_rate = 1;

		/**
		*	Method of payment that caused this transaction (Cheque, EFT, etc).
		*/
		private $payment_method;

		/**
		*	Currency of this transaction
		*/
		private $currency;
		
		/**
		*	Constructor\n
		*	As a parameter, the BankTransactionLine takes either a TransactionLine that it is associated with 
		*	(both for the same amount to the same account) or a GLTransaction it is part of.  
		*	The TransactionLine to be associated with is the preferred method.
		*	@param parent TransactionLine or GLTransaction that this BankTransactionLine is part of
		*/
		public function BankTransactionLine($parent, $line_no = false) {
			$config = CreateObject('phpgwapi.config', 'perp_ledger');
			$config->read_repository();
			$this->set_currency($config->config_data['gl_currency']);
			if(is_a($parent, 'TransactionLine')) {
				// An associated transaction
				$this->transaction = $parent->transaction;
				if(!$this->transaction->get_transaction_number()) {
					$this->set_account($parent->get_account());
					$this->set_amount($parent->get_amount());
					$this->narrative = $parent->get_narrative();
				}
			} elseif( is_a($parent, 'GLTransaction')) {
				// New transaction
				$this->transaction = $parent;
			}
			if($this->transaction->get_transaction_number()) {
				$this->read($this->transaction->get_transaction_number(), $line_no);
			}
		}

		/**
		*	Set the GL account that this transaction is to
		*
		*	@param account int General Ledger account number
		*/
		public function set_account($account) {
			if($this->transaction->get_transaction_number()) {
				// Read only
				return False;
			}

			$row = $GLOBALS['egw']->db->select('perp_bank_account',
				array('currency', 'exchange_rate'),
				array('bank_account_id' => $account),
				__LINE__, __FILE__, False, '', 'perp_ledger', False,
				'JOIN perp_currency ON perp_bank_account.currency = perp_currency.currency_id'
			)->fetch();
			$this->set_currency($row['currency']);
			$this->account = $account;
		}

		/**
		*	Get the amount of this transaction line in the bank account's currency.
		*	This may be different from the TransactionLine's amount, depending on
		*	exchange rate.
		*
		*	@return float Amount in the Bank Account's currency
		*/
		public function get_corrected_amount() {
			return $this->amount / $this->exchange_rate;
		}

		/**
		*	Get the exchange rate for this transaction line.  May be different from
		*	the current exchange rate.
		*	@return float Exchange rate for this transaction.
		*/
		public function get_exchange_rate() {
			return $this->exchange_rate;
		}
		/**
		*	Set the exchange rate for this transaction line.  Should be the current
		*	exchange rate.
		*	@param rate float Exchange rate for this transaction
		*/
		public function set_exchange_rate($rate) {
			if(!$this->transaction->get_transaction_number()) {
				$this->exchange_rate = $rate;
			}
		}

		/**
		*	Get the currency for this transaction line.  May be different from the GL
		*	currency, depending on the currency of the bank account
		*	@return string 3 letter currency ID
		*/
		public function get_currency() {
			return $this->currency;
		}
		/**
		*	set_urrency
		*	@param currency string Must be a valid 3 letter currency ID
		*	@return new exchange rate or False on error
		*/
		public function set_currency($currency) {
			if(isset($this->transaction) && $this->transaction->get_transaction_number()) {
				return False;
			}
			$exchange_rate = array();
			$boerp = CreateObject('perp_api.bo_perp_api');
			$query = array('currency_id' => $currency, 'template' => false);
			$boerp->get_currencies( $query, $exchange_rate);
			if(count($exchange_rate) == 1) {
				$this->currency = $currency;
				$this->exchange_rate = $exchange_rate[0]['exchange_rate'];
			} else {
				return False;
			}
		}

		/**
		*	Get the method of payment for this transaction (Cheque, EFT, etc.)
		*	@return int payment_method_id
		*/
		public function get_payment_method() {
			return $this->payment_method;
		}
		/**
		*	Set the method of payment for this transaction
		*	@param method int payment_method_id
		*/
		public function set_payment_method($method) {
			if(is_numeric($method) && !$this->transaction->get_transaction_number()) {
				$this->payment_method = $method;
			}
		}

		/**
		*	Read this BankTransactionLine from the Bank Ledger (if it's been posted already)
		*	This function is called as part of reading a posted GLTransaction.
		*/
		public function read($trans_id, $line_no) {
			$row = $GLOBALS['egw']->db->select(
				$this->table, 
				array(
					'trans_id', 
					'line_no',
					'bank_account',
					'exchange_rate',
					'payment_method',
					'amount',
					'currency'
				),
				array(
					'trans_id'	=>	$this->transaction->get_transaction_number(),
					'line_no'	=>	$line_no
				),
				__LINE__, __FILE__, 'perp_ledger')->fetch();

			$this->account = $row['bank_account'];
			$this->amount = $row['amount'];
			$this->exchange_rate = $row['exchange_rate'];
			$this->payment_method = $row['payment_method'];
			$this->currency = $row['currency'];
		}

		/**
		*	Write (or 'post') this TransactionLine to the Bank Ledger
		*	This function is called as part of posting a TransactionLine's associated transaction lines
		*/
		public function write($line_no) {
			$GLOBALS['egw']->db->insert($this->table, array(
				'trans_id'	=>	$this->transaction->get_transaction_number(),
				'line_no'	=>	$line_no,
				'bank_account'	=>	$this->account,
				'exchange_rate'	=>	$this->get_exchange_rate(),
				'payment_method'=>	$this->get_payment_method(),
				'amount'	=>	$this->get_corrected_amount(),
				'currency'	=>	$this->get_currency()),
				array('trans_id'=>	$this->transaction->trans_no),
				__LINE__, __FILE__, 'perp_ledger');
		}

		/**
		*	Prepare this object for serialization
		*	@return array of properties to preserve across serialization
		*/
		public function __sleep() {
			$save =  array('account', 'exchange_rate', 
				'payment_method', 'currency', 'transaction', 'narrative', 'amount');
			return $save;
		}

		/**
		*	Return a string representation of this BankTransactionLine.
		*	Very useful for debugging.
		*	@return string This object in string form.
		*/
		public function __toString() {
			$me = sprintf("BankTransactionLine Account: %s Amount: %s ", $this->account, $this->amount);
			return $me;
		}
	}
?>
