<?php
	require_once('class.BankTransactionLine.inc.php');

	/**
	*	@brief A transaction on the General Ledger
	*
	*	@author Nathan Gray
	*
	*	GLTransaction encapsulates a financial transaction.  
	*	A Transaction is made of several TransactionLines, one for each General Ledger account involved in
	*	the transaction.  If other ledgers are involved in the transaction, special lines are linked to 
	*	the TransactionLines.
	*
	*	@note All GLTransactions are in the currency of the General Ledger.
	*	@note Transactions must have a minumum of two TransactionLines.
	*	@note Transactions must be balanced, that is, the sum of the transaction lines must equal zero.
	*	@note Transactions cannot be altered or removed once posted.
	*/
	class GLTransaction {
		
		/**
		*	Database table for the general ledger
		*/
		private $table = 'perp_general_ledger';

		/**
		*	ID number for this transaction.
		*	Will be 0 until the transaction is posted
		*/
		protected $transaction_number;

		/**
		*	Type of transaction.  
		*	@see bo_perp_ledger::get_transaction_types()
		*/
		protected $transaction_type;

		/**
		*	Date of the transaction.  Not always the current date.
		*/
		protected $date;

		/**
		*	Array of TransactionLines that make up the transaction
		*/
		protected $lines;

		/**
		*	Constructor\n
		*	Create a transaction
		*	
		*	@param trans_number	int	If provided, the transaction will be loaded.  Otherwise,
		*				a new, empty transaction will be created
		*/
		public function GLTransaction($trans_number = null) {
			$this->lines = array();
			if($trans_number) {
				$this->read($trans_number);
			} else {
				$this->transaction_number = null;
				$this->transaction_type = 0; // General Ledger Journal entry
				$this->date = date('U');
			}
		}

		/**
		*	Read the given transaction from the database
		*
		*	@param trans_no	int	Transaction_number of an existing transaction
		*	@return string Error message on error, otherwise null
		*/
		public function read($trans_no) {
			if(!$GLOBALS['egw']->acl->check('ledger', EGW_ACL_READ, 'perp_ledger')) {
				return lang('Access denied');
			}

			$GLOBALS['egw']->db->select($this->table, array(
					'trans_no',
					'trans_type',
					'trans_date',
					'line_no',
					'amount', 
					'asset_or_liability',
					'account',
					'narrative'),
				array( 'trans_no' => $trans_no), 
				__LINE__, __FILE__, False, 
				'ORDER BY line_no', 'perp_ledger', False,
				'JOIN perp_account ON ' . $this->table . '.account = perp_account.account_id
				JOIN perp_account_group ON perp_account.account_group = perp_account_group.group_id');
			$lines = array();

			while($row = $GLOBALS['egw']->db->row(true)) {
				$lines[] = $row;
			}
			foreach($lines as $line) {
				$this->transaction_number = $line['trans_no'];
				$this->transaction_type = $line['trans_type'];
				$this->date = $GLOBALS['egw']->db->from_timestamp($line['trans_date']);
				$this->add_line($line['account'], $line['narrative'], $line['amount'], $line['line_no']);
			}
		}

		/**
		*	Writes the given transaction to the database, if it passes the check().
		*
		*	Checks to see if the GLTransaction should create its own
		*	database transaction for the write.  If there's something else being done, such as
		*	recording that an invoice was posted, the entire thing needs to be atomic.  
		*
		* 	@return the transaction number
		*	@throws UnbalancedTransaction {
		*		Transactions must balance, that is, the sum of the TransactionLines must equal 0.
		*		You should call check() before trying to write (also known as 'post') a transaction to 
		*		make sure the transaction is valid.
		*	}
		*/
		public function write() {
			if(!$GLOBALS['egw']->acl->check('ledger', EGW_ACL_ADD, 'perp_ledger')) {
				throw new egw_exception_no_permission(lang('Unable to post transaction, you need Add permission on "ledger"'));
			}
			$result = $this->check();
			if($result !== True) {
				throw new Exception('Transaction is unbalanced.  (' . $result . ')');
			}

			$GLOBALS['egw']->db->transaction_begin();

			// Get transaction number
			$trans_no = $GLOBALS['egw']->db->select($this->table, 'IF(max(trans_no) IS NULL, 0, max(trans_no)) + 1 AS trans_no', array(), __LINE__, __FILE__, 'perp_ledger')->fetchSingle();
			if($trans_no) {
				$this->transaction_number = $trans_no;
			} else {
				$GLOBALS['egw']->db->transaction_abort();
				throw new egw_exception('Could not find a transaction number');
			}


			// Write lines
			foreach($this->lines as $key => $line) {
				if(!$line->get_account()) {
					$GLOBALS['egw']->db->transaction_abort();

					// The transaction is only rolled back on commit
					$GLOBALS['egw']->db->transaction_commit();
					throw new Exception(lang("Unable to post transaction.  Line %1 (%2) was missing an account", $key, $line->get_amount()));
				}
				$line->write($key);
			}
	
			$GLOBALS['egw']->db->transaction_commit();

			return $this->transaction_number;
		}

		/** 
		*	Check that this transaction balances & that it's not trying to post
		*	before the start of the year
		*
		*	@return True if it balances, or outstanding amount if it doesn't
		*	@throws Exception {
		*		Transactions cannot be posted with a date earlier than the start of the current accounting year.
		*	}
		*/
		public function check() {
			// Check for backdating before year end
			$config = config::read('perp_ledger');
			if($config['last_year_end']) {
				if($config['last_year_end'] >= $this->date) {
					throw new Exception(lang('Transaction back-dated before start of year (%1)', $config['last_year_end']));
				}
			}

			$total = 0;
			foreach($this->lines as $key => $line) {
				$total = bcadd($total, $line->get_amount(), 3);
			}
			if($total == 0) {
				return True;
			} else {
				return $total;
			}
		}

		// Accessors

		/**
		*	Get the transaction id number.  It will be 0 until the transaction is posted.
		*
		*	@return int Transaction number
		*/
		public function get_transaction_number() {
			return $this->transaction_number;
		}

		/**
		*	Get the transaction type
		*
		*	@return int ID of the transaction type
		*/
		public function get_transaction_type() {
			return $this->transaction_type;
		}

		/**
		*	Set the transaction type for this transaction
		*
		*	@param trans_type	int	Transaction type
		*/
		public function set_transaction_type($trans_type) {
			$this->transaction_type = $trans_type;
		}
		
		/**
		*	Get the date of this transaction
		*
		*	@return Transaction date
		*/
		public function get_date() {
			return $this->date;
		}

		/**
		*	Set the date of this transaction.
		*	
		*	@param date	Date of this transaction
		*/
		public function set_date($date) {
			if(!$this->transaction_number && (int)$date) {
				$this->date = $date;
			}
		}
	
		/**
		*	Create a new TransactionLine and add it to this transaction
		*
		*	@param	account	int	General Ledger account number
		*	@param	narrative	string	Description of this line
		*	@param	amount	float	Amount involved in this line
		*	@param	line_no	int	If provided, the new line will be assigned this line number
		*	@return TransactionLine	A reference to the newly created TransactionLine
		*/
		public function add_line($account, $narrative, $amount, $line_no = null) {
			$line = new TransactionLine($this);
			if($line->set_account($account) == false) {
				throw new Exception(lang('Invalid account (%1) for line: %2', $account, $narrative));
			}
			$line->set_amount($amount);
			//$line->set_amount($line->get_corrected_amount());
			$line->set_narrative($narrative);

			if(!is_null($line_no)) {
				$this->lines[$line_no] = $line;
			} else {
				$this->lines[] = $line;
			}
			return $line;
		}

		/**
		*	Remove the given line from this transaction
		*
		*	@param line_number	int	Line number of the line to remove
		*/
		public function remove_line($line_number) {
			unset($this->lines[$line_number]);
		}

		/**
		*	This returns the lines on this transaction as an array
		*	@return the lines as an array
		*/
		public function get_lines() {
			$lines = array();
			foreach($this->lines as $key => $line) {
				$lines[$key] = array(
					'line_no'	=>	$key,
					'account'	=>	$line->get_account(),
					'amount'	=>	$line->get_amount(),
					'credit'	=>	$line->get_credit(),
					'debit'		=>	$line->get_debit(),
					'narrative'	=>	$line->get_narrative(),
				);
			}
			return $lines;
		}
		

		/**
		*	Returns the lines as an array
		*	The amount is corrected based on asset/expense/revenue/liability
		*	account type, and may be different from the amount as entered
		*	
		*	Depricated?  Used anywhere?
		*/
		public function get_corrected_lines() {
			$lines = array();
			foreach($this->lines as $key => $line) {
				$lines[$key] = array(
					'line_no'	=>	$key,
					'account'	=>	$line->get_account(),
					'amount'	=>	$line->get_corrected_amount(),
					'narrative'	=>	$line->get_narrative(),
				);
			}
			return $lines;
		}
	
		/**
		*	Returns a string representation of this transaction.
		*	Very useful for debugging purposes.  If echoing the transaction, it must be called on its own line:
		*	@code echo $transaction; @endcode
		*	Otherwise, this function can be called directly:
		*	@code echo '<pre>' . $transaction->__toString() . '</pre>'; @endcode
		*
		*	@return string String representation of this transaction
		*/
		public function __toString() {
			$string = sprintf("Transaction # %d\t Type: %d on %s\n\tLine\tAccount\tDebit\tCredit\tAmount\n", 
				$this->transaction_number,
				$this->transaction_type, 
				$this->date);
			foreach($this->lines as $no => $line) {
				$string .= "\t$no.\t" . $line->__toString();
			}
			return $string;
		}

		/**
		*	Preserve this transaction in the $_SESSION.
		*/
		public function __sleep() {
			$save = array('transaction_number', 'transaction_type', 'date', 'lines');
			return $save;
		}

		/**
		*	Restore connection to database
		*/
		public function __wakeup() {
		}

		/**
		*	Good rounding using strings to get 5 up rounding
		*	Thanks to udochen at gmail dot com who submitted this to
		*	PHP manual
		*	@param x Value to round
		*	@param p Round precision
		*/
		public static function roundbc($x, $p) {
			$x = trim($x);
			$data = explode(".",$x);

			if(substr($data[1],$p,1) >= "5") {
				//generate the add string.
				$i=0;
				$addString = "5";
				while($i < $p) {
					$addString = "0" . $addString;
					$i++;
				}//end while.
				$addString = "." . $addString;
				//now add the addString to the original fraction.
				if($x >= 0) {
					$sum = bcadd($data[0] . "." . $data   [1],$addString,$p+1);
				} else {
					$sum = bcsub($data[0] . "." . $data   [1],$addString,$p+1);
				}

				//explode the result.
				$sumData = explode(".",$sum);

				//now, return the correct precision on the rounded number.
				return (float)$sumData[0] . "." . substr($sumData[1],0,$p);

			} else {

				//don't round the value and return the orignal to the desired
				//precision or less.
				return (float)$data[0] . "." . substr($data[1],0,$p);

			}//end if/else.

		}//end roundbc.
	}

	/**
	*	@brief A single entry on a ledger
	*
	*	TransactionLine encapsulates a single entry onto a ledger, specifically the General Ledger.
	*	Specialized TransactionLines can be linked (via associated_lines) for other Ledgers.
	*/
	class TransactionLine {

		/**
		*	Reference to GLTransaction this line is a part of
		*/
		protected $transaction;

		/**
		*	General Ledger account for this line
		*/
		protected $account;

		/**
		*	Description of this line
		*/
		protected $narrative;

		/**
		*	Amount of this line
		*/
		protected $amount;

		/**
		*	If this line credits the account, the amount.  Otherwise, 0
		*/
		protected $credit;

		/**
		*	If this line debits the account, the amount.  Otherwise, 0
		*/
		protected $debit;

		/**
		*	Whether a positive amount is a debit (1) or a credit (-1)
		*	Deprecated?
		*/
		private $backwards_flag = 1; // 1 or -1

		/**
		*	Associated TransactionLines for other ledgers, such as bank transactions.
		*	They should extend TransactionLine and re-implement the read() and write() methods.
		*/
		private $associated_lines = array();

		/**
		*	Constructor\n
		*	@param transaction GLTransaction A reference to the GLTransaction this line is part of
		*/
		public function TransactionLine(GLTransaction $transaction) {
			$this->transaction = $transaction;
			$backwards_flag = 1;
		}

		// Accessors

		/**
		*	Get the General Ledger account for this TransactionLine
		*	@return	int	General ledger account
		*/
		public function get_account() {
			return $this->account;
		}

		/**
		*	Set the General Ledger account for this TransactionLine
		*	@param	account	int	General Ledger account for this TransactionLine
		*	@return boolean True on successfully changing the account
		*/
		public function set_account($account) {
			// 0 isn't a valid account and you can't change the account once set
			if($account == 0 || $this->get_account()) {
				return false;
			}

			/** 
			* And now for something completely different:
			* Debits increase assets & expenses, decrease revenue and liablilities
			* Credits reduce assets & expenses, increase revenue, liabilities & net worth
			* So, we multiply by a factor of +/- 1 (set by account group) to make things go up
			* or down accordingly.
			*/
			$GLOBALS['egw']->db->select('perp_account', 
				array('asset_or_liability'),
				array('account_id' => $account),
				__LINE__, __FILE__, False,'','perp_ledger',False,
				'JOIN perp_account_group ON perp_account_group.group_id = perp_account.account_group');
			$row = $GLOBALS['egw']->db->row(True);
			if(!$row) {
				return false;
			}

			// There's actually different options (-2, -1, 1, 2), we only want +/- 1
			$this->backwards_flag = $row['asset_or_liability'] / abs($row['asset_or_liability']);
			$this->account = $account;
			
			// Check to see if this is a bank account
			$row = $GLOBALS['egw']->db->select('perp_bank_account',
				array('exchange_rate'),
				array('bank_account_id' => $account),
				__LINE__, __FILE__, False, '', 'perp_ledger', False,
				'JOIN perp_currency ON perp_bank_account.currency = perp_currency.currency_id')->fetch();
			if($row) {
				$bank_trans = CreateObject('perp_ledger.BankTransactionLine', $this);
				$this->add_associated_line($bank_trans);
			}

			return True;
		}
		
		/**
		*	@return The reference description of this line
		*/
		public function get_narrative() {
			return $this->narrative;
		}

		/**
		*	Set the description of this line
		*
		*	@param narrative string Description of this line
		*/
		public function set_narrative($narrative) {
			$this->narrative = $narrative;
			foreach($this->associated_lines as $line) {
				$line->set_narrative($narrative);
			}
		}

		/**
		*	Get the amount multiplied by backwards flag
		*	@deprecated ?
		*
		*	@return float mangled line amount
		*/
		public function get_corrected_amount() {
			return $this->amount * $this->backwards_flag;
		}

		/**
		*	Return the amount as entered on the General Ledger
		*
		*	@return float line amount
		*/
		public function get_amount() {
			return $this->amount;
		}

		/**
		*	Set the amount for this line
		*
		*	@param amount float Amount for this TransactionLine
		*/
		public function set_amount($amount) {
			$amount = GLTransaction::roundbc($amount, 2);
			$this->amount = $amount;
			foreach($this->associated_lines as $line) {
				$line->set_amount($this->amount);
			}
			if($this->amount >= 0) {
				$this->debit = abs($this->amount);
			} else {
				$this->credit = abs($this->amount);
			}
		}

		/**
		*	Write this line as part of writing the GLTransaction
		*	This function is called by GLTransaction::write(), not anywhere else.
		*	It calls write() on any associated TransactionLine to write them.
		*
		*	@param line_number	The line number for this line
		*/
		public function write($line_number) {
			$GLOBALS['egw']->db->insert('perp_general_ledger', array(
				'trans_no'	=>	$this->transaction->get_transaction_number(),
				'line_no'	=>	$line_number,
				'trans_type'	=>	$this->transaction->get_transaction_type(),
				'trans_date'	=>	$GLOBALS['egw']->db->to_timestamp($this->transaction->get_date()),
				'account'	=>	$this->account,
				'amount'	=>	$this->amount, 
				'narrative'	=>	$this->narrative,
				), false, __LINE__, __FILE__, 'perp_ledger');
			foreach($this->associated_lines as $line) {
				$line->write($line_number);
			}
		}

		/**
		*	Add an associated TransactionLine for adding a line to a sub-ledger like BankLedger
		*
		*	@param line	Associated TransactionLine to another Ledger
		*/
		public function add_associated_line($line) {
			$this->associated_lines[] = $line;
		}

		/**
		*	Return a reference to the associated TransactionLine identified by $line.
		*
		*	@param line int Index of the TransactionLine
		*/
		public function get_associated_line($line) {
			return $this->associated_lines[$line];
		}

		/**
		*	Returns an array of references to all associated lines.
		*
		*	@return array All associated TransactionLines.
		*/
		public function get_associated_lines() {
			return $this->associated_lines;
		}

		/**
		*	Return the credit amount of this line.
		*	This will be 0 if the line is actually a debit to the account.
		*
		*	@return float Amount credited to the account.
		*/
		public function get_credit() {
			return $this->credit;
		}

		/**
		*	Return the debit amount of this line.
		*	This will be 0 if the line is actually a credit to the account.
		*
		*	@return float Amount debited to the account.
		*/
		public function get_debit() {
			return $this->debit;
		}

		/**
		*	Return a string representation of this line.  Includes associated lines.  
		*	Very useful for debugging purposes, not much else.
		*
		*	@return string Representation of this TransactionLine
		*/
		public function __toString() {
			$string = sprintf("%d\t%.2f\t%.2f\t%.2f\t%s\n", 
				$this->account, $this->get_debit(), $this->get_credit(),
				$this->amount, $this->narrative
			);
			foreach($this->associated_lines as $num => $line) {
				$string .= "\t" . $line->__toString() . "\n";
			}
			return $string;
		}

		/**
		*	Prepare this object for serialization
		*
		*	@return array Properties to preserve across serialization.
		*/
		public function __sleep() {
			return array('transaction', 'account', 'narrative', 'amount', 
				'credit', 'debit',
				'backwards_flag', 'associated_lines');
		}

		/** 
		*	Restore this object from serialization.
		*	@note This is a stub.
		*/
		public function __wakeup() {
		}

	}
?>
