<?php

	/**
	*	Income Statement Report
	*
	*	This report lists General Ledger Account Group balances.
	*	It's divided into several parts, according to the standard Income Statement report:
	*
	*	@verbatim
	*	1.  Sales			-
	*	2.  Cost of Sales		=
	*	3. Gross Margin			-
	*	4.  Expenses			=
	*	5. Gross Earnings		+
	*	6.  Other Income & Expenses	=
	*	7.  Earnings or Loss Before Taxes -
	*	8.  Provision for Taxes		=
	*	9. Net Earnings
	*	@endverbatim
	*	
	*	The user can specify which General Ledger account groups to use for 1,2,4, and 8.
	*	3, 5, 6, 7 and 9 are calculated.
	*	Each section may or may not be broken down by into it's accounts
	*
	*	@author Nathan Gray
	*	@ingroup perp_ledger
	*	@ingroup perp_report
	*/
	include_once(EGW_INCLUDE_ROOT . '/perp_api/inc/class.perp_report.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/phpgwapi/inc/class.config.inc.php');

	class income_statement extends perp_report {
		/**
		*	Constructor
		*	@param parameters array Optional.  Report parameters (end_date, income sequences, expense sequences)
		*/
		function income_statement(array $parameters = array()) {
			parent::perp_report($parameters);
			$this->template = 'income_statement.ods';

			// Get config
			$config = config::read('perp_ledger');
			if(count($this->parameters) == 0 && $config['income_statement']) {
				$this->parameters += $config['income_statement'];
			} elseif ($config['income_statement']) {
				$this->parameters = array_merge($config['income_statement'], $this->parameters);
			}
			if(!$this->parameters['start_date']) {
				$this->parameters['start_date']	=	strtotime("+1 day", strtotime("last year", $config['next_year_end']));
			}
			if(!$this->parameters['end_date']) {
				$this->parameters['end_date']	=	date('U');
			}

			$this->parameters += $this->get_options();

			if(!$config['income_statement']) {
				$this->parameters['show_settings'] = true;
			}
		}

		/**
		*	Get data for the report
		*
		*	@param	query	array	Report parameters: start_date, end_date, income, expense
		*/
		public function get_data($query = array()) {
			// Save the settings for next time (not the dates though)
			$save = array(
				'include_accounts'	=>	$query['include_accounts'],
				'income'		=>	$query['income'],
				'expense'		=>	$query['expense'],
				'sales'			=>	$query['sales'],
				'cost_of_sales'		=>	$query['cost_of_sales'],
				'expenses'		=>	$query['expenses'],
				'provision_for_taxes'	=>	$query['provision_for_taxes'],
				'percent_column'	=>	$query['percent_column']
			);
			$config = new config('perp_ledger');
			$config->read_repository();
			$config->save_value('income_statement', $save, 'perp_ledger');

			$data = array();

			// Explode into arrays
			$query['sales']			= $query['sales'] ? explode(',', $query['sales']) : array();
			$query['cost_of_sales']		= $query['cost_of_sales'] ? explode(',', $query['cost_of_sales']) : array();
			$query['expenses']		= $query['expenses'] ? explode(',', $query['expenses']) : array();
			$query['provision_for_taxes']	= $query['provision_for_taxes'] ? explode(',', $query['provision_for_taxes']) : array();
			$query['percent_column']	= $query['percent_column'] ? explode(',', $query['percent_column']) : array();
			
			// Determine calulated parameters
			$this->get_calculated_parameters($query, $data);

			// Build some headings
			$this->get_labels($query, $data);

			$this->get_section($query, $data, 'sales');
			$this->get_section($query, $data, 'cost_of_sales');

			// Gross margin
			$data['gross_margin'][] = array(
				'current'		=>	$data['sales_summary'][0]['current'] - $data['cost_of_sales_summary'][0]['current'],
				'previous'		=>	$data['sales_summary'][0]['previous'] - $data['cost_of_sales_summary'][0]['previous'],
				'current_percent'	=>	$data['sales_summary'][0]['current_percent'] - $data['cost_of_sales_summary'][0]['current_percent'],
				'previous_percent'	=>	$data['sales_summary'][0]['previous_percent'] - $data['cost_of_sales_summary'][0]['previous_percent']
			);

			$this->get_section($query, $data, 'expenses');

			// Gross Earnings
			$data['gross_earnings'][] = array(
				'current'		=>	$data['gross_margin'][0]['current'] - $data['expenses_summary'][0]['current'],
				'previous'		=>	$data['gross_margin'][0]['previous'] - $data['expenses_summary'][0]['previous'],
				'current_percent'	=>	$data['gross_margin'][0]['current_percent'] - $data['expenses_summary'][0]['current_percent'],
				'previous_percent'	=>	$data['gross_margin'][0]['previous_percent'] - $data['expenses_summary'][0]['previous_percent']
			);
			
			$this->get_section($query, $data, 'other');

			// Before income taxes
			$data['before_taxes'][] = array(
				'current'		=>	$data['gross_earnings'][0]['current'] + $data['other_summary'][0]['current'],
				'previous'		=>	$data['gross_earnings'][0]['previous'] + $data['other_summary'][0]['previous'],
				'current_percent'	=>	$data['gross_earnings'][0]['current_percent'] + $data['other_summary'][0]['current_percent'],
				'previous_percent'	=>	$data['gross_earnings'][0]['previous_percent'] + $data['other_summary'][0]['previous_percent']
			);

			$this->get_section($query, $data, 'provision_for_taxes');

			// Net Earnings
			$data['net_earnings'][] = array(
				'current'		=>	$data['before_taxes'][0]['current'] - $data['provision_for_taxes_summary'][0]['current'],
				'previous'		=>	$data['before_taxes'][0]['previous'] - $data['provision_for_taxes_summary'][0]['previous'],
				'current_percent'	=>	$data['before_taxes'][0]['current_percent'] - $data['provision_for_taxes_summary'][0]['current_percent'],
				'previous_percent'	=>	$data['before_taxes'][0]['previous_percent'] - $data['provision_for_taxes_summary'][0]['previous_percent']
			);

//			_debug_array($data);die();

			// Logo
			$config = config::read('perp_api');
			$data['images'] = array();
			$data['images'][] = array(
				'name'	=>	'company_logo',
			) + $config['company_logo'];

			return $data;
		}

		/**
		*	Build some translated headings, so this report is language independent
		*
		*	@param query Array of report parameters
		*	@param data Array of report data
		*/
		protected function get_labels(&$query, &$data) {
			$data['lang'][] = array(
				'percent_title'		=>	lang('% of Sales'),
				'income_statement'	=>	lang('Income Statement'),
				'sales'			=>	lang('Sales'),
				'cost_of_sales'		=>	lang('Cost of Sales'),
				'gross_margin'		=>	lang('Gross Margin'),
				'expenses'		=>	lang('Expenses'),
				'gross_earnings'	=>	lang('Gross Earnings'),
				'other'			=>	lang('Other Income / Expenses'),
				'before_taxes'		=>	lang('Earnings / Loss Before Taxes'),
				'provision_for_taxes'	=>	lang('Provision for Taxes'),
				'net_earnings'		=>	lang('Net Earnings')
			);
			$data['lang'][0]['date_range'] = lang('Between %1 and %2', 
				date($GLOBALS['egw_info']['user']['preferences']['common']['dateformat'], $query['start_date']),
				date($GLOBALS['egw_info']['user']['preferences']['common']['dateformat'], $query['end_date'])
			);
		}

		/**
		*	Determine parameters that can be calculated:
		*	- Current Year
		*	- Previous Year
		*	- Start Date (current year & previous year)
		*	- End Date (current year & previous year)
		*	- Other income & expense groups
		*	- % column total
		*
		*	@param query Array of report parameters
		*	@param data Array of report data
		*/
		protected function get_calculated_parameters(&$query, &$data) {
			$query['current']['year'] 	= date('Y', $query['start_date']);
			$query['current']['start_date']	= $query['start_date'];
			$query['current']['end_date']	= $query['end_date'];

			$query['previous']['year']	= date('Y', $query['start_date'] - 1);
			$query['previous']['start_date']	= strtotime('-1 year', $query['start_date']);
			$query['previous']['end_date']	= strtotime('-1 year', $query['end_date']);

			$data['current'][] = $query['current'];
			$data['previous'][] = $query['previous'];

			// Already selected groups
			$taken = array_merge($query['sales'], $query['cost_of_sales'], $query['expenses'], $query['provision_for_taxes']);
			$groups = ExecMethod('perp_ledger.bo_perp_ledger.get_account_groups', array(
				'sequence'	=>	array($query['income'], $query['expense'])
			));
			$query['other'] = array_diff(array_keys($groups), $taken);

			// % column total for calculating percentages
			$this->get_section($query, $data, 'percent_column');
			$data['percent_column'] = array(
				'current'	=>	$data['percent_column_summary'][0]['current'] ? (float)$data['percent_column_summary'][0]['current'] : 0,
				'previous'	=>	$data['percent_column_summary'][0]['previous'] ? (float)$data['percent_column_summary'][0]['current'] : 0,
			);
		}


		/**
		*	Get the data for one section
		*/
		protected function get_section(&$query, &$data, $section_name) {
			if(count($query[$section_name]) == 0) {
				return;
			}
			$result = $this->calculate($query, $data,
				array(), 
				array(
					'account_group IN (' . implode(',', $query[$section_name]) . ')'
				),
				array(
					'account_group IN (' . implode(',', $query[$section_name]) . ')'
				)
			);
			if($result) {
				$data[$section_name . '_summary'][] = $result[0];
			} else {
				$data[$section_name . '_summary'][] = array(
					'current'		=>	0,
					'current_percent'	=>	0,
					'previous'		=>	0,
					'previous_percent'	=>	0
				);
					
			}
				

			$data[$section_name . '_details'] = array();
			if($query['include_accounts'][$section_name]) {
				$data[$section_name . '_details'] = $this->calculate(
					$query, $data,
					array(
						'account_name'
					),
					array(
						'account_group IN (' . implode(',', $query[$section_name]) . ')'
					),
					array(
						'account_id',
					)
				);
			}
		}

		/**
		*	Get some numbers
		*/
		protected function calculate($query, $data, Array $select = array(), $where, $group_by = '') {
			$select[] = 'SUM(current.amount  * -1 * asset_or_liability) AS current';
			$select[] = 'SUM(previous.amount  * -1 * asset_or_liability) AS previous';
			if($data['percent_column']) {
				$select[] = 'SUM(current.amount  * -1 * asset_or_liability) / ' . $data['percent_column']['current'] . ' AS current_percent';
				$select[] = 'SUM(previous.amount  * -1 * asset_or_liability) / ' . $data['percent_column']['previous'] . ' AS previous_percent';
			}

			$join = 'LEFT JOIN (
					SELECT 
						account,
						SUM(amount) AS amount
					FROM perp_general_ledger
					WHERE
					trans_date BETWEEN \'' . $GLOBALS['egw']->db->to_timestamp($query['current']['start_date']) . '\' AND \'' . 
						$GLOBALS['egw']->db->to_timestamp($query['current']['end_date']) . '\' 
					GROUP BY account
				) AS current ON 
					perp_account.account_id = current.account
				LEFT JOIN (
					SELECT 
						account,
						SUM(amount) AS amount
					FROM perp_general_ledger
					WHERE
					trans_date BETWEEN \'' . $GLOBALS['egw']->db->to_timestamp($query['previous']['start_date']) . '\' AND \'' . 
						$GLOBALS['egw']->db->to_timestamp($query['previous']['end_date']) . '\'
					GROUP BY account
				) AS previous ON
					perp_account.account_id = previous.account
				JOIN perp_account_group ON
					perp_account.account_group = perp_account_group.group_id';
			$sql = 'SELECT ' . implode (',', $select) . ' FROM perp_account ' . $join;
			if($where) {
				$sql .= ' WHERE ' . implode(' AND ', $where);
			}
			$sql .= ' GROUP BY ' . ($group_by ? implode(', ', $group_by) : ' account_id ') . 
				' ORDER BY current DESC, account_id';

			$result = $GLOBALS['egw']->db->query($sql, __LINE__, __FILE__);

			$list = array();
			foreach($result as $row) {
				if($query['include_zero'] || (!$query['include_zero'] && ($row['current'] != 0 || $row['previous'] != 0))) {
					$list[] = $row;
				}
			}
			return $list;
		}

		/**
		*	Get selectbox options for the template(s)
		*/
		protected function get_options() {
			$sequences = ExecMethod('perp_ledger.bo_perp_ledger.get_sequences');
			$groups = ExecMethod('perp_ledger.bo_perp_ledger.get_account_groups');
			$income = ExecMethod('perp_ledger.bo_perp_ledger.get_account_groups', array('sequence' => $config['income_statement']['income']));

			return array(
				'options-income'		=>	$sequences,
				'options-expense'		=>	$sequences,
				'options-sales'			=>	$income,
				'options-cost_of_sales'		=>	$groups,
				'options-expenses'		=>	$groups,
				'options-provision_for_taxes'	=>	$groups,
				'options-percent_column'	=>	$groups
			);
		}

		/**
		*	Settings specific to being scheduled
		*/
		public function get_schedule_settings($async_id) {
			$settings = parent::get_schedule_settings($async_id);

			if(!$settings['start_date']) {
				$settings['start_date'] = 'this_month_start';
			}

			$config = config::read('perp_ledger');
			if($config['income_statement']) {
				$settings += $config['income_statement'];
			}

			$settings += $this->get_options();

			return $settings;
		}

	}

?>
