<?php
	/**
	*	SalesOrder encapsulates an order where we are the supplier, and we're selling something.
	*	SalesOrder extends perp_api::Order, and inherits much of its functionallity from there.
	*
	*	@author Nathan Gray
	*	@ingroup perp_orders
	*/

	require_once(EGW_INCLUDE_ROOT . '/perp_api/inc/class.Order.inc.php');

	require_once('class.SalesOrderLine.inc.php');

	class SalesOrder extends Order {
		const sales_order_table = 'perp_sales_order';

		public static $work_order_table = 'perp_sales_order_work_order';

		protected $client_account_id;
		protected $client_branch_id;

		public function __construct($so_id = null) {
			parent::__construct($so_id);
		}

		/**
		*	Read the given SalesOrder from the database
		*/
		public function read($so_id) {
			$GLOBALS['egw']->db->select(self::sales_order_table, array(
					'client_branch_id',
					'client_account_id',
				),
				array( 'order_id' => $so_id), 
				__LINE__, __FILE__, 'perp_orders');
			$row = $GLOBALS['egw']->db->row(True);
			$this->client_branch_id		=	$row['client_branch_id'];
			$this->client_account_id	=	$row['client_account_id'];

			parent::read($so_id);

			// Get price overrides
			$bo_infolog = CreateObject('infolog.infolog_bo');
			$query = array(
				'action'	=>	'perp_orders',
			);
			foreach($this->order_lines as $line_no => $line) {
				$query['action_id'] = 'SALES_ORDER-' . $so_id . ':' . $line_no;
				$log_keys = $bo_infolog->search($query);
				if(count($log_keys)) {
					$key = key($log_keys);
					$line->set_price_override($log_keys[$key]['info_id']);
				}
			}
		}

		/* Writes the given sales order to the database
		*  Returns the sales order ID (not the SO number)
		*/
		public function write() {
			$GLOBALS['egw']->db->transaction_begin();
			$result = parent::write();
	
			// Write SO
			$GLOBALS['egw']->db->insert(self::sales_order_table, array(
					'client_branch_id'	=>	$this->client_branch_id,
					'client_account_id'	=>	$this->client_account_id
				),
				array('order_id'	=>	$this->order_id),
				__LINE__, __FILE__, 'perp_orders'
			);

			$GLOBALS['egw']->db->transaction_commit();
			return $result;
		}

		/* Check that this sales order is valid.
		*	Returns True or error message
		*/
		public function check() {
			$check = parent::check();
			if($check !== true) {
				return $check;
			}
			// Check for valid client account
			if(!$this->client_account_id) {
				return lang('Missing client account');
			}
			
			if(!$this->client_branch_id) {
				return lang('Missing client branch');
			}

			// Check all lines for completeness
			if($this->approved_date) {
				$invoiced_date = null;
				$ship_date = null;
			
				foreach($this->order_lines as $line) {
					if($line->get_shipped_date()) {
						$shipped_date = max($shipped_date, $line->get_shipped_date());
					} else {
						$shipped_date = null;
					}
					if($line->get_invoiced_date()) {
						$invoiced_date = max($invoiced_date, $line->get_invoiced_date());
					} else {
						$invoiced_date = null;
					}
	
				}

				// If all the lines are received || invoiced, this SO is too
				if(!$this->ship_date) {
					$this->ship_date = $shipped_date;
				}
				if(!$this->invoiced_date) {
					$this->invoiced_date = $invoiced_date;
				}
			}

			return True;
		}


		/* Can $id deny the Order?
		*  Same as approve, as long as it hasn't been approved
		*/
		public function is_denier($id) {
			$err = $this->is_admin($id);
			return ($err === True && !$this->get_approved_date());
		}

		public function deny() {
			$denier = $GLOBALS['egw_info']['user']['account_id'];
			$check = $this->is_denier($denier);
			if($check === True) {
				return $this->cancel();
			} else {
				return $check;
			}
		}

		public function is_canceller($id) {
			$perm = $GLOBALS['egw']->acl->check('sales_order', EGW_ACL_CUSTOM_1, 'perp_orders');
			if(!$this->requires_approval || !$this->approved_date) {
				$perm = ($id == $this->initiator) || $perm;
			}
			return $perm;
		}

		public function cancel($cancelling_user_id) {

			if($this->is_canceller($id)) {
				$GLOBALS['egw']->db->transaction_begin();
				parent::cancel($cancelling_user_id);

				$GLOBALS['egw']->db->delete(SalesOrder::sales_order_table, 
					array('order_id' => $this->order_id),
					__LINE__, __FILE__, 'perp_ar'
				);
				$GLOBALS['egw']->db->delete(SalesOrderLine::sales_order_line_table,
					array('order_id' => $this->order_id),
					__LINE__, __FILE__, 'perp_ar'
				);
				$GLOBALS['egw']->db->delete(self::$work_order_table,
					array('order_id' => $this->order_id),
					__LINE__, __FILE__, 'perp_ar'
				);

				egw_link::unlink('', 'perp_orders', 'ORDER-' . $this->order_id);

				$GLOBALS['egw']->db->transaction_commit();
				return True;
			} else {
				return lang('Access not permitted');
			}
		}

		public function ship($shipment_id, $line_no, $quantity) {
			if(!$quantity) {
				return;
			}

			$ship_line = $this->get_line($line_no);

			if($ship_line) {
				$ship_line->ship($shipment_id, $quantity);
			}
			$this->write();
			return $ship_line;
		}

		public function can_invoice() {
			$test = false;
			foreach($this->order_lines as $line) {
				$test = $test || (boolean)$line->can_invoice();
			}
			$test = $test && !$this->get_approval_required();
			$test = $test && (boolean)$GLOBALS['egw']->acl->check('invoice', EGW_ACL_ADD, 'perp_ar');
			return $test;
		}

		/**
		*	Only updates sales_order_details, checks & updates invoice date
		*	@param line_no is either a single stock_id, or an array of stock_id => quantity
		*
		*	@todo Check if this is still used, I don't think so.
		*/
		public function invoice($line_no, $invoice_qty = '') {
			if(is_array($line_no)) {
				foreach($line_no as $line => $qty) {
					$this->_invoice($line, $qty);
				}
			} elseif($invoice_qty) {
				$this->_invoice($line_no, $invoice_qty);
			}
			$this->write();
		}
		private function _invoice($line_no, $invoice_qty) {
			$invoice_line = $this->get_line($line_no);
			if($invoice_line) {
				$invoice_line->invoice($invoice_qty);
			} else {
				throw new Exception(lang('Could not find line to invoice'));
			}
		}

		// Accessors
		public function get_client_branch_id() {
			return $this->client_branch_id;
		}
		public function set_client_branch_id($branch) {
			// Check for valid client branch
			$branches = ExecMethod('perp_ar.bo_perp_client_branch.get_client_branch_list', array('client_id' => $this->get_client_id()));

			if(!array_key_exists($branch, $branches)) {
				$this->client_branch_id = null;
				return lang('Invalid client branch');
			}
			$this->client_branch_id = $branch;
		}

		public function get_client_account_id() {
			return $this->client_account_id;
		}
		public function set_client_account_id($client_account) {
			$accounts = ExecMethod('perp_ar.bo_perp_client_account.get_client_account_list', array('client_id' => $this->get_client_id()));
			if(!array_key_exists($client_account, $accounts)) {
				$this->client_account_id = null;
				return lang('Invalid client account');
			}
			$this->client_account_id = $client_account;
		}
		
		public function get_requested_delivery_date() {
			if($this->requested_delivery_date) {
				return $this->requested_delivery_date;
			}

			$lead_time = 0;

			foreach($this->order_lines as $key => $line) {
				$lead_time = max($lead_time, $line->get_lead_time());
			}
			$start_time = $this->issue_date;
			return date('U', mktime(0,0,0, date('m', $start_time), date('d', $start_time) + $lead_time, date('Y', $start_time)));
		}

		/*
		*	Can this SalesOrder be shipped?  At all?
		*	Returns true if this SalesOrder can be at least partly shipped.
		*/
		public function can_ship() {
			$test = parent::can_ship();

			if($test !== true) {
				return $test;
			}

			if($this->is_complete()) {
				return lang('Closed');
			}

			if($this->get_approval_required()) {
				return lang('Waiting for approval');
			}
			
			return true;
		}

		/* Returns 2 if this SO can be fully invoiced,
		*  1, if this SO can be partially invoiced,
		*  0 if this SO cannot be invoiced (nothing shipped)
		*/
		public function is_invoicable() {
			if($this->invoiced_date) {
				return False;
			}

			// If terms are cash only or pre-pay, it's invoicable
			$terms = ExecMethod('perp_api.bo_perp_api.get_payment_terms', array('terms' => $this->get_terms()));
			$terms = $terms[0];
			$due = max($terms['days_before_due'], $terms['day_in_following_month']);
			if($due <= 1) {
				return True;
			}

			$invoicable = True;
			$partial = False;
			foreach($this->order_lines as $key => $line) {
				if($line->get_quantity_shipped() && !$line->get_invoiced_date()) {
					$partial |= True;
				} else {
					$invoicable = False;
				}
			}
			return $invoiceable + $partial;
		}
	

		/* remove the given line from this transaction
		*/
		public function remove_line($line_number) {
			$line = $this->get_line($line_number);
			if($line ) {
				// Remove price overrides
				if($line->get_price_override()) {
					ExecMethod2('infolog.infolog_bo.delete', $line->get_price_override(), True);
				}
			}
			parent::remove_line($line_number);
		}

		/**
		*	Determine the source address by whichever location has the highest number of unshipped goods.
		*
		*	Because each OrderLine can have a different source location,
		*/
		public function get_ship_from_address() {
			$locations = array();
			foreach($this->order_lines as $line) {
				$locations[$line->get_ship_location()] += $line->get_ship_quantity();
			}
			unset($locations[0]);
			arsort($locations);
			if(count($locations) > 0) {
				$location_id = key($locations);
				$location = ExecMethod('perp_inventory.bo_perp_location.get_location', $location_id);
				return $location['delivery_address'];
			} else {
				return null;
			}
		}

		/**
		*	return the lines as an array
		*/
		public function get_lines($options = False) {
	return parent::get_lines();
			$lines = array();
			if($options) {
				$bo_wo = CreateObject('perp_manufacturing.bo_perp_work_order');
				$bo_so = CreateObject('perp_orders.bo_perp_sales_order');
			}

			foreach($this->order_lines as $key => &$line) {
				$lines[$key] = $line->toArray();
				$lines[$key]['line_no'] = $key;
				$lines[$key]['extended_cost'] = sprintf("%.2f", $lines[$key]['quantity_ordered'] * $lines[$key]['unit_price']);
				
				if($options && $line->get_work_orders()) {
					$work_orders = $line->get_work_orders();
					foreach($work_orders as $wo_number => $wo) {
						// Get work order reference
						$work_order = $bo_wo->get_work_order($wo['work_order_id']);
						$lines[$key]['work_orders'][$wo_number]['work_order_reference'] = $work_order['work_order_reference'];

						// Get serial(s)
						/* You can uncomment this if you need
						$sn = $bo_wo->get_serial_number($wo['work_order_id'], $line->get_stock_id());
						if(count($sn)) {
							array_unshift($sn, false);
							$lines[$key]['serial_numbers'] = $sn;
						}
						*/
						// Get options
						$option_list = array();
						$query = array(
							'work_order_id'	=>	$wo['work_order_id'],
							'detail_id'	=>	$wo['detail_no'],
							'stock_id'	=>	$line->get_stock_id(),
							'client_account_id'	=>	$this->get_client_account()
						);
						$bo_so->get_detail_details(
							$query,
							$option_list
						);
						foreach($option_list as $opt_no => &$opt) {
							$opt['value']	=	$opt['options-value'][$opt['value']]['build_option_name'];
							unset($opt['options-value']);
						}
						if(count($option_list)) {
							array_unshift($option_list, false);
							$lines[$key]['options'] = $option_list;
						}
					}
					array_unshift($lines[$key]['work_orders'], false);
					$lines[$key]['details'] = $lines[$key]['serials'] || $lines[$key]['options'];
				}

				// Price overrides
				if($line->get_price_override()) {
					$lines[$key]['price_override'] = 'flag';
					$lines[$key]['info_id'] = $line->get_price_override();
				}
			}
			return $lines;
		}

		public function toArray() {
			$data = parent::toArray();
			$data['requested_delivery_date'] = $this->get_requested_delivery_date();
			return $data;
		}

		public function __toString() {
			$GLOBALS['egw']->accounts->get_account_name($this->originating_user, $id, $fname, $lname);
			$initiator_name = $fname . ' ' . $lname . ' [' . $id . ']';
			if($this->client_account_id) {
				$account_name = ExecMethod('perp_ar.bo_perp_client_account.link_title', $this->client_account_id);
			}
			if($this->client_branch_id) {
				$branch_name = ExecMethod('perp_ar.bo_perp_client_branch.link_title', $this->client_branch_id);
			}
			$string = sprintf("Sales Order for \nAccount %s\nBranch %s",
				$account_name, $branch_name
			);
			return $string . "\n" . parent::__toString();
		}

		public function __sleep() {
			$sleep = parent::__sleep();
			$sleep[] = 'client_branch_id';
			$sleep[] = 'client_account_id';
			return $sleep;
		}

		public function suggest_order_number() {

			// Get so number format
			$config = config::read('perp_orders');
			$so_format = $config['so_number'];

			// Get customer_num & next_customer_num
			$GLOBALS['egw']->db->select(self::sales_order_table,
				array('count(' . self::sales_order_table . '.order_id) + 1 AS next_customer_num',
					'perp_client.client_id'
				),
				array('perp_client.client_id = ' .$this->client_id),
				__LINE__, __FILE__, False, 
				'GROUP BY client_id ', 'perp_orders', 0,
				'JOIN ' . self::order_table . ' ON
					' . self::order_table . '.order_id = perp_order.order_id
				LEFT JOIN perp_client ON
					' . self::order_table . '.client_id = perp_client.client_id'
			);
			$row = $GLOBALS['egw']->db->row(True);
			$next_client_num = $row['next_client_num'];
			$client_account = ExecMethod('perp_ar.bo_perp_client_account.get_client_account', $client_account_id);
			$client = ExecMethod('perp_ar.bo_perp_client.get_client', $this->client_id);
			$client_num = $client['client_num'];

			// get next_so_id
			$GLOBALS['egw']->db->select(self::order_table,
				array('max(order_id) + 1 AS next_so_id'),
				array(),
				__LINE__, __FILE__, False, False, 'perp_orders');
			$row = $GLOBALS['egw']->db->row(True);
			$next_so_id = $row['next_so_id'];

			// get next_by_date
			$GLOBALS['egw']->db->select(self::order_table,
				array('count(order_id) + 1 AS next_by_date'),
				array(
					self::order_table . '.client_id = '. $this->client_id,
					'issue_date BETWEEN \'' . $GLOBALS['egw']->db->to_timestamp(date('U', mktime(0,0,0,date('m'), date('d'), date('Y')))) . 
						'\' AND \'' . $GLOBALS['egw']->db->to_timestamp(date('U', mktime(0,0,0,date('m'), date('d') + 1, date('Y')))) .'\''
				),
				__LINE__, __FILE__, False,
				'', 'perp_orders', 0,
				'LEFT JOIN perp_client ON
					' . self::order_table . '.client_id = perp_client.client_id'
			);
			$row = $GLOBALS['egw']->db->row(True);
			$next_by_date = $row['next_by_date'];
				
			ereg('date\([dDjlNSwzWFmMntYycU.]+\)', $so_format, $regs);
			if(count($regs)) {
				eval('$date = ' . $regs[0] . ';');
				$so_format = str_replace($regs[0], '%5$s', $so_format);
			}
			
			$so_num = sprintf($so_format, 
				$client_num,
				$next_so_id,
				$next_client_num,
				sprintf("%02d", $next_by_date),
				$date,
				$client_account_id
			);
			return $so_num;
		}
	}
?>
