<?php

	/**
	*	Logic layer for automatically adjusting prices.
	*
	*	Based on a user-defined calculation, all the prices in the specified price list(s)
	*	will be updated.
	*
	*	@author Nathan Gray
	*/

	require_once( EGW_INCLUDE_ROOT . '/perp_orders/inc/class.so_perp_price_calculation.inc.php');

	class bo_perp_price_calculation {

		/**
		*	The name of the custom variable that will be prompted for when the calculation is run
		*/
		public static $custom_variable_name = '$custom';

		/**
		*	Allowed functions, with an explaination
		*/
		public static $allowed_functions = array(
			'if'		=>	'( check ? true value : false value)',
			'round'		=>	'Round to the specified number of decimals, using standard rounding rules',
			'ceil'		=>	'(Ceiling) Adjust price to the next highest whole number',
			'floor'		=>	'Adjust price to the next lowest whole number',
			'min'		=>	'The smallest number between a list of numbers',
			'max'		=>	'The largest number between a list of numbers',
			'rand'		=>	'Generate a random integer between two limits',
		);

		/**
		*	Allowed variables, with an explaination
		*
		*	Some constants are added in the constructor
		*/
		public static $allowed_variables = array(
			'$current_fifo'		=>	'The current cost for the item, calculated using FIFO',
			'$current_catalog'	=>	"The current cost for the item, calculated by (supplier's catalog cost + supplier's indirect cost)",
			'$current_average'	=>	'The current cost for the item, calculated by an average over current stock',
			'$exchange_rate'	=>	'The current exchange rate, from your GL currency to the price list currency',
			'$current_price'	=>	'The current price for the item on the price list, at the time the calculation is done',
		);

		/**
		*	Allowed strings, usually keywords the parser interprets as strings
		*/
		protected static $allowed_strings = array(
			'true', 'false', 'null'
		);

		/**
		*	Store any errors when executing
		*/
		protected static $error = null;

		/**
		*	Storage object
		*/
		private static $so;

		public function __construct() {
			self::static_init();
		}

		public static function static_init() {
			self::$so = new so_perp_price_calculation();
			self::$allowed_variables[self::$custom_variable_name]	=	'A variable that will be prompted for when the calculation is done.';
			self::$allowed_variables['\'' . so_perp_price::NOT_ALLOWED . '\'']	=	'The item is not sold on the price list';
		}

		/**
		*	Get a list of calculations
		*/
		public function search(&$query, &$rows, &$readonly) {
			if(!$GLOBALS['egw']->acl->check('calculation', EGW_ACL_READ, 'perp_orders')) {
				$rows = array();
				return 0;
			}

			if(!$query['order']) {
				$query['order'] = 'id';
			}

			$rows = self::$so->search(array(), false, $query['order'] . ' ' . $query['sort'], 
				'', '', false, false, array($query['start'], $query['num_rows']));

			if(!is_array($rows)) {
				return 0;
			}

			// Readonlys and schedules
			foreach($rows as $key => &$row) {
				// Readonlys
				if(!$GLOBALS['egw']->acl->check('price_list', EGW_ACL_EDIT, 'perp_orders')) {
					$readonly[$key]['run[' . $row['id'] . ']'] = true;
					$readonly['schedule[' . $row['id'] . ']'] = true;
				}
				if(!$GLOBALS['egw']->acl->check('calculation', EGW_ACL_DELETE, 'perp_orders')) {
					$readonly['delete[' . $row['id'] . ']'] = true;
				}

				// Schedules
				$search_id = 'price_adjust.' . $row['id'] . '.%';
				$schedules = ExecMethod('phpgwapi.asyncservice.read', $search_id);
				if(is_array($schedules)) {
					foreach($schedules as $async_id => $schedule) {
						$row['schedules'][] = array(
							'async_id'		=>	$async_id,
							'escaped_async_id'	=>	urlencode($async_id),
							'next'			=>	$schedule['next'],
							'price_list_id'		=>	$schedule['data']['price_list_id'],
							// Prompt is the same for all schedules, but it makes display easier
							'variable_prompt'	=>	$row['variable_prompt'],
							'custom'		=>	$schedule['data']['custom'],
							'formatted_next'	=>	date(
								$GLOBALS['egw_info']['user']['preferences']['common']['dateformat'] . ' ' .
								($GLOBALS['egw_info']['user']['preferences']['common']['timeformat'] == '24' ? 'H:i' : 'h:i a'),
								$schedule['next']
							),
						);
						$can_cancel = $GLOBALS['egw_info']['user']['account_id'] == $schedule['account_id'] || $GLOBALS['egw']->acl->check('run', EGW_ACL_READ, 'admin');
						$readonly['cancel[' . urlencode($async_id) . ']'] = !$can_cancel;
					}
				} else {
					$readonly[$key+1]['schedules'] = true;
				}
			}

			return self::$so->total;
		}

		/**
		*	Read an existing calculation.
		*/
		public function read($calculation_id) {
			if(!$GLOBALS['egw']->acl->check('calculation', EGW_ACL_READ, 'perp_orders')) {
				throw new egw_exception_no_permission(lang('You do not have permission to view pricing calculations'));
			}
			return self::$so->read($calculation_id);
		}

		/**
		*	Save a calculation
		*/
		public function save($calculation) {
			// Permission check
			if(!$GLOBALS['egw']->acl->check('calculation', EGW_ACL_EDIT, 'perp_orders')) {
				throw new egw_exception_no_permission(lang('You do not have permission to edit pricing calculations'));
			}

			// Check for a valid calculation
			$check = self::check($calculation);
			if($check === true) {
				self::$so->save($calculation);
				return self::$so->data[self::$so->autoinc_id];
			} else {
				throw new egw_exception(lang('Invalid calculation: %1', $check));
			}
			return false;
		}

		/**
		*	Delete a calculation
		*/
		public function delete($calculation_id) {
			if(!$GLOBALS['egw']->acl->check('calculation', EGW_ACL_DELETE, 'perp_orders')) {
				throw new egw_exception_no_permission(lang('You do not have permission to delete pricing calculations'));
			}
			return self::$so->delete($calculation_id);
		}
			

		/**
		*	Check a provided calculation for validity.
		*
		*	Some things that get checked:
		*	- Matching brackets
		*	- Limited to the function list: IF, CASE, ROUND, CEILING, FLOOR, MIN, MAX, RANDOM
		*	- Using the custom variable without a prompt
		*	- Using a prompt and not using the custom variable
		*	- Calculation can be executed
		*
		*	@param calculation Array of calculation fields, as provided by read()
		*	@return true or error message
		*/
		public static function check(Array $calculation) {
			$result = self::bracket_check($calculation['calculation']);
			if($result !== true) {
				return lang('Brackets do not match');
			}

			/*
			$result = self::variable_check($calculation['calculation']);
			if($result !== true) {
				return lang('Using an invalid variable: %1', $result);
			}
			*/

			if(strpos($calculation['calculation'], self::$custom_variable_name) !== false && trim($calculation['variable_prompt']) == '') {
				return lang('Missing prompt for custom variable');
			}
			if(trim($calculation['variable_prompt']) != '' && strpos($calculation['calculation'], self::$custom_variable_name) === false) {
				return lang('You gave a prompt, but are not using the variable anywhere');
			}

			$result = self::calculation_check($calculation['calculation']);
			if($result !== true) {
				return $result;
			}

			return true;
		}

		/**
		*	Check to see if brackets match
		*	
		*	Allowed brackets are (), [], {}
		*
		*	@param string String to check
		*	@return boolean
		*/
		protected static function bracket_check($string) {
	
			$brackets = array(
				'('	=>	')',
				'['	=>	']',
				'{'	=>	'}',
			);

			$bracket_stack = array();

			for($i = 0; $i < strlen($string); $i++) {
				if(in_array($string[$i], array_keys($brackets))) {
					array_push($bracket_stack, $brackets[$string[$i]]);
				} elseif (in_array($string[$i], $brackets)) {
					if(array_pop($bracket_stack) != $string[$i]) {
						return false;
					}
				}
			}
			return count($bracket_stack) == 0;
		}

		/**
		*	Check to make sure any functions used are in the allowed list, and that
		*	the calculation can be executed.
		*
		*	@param string String to check
		*	@return boolean
		*/
		protected static function calculation_check($string) {
			// Start buffering, so we can catch any parse errors
			ob_start();

			// Need the php tags so the parser knows this is PHP
			$parse_string = '<?php ' . $string . '; ?>';

			$tokens = token_get_all($parse_string);
			$parsed = '';

			foreach($tokens as $token) {
				if(is_array($token)) {
					$id = $token[0];
					switch ($id) {
				// Accept these tokens
					case T_WHITESPACE:
					case T_CONSTANT_ENCAPSED_STRING:
					case T_COMMENT:
					case T_LNUMBER:	case T_DNUMBER:
					case T_IS_EQUAL: case T_IS_SMALLER_OR_EQUAL: case T_IS_GREATER_OR_EQUAL: case T_IS_NOT_EQUAL:
					case T_PLUS_EQUAL: case T_MINUS_EQUAL: case T_MUL_EQUAL: case T_DIV_EQUAL: case T_MOD_EQUAL:
					case T_IF: case T_ELSEIF: case T_ELSE:
						$parsed .= $token[1];
						break;

				// These tokens need to be checked
					case T_STRING:
					case T_FUNCTION:
						if($id == T_STRING && array_search(strtolower($token[1]), self::$allowed_strings)) {
							$parsed .= $token[1];
							break;
						}
						if(array_search($token[1], array_keys(self::$allowed_functions)) === false) {
							// Invalid function
							return lang('%1 is not an allowed function', $token[1]);
							break;
						}
						$parsed .= $token[1];
						break;
					case T_VARIABLE:
						if(array_search($token[1], array_keys(self::$allowed_variables)) === false && $token[1] != '$price') {
							// Invalid function
							return lang('%1 is not an allowed variable', $token[1]);
							break;
						}
						$parsed .= $token[1];
						break;

				// Ignore these tokens
					case T_OPEN_TAG:
					case T_CLOSE_TAG:
						break;
					default:
						return lang('Unknown token: %1', token_name($id));
						break;
					}
				} else {
					$parsed .= $token;
				}
			}


			// Check to see if it can be executed
			eval('$price = ' . $string . ';');

			$buffer = ob_get_clean();
			if($buffer) {
				$error = error_get_last();
				switch ($error['type']) {
					case E_PARSE:
						return lang('Not a valid formula.  Forget something?');
						break;
					default:
						return $buffer;
				}
			}

			return true;
		}

		/**
		*	Adjust prices on certain price lists using the given calculation.
		*
		*	@param parameters An array with the following keys:
		*	- id:			ID of the calculation to use
		*	- price_list_id:	Array of price lists to adjust
		*	- effective_date:	Date and time to make the new prices take effect.  Must be in the future.
		*	- custom:		Optional, if the calculation uses a custom variable
		*/
		public static function run(Array $parameters, $make_changes = true) {
			if(!$parameters['id'] && !$parameters['calculation']) {
				throw new egw_exception_wrong_parameter(lang('Missing calculation ID or formula'));
			}
			if(!$parameters['price_list_id']) {
				throw new egw_exception_wrong_parameter(lang('Missing price lists'));
			}
			if(!$parameters['effective_date']) {
				throw new egw_exception_wrong_parameter(lang('Missing effective date'));
			}
			if(!is_numeric($parameters['effective_date'])) {
				// Effective date is probably a relative date
				$parameters['effective_date'] = ExecMethod2('perp_api.perp_report.convert_date', $parameters['effective_date']);
			}
			if($parameters['effective_date'] <= time()) {
				throw new egw_exception_wrong_userinput(lang('Effective date must be in the future'));
			}
				
			if($parameters['id'] && !$parameters['calculation']) {
				$calculation = self::read($parameters['id']);
				$parameters['calculation'] = $calculation['calculation'];
			}

			if($parameters['calculation'] && self::calculation_check($parameters['calculation']) === true) {
				return self::$so->run($parameters, $make_changes);
			} else {
				return lang('Unable to run calculation, it may not be valid');
			}
		}

		/**
		*	Schedule a calculation for recurring execution
		*
		*	@param parameters Array with the following keys:
		*	- id
		*	- price_list_id
		*	- schedule_time
		*	- effective_date
		*
		*	@return async_id, or false
		*	@throws egw_exception_wrong_parameter on missing required parameters
		*/
		public static function schedule($parameters) {
			if(!$parameters['id']) {
				throw new egw_exception_wrong_parameter(lang('Missing calculation ID'));
			}
			if(!$parameters['price_list_id']) {
				throw new egw_exception_wrong_parameter(lang('Missing price lists'));
			}
			if(!$parameters['effective_date']) {
				throw new egw_exception_wrong_parameter(lang('Missing effective date'));
			}

			// ID is price_adjust.id.price_list_id
			$async_id = "price_adjust.{$parameters['id']}." . implode(',', $parameters['price_list_id']);
			if(!$parameters['schedule']) {
				// No time set, cancel the report
				return ExecMethod2('phpgwapi.asyncservice.cancel_timer', $async_id);
			} elseif ($schedules = ExecMethod('phpgwapi.asyncservice.read', $async_id)) {
				// report already scheduled, cancel it and save the updated version
				foreach($schedules as $id => $schedule) {
					ExecMethod('phpgwapi.asyncservice.cancel_timer', $id);
				}
			}
			
			$method = 'perp_orders.bo_perp_price_calculation.run';
			$account_id = $GLOBALS['egw_info']['user']['account_id'];

			$result = ExecMethod2('phpgwapi.asyncservice.set_timer', $parameters['schedule'], $async_id, $method, $parameters, $account_id);

			return $result ? $async_id : false;
		}
	}

	bo_perp_price_calculation::static_init();
?>
