<?php

	/**
	*	Storage layer for prices of stock kit components.
	*
	*	Because every component of a stock kit could potentially be sold separately with a different
	*	(usually higher) price, we need to keep track of what the price is when the item is sold as
	*	part of a kit.  Regular pricing and option pricing are tracked through the regular methods.
	*
	*	@author Nathan Gray
	*	@package perp_orders
	*/
	class so_perp_kit_price {

		/**
		*	Get a list of stock items that are components of a stock_kit and their prices
		*
		*	@nextmatch
		*
		*	The following extra query keys are available:
		*	- price_list_id
		*	- stock_id
		*	- component_id
		*	- effective_date
		*	- include_not_allowed
		*
		*/
		public static function get_kit_price_list(&$query) {
			$fields = array(
				'perp_stock_kit_price.price_list_id',
				'IF(perp_stock_kit_price.stock_id IS NULL, ' . $query['stock_id'] . ', perp_stock_kit_price.stock_id) AS stock_id',
				'perp_bill_of_material.stock_id AS kit_stock_id',
				'perp_bill_of_material_detail.component AS component_id',
				'perp_bill_of_material_detail.quantity', 
				'perp_stock_kit_price.price',
				'perp_stock_kit_price.effective_date',
				'perp_stock_kit_price.modified_date',
				'perp_stock_kit_price.modified_by',
				'component.stock_num',
				'component.short_description',
				'component.decimal_places',
				'uom_name AS unit_of_measure_name',
			);

			// Which components go into a kit are recorded in the bill of material
			$from = 'perp_bill_of_material';

			$where = array(
				'perp_stock.stock_type_flag = \'K\''
			);

			$join = 'JOIN perp_bill_of_material_detail ON
				perp_bill_of_material.bill_of_material_id = perp_bill_of_material_detail.bill_of_material_id
			JOIN perp_stock ON
				perp_bill_of_material.stock_id = perp_stock.stock_id';

			if($query['price_list_id']) {
				$where[] = '(
					perp_stock_kit_price.price_list_id IN (' . implode(',', (array)$query['price_list_id']) . ') OR
					perp_stock_kit_price.price_list_id IS NULL
				)';
			}
			if($query['kit_stock_id']) {
				$where[] = 'perp_bill_of_material.stock_id = ' . (int)$query['kit_stock_id'];
			}
			if($query['stock_id']) {
				$where[] = '(perp_stock_kit_price.stock_id = ' . (int)$query['stock_id'] .
					($query['include_not_allowed'] ? ' OR perp_stock_kit_price.stock_id IS NULL' : '') . 
					')';
			}
			if($query['component_id']) {
				$where[] = 'perp_bill_of_material_detail.component= ' . (int)$query['component_id'];
			}
			if($query['effective_date']) {
				$date = $GLOBALS['egw']->db->to_timestamp($query['effective_date']);

				// Fancyness to get only the price effective on the date given instead of all prices
				$where[] = "'$date' >= perp_bill_of_material.effective_after_date AND (
					'$date' <= perp_bill_of_material.effective_to_date OR perp_bill_of_material.effective_to_date IS NULL
				)";
				$join .= ' LEFT JOIN (
					SELECT *
					FROM ( 
						SELECT *
						FROM perp_stock_kit_price
						WHERE effective_date <= \'' . $date . '\'' . 
						($query['price_list_id'] ? ' AND price_list_id IN (' . implode(',', (array)$query['price_list_id']) .')' : '') . '
						ORDER BY effective_date DESC
					) AS dated_prices
					GROUP BY price_list_id, stock_id, component_id
				) AS perp_stock_kit_price ON
					perp_stock_kit_price.kit_stock_id = perp_bill_of_material.stock_id AND
					perp_stock_kit_price.component_id = perp_bill_of_material_detail.component ';
			} else {
				$join .= ' LEFT JOIN perp_stock_kit_price ON
					perp_stock_kit_price.kit_stock_id = perp_bill_of_material.stock_id AND
					perp_stock_kit_price.component_id = perp_bill_of_material_detail.component ';
			}
			if(!$query['include_not_allowed']) {
				$where[] = 'price IS NOT NULL';
			}

			$join .= ' JOIN perp_stock AS component ON
				perp_bill_of_material_detail.component = component.stock_id
				LEFT JOIN perp_price_list ON
					perp_stock_kit_price.price_list_id = perp_price_list.price_list_id
				LEFT JOIN perp_unit_of_measure ON
					component.unit_of_measure = perp_unit_of_measure.uom_id';
			

			if($where) {
				$where_string .= "WHERE " . implode(' AND ', $where) . "\n";
			}
			$order = 'ORDER BY ' . ($query['order'] ? $query['order'] : 'perp_bill_of_material_detail.component') . ' ' . $query['sort'];

			// Number of rows
			$query['total'] = $GLOBALS['egw']->db->query("SELECT count(*) FROM $from $join $where_string")->fetchSingle();

			// Results
			$result = $GLOBALS['egw']->db->query('SELECT ' . implode(',', $fields) . " FROM $from $join $where_string $order", __LINE__, __FILE__, $query['start'], $query['num_rows']);

			$list = array();
			foreach($result as $row) {
				$row['effective_date'] = $GLOBALS['egw']->db->from_timestamp($row['effective_date']);
				$row['modified_date'] = $GLOBALS['egw']->db->from_timestamp($row['modified_date']);
				$row['quantity'] = round($row['quantity'], $row['decimal_places']);

				if(is_null($row['price'])) {
					$row['price'] = so_perp_price::NOT_ALLOWED;
				} elseif ($query['format']) {
					$row['price'] = number_format($row['price'], 2);
				}
				$list[] = $row;
			}
			return $list;
		}

		/**
		*	Set the price for a component of a stock kit.  
		*
		*	The new price takes effect at the given effective date (& time).  
		*	If the component is not allowed on the given
		*	price list (N/A), the price should be null.  Clients using that price list will
		*	not get that component.
		*
		*	@param price_list_id
		*	@param stock_id		The parent item being priced
		*	@param kit_stock_id	The immediate parent of the component.  Most cases, this will be the same as stock_id
		*	@param component_id	The stock_id of the component you are setting the price for
		*	@param effective_date	Date the price comes into effect
		*
		*	@return true
		*
		*	@throws egw_exception_wrong_parameter When the effective date is not set
		*/
		public static function set_price($price_list_id, $stock_id, $kit_stock_id, $component_id, $price, $effective_date) {
			if(!$effective_date) {
				throw new egw_exception_wrong_parameter('Missing effective date');
			}

			$db_effective_date = is_null($effective_date) ? null : $GLOBALS['egw']->db->to_timestamp($effective_date);
			$db_modified_date = $GLOBALS['egw']->db->to_timestamp(time());

			$GLOBALS['egw']->db->transaction_begin();

			// Update component price
			$result = $GLOBALS['egw']->db->insert(
				'perp_stock_kit_price',
				array(
					'price_list_id'		=>	(int)$price_list_id,
					'stock_id'		=>	(int)$stock_id,
					'kit_stock_id'		=>	(int)$kit_stock_id,
					'component_id'		=>	(int)$component_id,
					'effective_date'	=>	$db_effective_date,
					'price'			=>	$price,
					'modified_date'		=>	$db_modified_date,
					'modified_by'		=>	$GLOBALS['egw_info']['user']['account_id']
				),
				array(
					'price_list_id'		=>	(int)$price_list_id,
					'stock_id'		=>	(int)$stock_id,
					'kit_stock_id'		=>	(int)$kit_stock_id,
					'component_id'		=>	(int)$component_id,
					'effective_date'	=>	$db_effective_date
				),
				__LINE__, __FILE__, 'perp_orders'
			);

			// Update kit price
			/* This is done by so_perp_price_list.set_price_list()
			$component_query = array(
				'price_list_id'		=>	$price_list_id,
				'stock_id'		=>	$stock_id,
				'kit_stock_id'		=>	$kit_stock_id,
				'effective_date'	=>	$effective_date
			);
			$component_list = self::get_kit_price_list($query);
			$total = 0;
			foreach($component_list as $component) {
				if($component['price']) {
					$total += $component['price'];
				}
			}
			$result = ExecMethod2('perp_orders.so_perp_price.set_price', $price_list_id, $kit_stock_id, $total, $effective_date, null);
			if(!$result) {
				$GLOBALS['egw']->db->transaction_abort();
			}
			*/

			$GLOBALS['egw']->db->transaction_commit();

			if($result) {
				return true;
			} else {
				return false;
			}
		}

		/**
		*	Get the current price for a component of a stock kit, as of the given effective date
		*
		*	@param price_list_id
		*	@param stock_id		The parent item being priced
		*	@param kit_stock_id	The immediate parent of the component.  Most cases, this will be the same as stock_id
		*	@param component_id	The stock_id of the component you want the price for
		*	@param effective_date	Date to find the price 
		*
		*	@return double price in the price list's currency
		*/
		public static function get_price($price_list_id, $stock_id, $kit_stock_id, $component_id, $effective_date) {
			$db_effective_date = $GLOBALS['egw']->db->to_timestamp($effective_date);

			// Get component price
			$price = $GLOBALS['egw']->db->select(
				'perp_stock_kit_price',
				array(
					'price'
				),
				array(
					'price_list_id'		=>	(int)$price_list_id,
					'stock_id'		=>	(int)$stock_id,
					'kit_stock_id'		=>	(int)$kit_stock_id,
					'component_id'		=>	(int)$component_id,
					'effective_date <= \'' . $db_effective_date . '\''
				),
				__LINE__, __FILE__, false,
				'ORDER BY effective_date DESC', 'perp_orders', 1
			)->fetchSingle();
			if(is_null($price) || $price === false) {
				$price = so_perp_price::NOT_ALLOWED;
			}
			return $price;
		}
	}
?>
