<?php


	/**
	*	Canada pension plan (CPP) contributions
	*/

	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.Calculation.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.PayrollUtility.inc.php');

	class CPP extends Calculation {

		public static function getName() {
			return 'Canada Pension Plan';
		}

		public static function calculate(PaystubLine &$line, $value) {
			/*
				Value contains the keys:
				'types'	- pensionable income type_ids
			*/
		
			/*********
			From Canada Revenue Agency form T4127 Part B - Canada Pension Plan

			Formula to determine CPP contributions for employees receiving salary or wages 
			C = The lesser of: 
			(i) $1,989.90 minus D; and 
			(ii) 0.0495 multiplied by [PI minus ($3,500 divided by P)] 
			If the result is negative, C = $0. 

			Where: 
			C = Canada Pension Plan contribution (deduction) for the pay period. 
			D = Year-to-date CPP contribution (deduction) with the employer. This amount 
			cannot be greater than the annual maximum. 
			PI = Pensionable income for the pay period, or the gross income plus any 
			taxable benefits for the pay period, including bonuses and retroactive pay 
			increases where applicable. For more information about income subject to 
			Canada Pension Plan contributions, see the employers' guide called Payroll 
			Deductions (Basic Information). 
			P = The number of pay periods in the year. 

			***********/

			// Set precision
			bcscale(3);

			// 2007
			$basic_exemption = 3500;
			$contribution_rate = 0.0495;

			$staff 	=	new Staff($line->get_paystub()->get_staff_id());
			$P	=	PayrollUtility::get_pay_frequency($staff);

			$PI = parent::total_lines($line, $value['types']);
			$C = min( self::get_maximum_CPP($line, $staff), bcmul($contribution_rate, bcsub( $PI, bcdiv( $basic_exemption, $P ) )));
			$C = max(0, round($C, 2));
			return $C;
		}

		/**
		*	Max contribution for the year is 1989.90, unless employee age > 70 or < 18 at any point
		*	during the calendar year, in which case the max is prorated by month.
		*/
		protected static function get_maximum_CPP(&$line, &$staff) {
			$maximum_pensionable_earnings = 43700;
			$maximum_contribution = 1989.90;

			
			$address = ExecMethod('addressbook.addressbook_bo.read', $staff->get_addressbook_id());
			if(!$address) {
				throw new Exception(lang('Could not find address for staff_id %1', $staff->get_staff_id()));
			} elseif (!$address['bday']) {
				$address_link = $GLOBALS['egw']->link('/index.php?menuaction=addressbook.addressbook_ui.edit&contact_id=' . $staff->get_addressbook_id());
				$GLOBALS['egw']->log->error(array(
					'text'	=>	'E-Paystub, %1 %2.  (<a href=%3 target=\'_blank\'>Fix</a>)',
					'p_1'	=>	'Birthday is not set for',
					'p_2'	=>	$address['n_fn'],
					'p_3'	=>	$address_link,
					'file'	=>	__FILE__,
					'line'	=>	__LINE__
				));
			}

			/*
			in the case of an employee becoming 18 years old during the year, the 
			number of months in the year following the month the employee reaches age 18. 
			For example, if the date of birth is March 8, CPP contributions will start 
			April 1, and the number of months is 9

			in the case of an employee becoming 70 years old during the year, the 
			number of months before and including the month the employee reaches age 70. 
			For example, if the date of birth is November 9, CPP contributions will stop 
			November 30, and the number of months is 11. (The maximum contribution for 
			the year could be reached before November 30.) 
			*/
			$bday = $GLOBALS['egw']->db->from_timestamp($address['bday']);
			$max_age = date('Y') - date('Y', $bday);
			$months = 12;
			if($max_age <= 18 || $max_age >= 70) {
				if($max_age == 18) {
					$months = max(0, 12 - (date('n', $bday) ));
				} elseif ($max_age == 70) {
					$months = date('n', $bday);
				} else {
					$maximum_contribution = 0;
				}
			}
			$maximum_contribution = bcmul(bcdiv($maximum_contribution, 12), $months);
			$YTD = PayStubEntryType::get_YTD($line->get_type_id(), $line->get_paystub()->get_pay_period_id(), $line->get_paystub()->get_staff_id());
			$D = min($maximum_pensionable_earnings, $YTD['ytd_amount']);
			return bcsub($maximum_contribution, $D);
		}

		public static function getTemplate() {
			return 'perp_payroll.DeductionValue.source_lines';
		}

		public static function getOverrideTemplate() {
			return '';
		}

		public static function getOptions() {
			return array(
				'types'		=>	ExecMethod('perp_payroll.bo_perp_payroll.get_line_types')
			);
		}
	}
?>
