<?php


	/**
	*	Canada Unemployment Insurance (EI) contributions
	*/

	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.Calculation.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.PayrollUtility.inc.php');

	class EI extends Calculation {

		public static function getName() {
			return 'Employment Insurance';
		}

		public static function calculate(PaystubLine &$line, $value) {
			/*
				Value contains the keys:
				'types'	- pensionable income type_ids
			*/
		
			/*********
			From Canada Revenue Agency form T4127 Part C - Employment Insurance (EI)

			As an employer, your contribution is 1.4 times the amount of the employee's 
			premiums for the pay period. 
			The formula below will allow computer users to determine, in their payroll 
			calculations, the premium payable by an insured person under the Employment 
			Insurance Act. The formula is: 

			EI = the lesser of: 
			(i) $720.00 minus D1; and 
			(ii) 0.0180 multiplied by IE 

			For employees in Quebec only: 

			EI = the lesser of 
			(i) $584.00 minus D1; and 
			(ii) 0.0146 multiplied by IE* 

			* Round the resulting amount(s) in (ii) to the nearest $0.01. 
			Where: 

			EI = Employee's Employment Insurance premium deduction for the pay period. 

			D1 = Employee's year-to-date Employment Insurance premium with the employer 
			(D1 cannot be more than the maximum annual premium.) 

			IE = Insurable earnings for the pay period including insurable taxable 
			benefits, bonuses, and retroactive pay increases. 

			An employer's contribution is 1.4 times the amount of the employee's premiums, unless a 
			reduced rate applies.
			***********/

			// Set precision
			bcscale(3);

			// 2007
			$maximum_insurable_earnings = 40000;
			$contribution_rate = 0.0180;

			$staff 	=	new Staff($line->get_paystub()->get_staff_id());
			$P	=	PayrollUtility::get_pay_frequency($staff);

			$YTD = PayStubEntryType::get_YTD($line->get_type_id(), $line->get_paystub()->get_pay_period_id(), $line->get_paystub()->get_staff_id());
			$D1 = min(bcmul($maximum_insurable_earnings, $contribution_rate), $YTD['ytd_amount']);

			$IE = parent::total_lines($line, $value['types']);

			$EI = min(bcsub(bcmul($maximum_insurable_earnings, $contribution_rate), $D1), bcmul($contribution_rate, $IE));

			if($line->get_type()->get_type() == DeductionType::EMPLOYER) {
				return round(self::get_employer_contribution($EI, $value), 2);
			} else {
				return round($EI, 2);
			}
		}

		/**
		*	An employer's EI contribution is 1.4 times the amount of the employee's premiums, unless a 
		*	reduced rate applies.
		*/
		protected static function get_employer_contribution($EI, $value) {
			if($value['employer_rate']) {
				return bcmul($value['employer_rate'], $EI);
			} else {
				return bcmul(1.4, $EI);
			}
		}

		public static function getTemplate() {
			return 'perp_payroll.DeductionValue.source_lines';
		}

		public static function getOverrideTemplate() {
			return '';
		}

		public static function getOptions() {
			return array(
				'types'		=>	ExecMethod('perp_payroll.bo_perp_payroll.get_line_types')
			);
		}
	}
?>
