<?php


	/**
	*	Canadian Federal income tax.
	*/

	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.Calculation.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.PayrollUtility.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/calculation/CA/CPP.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/calculation/CA/EI.inc.php');

	class Federal extends Calculation {

		const FEDERAL = 1;
		const REGIONAL = 2;
		const LOCAL = 3;

		public static function getName() {
			return 'Canadian income tax';
		}

		public static function calculate(PaystubLine &$line, $value) {

			// Watch for rounding errors.  bcmath can't round.
	
			/************
			The following is from Canada Renvenue Agency T4127 84th edition

			We use Option 1, with the exception of a few factors, to determine the tax 
			deductions amounts contained in the Payroll Deductions Tables (T4032) and 
			Payroll Deductions Supplementary Tables (T4008) for each province and 
			territory, and in Canada beyond the limits of any province or outside Canada. 

			Outline of Option 1 

			In general, the Option 1 steps are as follows: 

			1. Determine the net taxable income for the pay period (remuneration minus 
			allowable deductions) and multiply it by the number of pay periods in the 
			year to get an estimated annual taxable income amount. This annual taxable 
			income amount is factor A. 

			2. Calculate the basic federal tax on the estimated annual taxable income, 
			after allowable federal personal tax credits. The basic federal tax is factor 
			T3. 

			3. Calculate the annual federal tax payable. This is factor T1. 

			4. Calculate the basic provincial or territorial tax on the estimated annual 
			taxable income, after allowable provincial or territorial personal tax 
			credits. The annual basic provincial or territorial tax is factor T4. 

			5. Calculate the annual provincial or territorial tax deduction. This is 
			factor T2. 

			6. Add the federal and provincial or territorial tax and divide the result by 
			the number of pay periods to find the estimated federal and provincial or 
			territorial tax deductions for a pay period. This is factor T. 

			Special rules apply to determine the annual income for employees remunerated 
			by commissions. An optional calculation is provided to determine the tax 
			deductions for bonuses, retroactive pay increases, and other non-periodic 
			payments. 

			******************/

			// Set precision
			bcscale(2);
			$staff 	=&	new Staff($line->get_paystub()->get_staff_id());

			$P	=	PayrollUtility::get_pay_frequency($staff);
			$L	=	$value['additional_deductions'];
			$A	=	self::get_net_taxable($line, $value, $P);
			if($A < 0) {
			// $L is additional deductions requested by employee
				return $L;
			}

			$T3	=	self::get_basic_federal_tax($line, $staff, $A, $P, $value);
			$T1	=	self::get_annual_federal_payable($line, $T3);
			$T4	=	self::get_basic_regional_tax($line, $staff, $A, $P, $value);
			$T2	=	self::get_annual_regional_payable($line, $T4);

			$T	=	bcadd( bcdiv( bcadd($T1, $T2), $P ), $L);
//echo "T	=	(($T1 + $T2) / $P ) + $L = $T<br />";
			return round($T, 2);	
		}

		public static function getTemplate() {
			return 'perp_payroll.DeductionValue.canadian_tax';

		}

		public static function getOverrideTemplate() {
			return 'perp_payroll.DeductionValue.canadian_tax_override';
		}

		public static function getOptions() {
			$line_types = ExecMethod('perp_payroll.bo_perp_payroll.get_line_types');
			return array(
				'gross_renumeration'	=>	$line_types,
				'payroll_deductions'	=>	$line_types
			);
		}

		/**
		*	A = Annual taxable income = [P * (I - F - F2 - U1 )] - HD - F1 
		*	If the result is negative, T = L. 
		*
		*	Where:
		*	T = Estimated federal and provincial or territorial tax deductions for the pay period. 
		*	L = Additional tax deductions for the pay period requested by the employee or pensioner as indicated on Form TD1.
		*	P = The number of pay periods in the year: 
		*		 Weekly P = 52 (or 53 where applicable) 
		*		 Bi-weekly P = 26 (or 27 where applicable) 
		*		 Semi-monthly P = 24 
		*		 Monthly P = 12 
		*		 Other P = 10, 13, 22, or any other number of pay periods of the employer or 
		*		 payer for the year 
		*
		*	I = Gross remuneration for the pay period. This includes overtime earned and 
		paid in the same pay period, pension income, qualified pension income, and 
		taxable benefits, but does not include bonuses, retroactive pay increases, or 
		other non-periodic payments. 
		*
		*	F = Payroll deductions for the pay period for employee contributions to a 
		registered pension plan (RPP) for current and past services, a registered 
		retirement savings plan (RRSP), or a retirement compensation arrangement 
		(RCA). 
		*	F2 = Alimony or maintenance payments required by a legal document
		*	U1 = Union dues for the pay period paid to a trade union, an association of 
		public servants, or dues required under the law of a province to a parity or 
		advisory committee or similar body. 
		*	HD = Annual deduction for living in a prescribed zone as indicated on Form 
		TD1. 
		*	F1 = Annual deductions (such as child-care expenses, and alimony and 
		maintenance payments not deducted from the employee's salary), requested by 
		an employee or pensioner and authorized by a tax services office or tax 
		centre. The employer or payer will be informed by a tax services office or 
		tax centre of the amount to be used with F1, when it is applicable. 
		*/
		protected static function get_net_taxable(&$line, &$value, $P) {
			$I = self::total_lines($line, $value['gross_renumeration']);
			$F = self::total_lines($line, $value['payroll_deductions']);

			$F2 = $value['alimony'];

			$U1 = self::total_lines($line, $value['union_dues']);

			$HD = $value['hd'];

			$F1 = $value['annual_garnish'];

			//A = $P * ($I - $F - $F2 - $U1)) - $HD - $F1;
			$A = bcsub(bcsub(bcmul($P,  bcsub($I, bcsub($F, bcsub($F2, $U1)))), $HD), $F1);
//echo "A = $P * ($I - $F - $F2 - $U1)) - $HD - $F1 = $A<br />";
			return $A;
		}


		/**
		*	T3 = Annual basic federal tax 
		*	= (R * A) - K - K1 - K2 - K3 - K4 
		*	If the result is negative, T3 = $0. 
		*
		*	Where:
		*	R = Federal tax rate applicable to annual taxable income A. 
		*	K = Federal constant to adjust the application of the federal tax rate to the annual taxable income A.
		*	K1 = Federal non-refundable personal tax credit = 0.155 multiplied by TC 
		*	TC = The total claim amount reported on Form TD1. If Form TD1 is not filed 
		by the employee or pensioner, TC is $8,929, and for non-resident individuals, TC is $0. If the claim code is E, T = $0.
		*	K2 = Canada Pension Plan contributions and Employment Insurance premiums federal tax credits for the year. 
		*	= [(0.155 * (P * C, maximum $1,989.90)) + (0.155 * (P * EI, maximum $720.00))]
		*	Where: 
		*	P = The number of pay periods in the year. 
		*	C = Canada (or Quebec) Pension Plan contribution for the pay period. See the 
		*	EI = Employment Insurance premium for the pay period.
		*	K3 = Other federal tax credits, such as medical expenses and charitable 
		donations requested by an employee or pensioner and authorized by a tax 
		services office or tax centre. The employer or payer will be informed by a 
		tax services office or tax centre of the amount to be used as K3, when 
		applicable. 
		*	K4 = Canada Employment Credit = The lesser of: 
		*		(i) 0.155 multiplied by A; and 
		*		(ii) 0.155 multiplied by $1,000 
		*
		*/
		protected static function get_basic_federal_tax(&$line, &$staff, $A, $P, $value) {
			$tax = self::get_tax_rate(FEDERAL, $line, $staff, $A);
			$R = $tax['rate'];
			$K = $tax['constant'];

			$tax = self::get_tax_rate(FEDERAL, $line, $staff, 0);
			$min_rate = $tax['rate'];

			$TC = self::get_total_claim_amount($line, $staff);

			$K1 = bcmul($min_rate, $TC);

			$C  = CPP::calculate($line, array('types' => $value['gross_renumeration']));
			$EI = EI::calculate($line, array('types' => $value['gross_renumeration']));
			$K2 = bcadd( bcmul($min_rate, min( bcmul($P, $C), 1989.9)), bcmul($min_rate, min(bcmul($P, $EI), 720)));

			$K3 = $value['other_tax_credits'];
			$K4 = min( bcmul($min_rate, $A), bcmul($min_rate, 1000));

			// T3 =  ($R * $A) - $K - $K1 - $K2 - $K3 - $K4);
			$T3 = bcsub( bcsub( bcsub( bcsub( bcsub( bcmul($R, $A), $K), $K1), $K2), $K3), $K4);
//echo "T3 =  ($R * $A) - $K - $K1 - $K2 - $K3 - $K4) = $T3<br />";
			return max(0, $T3);
		}

		/**
		*	T1 = Annual federal tax deduction, except for employees in Quebec, outside Canada, and in Canada beyond the limits of any province
		*	= (T3 - LCF) or 0 if the result is negative
		*
		*	T3 = Annual basic federal tax.
		*	LCF = Labour-sponsored funds federal tax credit for the year. = The lesser of: 
		*		(i) $750; and 
		*		(ii) 15% of the amount deducted or withheld during the year for the acquisition, by the employee, of approved shares of the capital stock of a 
		*			prescribed labour-sponsored venture capital corporation. 
		*	@todo calculate LCF properly
		*/
		protected static function get_annual_federal_payable(&$line, $T3) {
			$LCF = min(750, bcmul(0.15, 0));
//echo "T1 = min(750, 0.15 * 0) = " . max(0, bcsub($T3, $LCF)) . '<br />';
			return max(0, bcsub($T3, $LCF));
		}


		/**
		*	Formula to calculate annual basic provincial or territorial tax (T4) 
		*
		*	T4 = Annual basic provincial or territorial tax = (V * A) - KP - K1P - K2P - K3P - K4P 
		*	Where:
		*	V = Provincial or territorial tax rate for the year (not applicable to Quebec, outside Canada, and in Canada beyond the limits of any province). 
		*	A = Annual taxable income. 
		*	KP = Provincial or territorial constant. 
		*	K1P = Provincial or territorial non-refundable personal tax credit. 
		*	K2P = Canada (or Quebec) Pension Plan contributions and Employment Insurance premiums provincial or territorial tax credits for the year. 
		*	K3P = Other provincial or territorial tax credits, such as medical expenses and charitable donations authorized by a tax services office or tax centre. 
		*	K4P = Provincial or territorial Canada Employment Credit (applies to Yukon only). 
		*/
		protected static function get_basic_regional_tax(&$line, &$staff, $A, $P, $value) {
			$tax = self::get_tax_rate(REGIONAL, $line, $staff, $A);
			$V = $tax['rate'];
			$KP = $tax['constant'];

			$tax = self::get_tax_rate(REGIONAL, $line, $staff, 0);
			$min_rate = $tax['rate'];  // Should be 0.1 for Alberta

			$K1P = bcmul($min_rate, 15435);

			$C  = CPP::calculate($line, array('types' => $value['gross_renumeration']));
			$EI = EI::calculate($line, array('types' => $value['gross_renumeration']));

			$K2P = bcadd( bcmul($min_rate, min( bcmul($P, $C), 1989.9)), bcmul($min_rate, min(bcmul($P, $EI), 720)));
			//T4 = (V * A) - KP - K1P - K2P - K3P - K4P 
			$T4 = bcsub( bcsub( bcsub( bcsub( bcsub( bcmul($V, $A), $KP), $K1P), $K2P), $K3P), $K4P); 
//echo "T4 = ($V * $A) - $KP - $K1P - $K2P - $K3P - $K4P = " .$T4 . '<Br />';
			return $T4;
		}

		/**
		*	Formulas to calculate annual provincial or territorial tax deduction (T2) 
		*
		*	T2 = Annual provincial or territorial tax deduction = T4 + V1 + V2 - S - LCP 
		*	If the result is negative, T2 = $0. 
		*	For employees in Quebec only: 	T2 = $0 
		*	For employees outside Canada and in Canada beyond the limits of any province only: T2 = $0 
		*
		*	Where:
		*	T4 = Annual basic provincial or territorial tax. 
		*	V1 = Provincial or territorial surtax calculated on the basic provincial or territorial tax (applicable to NL, NS, PE, ON, and YT only). 
		*	V2 = Additional tax calculated on taxable income (applies to Ontario Health Premium only) 
		*	S = Provincial tax reduction (applicable to ON, MB and BC only). 
		*	LCP = Provincial or territorial labour-sponsored funds tax credit (applicable to NL, NS, NB, ON, MB, SK, BC, and YT only). 
		*/
		 
		protected static function get_annual_regional_payable(&$staff, $T4) {
			$T2 = max(0, bcadd( $T4, bcadd( $V1, bcsub( $V2, bcsub( $S, $LCP)))));
//echo "T2 = $T4 + $V1 + $V2 - $S - $LCP = $T2<br />";
			return $T2;
		}

		/*
		*	TC = The total claim amount reported on Form TD1. If Form TD1 is not filed 
		by the employee or pensioner, TC is $8,929, and for non-resident individuals, TC is $0. If the claim code is E, T = $0.
		*/
		protected static function get_total_claim_amount(&$line, &$staff) {
			return 8929;
		}


		/**
		*	Get federal tax rates
		*/
		protected static function get_tax_rate($type, &$line, &$staff, $income) {
			$pay_period = new PayPeriod($line->get_paystub()->get_pay_period_id());

			$where = array(
				'effective_date' => $pay_period->start_date,
				'income' => $income,
			);
			$address = ExecMethod('addressbook.addressbook_bo.read', $staff->get_addressbook_id());
			if(!$address) {
				throw new Exception(lang('Could not find address for staff_id %1', $staff->get_staff_id()));
			}
			$address_link = $GLOBALS['egw']->link('/index.php?menuaction=addressbook.addressbook_ui.edit&contact_id=' . $staff->get_addressbook_id());
			switch ($type) {
				case LOCAL:
					if(!$address['adr_two_locality']) {
						$GLOBALS['egw']->log->error(array(
							'text'	=>	'E-Paystub, %1 %2.  (<a href=%3 target=\'_blank\'>Fix</a>)',
							'p_1'	=>	'Could not find locality for',
							'p_2'	=>	$address['n_fn'],
							'p_3'	=>	$address_link,
							'file'	=>	$e->getFile(),
							'line'	=>	$e->getLine()
						));
					}
					$where['locality'] = $address['adr_two_locality'];
					// fall through
				case REGIONAL:
					if(!$address['adr_two_region']) {
						$GLOBALS['egw']->log->error(array(
							'text'	=>	'E-Paystub, %1 %2.  (<a href=%3 target=\'_blank\'>Fix</a>)',
							'p_1'	=>	'Could not find region for',
							'p_2'	=>	$address['n_fn'],
							'p_3'	=>	$address_link,
							'file'	=>	__FILE__,
							'line'	=>	__LINE__
						));
					}
					$where['region'] = $address['adr_two_region'];
					// fall through
				case FEDERAL:
					if(!$address['adr_two_countryname']) {
						$GLOBALS['egw']->log->error(array(
							'text'	=>	'E-Paystub, %1 %2.  (<a href=%3 target=\'_blank\'>Fix</a>)',
							'p_1'	=>	'Could not find country for',
							'p_2'	=>	$address['n_fn'],
							'p_3'	=>	$address_link,
							'file'	=>	__FILE__,
							'line'	=>	__LINE__
						));
					}
					$where['country'] = 'CA';
					break;					
			}
			return bo_perp_rate::get_rate($where);
		}
	
	}
?>
