<?php


	/**
	*	Just a constant amount.  The value is what amount to return.
	*
	*	Sounds easy, right?  Unfortunately, once we add the question of 
	*	over what time period, things get a little complicated.  The easy case is a 
	*	constant amount on each pay period.  This is useful for regular deductions such as
	*	$2 each pay period for union dues.
	*
	*	It gets a little more complicated if we want a constant $40 a month for parking, and
	*	the pay period is bi-weekly.  For 10 months out of the year, each paycheque will have
	*	a $20 deduction, because there's 2 pay periods a month.  The other 2 months, there are
	*	three paycheques in the month.  To handle this, we check the total per month against 
	*	the amount paid so far in the month.  On the last pay period in the month, the amount
	*	deducted is adjusted to meet the requirments, in this case it is set to $0.
	*
	*	In our parking example, if you want to park, you pay the monthly fee.  This is a poor
	*	deal if you start near the end of the month, because you still have to pay $40.  The
	*	last pay period in the month will start with $20, check the requirements and then 
	*	increase your deduction to the needed $40.
	*
	*	Other calculation periods work similarly, and this calculation is also used for
	*	salaries.  
	*
	*	Pro-rating will only calculate over days in the pay period, so you only pay parking
	*	for the pay period, about $18.67 for a 2 week pay period in a 30 day month, 
	*	instead of all days in the month.
	*	
	*/

	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.Calculation.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.PayrollUtility.inc.php');

	class Constant extends Calculation {

		private static $debug = false;

		public static function getName() {
			return 'Constant amount';
		}

		/**
		*	Calculate the constant amount
		*	Value is an array with keys amount, per and pro_rate.
		*	Amount is the constant amount, per is the time period, one of:
		*	- period - Amount is per pay period
		*	- day - Amount is per day in the pay period
		*	- week - Amount is per week in the pay period
		*	- month - Amount is per month in the pay period
		*	- year - Amount is per year in the pay period
		*	Pro rate uses only the days worked in the pay period, and pro-rates the amount by the number of days.
		*
		*	Some examples for a bi-weekly pay schedule:
		*	- 100 per period = every paystub has a $100 deduction, $2600 per year
		*	- 100 per day = every paystub has a $1400 deduction, ~$36500 per year
		*	- 100 per month = Most paystubs have a $50 deduction, but two months of the year, there is a paystub where this deduction is $0
		*	- 100 per year = Most paystubs have a $3.85 deduction.
		*/
		public static function calculate(PaystubLine &$line, $value) {
			if(!is_array($value)) {
				return $value;
			}
			$pay_period =& new PayPeriod($line->get_paystub()->get_pay_period_id());
			$pay_period_info = PayrollUtility::get_period_info($pay_period->start_date, $pay_period->end_date);

			$pay_per = $value['per'];

			// Salary style, where someone is paid 5000 / month, but paid biweekly
			if($value['pay_unit'] && $value['pay_rate']) {
				$pay_per = $value['pay_unit'];
			}
			switch ($pay_per) {
				case 'h':
				case 'hour':
					// Not allowed
					$GLOBALS['egw']->log->error(array(
						'text'  =>      'D-Paystub, %2 Hourly constant amount for %1 - use a Wage instead of Salary. ',
						'p_1'   =>      $address['n_fn'],
						'p_2'	=>	$line->get_type()->get_name(),
						'file'  =>      __FILE__,
						'line'  =>      __LINE__
					));
					return;
				case 'period':
					return $value['amount'];
					break;
				case 'day':
					$per_frequency = $pay_period_info['days_in_year'];
					$per_days = 1;
					$per_start = $pay_period->start_date;
					$per_end = strtotime("+$per_days days", $per_start);
					$check_last = (date('z', $per_start) != date('z', strtotime('+' . $pay_period_info['days'] . ' days', $pay_period->transaction_date) ));
					break;
				case 'week':
					$per_frequency = 52;
					$per_days = 7;

					$per_start = strtotime('this Monday', $pay_period->transaction_date);
					$per_end = strtotime("+$per_days days", $per_start);
					$check_last = (date('W', $per_start) != date('W', strtotime('+' . $pay_period_info['days'] . ' days', $pay_period->transaction_date) ));
					break;
				case 'm':
				case 'month':
					$per_frequency = 12;
					$per_days = date('t', $pay_period->transaction_date);
					$per_start = mktime(0,0,0, date('m', $pay_period->transaction_date), 1, date('Y', $pay_period->transaction_date));
					$per_end = mktime(0,0,0, date('m', $pay_period->transaction_date), date('t', $pay_period->transaction_date), date('Y', $pay_period->transaction_date));
					$check_last = (date('m', $per_start) != date('m', strtotime('+' . $pay_period_info['days'] . ' days', $pay_period->transaction_date) ));
					break;
				case 'y':
				case 'year':
					$per_frequency = 1;
					$per_days = $pay_period_info['days_in_year'];
					$per_start = mktime(0,0,0, 1, 1, date('Y', $pay_period->transaction_date));
					$per_end = mktime(0,0,0, 12, 31, date('Y', $pay_period->transaction_date));
					$check_last = (date('Y', $per_start) != date('Y', strtotime('+' . $pay_period_info['days'] . ' days', $pay_period->transaction_date) ));
					break;
				default:
					return 0;
			}

			if($value['pro_rate']) {
				$base_amount = bcmul($value['amount'], bcmul($per_frequency, bcdiv($pay_period_info['days'],  $pay_period_info['days_in_year'], 10), 2), 2);
				$check_last = false;
			} else if($pay_period_info['days'] > $per_days) {
				$base_amount = bcmul($value['amount'], ceil($pay_period_info['days'] / $per_days), 2);
			} else if ($pay_period_info['days'] <= $per_days) {
				$base_amount = bcdiv($value['amount'], round($per_days / $pay_period_info['days']), 2);
			}

			if(self::$debug) {
				echo 'Pay period: ' . date('Y-m-d', $pay_period->start_date) . ' -> ' . date('Y-m-d', $pay_period->end_date) . ' [' . date('Y-m-d', $pay_period->transaction_date) . ']<br />';
				echo "Per Start: " . date('Y-m-d', $per_start) ."<br />Per End: " . date('Y-m-d', $per_end) . "<br />";
				echo "Check Last: " . date('Y-m-d', $per_start) . ' != ' . date('Y-m-d', strtotime('+' . $pay_period_info['days'] . ' days', $pay_period->transaction_date)) . '<Br />';
				echo 'Constant: $' . $value['amount'] . ' / ' . $value['per'] . '<br />';
				echo "Base amount: $base_amount<br />";
			}

			// Now we need to check to make sure things actually line up at the end of the time period due to rounding errors
			if($check_last || !$value['pro_rate'] && ($value['pay_unit'] && $value['pay_unit'] != $value['per'])) {
				// How much should have been paid in the $value['per'] is $value['amount'];
				$actual = PayStubEntryType::get_ledger_balance($line->get_type_id(), $line->get_paystub()->get_staff_id(), $per_end, $per_start);
				if(self::$debug) {
					echo "Paid in constant period: $actual<br />";
				}

				if($pay_period_info['days'] > $per_days) {
					$final_amount = $base_amount - $actual;
				} else {
					$final_amount = $value['amount'] - $actual;
				}
				if($final_amount != $base_amount) {
					$GLOBALS['egw']->log->error(array(
						'text'  =>      'D-Paystub, Constant amount for %1 - %2 adjusted from %3 to %4. ',
						'p_1'   =>      $address['n_fn'],
						'p_2'	=>	$line->get_type()->get_name(),
						'p_3'	=>	$base_amount,
						'p_4'	=>	$final_amount,
						'file'  =>      __FILE__,
						'line'  =>      __LINE__
					));
				}
			} else {
				$final_amount = $base_amount;
			}
			if(self::$debug) {
				echo "Final amount: $final_amount<br />";
			}
			return $final_amount;
		}

		public static function getTemplate() {
			return 'perp_payroll.Calculation.Constant';
		}
		public static function getOverrideTemplate() {
			return 'perp_payroll.Calculation.ConstantOverride';
		}
		public static function getOptions() {
			return array(
				'per'	=> array(
					'period'	=>	lang('pay period'),
					'day'		=>	lang('day'),
					'week'		=>	lang('week'),
					'month'		=>	lang('month'),
					'year'		=>	lang('year'),
				)
			);
		}
	}
?>
