<?php

	/**
	*	A special type of PaystubLineType that builds a balance instead of adding to earnings.
	*
	*	Accruals accrue a balance at a given rate, to be paid out in the future.
	*/

	require_once('class.PaystubEntryType.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.Calculation.inc.php');

	class AccrualType extends PaystubEntryType {

		/**
		*	If the balance is over the max or under the min,
		*	the max or min is ignored, and it continues to accrue
		*/
		const NOTHING	=	0;

		/**
		*	If the balance is over the max or under the min, 
		*	no more can be added (or removed).  This is the normal
		*	behaviour for things like vacation time with a limit on how
		*	much you can save up.
		*/
		const CUT	=	1;

		/**
		*	The default rate at which the balance accumulates.
		*
		*	This is a percentage of the base types for this pay period.
		*/
		protected $rate = 0;

		/**
		*	Minimum value.  
		*	If the employee's balance of this accrual type falls below this minimum, something will happen
		*/
		protected $minimum = 0;

		/**
		*	Maximum value.
		*	If the employee's balance of this accrual type builds to above this maximum, something will happen
		*/
		protected $maximum = null;

		/**
		*	If the balance is below the minimum, this is what happens.
		*/
		protected $under_min = self::CUT;

		/**
		*	If the balance is above the maximum, this is what happens.
		*/
		protected $over_max = self::NOTHING;

		/**
		*	The pay stub entry types that this accrual is based on
		*/
		protected $base_types = '';

		/**
		*	The entry type that pulls money out of the balance for this type
		*/
		protected $draw_type = null;
		
		/**
		*	Get the accrual amount on the given value
		*
		*	@param line PaystubLine to calculate accrual on
		*	@param staff Staff of the employee to calculate accrual for
		*	@return accrued amount
		*/
		public function get_amount(PaystubLine &$line, Staff &$staff) {
			$base = Calculation::total_lines($line, $this->base_types);

			$rate = $this->get_rate($staff);

			$accrual = bcmul($rate, $base, 2);
			if(!is_null($this->maximum) && $this->over_max != self::NOTHING) {
				$accrual = min( $accrual + $balance, $this->maximum);
			}
			return $accrual;
		}

		/**
		*	Set the entry types used as the base of the accrual
		*
		*	@param types
		*/	
		public function set_base_types($types) {
			// Store them as CSV
			if(is_array($types)) {
				$types = implode(',', $types);
			}
			$this->types = $types;
		}

		/**
		*	Set the accrual rate
		*
		*	@param new accrual rate
		*/
		public function set_rate($rate) {
			if($rate != 0) {
				$this->rate = $rate;
			}
		}

		/**
		*	Get the accrual rate
		*
		*	If a Staff object is provided, the rate will be calculated based on 
		*	the breakpoints and that Staff person.  Otherwise, the base rate
		*	will be returned.
		*/
		public function get_rate(Staff &$staff = null) {
			if(is_null($staff) || !$staff->get_start_date()) {
				return $this->rate;
			}

			$rate = $this->rate;
			$breakpoints = $this->get_breakpoints();
			
			foreach($breakpoints as $breakpoint) {
				$breaktime = strtotime('+' . $breakpoint['service_length'] . ' ' . Paystub::$pay_unit[$breakpoint['service_units']], $staff->get_start_date());
				if($breaktime !== false && $breaktime <= time()) {
					$rate = $breakpoint['rate'];
				} else {
					break;
				}
			}
			return $rate;
		}

		/**
		*	Set the draw type.
		*
		*	The draw type is the line type that makes the balance go down.
		*/
		public function set_draw_type($type_id) {
			$this->draw_type = (int) $type_id;
		}

		/**
		*	Set the minimum limit
		*/
		public function set_minimum($minimum) {
			$this->minimum = (float)$minimum;
		}

		/**
		*	Set the maximum limit
		*/
		public function set_maximum($maximum) {
			$this->maximum = (float)$maximum;
		}

		/**
		*	Set the accrual breakpoints.
		*
		*	Breakpoints allow you to set a rate for a period of time,
		*	after which the rate changes to the next rate.  For example,
		*	in Canada, employees accrue 4% vacation pay (2 weeks) for the 
		*	first two years they work for their employer.  After two years, they
		*	accrue at 6% (3 weeks), and after five years, 8%.
		*
		*	The rate as set with set_rate() is in effect until the first breakpoint.
		*
		*	@param breakpoints An array with one array for each breakpoint.
		*		The breakpoint array has keys of service length,
		*		service_units (h,m,y) and rate.
		*/
		public function set_breakpoints(Array $breakpoints) {
			$GLOBALS['egw']->db->delete(
				'perp_accrual_breakpoint',
				array(
					'type_id'	=>	$this->get_id(),
				),
				__LINE__, __FILE__
			);
			foreach($breakpoints as $breakpoint) {
				$breakpoint['type_id'] = $this->get_id();
				$GLOBALS['egw']->db->insert(
					'perp_accrual_breakpoint',
					$breakpoint,
					array(),
					__LINE__, __FILE__, 'perp_payroll'
				);
			}
		}

		/**
		*	Get the accural breakpoints.
		*
		*/
		public function get_breakpoints() {
			$breakpoints = array();
			$result =& $GLOBALS['egw']->db->select(
				'perp_accrual_breakpoint',
				array('*'),
				array(
					'type_id'	=>	$this->get_id()
				),
				__LINE__, __FILE__,false, 
				'ORDER BY rate', 'perp_payroll'
			);
			foreach($result as $row) {
				$breakpoints[] = $row;
			}
			return $breakpoints;
		}
	
		public function get_value($staff_id) {
			$GLOBALS['egw']->db->select(
				'perp_staff_accrual',
				array(
					'balance',
					'accrues'
				),
				array(
					'type_id'	=>	$this->id,
					'staff_id'	=>	$staff_id
				),
				__LINE__, __FILE__, 0,
				'', 'perp_payroll'
			);
			$staff = $GLOBALS['egw']->db->row(true);
			if(is_array($staff) ) {
				return $staff;
			} else {
				return array(
					'balance'	=>	0,
					'accrues'	=>	false
				);
			}
		}

		public function get_balance($staff_id) {
			$staff = $this->get_value($staff_id);
			return $staff['balance'];
		}

		/**
		*	Check a draw against balance, and correct the line
		*	if needed.
		*
		*	@param line PaystubLine for the pay out.
		*/
		public function draw_check(PaystubLine &$line) {
			$balance = $this->get_balance($line->get_paystub()->get_staff_id());
			// Check balance vs draw
			$requested = $line->get_amount();
			if(!is_null($this->maximum) && $balance - $line->get_amount() >= $this->maximum) {
				switch ($this->over_max) {
					case self::NOTHING:
						break;
					case self::CUT:
						$line->set_quantity(bcdiv($this->maximum, $line->get_rate()));
						break;
				}
			} 
			if ($balance - $line->get_amount() <= $this->minimum) {
				switch ($this->under_min) {
					case self::NOTHING:
						break;
					case self::CUT:
						$line->set_rate($balance);
						$line->set_quantity(1);
						break;
				}
			}

			if($requested != $line->get_amount()) {
				$line->add_note(null, lang('The requested amount of %1 would exceed the limits, a value of %2 was used instead.', 
					ExecMethod2('perp_api.bo_perp_api.get_formatted_price', $line->get_paystub()->get_currency(), $requested), 
					ExecMethod2('perp_api.bo_perp_api.get_formatted_price', $line->get_paystub()->get_currency(), $line->get_amount())
				));
				$staff = new Staff($line->get_paystub()->get_staff_id());
				$address = ExecMethod('addressbook.addressbook_bo.read', $staff->get_addressbook_id());
				$GLOBALS['egw']->log->error(array(
					'text'	=>	'W-Paystub, %2 draw tried to exceed limits for %1.',
					'p_1'	=>	$address['n_fn'],
					'p_2'	=>	$this->get_name(),
					'file'	=>	__FILE__,
					'line'	=>	__LINE__
				));
			}
		}

		/**
		*	Override add_YTD to update the balance
		*/
		public static function add_YTD($type_id, $pay_period_id, $staff_id, $quantity, $amount) {
			parent::add_YTD($type_id, $pay_period_id, $staff_id, $quantity, $amount);

			// Add to balance
			self::add_balance($type_id, $staff_id, $amount);
		}

		public static function add_balance($type_id, $staff_id, $amount) {
			$GLOBALS['egw']->db->query(
				"UPDATE perp_staff_accural SET balance = balance + " . (float)$amount . 
				'WHERE staff_id = ' . (int)$staff_id . ' AND
					type_id = ' . (int)$type_id,
				__LINE__, __FILE__
			);
		}

		/**
		*	Set the staff members that this AccrualType applies to
		*
		*	@param data array of data for staff people to change, 
		*	each array having keys staff_id, balance, accrues
		*/
		public function set_staff(Array $data) {
			foreach($data as $staff_data) {
				if(!$staff_data['staff_id']) {
					continue;
				}
				// If accrues is not set, default to true
				if(!array_key_exists('accrues', $staff_data)) {
					$staff_data['accrues'] = true;
				}
				if(isset($staff_data['value'])) {
					$staff_data += $staff_data['value'];
				}
				$GLOBALS['egw']->db->insert(
					'perp_staff_accrual',
					array(
						'staff_id'	=>	$staff_data['staff_id'],
						'type_id'	=>	$this->id,
						'balance'	=>	$staff_data['balance'],
						'accrues'	=>	$staff_data['accrues']
					),
					array(
						'staff_id'	=>	$staff_data['staff_id'],
						'type_id'	=>	$this->id
					),
					__LINE__, __FILE__, 'perp_payroll'
				);
			}
		}

		/**
		*	Gets the staff members that this AccrualType applies to
		*	@return an array of data for each staff person, having the keys
		*	staff_id, type_id, balance, accrues
		*	@return array
		*/
		public function get_staff() {
			$GLOBALS['egw']->db->select(
				'perp_staff_accrual',
				array('*'),
				array('type_id' => $this->id),
				__LINE__, __FILE__, 0,
				'ORDER BY n_family, n_given', 'perp_payroll', $query['num_rows'],
				'JOIN perp_staff ON
					perp_staff_accrual.staff_id = perp_staff.staff_id
				JOIN egw_addressbook ON
					perp_staff.addressbook_id = egw_addressbook.contact_id'
			);
			$staff = array();
			$staff_object = null;
			while($row = $GLOBALS['egw']->db->row(true)) {
				$staff[] = $row;
			}

			// Separate loop because new Staff() will kill our $GLOBALS['egw']->db results,
			// giving us only the first result.
			foreach($staff as $key => &$staff_row) {
				$staff_object =& new Staff($staff_row['staff_id']);
				$staff_row['rate'] = $this->get_rate($staff_object);
				$staff_row['formatted_rate'] = $staff_row['rate'] * 100;
			}
			return $staff;
		}


		public function remove_staff($staff) {
			if($staff instanceof Staff) {
				$staff_id = $staff->get_staff_id();
			} else {
				$staff_id = $staff;
			}
			$balance = $this->get_balance($staff_id);
			if($balance == 0) {
				$GLOBALS['egw']->db->delete(
					'perp_staff_accrual',
					array(
						'staff_id'	=>	$staff_id,
						'type_id'	=>	$this->id,
					),
					__LINE__, __FILE__, 'perp_payroll'
				);
			} else {
				$GLOBALS['egw']->db->update(
					'perp_staff_accrual',
					array(
						'accrues'	=>	false
					),
					array(
						'staff_id'	=>	$staff_id,
						'type_id'	=>	$this->id,
					),
					__LINE__, __FILE__, 'perp_payroll'
				);
			}
		}

		public function read() {
			parent::read();

			$GLOBALS['egw']->db->select(
				'perp_accrual_type',
				array('*'),
				array('type_id'	=>	$this->id),
				__LINE__, __FILE__
			);
			$row = $GLOBALS['egw']->db->row(true);
			$this->id		=	$row['type_id'];
			$this->rate		=	$row['rate'];
			$this->minimum		=	$row['minimum'];
			$this->maximum		=	$row['maximum'];
			$this->under_min	=	$row['under_min'];
			$this->over_max		=	$row['over_max'];
			$this->base_types 	=	$row['base_types'];
			$this->draw_type	=	$row['draw_type'];
		}

		public function save() {
			parent::save();

			$GLOBALS['egw']->db->insert(
				'perp_accrual_type',
				array(
					'rate'		=>	$this->rate,
					'minimum'	=>	$this->minimum,
					'maximum'	=>	$this->maximum,
					'under_min'	=>	$this->under_min,
					'over_max'	=>	$this->over_max,
					'base_types'	=>	$this->types,
					'draw_type'	=>	$this->draw_type,
				),
				array(
					'type_id'	=>	$this->id
				),
				__LINE__, __FILE__, 'perp_payroll'
			);
		}

		public function delete() {
			if(parent::delete()) {
				$GLOBALS['egw']->db->delete(
					'perp_accrual_type',
					array('type_id'	=>	$this->id),
					__LINE__, __FILE__, 'perp_payroll'
				);

				return true;
			} else {
				return false;
			}
		}

		public function toArray() {
			return parent::toArray() + array(
				'rate'		=>	$this->rate,
				'minimum'	=>	$this->minimum,
				'maximum'	=>	$this->maximum,
				'under_min'	=>	$this->under_min,
				'over_max'	=>	$this->over_max,
				'base_types'	=>	$this->base_types,
				'draw_type'	=>	$this->draw_type,
				'breakpoints'	=>	$this->get_breakpoints()
			);
		}
	}
?>
