<?php

	/**
	*	DeductionType is a type of PayStubEntryType for removing amounts from a PayStub.
	*
	*	DeductionTypes are used for things like taxes, payroll deductions and employer contributions.
	*
	*	@package perp_payroll
	*	@author Nathan Gray
	*/

	require_once('class.PaystubEntryType.inc.php');

	class DeductionType extends PaystubEntryType {

		/**
		*	Constants that affect behaviour and groupings
		*	@{
		*/

		/**
		*	This line has something to do with taxes.  It probably has some complex calculation,
		*	and changes every year according to the government.
		*	The amount is deducted from the employee's pay.
		*/
		const TAX	=	'T';

		/**
		*	This line is contributed by the employer, such as with benefits.
		*	@note It doesn't actually remove money from the Paystub
		*/
		const EMPLOYER	=	'C';

		/**
		*	This line is a deduction from the Paystub that is not a tax, such
		*	as union dues, employee purchases or manditory contributions.
		*	The amount is deducted from the employee's pay.
		*/
		const EMPLOYEE	=	'D';

		/**
		*	@}
		*/

		public static $type_array = array(
			self::TAX	=>	'tax',
			self::EMPLOYER	=>	'employer contribution',
			self::EMPLOYEE	=>	'employee deduction'
		);

		/**
		*	Type of deduction (tax, employer contribution or employee deduction).
		*	Default is EMPLOYEE.
		*/
		protected $type = self::EMPLOYEE;

		public function __construct($id) {
			parent::__construct($id);
		}

		/**
		*	Get the deduction type
		*	@return string
		*/
		public function get_type() {
			return $this->type;
		}

		/**
		*	Set the deduction type
		*	Must be one of the defined type constants
		*	@param type String
		*/
		public function set_type($type) {
			if($type == self::TAX || $type == self::EMPLOYER || $type == self::EMPLOYEE) {
				$this->type = $type;
			} else {
				throw new Exception(lang('Invalid parameter.  ') .  lang('Type must be one of the defined types'));
			}
		}

		/**
		*	Set the staff members that this DeductionType applies to
		*
		*	@param data array of data for staff people to change, 
		*	each array having keys staff_id, value
		*/
		public function set_staff(Array $data) {
			foreach($data as $staff_data) {
				if(!$staff_data['staff_id'] || $staff_data['type_id'] != $this->id) {
					continue;
				}

				// Check for empty values, set to null
				if(trim($staff_data['value']) == '') {
					$staff_data['value'] = null;
				} elseif (is_array($staff_data['value'])) {
					$is_null = true;
					foreach($staff_data['value'] as $key => $value) {
						$is_null = ($is_null && (is_null($value) || trim($value) == ''));
					}
					if($is_null) {
						$staff_data['value'] = null;
					} else {
						$staff_data['value'] = serialize($staff_data['value']);
					}
				} else {
					$staff_data['value'] = serialize($staff_data['value']);
				}
				$GLOBALS['egw']->db->insert(
					'perp_staff_entry_type',
					array(
						'staff_id'	=>	$staff_data['staff_id'],
						'type_id'	=>	$this->id,
						'value'		=>	$staff_data['value']
					),
					array(
						'staff_id'	=>	$staff_data['staff_id'],
						'type_id'	=>	$this->id
					),
					__LINE__, __FILE__, 'perp_payroll'
				);
			}
		}

		/**
		*	Gets the staff members that this DeductionType applies to
		*	@return an array of data for each staff person, having the keys
		*	staff_id, type_id, value
		*	@return array
		*/
		public function get_staff() {
			$GLOBALS['egw']->db->select(
				'perp_staff_entry_type',
				array('*'),
				array('perp_staff_entry_type.type_id = ' . (int)$this->id),
				__LINE__, __FILE__, 0,
				'ORDER BY n_family, n_given', 'perp_payroll', -1,
				' JOIN perp_pay_stub_entry_type ON
					perp_pay_stub_entry_type.type_id = perp_staff_entry_type.type_id
				JOIN perp_staff ON
					perp_staff_entry_type.staff_id = perp_staff.staff_id
				JOIN egw_addressbook ON
					perp_staff.addressbook_id = egw_addressbook.contact_id'
			);
			if(!$this->calculation) {
				return array();
			}
			$staff = array();
			$calc = CalculationFactory::getCalculation($this->calculation);
			$staff['override_template'] = $calc->getOverrideTemplate();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$row['default_value'] = unserialize($row['default_value']);
				$row['value'] = unserialize($row['value']);
				if(is_array($row['default_value']) && is_array($row['value'])) {
					$row['value'] = array_merge($row['default_value'], $row['value']);
				}
				$staff[] = $row;
			}
			return $staff;
		}

		public function remove_staff($staff) {
			if($staff instanceof Staff) {
				$staff_id = $staff->get_staff_id();
			} else {
				$staff_id = $staff;
			}
			$GLOBALS['egw']->db->delete(
				'perp_staff_entry_type',
				array(
					'staff_id'	=>	$staff_id,
					'type_id'	=>	$this->id
				),
				__LINE__, __FILE__, 'perp_payroll'
			);
		}

		/**
		*	Read this Deduction
		*/
		public function read() {
			parent::read();

			$GLOBALS['egw']->db->select(
				'perp_deduction_type',
				array(
					'deduction_type',
				),
				array('type_id'	=>	$this->id),
				__LINE__, __FILE__
			);
			if($row = $GLOBALS['egw']->db->row(true)) {
				$this->set_type($row['deduction_type']);
				return true;
			} else {
				return false;
			}
		}

		public function save() {
			parent::save();

			$GLOBALS['egw']->db->insert(
				'perp_deduction_type',
				array(
					'type_id'		=>	$this->get_id(),
					'deduction_type'	=>	$this->type,
				),
				array(
					'type_id'	=>	$this->get_id()
				),
				__LINE__,__FILE__, 'perp_payroll'
			);
		}

		public function delete() {
			if(parent::delete()) {
				$GLOBALS['egw']->db->delete(
					'perp_deduction_type',
					array('type_id'	=>	$this->id),
					__LINE__, __FILE__, 'perp_payroll'
				);

				return true;
			} else {
				return false;
			}
		}

		public function toArray() {
			return parent::toArray() + array(
				'type'			=>	$this->type,
			);
		}
	}
?>
