<?php

	/**
	*	A special type of PaystubLineType that builds a balance instead of adding to earnings.
	*
	*	Accruals accrue a balance at a given rate, to be paid out in the future.
	*/

	require_once('class.PaystubEntryType.inc.php');

	class EarningType extends PaystubEntryType {

		protected $rate = 0;
		protected $pay_type = 'W';

		
		/**
		*	Types of payment.
		*	Wage is payment for time worked, on an hourly basis.
		*	Salary is payment regardless of time worked.
		*	Accrued time payout reduces an accrual balance.
		*	One-time is a one time payout.
		*/
		public static $pay_type_array = array(
			'W'	=>	'Wage',
			'S'	=>	'Salary',
			'A'	=>	'Accrued time payout',
			'O'	=>	'One time payout',
		);

		/**
		*	Calculations used to calculate the different earning types.
		*
		*	The pay_type_array is mostly for human reference, it's these calculations that are used to 
		*	figure out the actual values for the pay stub.
		*/
		public static $pay_type_calculation = array(
			'W'	=>	'TimeSheet',
			'S'	=>	'Constant',
			'A'	=>	'TimeSheet',
			'O'	=>	'Constant',
		);

		/**
		*	Set the associated timesheet category for this pay type.
		*	Time logged on the timesheet under this category will
		*	be paid on the paystub with this type.
		*
		*	@param cat_id The ID of the timesheet category
		*	@todo parameter checking
		*/
		public function set_timesheet($cat_id) {
			$this->default_value['timesheet_category_id'] = $cat_id;
		}

		/**
		*	Set the rate multiplier for this earning type.
		*	This is used for things like overtime (1.5).  For regular
		*	time, it should be 1.0.
		*
		*	@param rate base pay rate multiplier
		*/
		public function set_rate($rate) {
			$this->rate = $rate;
			$this->default_value['rate_multiplier'] = $rate;
		}

		/**
		*	Override get_calculation_name() from parent, to return calculation base
		*	on earning type
		*/
		public function get_calculation_name() {
			if($this->calculation) {
				return $this->calculation;
			} else {
				return self::$pay_type_calculation[$this->get_pay_type()];
			}
		}

		/**
		*	Get the pay type
		*/
		public function get_pay_type() {
			return $this->pay_type;
		}

		/**
		*	Set the pay type, either wage or salary.  
		*	Wage uses the timesheet, salary is time independant.
		*	@param type { W | S | A | O }
		*/
		public function set_pay_type($type) {
			if(!in_array( $type, array_keys(self::$pay_type_array) ) ) {
				throw new Exception(lang('Invalid parameter.  ') . ' ' . 
					lang('Pay type must be one of [ %1 ]', implode(', ', self::$pay_type_array)));
			}
			$this->pay_type = $type;
		}

		/**
		*	Set the staff members that this EarningType applies to
		*
		*	@param data array of data for staff people to change, 
		*	each array having keys staff_id, pay_rate
		*/
		public function set_staff(Array $data) {
			foreach($data as $staff_data) {
				$this->_set_staff($staff_data);
			}
		}

		/**
		*	Gets the staff members that this EarningType applies to
		*	@return an array of data for each staff person, having the keys
		*	staff_id, type_id, pay_rate, pay_unit, effective_date
		*	@return array
		*/
		public function get_staff() {
			$GLOBALS['egw']->db->select(
				'perp_staff_entry_type',
				array(
					'perp_staff_entry_type.*',
					'perp_staff_pay_detail.*',
					'IF(perp_staff_pay_detail.staff_id IS NOT NULL, perp_staff_pay_detail.staff_id, perp_staff_entry_type.staff_id) AS staff_id',
					'IF(perp_staff_pay_detail.type_id IS NOT NULL, perp_staff_pay_detail.type_id, perp_staff_entry_type.type_id) AS type_id',
					'n_family',
					'n_given',
				),
				array('perp_staff_entry_type.type_id = ' . (int)$this->id),
				__LINE__, __FILE__, 0,
				'GROUP BY perp_staff_entry_type.staff_id, perp_staff_entry_type.type_id
				ORDER BY n_family, n_given, effective_date DESC ',
				'perp_payroll', -1, 
				'LEFT JOIN perp_staff_pay_detail ON
					perp_staff_entry_type.type_id = perp_staff_pay_detail.type_id
					AND perp_staff_entry_type.staff_id = perp_staff_pay_detail.staff_id
				LEFT JOIN perp_pay_stub_entry_type ON perp_staff_pay_detail.type_id = perp_pay_stub_entry_type.type_id
				JOIN perp_staff ON
					perp_staff_entry_type.staff_id = perp_staff.staff_id
				JOIN egw_addressbook ON
					perp_staff.addressbook_id = egw_addressbook.contact_id'
			);
			$staff = array();
			$calc = CalculationFactory::getCalculation($this->get_calculation_name());
			if($calc instanceof Calculation) {
				$staff['override_template'] = $calc->getOverrideTemplate();
			}
			while($row = $GLOBALS['egw']->db->row(true)) {
				$row['value'] = unserialize($row['value']);
				$row['default_value'] = unserialize($row['default_value']);
				if(is_array($row['value']) && is_array($row['default_value'])) {
					$row['value'] = array_merge($row['default_value'], $row['value']);
				}
				$staff[] = $row;
			}
			return $staff;
		}

		public function remove_staff($staff) {
			// Earnings are either time-based (timesheet) or constant, with settings elsewhere
			
			if($staff instanceof Staff) {
				$staff_id = $staff->get_staff_id();
			} else {
				$staff_id = $staff;
			}
			$GLOBALS['egw']->db->delete(
				'perp_staff_entry_type',
				array(
					'staff_id'	=>	$staff_id,
					'type_id'	=>	$this->id
				),
				__LINE__, __FILE__, 'perp_payroll'
			);
		}

		/**
		*	Set the earning info for one staff person
		*
		*	@todo backpay if the effective date is in the past
		*	@param staff_data and array of data for the staff person, having the keys
		*		staff_id, pay_rate, [effective_date, type]
		*/
		private function _set_staff(Array $staff_data) {
			if(!$staff_data['staff_id']) {
				return;
			}

			// Get current pay
			$GLOBALS['egw']->db->select(
				'perp_staff_pay_detail',
				array('*'),
				array(
					'staff_id'	=>	(int)$staff_data['staff_id'],
					'type_id'	=>	$this->id
				),
				__LINE__, __FILE__, 0,
				'ORDER BY effective_date DESC', 'perp_payroll',
				1
			);
			$pay = $GLOBALS['egw']->db->row(true);
			// If no effective date, it's today
			if(!$staff_data['effective_date']) {
				$staff_data['effective_date'] = time();
			}
			// If the rate has changed, create a record
			if($staff_data['pay_rate'] != $pay['pay_rate'] || $staff_data['type_id'] != $pay['type_id']) {
				// Set data
				$staff_data['effective_date'] = $GLOBALS['egw']->db->to_timestamp($staff_data['effective_date']);
				$GLOBALS['egw']->db->insert(
					'perp_staff_pay_detail',
					$staff_data,
					array(
						'staff_id'	=>	$staff_data['staff_id']
					),
					__LINE__, __FILE__, 'perp_payroll'
				);

			// If it's backdated, we need to do backpay
			}

			// Update the value
			if(trim($staff_data['value']) == '') {
				$staff_data['value'] = null;
			} elseif (is_array($staff_data['value'])) {
				$is_null = true;
				foreach($staff_data['value'] as $key => $value) {
					$is_null = ($is_null && (is_null($value) || trim($value) == ''));
				}
				if($is_null) {
					$staff_data['value'] = null;
				} else {
					$staff_data['value'] = serialize($staff_data['value']);
				}
			} else {
				$staff_data['value'] = serialize($staff_data['value']);
			}
			$GLOBALS['egw']->db->insert(
				'perp_staff_entry_type',
				array(
					'staff_id'      =>      $staff_data['staff_id'],
					'type_id'       =>      $this->id,
					'value'         =>      $staff_data['value']
				),
				array(
					'staff_id'      =>      $staff_data['staff_id'],
					'type_id'       =>      $this->id
				),
				__LINE__, __FILE__, 'perp_payroll'
			);
		}

		/**
		*	Get the value for a given staff.
		*	
		*	Merges parent info with pay info
		*
		*	@param staff_id
		*	@return mixed, but probably an array
		*/
		public function get_value($staff_id) {
			$parent = parent::get_value($staff_id);

			if($this->pay_type == 'S' && $value['pay_rate']) {
				// The calculation for salary uses 'amount', rather than 'pay_rate'
				$value['amount'] = $value['pay_rate'];

				// As nice as it would be to be paid the same if you start in January or December, 
				// Salary is pro-rated
				$value['pro_rate'] = true;
			}
			
			return is_array($value) ? array_merge($parent, $value) : $parent;
		}

		/**
		*	Get the type_id of the associated accrual entry if this is
		*	an accrual payout entry type.
		*
		*	@return array of type_ids or false if this is not an accrual payout type
		*/
		public function get_accrual_type() {
			if($this->pay_type != 'A') {
				return false;
			}
			$GLOBALS['egw']->db->select(
				'perp_accrual_type',
				array('type_id'),
				array(
					'draw_type'	=>	$this->id
				),
				__LINE__, __FILE__, false,
				'', 'perp_payroll'
			);
			$list = array();
			if($row = $GLOBALS['egw']->db->row(true)) {
				return $row['type_id'];
			} else {
				return false;
			}
		}

		public function read() {
			parent::read();

			$GLOBALS['egw']->db->select(
				'perp_earning_type',
				array('*'),
				array('type_id'	=>	$this->id),
				__LINE__, __FILE__, false,
				'', 'perp_payroll'
			);
			$row = $GLOBALS['egw']->db->row(true);
			$this->default_value['timesheet_category_id'] = $row['timesheet_category_id'];
			$this->rate = $row['rate_multiplier'];
			$this->pay_type = $row['pay_type'];
		}

		public function save() {
			parent::save();

			$GLOBALS['egw']->db->insert(
				'perp_earning_type',
				array(
					'rate_multiplier'	=>	$this->default_value['rate_multiplier'],
					'timesheet_category_id'	=>	$this->default_value['timesheet_category_id'],
					'pay_type'		=>	$this->pay_type
				),
				array(
					'type_id'	=>	$this->id
				),
				__LINE__, __FILE__, 'perp_payroll'
			);
		}

		public function delete() {
			if(parent::delete()) {
				$GLOBALS['egw']->db->delete(
					'perp_earning_type',
					array('type_id'	=>	$this->id),
					__LINE__, __FILE__, 'perp_payroll'
				);

				return true;
			} else {
				return false;
			}
		}

		/**
		*	Override add_YTD to decrement the balance for accrual draw types
		*/
		public static function add_YTD($type_id, $pay_period_id, $staff_id, $quantity, $amount) {
			parent::add_YTD($type_id, $pay_period_id, $staff_id, $quantity, $amount);

			$type = PayStubEntryType::factory($type_id);
			$accrual_id = $type->get_accrual_type();
			if($accrual_id !== false) {
				AccrualType::add_balance($accrual_id, $staff_id, -$amount);
			}
		}

		public function toArray() {
			return parent::toArray() + array(
				'timesheet_category_id'	=>	$this->timesheet_category_id,
				'rate_multiplier'	=>	$this->rate,
				'pay_type'		=>	$this->pay_type
			);
		}
	}
?>
