<?php

	/**
	*	Pay Period is from one date to another, with a payment
	*	made after the end of the pay period.  Each Pay Period
	*	applies to certain (not all) employees.  Those employees have a PayStub
	*	generated for the PayPeriod.  Changes cannot be
	*	made to the PayStubs after the payment.
	*
	*	@package perp_payroll
	*	@author Nathan Gray
	*/

	require_once( EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.Paystub.inc.php' );
	require_once( EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.PayrollTransactionLine.inc.php' );
	require_once( EGW_INCLUDE_ROOT . '/etemplate/inc/class.so_sql2.inc.php' );
	require_once( EGW_INCLUDE_ROOT . '/perp_ledger/inc/class.GLTransaction.inc.php' );
	require_once(EGW_API_INC . '/class.asyncservice.inc.php');

	// so_sql2 is a generalized db access object
	class PayPeriod extends so_sql2 {

		/**
		*	These variables are actually handled by so_sql2, 
		*	but I left them here so I know what variables should be around.
		*/

		/**
		*	Unique ID for the pay period
		*/
		//protected $pay_period_id;

		/**
		*	Start date for this pay period
		*/
		//protected $start_date;

		/**
		*	End date 
		*/
		//protected $end_date;

		/**
		*	Payday
		*/
		//protected $transaction_date;

		public static $async = null;

		/**
		*	The constructor is overloaded.
		*	Passing just 1 int will load the matching period.
		*	Passing schedule_id, start date, end date and pay day will create a new pay period.
		*/
		public function __construct($schedule_id, $start_date = null, $end_date = null, $pay_day = null) {
			// Set up DB access
			$this->so_sql2('perp_payroll', 'perp_pay_period');

			if(is_numeric($schedule_id) && $schedule_id != 0 && is_null($start_date)) {
				// Load an existing pay period
				$this->read(array('pay_period_id' => $schedule_id));
			} elseif ($start_date && $end_date && $pay_day) {
				$this->start_date = $start_date;
				$this->end_date = $end_date;
				$this->transaction_date = $pay_day;
				if($this->save() != 0) {
					throw new Exception(lang('Unable to create pay period, it may already exist'));
				}

				$schedule = new PaySchedule($schedule_id);

				// Schedule stub & pay days
				if($schedule instanceof PaySchedule) {
					$this->schedule_stub_day($schedule->get_schedule_id(), $schedule->get_next_paystub($start_date));
					$this->schedule_pay_day($pay_day ? $pay_day : $schedule->get_next_payday($start_date));

					// Create next pay period
					$schedule->schedule($this->end_date);
				}
			}
		}

		/**
		*	Delete this pay period
		*	
		*	If this pay period has not been posted yet, 
		*	delete it.  This will remove any created paystubs,
		*	and the pay period ID will not be reused.
		*/
		public function delete() {
			if(!$this->isEditable()) {
				throw new egw_exception('Unable to delete pay period, it has already been posted');
			}
			if(!$this->pay_period_id) {
				throw new egw_exception('Unable to delete pay period, it has not been saved yet');
			}
			if(!$GLOBALS['egw']->acl->check('pay_period', EGW_ACL_DELETE, 'perp_payroll')) {
				throw new egw_exception_no_permission('Unable to delete pay period, you do not have permission');
			}
			$result = parent::delete($this->pay_period_id);

			$this->delete_paystubs();

			return ($result == 1);
		}

		/**
		*	Revert this pay period
		*
		*	If you need to use this, you've really screwed something up, and a permanent record
		*	has been made.  Reverting will create another pay period for the same dates, with
		*	all the numbers negated.  This will decrement YTD totals, and reverse any transactions.
		*	You can't claw back any amounts paid, so it'll probably screw up all sorts of other stuff
		*	like payments.  You should have deleted it before it got posted.
		*/
		public function revert() {
			if(!$GLOBALS['egw']->acl->check('pay_period', EGW_ACL_CUSTOM_1, 'perp_payroll')) {
				throw new egw_exception_no_permission('Unable to revert pay period, you do not have permission');
			}

			$GLOBALS['egw']->db->transaction_begin();

			$revert = new PayPeriod(null, $this->start_date, $this->end_date, time());

			$query = array();
			$stubs = $this->get_paystubs($query);
			foreach($stubs as &$paid_stub) {
				$revert->add_staff($paid_stub->get_staff_id(), true);
			}
			$revert->create_paystubs();
			$revert->reverted_id = 0;
			$revert->save();
			$revert_stubs =& $revert->get_paystubs($query);
			foreach($revert_stubs as &$revert_stub) {
				$lines =& $revert_stub->get_lines();
				foreach($lines as &$line) {
					$line->set_rate(-$line->get_rate());
				}
				$revert_stub->save();
			}
			$revert->post_paystubs();

			$this->reverted_id = $revert->transaction_id;
			$this->save();

			$GLOBALS['egw']->db->transaction_commit();
			return true;
		}

		/** 
		*	Translate local data array into database specific
		*/
		public function db2data() {
			$data =& $this->data;
			$data['start_date'] = is_null($data['start_date']) ? null : $GLOBALS['egw']->db->from_timestamp($data['start_date']);
			$data['end_date'] = is_null($data['end_date']) ? null: $GLOBALS['egw']->db->from_timestamp($data['end_date']);
			$data['transaction_date'] = is_null($data['transaction_date']) ? null : $GLOBALS['egw']->db->from_timestamp($data['transaction_date']);
		}

		/** 
		*	Translate database specific data array into usable values
		*/
		public function data2db($data = null) {
			if(!$data) {
				$data =& $this->data;
			}
			$data['start_date'] = is_null($data['start_date']) ? null : $GLOBALS['egw']->db->to_timestamp($data['start_date']);
			$data['end_date'] = is_null($data['end_date']) ? null : $GLOBALS['egw']->db->to_timestamp($data['end_date']);
			$data['transaction_date'] = is_null($data['transaction_date']) ? null : $GLOBALS['egw']->db->to_timestamp($data['transaction_date']);

			return $data;
		}

		/**
		*	Can this pay period be edited?
		*	If the pay period can't be edited, the pay stubs shouldn't either.
		*
		*	@return boolean
		*/
		public function isEditable() {
			return (is_null($this->transaction_id));
		}

		public function schedule_stub_day($schedule_id, $date) {
			if(!$this->isEditable()) {
				throw new Exception(lang('Cannot reschedule stub creation, pay period is not editable'));
			}

			//echo "schedule_stub_day($schedule_id, " . date('Y-m-d H:i', $date) . ')<br />';
			$id = 'perp_payroll-' . $this->pay_period_id . '-stub_day';

			// Cancel any pending runs
			self::$async->cancel_timer($id);

			return self::$async->set_timer(
				$date,
				$id,
				'perp_payroll.bo_perp_payroll.create_paystubs',
				array(
					'pay_schedule_id'	=>	$schedule_id,
					'pay_period_id'	=>	$this->pay_period_id,
				),
				false
			);
		}

		public function schedule_pay_day($date) {
			if(!$this->isEditable()) {
				throw new Exception(lang('Cannot reschedule payday, pay period is not editable'));
			} elseif (!$date || $date < time()) {
				throw new Exception(lang('Cannot schedule payday in the past (%1)', date($GLOBALS['egw_info']['user']['preferences']['common']['dateformat'], $date)));
			}

			//echo 'schedule_pay_day(' . date('Y-m-d H:i', $date) . ')<br />';
			$id = 'perp_payroll-' . $this->pay_period_id . '-pay_day';

			// Cancel any pending runs
			self::$async->cancel_timer($id);

			return self::$async->set_timer(
				$date,
				$id,
				'perp_payroll.bo_perp_payroll.process_payday',
				array(
					'pay_period_id'		=>	$this->pay_period_id,
				),
				false
			);
		}

		/**
		*	Get a list of paystubs for this pay period
		*/
		public function &get_paystubs(&$query) {
			$stubs = array();
			$where = array(
				'pay_period_id'	=>	$this->pay_period_id
			);
			if($query['staff_id']) {
				if(is_array($query['staff_id'])) {
					$where[] = 'perp_pay_stub.staff_id IN (' . implode(',', $query['staff_id']) . ')';
				} else {
					$where[] = 'perp_pay_stub.staff_id = ' . (int)$query['staff_id'];
				}
			}
			$join = 'JOIN perp_staff ON
					perp_pay_stub.staff_id = perp_staff.staff_id
				JOIN egw_addressbook ON 
					egw_addressbook.contact_id = perp_staff.addressbook_id';

			if(!$query['order']) {
				$query['order'] = 'staff_id';
			}

			$GLOBALS['egw']->db->select(
				'perp_pay_stub',
				array('count(perp_pay_stub.staff_id) AS count'),
				$where,
				__LINE__, __FILE__, 0, 
				'', 'perp_payroll', 1,
				$join
			);
			if($row = $GLOBALS['egw']->db->row(true)) {
				$query['count'] = $row['count'];
			}

			$GLOBALS['egw']->db->select(
				'perp_pay_stub',
				array(
					'perp_pay_stub.staff_id'
				),
				$where,
				__LINE__, __FILE__, $query['start'],
				'ORDER BY ' . $query['order'] . ' ' . $query['sort'], 
				'perp_payroll', $query['num_rows'],
				$join
			);
			$rows = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$rows[] = $row;
			}

			if(!array_key_exists('make_paystubs', $query) || $query['make_paystubs']) {
				foreach($rows as $row) {
					$stubs[] =& new Paystub($row['staff_id'], $this);
				}
			}
			return $stubs;
		}

		/**
		*	Create paystubs for this pay period
		*
		*	@param schedule The PaySchedule that generated this pay period.  This is needed because the PaySchedule determines who gets paid.
		*		If schedule is null, this will be a 1 time pay period, not according to a regular schedule.
		*/
		public function create_paystubs(PaySchedule $schedule = null) {
			$staff = null;
			$stub = null;

			$this->data['created_paystubs'] = false;

			if(!$this->isEditable()) {
				throw new Exception(lang('Pay period is not editable'));
			}

			// This may timeout, so we need to make sure the transaction is aborted
			register_shutdown_function(array($this, 'create_timeout'));

			$GLOBALS['egw']->db->transaction_begin();

			// Update that we tried to create paystubs
			config::save_value('last_create_paystub_date', time(), 'perp_payroll');

			// Remove any existing paystubs for this pay period
			$this->delete_paystubs();

			// Get notes
			$GLOBALS['egw']->db->select(
				'perp_pay_stub_note',
				array('note'),
				array(
					'pay_period_id'	=>	$this->pay_period_id,
					'staff_id'	=>	0
				), __LINE__, __FILE__, false, '', 'perp_payroll'
			);
			$notes = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$notes[] = $row['note'];
			}

			if(!is_null($schedule)) {
				$staff =& $schedule->get_staff($this->start_date);
			} else {
				$staff =& $this->get_staff();
			}
			$config = config::read('perp_ledger');
			$currency = $config['gl_currency'];

			foreach($staff as $staff_id) {
				// Give each paystub 20 seconds to be created
				set_time_limit(20);
				$stub = new Paystub($staff_id, $this, $currency, 1, true);

				foreach($notes as $note) {
					$stub->add_note($note);
				}
				$stub->save();
				$stub = null;
			}

			$GLOBALS['egw']->db->transaction_commit();
			$this->data['created_paystubs'] = true;

		}

		/**
		*	Delete all paystubs associated with this pay period.
		*/
		protected function delete_paystubs() {
			if(!$this->isEditable()) {
				throw new egw_exception('Unable to delete pay stubs, they have already been posted');
			}
			$GLOBALS['egw']->db->delete(
				'perp_pay_stub',
				array(
					'pay_period_id'	=>	$this->pay_period_id
				),
				__LINE__,__FILE__
			);
			$GLOBALS['egw']->db->delete(
				'perp_pay_stub_entry',
				array(
					'pay_period_id'	=>	$this->pay_period_id
				),
				__LINE__,__FILE__
			);
			$GLOBALS['egw']->db->delete(
				'perp_pay_stub_note',
				array(
					'pay_period_id'	=>	$this->pay_period_id,
					'entry_id IS NOT NULL'
				),
				__LINE__, __FILE__
			);
		}

		/**
		*	Post previously created paystubs.
		*	We go through all the paystubs, and create 1 GLTransaction for all of them.
		*	We add a debit and a credit line to the GLTransaction for each
		*	line on the paystub.  Then we check the GLTransaction.
		*	If the transaction is good, then we'll post it.  
		*	We also add a line to the payroll_ledger for each line on each stub, 
		*	Because of the one GL transaction vs many payroll transactions, the payroll_ledger
		*	is just a record.
		*
		*	This should make the pay run atomic.  Of course, if something
		*	goes wrong during the posting, it's still borked...
		*/
		public function post_paystubs() {

			$GLOBALS['egw']->db->transaction_begin();

			// Get config for other transaction properties
			$transaction = new GLTransaction();
			$transaction->set_transaction_type($trans_type);
			$transaction->set_date($this->transaction_date);

			$GLOBALS['egw']->db->select(
				'perp_pay_stub_entry',
				array(
					'name',
					'type', 
					'SUM(rate * quantity) AS amount',
					gl_debit_account,
					gl_credit_account
				),
				array(
					'pay_period_id'	=>	$this->pay_period_id
				),
				__LINE__, __FILE__, 0,
				'GROUP BY type_id ORDER BY sort_order, name', 'perp_payroll', -1,
				'JOIN perp_pay_stub_entry_type ON
					perp_pay_stub_entry.type = perp_pay_stub_entry_type.type_id'
			);
			$gl_lines = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$gl_lines[] = $row;
			}

			// Link in PayrollTransactionLines
			$query = array();
			$stubs = $this->get_paystubs($query);
			foreach($gl_lines as $transaction_line) {
				if($transaction_line['gl_credit_account']) {
					$line = $transaction->add_line(
						$transaction_line['gl_credit_account'],
						$transaction_line['name'],
						-$transaction_line['amount']
					);
				} 
				if ($transaction_line['gl_debit_account']) {
					$line = $transaction->add_line(
						$transaction_line['gl_debit_account'],
						$transaction_line['name'],
						$transaction_line['amount']
					);
				}
				if(!$line) {
					throw new egw_exception(lang('Missing transaction line, you need a General Ledger account for %1', $transaction_line['name']));
					$GLOBALS['egw']->db->transaction_abort();
				}
				foreach($stubs as &$paystub) {
					$stub_lines =& $paystub->get_lines($transaction_line['type']);
					foreach($stub_lines as &$paystub_line) {
						new PayrollTransactionLine($line, $paystub_line);
					}
				}
			}
					
			// Check all transactions first
			if($transaction->check() !== true) {
				$GLOBALS['egw']->log->error(array(
					'text'	=>	'D-Paystub, Transaction: <pre>' . $transaction->__toString() . '</pre>',
					'line'	=>	__LINE__,
					'file'	=>	__FILE__
				));
				$GLOBALS['egw']->log->error(array(
					'text'	=>	'E-Paystub, Transaction imbalanced by %1',
					'p_1'	=>	$transaction->check(),
					'line'	=>	__LINE__,
					'file'	=>	__FILE__
				));
				throw new Exception('Transaction failed check because of imbalance');
			}

			// All transactions passed, we'll post and update YTD
			$this->transaction_id = $transaction->write();
			$this->save();
			$stubs =& self::get_paystubs($query);
			foreach($stubs as &$paystub) {
				$paystub->update_YTD();
			}

			$GLOBALS['egw']->db->transaction_commit();
		}

		/**
		*	Get a list of staff specifically assigned to this pay period.
		*
		*	Usually this is not needed, as the PaySchedule provided to create_paystubs() is used
		*	to pull the staff from the perp_staff table.  The 1 time pay periods need this though.
		*	If you call this function for a regular pay period created according to a schedule,
		*	it will return an empty array.
		*
		*	@return an array of staff_ids
		*/
		public function get_staff() {
			$GLOBALS['egw']->db->select(
				'perp_pay_period_staff',
				array(
					'staff_id'
				),
				array(
					'pay_period_id'	=>	$this->pay_period_id
				),
				__LINE__, __FILE__, 0, '',
				'perp_payroll', -1
			);
			$staff = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$staff[] = $row['staff_id'];
			}
			return $staff;
		}

		/**
		*	Add a staff person to this pay period.
		*
		*	If this pay period is a regurlar pay period created according to a schedule, 
		*	any added staff will be ignored and the schedule followed instead.
		*
		*	@param staff_id Integer ID of staff person to add
		*	@param force boolean Add staff even if there is an overlap
		*	@param boolean True on success, pay period ID if staff overlaps with another (double pay for same date)
		*/
		public function add_staff($staff_id, $force = false) {
			// Check for overlap
			$result = $GLOBALS['egw']->db->select(
				'perp_pay_period',
				array(
					'perp_pay_period.pay_period_id'
				),
				array(
					'staff_id = ' . (int)$staff_id,
					'(( start_date <= ' . $GLOBALS['egw']->db->quote($GLOBALS['egw']->db->to_timestamp($this->start_date)) . ' AND ' .  
					'end_date >= ' . $GLOBALS['egw']->db->quote($GLOBALS['egw']->db->to_timestamp($this->start_date)) . ') OR ('  .
					'start_date <= ' . $GLOBALS['egw']->db->quote($GLOBALS['egw']->db->to_timestamp($this->end_date)) . ' AND ' .  
					'end_date >= ' . $GLOBALS['egw']->db->quote($GLOBALS['egw']->db->to_timestamp($this->end_date)) . '))', 
				),
				__LINE__, __FILE__, false,
				'', 'perp_payroll', -1,
				'JOIN perp_pay_stub ON 
					perp_pay_period.pay_period_id = perp_pay_stub.pay_period_id'
			);
			if($result->NumRows() > 0) {
				return $result->fetchSingle();
			}
			return $GLOBALS['egw']->db->insert(
				'perp_pay_period_staff',
				array(
					'pay_period_id'	=>	$this->pay_period_id,
					'staff_id'	=>	(int)$staff_id
				),
				array(
					'pay_period_id'	=>	$this->pay_period_id,
					'staff_id'	=>	(int)$staff_id
				),
				__LINE__, __FILE__, 'perp_payroll'
			);
		}

		/**
		*	Remove a staff person from this pay period.
		*	If this pay period is a regurlar pay period created according to a schedule, 
		*	any added staff will be ignored and the schedule followed instead.
		*/
		public function remove_staff($staff_id) {
			$GLOBALS['egw']->db->delete(
				'perp_pay_period_staff',
				array(
					'staff_id'	=>	(int)$staff_id,
					'pay_period_id'	=>	$this->pay_period_id
				),
				__LINE__, __FILE__
			);
		}


		/**
		*	Add a note to every paystub in this pay period.
		*
		*	This can only be undone by removing the note from
		*	every paystub individually.
		*
		*	@param note The text to add to each paystub
		*/
		public function add_note($note) {
			$query = array();
			$success = true;
			$stubs = $this->get_paystubs($query);
			if(count($stubs) > 0) {
				foreach($stubs as &$stub) {
					$success = $success && $stub->add_note($note);
					$stub->save();
				}
			} else {
				$GLOBALS['egw']->db->insert(
					'perp_pay_stub_note',
					array(
						'pay_period_id'	=>	$this->pay_period_id,
						'staff_id'	=>	null,
						'entry_id'	=>	null,
						'note'		=>	$note
					),
					array(
						'pay_period_id'	=>	$this->get_pay_period_id,
						'staff_id'	=>	null,
						'entry_id'	=>	null,
						'note'		=>	$note
					),
					__LINE__, __FILE__, 'perp_payroll'
				);
				$success = true;
			}
			return $success;
		}

		public function toArray() {
			$array = $this->data;
			return $array;
		}

		public function __toString() {
			return lang(
				'Pay Period %1 (%2 -> %3)', 
				$this->pay_period_id, 
				date($GLOBALS['egw_info']['user']['preferences']['common']['dateformat'], $this->start_date),
				date($GLOBALS['egw_info']['user']['preferences']['common']['dateformat'], $this->end_date)
			);
		}

		public function create_timeout() {
			if(isset($this->data['created_paystubs']) && !$this->data['created_paystubs']) {
				$GLOBALS['egw']->db->transaction_abort();
				$GLOBALS['egw']->db->transaction_commit();
			}
		}
	}

	PayPeriod::$async = new asyncservice();
?>
