<?php

	/**
	*	Pay schedule keeps track of when to create pay periods, and pay stubs
	*
	*	Schedules are tied into the calendar, and use all its logical goodness for
	*	repeating.
	*
	*	@package perp_payroll
	*	@author Nathan Gray
	*/

	require_once(EGW_INCLUDE_ROOT . '/calendar/inc/class.calendar_bo.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_payroll/inc/class.PayPeriod.inc.php');
	require_once(EGW_INCLUDE_ROOT . '/perp_human_resources/inc/class.Staff.inc.php');
	require_once(EGW_API_INC . '/class.asyncservice.inc.php');

	class PaySchedule {

		/**
		*	cal_id of the Calendar entry used for this pay schedule
		*	We use the attendees to track who is in the pay period,
		*	and the title as the name of the schedule
		*/
		protected $calendar_id = null;

		/**
		*	How many days after the end of the pay period do we wait for
		*	supervisors and administrators to verify and correct source data
		*	such as timesheets and such before
		*	creating the PayStubs for the period
		*
		*	The default is 3, so 2 days to validate pay stubs before
		*	the default payday (5 day delay)
		*/
		protected $paystub_delay = 3;

		/**
		*	How many days after the end of the pay period do we wait for
		*	supervisors and administrators to verify and correct before
		*	we post the transaction.
		*
		*	The default is 5, which will give a payday on the following Friday for a 
		*	Monday - Sunday pay period.
		*/
		protected $transaction_delay = 5;

		/**
		*	BO object for accessing calendar entries
		*	
		*	Initialized at the bottom of this file
		*/
		public static $bo_cal;

		/**
		*	Async services for eGW
		*	Initialized at the bottom of this file
		*/
		public static $async;

		/**
		*	Constructor
		*
		*	@param calendar_id Provide the calendar ID to load an existing schedule
		*/
		public function __construct($calendar_id = null) {
			$this->calendar_id = $calendar_id;

			// Existing schedule?
			$GLOBALS['egw']->db->select(
				'perp_pay_schedule',
				array('*'),
				array(
					'schedule_id' => $this->calendar_id
				),
				__LINE__, __FILE__
			);
			if($row = $GLOBALS['egw']->db->row(true)) {
				$this->paystub_delay = $row['stub_delay_days'];
				$this->transaction_delay = $row['pay_delay_days'];
			}
		}

		/**
		*	Accessors
		*	@{
		*/

		/**
		*	Get the id of this schedule
		*/
		public function get_schedule_id() {
			return $this->calendar_id;
		}

		/**
		*	Get the number of days to wait after the end of the pay period
		*	before the payments are posted.
		*	@return int days to wait
		*/
		public function get_transaction_delay() {
			return $this->transaction_delay;
		}
		/**
		*	Set the number of days to wait after the end of the pay period
		*	before the payments are posted.
		*	@param delay int days to wait
		*/
		public function set_transaction_delay($delay) {
			if(is_numeric($delay) && $delay >= 0) {
				$this->transaction_delay = $delay;
			}
		}

		/**
		*	Set the number of days to wait after the end of the pay period
		*	before the paystubs are created.
		*	@param delay int days to wait
		*/
		public function get_paystub_delay() {
			return $this->paystub_delay;
		}
		/**
		*	Set the number of days to wait after the end of the pay period
		*	before the paystubs are created.
		*	@param delay int days to wait
		*/
		public function set_paystub_delay($delay) {
			if(is_numeric($delay) && $delay >= 0) {
				$this->paystub_delay = $delay;
			}
		}

		/**
		*	Get the next pay period start date, according to this schedule
		*	@return int timestamp
		*/
		public function get_next_start($date = null) {
			if($date == null) {
				$date = time();
			}
			$calendar = self::$bo_cal->read( $this->calendar_id, $date, true);
			return $calendar['start'];
		}

		/**
		*	Get the next pay period end date, according to this schedule
		*	@return int timestamp
		*/
		public function get_next_end($date = null) {
			if($date == null) {
				$date = time();
			}
			$calendar = self::$bo_cal->read( $this->calendar_id, $date, true);
			return $calendar['end'];
		}

		/**
		*	Calculate the next payday from a given date
		*/
		public function get_next_paystub($date = null) {
			// Strip off the time
			$end_date = $this->get_next_end($date);
			$end_date = mktime(0,0,0, date('m', $end_date), date('d', $end_date), date('Y', $end_date));
			$next_date = strtotime('+' . $this->paystub_delay . ' days', $end_date);
			return $next_date;
		}

		/**
		*	Calculate the next payday from a given date
		*/
		public function get_next_payday($date = null) {
			$end_date = $this->get_next_end($date);
			$end_date = mktime(0,0,0, date('m', $end_date), date('d', $end_date), date('Y', $end_date));
			$transaction_date = strtotime('+' . $this->transaction_delay . ' days', $end_date);
			return $transaction_date;
		}

		/**
		*	Get the name of this pay schedule
		*	@return name
		*/
		public function get_name() {
			$calendar = self::$bo_cal->read( $this->calendar_id, $date, true);
			return $calendar['title'];
		}

		/**
		*	@}
		*/

		/**
		*	Save this pay schedule to the database
		*/
		public function save() {
			$GLOBALS['egw']->db->insert(
				'perp_pay_schedule',
				array(
					'schedule_id'		=>	$this->calendar_id,
					'stub_delay_days'	=>	$this->paystub_delay,
					'pay_delay_days'	=>	$this->transaction_delay
				),
				array(
					'schedule_id'	=>	$this->calendar_id
				),
				__LINE__, __FILE__, 'perp_payroll'
			);

			$this->schedule();
		}		

		/**
		*	Delete all traces of this pay schedule
		*/
		public function delete() {
			if(!$GLOBALS['egw']->acl->check('pay_period', EGW_ACL_DELETE, 'perp_payroll')) {
				throw new egw_exception_no_permission();
			}


			// Cancel any pending runs
			$period_id = 'perp_payroll-' . $this->calendar_id . '-create_pay_period';
			PaySchedule::$async->cancel_timer($period_id);

			// Delete calendar
			self::$bo_cal->so->delete($this->calendar);

			// Delete from database
			$GLOBALS['egw']->db->delete(
				'perp_pay_schedule',
				array(
					'schedule_id'	=>	$this->calendar_id
				),
				__LINE__, __FILE__, 'perp_payroll'
			);
		}

		/**
		*	Use async services to schedule the payperiod creation
		*/
		public function schedule($date = null) {
			$period_id = 'perp_payroll-' . $this->calendar_id . '-create_pay_period';

			// Cancel any pending runs
			PaySchedule::$async->cancel_timer($period_id);

			// Schedule next run
			//echo 'make_payperiod on ' . date('Y-m-d H:i', $this->get_next_start($date)) . ' (' . $this->calendar_id . ') <br />';
			return PaySchedule::$async->set_timer(
				$this->get_next_start($date),
				$period_id,
				'perp_payroll.bo_perp_payroll.make_payperiod',
				array(
					'schedule_id'	=>	$this->calendar_id,
					'start'		=>	$this->get_next_start($date)
				),
				false
			);
		}

		/**
		*	Create the next pay period for the given schedule
		*	
		*	@param time Timestamp used to search for the next pay period
		*	@throws Exception if it can't get the calendar entry
		*	@throws Exception if the pay period has already ended
		*/
		public function create_pay_period($time = null) {
			if($time == null) {
				$time = time();
			}
			$calendar = self::$bo_cal->read( $this->calendar_id, $time, true);
			if(!$calendar) {
				self::$bo_cal->check_move_horizont(strtotime('+2 months', $time));
				$calendar = self::$bo_cal->read( $this->calendar_id, $time, true);
			}
			if(!is_array($calendar) || $calendar['end'] <= time()) {
				throw new Exception(lang('Unable to create pay period, %1',
					is_array($calendar) ? lang('Pay period would have already ended') : lang('Could not find calendar entry %1', $this->calendar_id)
				));
			}
			$pay_period = new PayPeriod($this, $calendar['start'], $calendar['end'], $this->get_next_payday($time));

			return $pay_period->pay_period_id;
		}

		public function get_staff($start_date) {
			if(!$this->calendar_id) {
				return array();
			}

			$GLOBALS['egw']->db->select(
				'perp_staff',
				array('staff_id'),
				array(
					'pay_schedule'	=>	$this->calendar_id,
					'status'	=>	Staff::ACTIVE
				),
				__LINE__, __FILE__, false,
				'', 'perp_staff'
			);
			$staff = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$staff[] = $row['staff_id'];
			}
			return $staff;
		}

		public function __toString() {
			return get_class($this);
		}

		/**
		*	Get a list of schedules
		*
		*	@nextmatch
		*/
		public static function get_rows(&$query = array(), &$rows = array(), &$readonlys = array()) {
			$where = array();
			if(!$query['order']) {
				$query['order'] = 'schedule_id';
			}
			if(!$query['time']) {
				$query['time'] = time();
			}
			$join = 'JOIN egw_cal ON egw_cal.cal_id = perp_pay_schedule.schedule_id';

			// Get number of rows
			$GLOBALS['egw']->db->select(
				'perp_pay_schedule',
				array('count(schedule_id) AS count'),
				$where,
				__LINE__, __FILE__, false, '',
				'perp_payroll', 1, 
				$join
			);
			if($row = $GLOBALS['egw']->db->row(true)) {
				$query['count'] = $row['count'];
			}

			$GLOBALS['egw']->db->select(
				'perp_pay_schedule',
				array(
					'*',
					'schedule_id', 
					'stub_delay_days',
					'pay_delay_days',
				), 
				$where,
				__LINE__, __FILE__, $query['start'],
				'ORDER BY ' . $query['order'] . ' ' . $query['sort'], 
				'perp_payroll', $query['num_rows'],
				$join
			);
			$rows = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$rows[] = $row;
			}
			$config = config::read('perp_payroll');
			foreach($rows as &$row) {
				// This gets the NEXT one, for some reason
				$next_calendar = self::$bo_cal->read( $row['cal_id'], $query['time'], true);
				if(!$next_calendar) {
					self::$bo_cal->check_move_horizont(strtotime('+2 months', $query['time']));
					$next_calendar = self::$bo_cal->read( $row['cal_id'], $query['time'], true);
				}

				// Find the current pay period
				$query = array(
					'end'		=>	$next_calendar['end'],
					'cat_id'	=>	$config['pay_schedule_category'],
					'ignore_acl'	=>	true,
					'users'		=>	$next_calendar['owner']
				);
				$calendar_list = self::$bo_cal->search($query);
				$current_calendar = array();
				foreach($calendar_list as $key => $calendar_entry) {
					list($cal_id, $timestamp) = explode('-', $key);
					if($cal_id != $row['cal_id']) {
						continue;
					}

					if($timestamp >= $next_calendar['start']) {
						break;
					}
					$current_calendar = $calendar_entry;
				}

				if($next_calendar['start']) {
					$row['next']['start_date'] = $next_calendar['start'];
					$row['next']['end_date'] = $next_calendar['end'];
					$row['next']['post_date'] = strtotime('+' . $row['pay_delay_days'] . ' days', $next_calendar['end']);
					$row['next']['paystub_date'] = strtotime('+' . $row['stub_delay_days'] . ' days', $next_calendar['end']);
				}

				if($current_calendar['start']) {
					$row['current']['start_date'] = $current_calendar['start'];
					$row['current']['end_date'] = $current_calendar['end'];
					$row['current']['post_date'] = strtotime('+' . $row['pay_delay_days'] . ' days', $current_calendar['end']);
					$row['current']['paystub_date'] = strtotime('+' . $row['stub_delay_days'] . ' days', $current_calendar['end']);
				}
			}

			return $query['count'];
		}
	}

	PaySchedule::$bo_cal = new calendar_bo();
	PaySchedule::$async = new asyncservice();
?>
