<?php

	/**
	*	Utility class full of handy goodness functions
	*/

	require_once(EGW_INCLUDE_ROOT . '/perp_human_resources/inc/class.Staff.inc.php');

	class PayrollUtility {

		protected static $config;

		// Pay frequency can be an expensive calculation, so we cache the result
		protected static $pay_frequency_cache = array();

		public function __construct() {
			self::$config = config::read('perp_payroll');
		}

		/**
		*	Get the number of hours the given staff person is / was scheduled to work
		*	between the start date and the end date.  Uses the work shift to determine
		*	how many hours.  It's probably best to use the pay schedule to determine
		*	the start date and end date.
		*
		*	@param staff		Staff person to perform calculations on
		*	@param start_date	Start counting from
		*	@param end_date		Count hours to
		*
		*	@return array(
		*		'seconds' => Number of seconds scheduled to work in the given period
		*		'days'	=> Days scheduled to work in the given period
		*	)
		*/
		public static function get_scheduled_time(Staff &$staff, $start_date, $end_date) {
			// For each day between start_date and end date, get the work schedule.  Add the number of hours to the total.
			// Get the breaks
			if(!$start_date) {
				return array('seconds' => 0, 'days' => 0);
			}

			$total_time = 0;
			$total_days = 0;

			$date = $start_date;
			do {
				$total_days++;
				$daily_time = 0;
				$work_shifts = $staff->get_work_schedule($date);
				foreach($work_shifts as $shift) {
					$daily_time += ($shift['end'] - $shift['start'] - self::get_total_break_time($shift));
				}
				$work_shifts = null;
				$total_time += $daily_time;
				$date = strtotime("+1 day", $date);
			} while($date <= $end_date);
			return array(
				'seconds'	=>	$total_time,
				'days'		=>	$total_days
			);
		}


		/**
		*	Get the total number of seconds a person should be on unpaid break
		*	during the shift.
		*
		*	@param shift ID of Calendar entry for the shift.
		*/
		public static function get_total_break_time($shift) {
			if(!is_array(self::$config)) {
				new PayrollUtility();
			}

			$break_time = 0;
			if(!is_array(self::$config['breaks']) || count(self::$config['breaks']) == 0) {
				return $break_time;
			}
			foreach(self::$config['breaks'] as $break) {
				$start = (int)$shift['start'];
				$end = (int)$shift['end'];
				do {
					$break_start = $start + ($break['work_length'] * 3600);
					$break_end = $break_start + ($break['break_length'] * 60);
					// If they're here for the break, add it in
					if($break_end < $end) {
						$break_time += ($break_end - $break_start);
					}
					$start = $break_start;
				} while($start < $break_start);
			}
			return $break_time;
		}

		/**
		*	Get the pay frequency based on a payment schedule.  A fairly easy calculation,
		*	365 / (schedule days)
		*
		*	@param staff Staff person to calculate pay frequency for
		*	@return int Number of times in a year the person should be paid
		*/
		public static function get_pay_frequency(Staff &$staff) {
			if(!self::$pay_frequency_cache[$staff->get_staff_id()]) {
				$pay_schedule = $staff->get_payment_schedule();

				$time = self::get_scheduled_time($staff, $pay_schedule['start'], $pay_schedule['end']);

				if($time['days'] == 0) {
					self::$pay_frequency_cache[$staff->get_staff_id()] = 0;
				} else {
					self::$pay_frequency_cache[$staff->get_staff_id()] = floor(365 / $time['days']);
				}
			}
			return self::$pay_frequency_cache[$staff->get_staff_id()];
		}

		/**
		*	Convert a salary to an rate in the given units
		*
		*	@param staff Staff person to calculate for
		*	@param unit a single char, 'h' for hourly, or null for per pay period
		*
		*	@return double hourly rate
		*/
		public static function get_salaried_rate(Staff $staff, $unit = null, $pay_period_id = null) {
			$payment_info = $staff->get_payroll_info();
			if($pay_period_id == null) {
				$pay_schedule = $staff->get_payment_schedule();
				$time = self::get_scheduled_time($staff, $pay_schedule['start'], $pay_schedule['end']);
			} else {
				$pay_period = new PayPeriod($pay_period_id);
				$time = self::get_scheduled_time($staff, $pay_period->start_date, $pay_period->end_date);
			}

			$frequency = self::get_pay_frequency($staff);
			if($frequency == 0) {
				// Never paid
				return 0;
			}
			
			if($payment_info['pay_unit'] == 'h') {
				$rate = $payment_info['pay_rate'] * (is_null($unit) ? ($time['seconds']/3600) : 1);
			} elseif ($payment_info['pay_unit'] == 'm') {
				$rate = (($payment_info['pay_rate'] / $frequency * 12)) / (is_null($unit) || $time['seconds'] == 0 ? 1 : ($time['seconds'] / 3600));
			} elseif ($payment_info['pay_unit'] == 'y') {
				$rate = ($payment_info['pay_rate'] / $frequency) / (is_null($unit) || $time['seconds'] == 0 ? 1 : ($time['seconds'] / 3600));
			}
			return round($rate, 2);
		}

		/**
		*	Get date calculations on a given period
		*
		*	@param start_date the beginning of the period
		*	@param end_date the end of the period
		*
		*	@return array of date information about that period
		*/
		public static function get_period_info($start_date, $end_date) {
			if(!(int)$start_date || !(int)$end_date) {
				throw new Exception(lang('Invalid parameter: %1', (int)$start_date == 0 ? $start_date : $end_date));
			}
			$period_info = array(
				'start_date'	=>	$start_date,
				'end_date'	=>	$end_date
			);

			$date_info = getdate(mktime(0,0,0,12,31,date('Y', $pay_period->start_date)));
			$period_info['days_in_year'] = $date_info['yday'] + 1;
			
			if(date('H:i', $end_date) == '00:00') {
				// End date is probably short by 23:59 because the time is missing
				$period_info['end_date'] += (23 * 3600) + (59 * 60);
			}
			
			// Duration and frequency
			$period_info['days'] = date('z', $period_info['end_date']) - date('z', $period_info['start_date']) + (date('Y', $period_info['end_date']) - date('Y', $period_info['start_date'])) * $period_info['days_in_year'];
			$period_info['days'] = round(($period_info['end_date'] - $period_info['start_date']) / 86400);
			$period_info['frequency'] = round($period_info['days_in_year'] / $period_info['days']);

			// Monthly is a little screwy
			if(date('m', $period_info['start_date']) == 2 && date('m', $period_info['end_date']) == 2 && $period_info['frequency'] == 13) {
				// January, monthly (period = 13)
				$period_info['frequency'] = 12;
			}
			$period_info['frequency_name'] = Paystub::$pay_frequency[$period_info['frequency']];

			// Significant dates
			$period_info['start_week'] = date('W', $period_info['start_date']);

			// First in year
			switch ($period_info['frequency_name']) {
				case 'Monthly':
					// First in year is Jan date('d', $start_date);
					$period_info['first_in_year']['start_date'] = mktime(
						date('H', $period_info['start_date']), 
						date('i', $period_info['start_date']), 
						date('s', $period_info['start_date']),
						1, date('d', $period_info['start_date']),
						date('Y', $period_info['start_date'])
					);
					$period_info['first_in_year']['end_date'] = mktime(
						date('H', $period_info['end_date']), 
						date('i', $period_info['end_date']), 
						date('s', $period_info['end_date']),
						date('m', $period_info['first_in_year']['start_date']),
						date('d', $period_info['end_date']) == date('t', $period_info['end_date']) ? date('t', strtotime('Jan 1 ' . date('Y', $period_info['end_date']))) : date('d', $period_info['end_date']),
						date('Y', $period_info['start_date'])
					);
					$period_info['last_in_year']['start_date'] = mktime(
						date('H', $period_info['start_date']),
						date('i', $period_info['start_date']),
						date('s', $period_info['start_date']),
						12, date('d', $period_info['start_date']),
						date('Y', $period_info['start_date'])
					);
					$period_info['last_in_year']['end_date'] = mktime(
						date('H', $period_info['start_date']),
						date('i', $period_info['start_date']),
						date('s', $period_info['start_date']),
						date('m', $period_info['last_in_year']['start_date']), 
						date('d', $period_info['end_date']) == date('t', $period_info['end_date']) ? date('t', strtotime('Jan 31 ' . date('Y', $period_info['end_date']))) : date('d', $period_info['end_date']),
						date('Y', $period_info['end_date'])
					);
					if($period_info['last_in_year']['end_date'] < $period_info['last_in_year']['start_date']) {
						$period_info['last_in_year']['end_date'] = strtotime('+1 Month', $period_info['last_in_year']['end_date']);
					}

					break;
				default:
					$temp_date = $period_info['start_date'];
					while(date('W', $temp_date) > 1) {
						$temp_date = strtotime('-' . $period_info['days'] . ' Day', $temp_date);
					}
					$period_info['first_in_year']['start_date'] = $temp_date;
					$period_info['first_in_year']['end_date'] = strtotime('+' . $period_info['days'] . ' Day', $temp_date);
			}

			return $period_info;
		}
	}
?>
