<?php

	/**
	*	A type of PayStubEntry, such as Regular Time, Overtime, Federal Tax, etc.
	*
	*	Each line on a PayStub will refer to a PayStubEntryType.
	*
	*	@package perp_payroll
	*	@author Nathan Gray
	*/

	abstract class PayStubEntryType {

		protected $id;
		protected $name;
		protected $disabled = false;
		protected $gl_credit_account;
		protected $gl_debit_account;
		protected $sort_order;
		protected $default_value;

		/**
		*	Some lines tend to have complicated calculations
		*	This is the name of the calculation
		*	@see Calculation
		*/
		protected $calculation;

		public function __construct($id) {
			if($id != null) {
				$this->id = $id;
				if(!$this->read()) {
				}
			}
		}

		/**
		*	Get the unique identifier for this type.
		*	@return int
		*/
		public function get_id() {
			return $this->id;
		}

		/**
		*	Get the human readable name for this type.  
		*	@return string
		*/
		public function get_name() {
			return $this->name;
		}

		/**
		*	Set the human readable name for this type.
		*	It should be unique to reduce confusion.
		*
		*	@param name 
		*/
		public function set_name($name) {
			if(trim($name) == '') {
				throw new Exception(lang('Invalid parameter.') . ' ' . lang('Name cannot be blank'));
			}
			$this->name = $name;
		}

		/**
		*	Check to see if this type has been disabled
		*/
		public function is_disabled() {
			return $this->disabled;
		}

		/**
		*	Get the General Ledger account used for credit transactions
		*
		*	@return int
		*/
		public function get_credit_account() {
			return $this->gl_credit_account;
		}

		/**
		*	Set the General Ledger account used for credits using this type
		*
		*	@param gl_account
		*/
		public function set_credit_account($gl_account) {
			$this->gl_credit_account = $gl_account;
		}

		/**
		*	Get the General Ledger account used for debit transactions
		*
		*	@return int
		*/
		public function get_debit_account() {
			return $this->gl_debit_account;
		}

		/**
		*	Set the General Ledger account used for debits using this type
		*
		*	@param gl_account
		*/
		public function set_debit_account($gl_account) {
			$this->gl_debit_account = $gl_account;
		}

		/**
		*	Sort order is used to determine the order (within similar.. types of type)
		*	that lines go on a PayStub.
		*
		*	@param order (int)
		*/
		public function set_sort_order($order) {
			$this->sort_order = $order;
		}

		/**
		*	Sort order is used to determine the order (within similar.. types of type)
		*	that lines go on a PayStub.
		*
		*	@return order (int)
		*/
		public function get_sort_order() {
			return $this->sort_order;
		}

		/**
		*	Set extra data used for calculating this line
		*
		*	@param data
		*/
		public function set_default_value($data) {
			$this->default_value = $data;
		}

		/**
		*	Get an array of extra data used for calculating this line
		*
		*	@return array
		*/
		public function get_default_value() {
			return $this->default_value;
		}

		/**
		*	Get the value for a given staff.
		*	
		*	Pulls the staff value from perp_staff_entry_type, and merges it with
		*	the default value for the entry type.
		*
		*	@param staff_id
		*	@return mixed, but probably an array
		*/
		public function get_value($staff_id) {
			$GLOBALS['egw']->db->select(
				'perp_staff_entry_type',
				array('value'),
				array(
					'type_id'       =>      $this->id,
					'staff_id'      =>      $staff_id
				),
				__LINE__, __FILE__, 0,
				'', 'perp_payroll'
			);
			$row = $GLOBALS['egw']->db->row(true);
			if(is_array($row) && $row['value'] != null) {
				$value = unserialize($row['value']);
				$default = $this->get_default_value();
				if($default) {
					if(is_array($default) && is_array($value)) {
						return array_merge($default, $value);
					} elseif (is_array($default)) {
						$default['override'] = $value;
						return $default;
					} else {
						return $value;
					}
				} else {
					return $value;
				}
			} else {
				return $this->get_default_value();
			}
		}

		/**
		*	Get the name of the calculation used for this line
		*
		*	@return String 
		*/
		public function get_calculation_name() {
			return $this->calculation;
		}

		/**
		*	Set the name of the calculation used for this line
		*
		*	@param name
		*/
		public function set_calculation_name($name) {
			$this->calculation = $name;
		}

		/**
		*	Set the employees that this line type applies to
		*
		*	@param data array of user ids and other associated data, which depends
		*	on the extending class' implementation.
		*/
		public abstract function set_staff(Array $staff);

		/**
		*	Get a list of employees that this line type applies to, and associated data
		*
		*	@return array of arrays, each sub-array having at least a staff_id key.  Other 
		*		fields are sub-class dependant.
		*/
		public abstract function get_staff();

		/**
		*	Remove a staff member from this entry type, so this entry type will
		*	not be added to any paystubs created in the future.
		*	@param staff Either the staff_id or Staff object
		*/
		public abstract function remove_staff($staff);

		/**
		*	Read this type from the database
		*/
		public function read() {
			$row = $GLOBALS['egw']->db->select(
				'perp_pay_stub_entry_type',
				array('*'),
				array('type_id'	=>	$this->id),
				__LINE__, __FILE__, false,
				'', 'perp_payroll'
			)->fetch();
			$this->id 			=	$row['type_id'];
			$this->name			=	$row['name'];
			$this->gl_debit_account		=	$row['gl_debit_account'];
			$this->gl_credit_account	=	$row['gl_credit_account'];
			$this->sort_order		=	$row['sort_order'];
			$this->default_value		=	unserialize($row['default_value']);
			$this->calculation		=	$row['calculation'];
			$this->disabled			=	$row['disabled'];
			return ($this->id != 0);
		}

		/**
		*	Save this type to the database
		*/
		public function save() {
			$GLOBALS['egw']->db->insert(
				'perp_pay_stub_entry_type',
				array(
					'type_id'	=>	$this->id,
					'name'		=>	$this->name,
					'gl_debit_account'	=>	$this->gl_debit_account,
					'gl_credit_account'	=>	$this->gl_credit_account,
					'sort_order'	=>	$this->sort_order,
					'calculation'	=>	$this->get_calculation_name(),
					'default_value'	=>	serialize($this->default_value),
				),
				array(
					'type_id'	=>	$this->id
				),
				__LINE__, __FILE__, 'perp_payroll'
			);
			$this->id = $this->id ? $this->id : $GLOBALS['egw']->db->get_last_insert_id('perp_pay_stub_entry_type', 'type_id');
		}

		/**
		*	Remove this type from the database
		*/
		public function delete() {
			if(!$GLOBALS['egw']->acl->check('earning_type', EGW_ACL_DELETE, 'perp_payroll')) {
				throw new egw_exception_no_permission(lang('You do not have permission to delete pay stub entry types'));
			}

			// Check to make sure it's not used anywhere
			$count = $GLOBALS['egw']->db->select(
				'perp_pay_stub_entry',
				array('count(type)'),
				array('type'	=>	$this->id),
				__LINE__, __FILE__, false,
				'', 'perp_payroll'
			)->fetchSingle();

			if($count > 0) {
				// Can't be deleted, disable it instead
				$GLOBALS['egw']->db->update(
					'perp_pay_stub_entry_type',
					array(
						'disabled'	=>	true,
					),
					array(
						'type_id'	=>	$this->id
					),
					__LINE__, __FILE__, 'perp_payroll'
				);

				$this->disabled = true;

				// Return false because it wasn't actually deleted
				return false;
			}

			$GLOBALS['egw']->db->delete(
				'perp_pay_stub_entry_type',
				array('type_id'	=>	$this->id),
				__LINE__, __FILE__, 'perp_payroll'
			);
			return true;
		}

		/**
		*	Convert to a handy array
		*/
		public function toArray() {
			return array(
				'type_id'		=>	$this->id,
				'name'			=>	$this->name,
				'gl_debit_account'	=>	$this->gl_debit_account,
				'gl_credit_account'	=>	$this->gl_credit_account,
				'sort_order'		=>	$this->sort_order,
				'calculation'		=>	$this->get_calculation_name(),
				'class'			=>	get_class($this),
				'default_value'		=>	$this->get_default_value()
			);
		}

		/**
		*	Using the given ID, returns an object with the correct sub-type,
		*	if it exists, or null if it doesn't.
		*	@param id int ID of an existing type
		*	@return PaystubEntryType, DeductionType, AccrualType or null
		*/
		public static function factory($id) {
			$row = $GLOBALS['egw']->db->select(
				'perp_pay_stub_entry_type',
				array(
					'type_id',
					'calculation'
				),
				array('type_id'	=>	$id),
				__LINE__, __FILE__, 0,
				'', 'perp_payroll'
			)->fetch();
			if(!$row) {
				return null;
			} elseif($row['calculation'] == 'Total') {
				return new TotalType($id);
			}
			$row = $GLOBALS['egw']->db->select(
				'perp_earning_type',
				array('type_id'),
				array('type_id'	=>	$id),
				__LINE__, __FILE__, 0,
				'', 'perp_payroll'
			)->fetch();
			if($row) {
				return new EarningType($id);
			}
			$row = $GLOBALS['egw']->db->select(
				'perp_deduction_type',
				array('type_id'),
				array('type_id'	=>	$id),
				__LINE__, __FILE__, 0,
				'', 'perp_payroll'
			)->fetch();
			if($row) {
				return new DeductionType($id);
			}
			$row = $GLOBALS['egw']->db->select(
				'perp_accrual_type',
				array('type_id'),
				array('type_id'	=>	$id),
				__LINE__, __FILE__, 0,
				'', 'perp_payroll'
			)->fetch();
			if($row) {
				return new AccrualType($id);
			}
			return null;
		}

		/**
		*	Get a list of PaystubEntryTypes
		*
		*	@param query
		*/
		public static function get_rows(&$query, &$rows, &$readonlys) {
			$where = array();
			$join = '';

			if($query['search']) {
				$search = $GLOBALS['egw']->db->quote('%'.$query['search'].'%');
				$where[] = "(name like $search OR gl_credit_account like $search OR gl_debit_account like $search)";
			}
			if($query['col_filter']['calculation']) {
				$where['calculation'] = $query['col_filter']['calculation'];
			}
			if(array_key_exists('filter', $query)) {
				$where['disabled'] = $query['include_disabled'];
			}

			if($query['staff_id']) {
				$join .= 'LEFT JOIN perp_staff_entry_type ON 
					perp_staff_entry_type.type_id = perp_pay_stub_entry_type.type_id AND
					perp_staff_entry_type.staff_id = ' . (int)$query['staff_id'];
				switch ($query['type']) {
					case 'EarningType':
						$table = 'perp_staff_pay_detail';
						break;
					case 'AccrualType':
						$table = 'perp_staff_accrual';
						break;
				}
				if($table) {
					$join .= " JOIN $table ON 
						$table.type_id = perp_pay_stub_entry_type.type_id AND 
						$table.staff_id = " . (int)$query['staff_id'];
				}
			}
			// Get rows
			$GLOBALS['egw']->db->select(
				'perp_pay_stub_entry_type',
				array('count(perp_pay_stub_entry_type.type_id) AS count'),
				$where,
				__LINE__, __FILE__, false,
				'', 'perp_payroll', 1,
				$join
			);
			if($row = $GLOBALS['egw']->db->row(true)) {
				$query['count'] = $row['count'];
			}
			$GLOBALS['egw']->db->select(
				'perp_pay_stub_entry_type',
				array(
					'*',
					'perp_pay_stub_entry_type.type_id'
				),
				$where,
				__LINE__, __FILE__, $query['start'],
				'ORDER BY ' . $query['order'] . ' ' . $query['sort'],
				'perp_payroll', $query['num_rows'],
				$join
			);
			$rows = array();
			while($row = $GLOBALS['egw']->db->row(true)) {
				$row['default_value'] = unserialize($row['default_value']);
				$rows[] = $row;

				if(!$GLOBALS['egw']->acl->check('earning_type', EGW_ACL_DELETE, 'perp_payroll')) {
					$readonlys['delete[' . $row['type_id'] . ']'] = true;
				}
			}
			return $query['count'];
		}

		/**
		*	Get year to date quantity and amount for the given staff person
		*	Note that this function always returns the year to date, regardless
		*	of the pay period passed.  This may differ from the YTD as at a certain date,
		*	so if you're looking at an old stub, YTD will still show current values.
		*	Also, YTD is updated when posted, so if you're looking at an unposted
		*	paystub, the YTD will not include the amounts on the stub.
		*
		*	@see get_ledger_balance()
		*
		*	@param type_id Int ID of the line type
		*	@param pay_period_id id of the pay period you're looking for.  
		*	@param staff_id The ID of the staff person 
		*	@return array with keys ytd_quantity and ytd_amount
		*/
		public static function get_YTD($type_id, $pay_period_id, $staff_id) {
			$GLOBALS['egw']->db->select(
				'perp_pay_type_ytd',
				array(
					'ytd_quantity',
					'ytd_amount'
				),
				array(
					'type_id'	=>	$type_id,
					'staff_id'	=>	$staff_id,
					'first_pay_period <= ' . $pay_period_id,
					'(last_pay_period >= ' . $pay_period_id . ' OR ' .
					'last_pay_period IS NULL)'
				),
				__LINE__, __FILE__, 0,
				'', 'perp_payroll', 1
			);
			if($row = $GLOBALS['egw']->db->row(true)) {
				return $row;
			} else {
				return null;
			}
		}

		/**
		*	Increase or decrease the Year to Date amount of a pay stub entry type for
		*	a given employee.
		*	@param type_id ID of the entry type to increase
		*	@param pay_period_id ID of the pay period that caused the change
		*	@param staff_id ID of the staff person the change is for
		*	@param amount Amount to add (or remove) from the YTD amount
		*/
		public static function add_YTD($type_id, $pay_period_id, $staff_id, $quantity, $amount) {
			// Check to make sure there is a record, create one if not
			if(self::get_YTD($type_id, $pay_period_id, $staff_id) == null) {
				$GLOBALS['egw']->db->insert(
					'perp_pay_type_ytd',
					array(
						'type_id'	=>	$type_id,
						'staff_id'	=>	$staff_id,
						'first_pay_period'	=>	$pay_period_id,
						'ytd_quantity'	=>	$quantity,
						'ytd_amount'	=>	$amount
					),
					array(),
					__LINE__, __FILE__, 'perp_payroll'
				);
			} else {
				$GLOBALS['egw']->db->update(
					'perp_pay_type_ytd',
					array(
						'ytd_quantity = ytd_quantity + ' . (double)$quantity,
						'ytd_amount = ytd_amount + ' . (double)$amount,
					),
					array(
						'type_id'	=>	(int)$type_id,
						'staff_id'	=>	(int)$staff_id,
						'first_pay_period >= ' . (int)$pay_period_id,
						'(last_pay_period <= ' . (int)$pay_period_id . ' OR ' .
						'last_pay_period IS NULL)'
					),
					__LINE__, __FILE__, 'perp_payroll'
				);
			}
		}

		/**
		*	Get the balance of a pay stub entry type
		*
		*	@param type_id ID of the type to get the balance for
		*	@param staff_id ID of the employee, or null for all
		*	@param end_date int end date
		*	@param start_date int Start date, or null for no start date
		*
		*	@return The balance for the specified parameters
		*/
		public static function get_ledger_balance($type_id, $staff_id, $end_date, $start_date = null) {
			$where['type_id'] = (int)$type_id;
			if(!is_null($staff_id)) {
				$where['staff_id'] = (int)$staff_id;
			}
			if((int)$end_date) {
				$where[] = 'trans_date <= ' . $GLOBALS['egw']->db->quote($GLOBALS['egw']->db->to_timestamp((int)$end_date));
			}
			if(!is_null($start_date)) {
				$where[] = 'trans_date >= ' . $GLOBALS['egw']->db->quote($GLOBALS['egw']->db->to_timestamp((int)$start_date));
			}
			$GLOBALS['egw']->db->select(
				'perp_payroll_ledger',
				array(
					'sum(perp_payroll_ledger.amount) as amount'
				),
				$where,
				__LINE__, __FILE__, false,
				'GROUP BY type_id ' . (is_null($staff_id) ? '' : ', staff_id'),
				'perp_payroll', -1, 
				'JOIN perp_general_ledger ON 
					perp_general_ledger.trans_no = perp_payroll_ledger.transaction_id'
			);
			$balance = 0;
			if($row = $GLOBALS['egw']->db->row(true)) {
				$balance = $row['amount'];
			}
			return $balance;
		}

		public function __sleep() {
			return array_keys($this->toArray());
		}
	}
?>
