<?php

	/**
	*	Base class for a warranty claim.
	*	
	*	This class will handle all the basic stuff of a warranty claim, such as
	*	status handling and making sure we don't get ahead of ourselves.
	*
	*	@author Nathan Gray
	*	@package perp_warranty
	*/

	require_once( EGW_INCLUDE_ROOT . '/perp_api/inc/class.Invoicable.inc.php');

	class Claim implements Invoicable {

		/**
		*	Basic header information about the claim
		*	@{
		*/
		protected $claim_id = null;
		protected $registration_id = null;
		protected $client_id = 0;
		protected $supplier_id = 0;
		/**
		*	@}
		*/

		/**
		*	Dates and times for changes
		*	@{
		*/
		protected $created_date;
		protected $claim_date;
		protected $last_modified_date;
		protected $last_modified_by;
		/**
		*	@}
		*/

		/**
		*	Actual claim information
		*	@{
		*/
		protected $description = '';
		protected $technician = '';  // Who actually did the work

		protected $time_claim = 0; // minutes
		protected $travel_claimed = 0; // km

		// Problems with the product
		protected $issues = array();

		/**
		*	@}
		*/

		/**
		*	Resolution constants
		*
		*	These are actually assigned on a per-issue basis, but 
		*	this is where everyone is going to look for them
		*
		*	@{
		*/
		const UNRESOLVED	=	' ';
		const ACCEPTED		=	'a';
		const REJECTED		=	'r';
		const DUPLICATE		=	'd';	// Implies rejected
		const INVALID		=	'i';	// Implies rejected
		/**
		*	@}
		*/

		// Database table
		const DB_TABLE		=	'perp_warranty_claim';



		/**
		*	Constructor 
		*
		*	@param claim_id Load an existing claim OR an array of claim information
		*/
		public function __construct($claim_id = null) {
			
			// Initialize times to now.
			$created_date = time();
			$this->created_by = $GLOBALS['egw_info']['user']['account_id'];
			$claim_date = time();
			$last_modified_date = time();
			$last_modified_by = $GLOBALS['egw_info']['user']['account_id'];

			if($claim_id) {
				return $this->read($claim_id);
			}
		}

		/**
		*	Read a claim
		*
		*	Overloaded to be able to handle reading from the database, or an array
		*	of values to be set.
		*/
		public function read($read_from) {
			if(is_numeric($read_from)) {
				return $this->read_from_db($read_from);
			} elseif(is_array($read_from)) {
				return $this->read_from_array($read_from);
			}
		}

		/**
		*	Read an existing claim from the database
		*/
		protected function read_from_db($claim_id) {
			$claim = $GLOBALS['egw']->db->select(
				'perp_warranty_claim',
				array('*'),
				array('claim_id' => $claim_id),
				__LINE__, __FILE__, false,
				'', 'perp_warranty'
			)->fetch();

			$claim['created_date'] = $GLOBALS['egw']->db->from_timestamp($claim['created_date']);
			$claim['claim_date'] = $GLOBALS['egw']->db->from_timestamp($claim['claim_date']);
			$claim['last_modified_date'] = $GLOBALS['egw']->db->from_timestamp($claim['last_modified_date']);

			$query = array(
				'claim_id'	=>	$claim_id
			);
			bo_perp_claim_issue::get_claim_issues($query, $claim['issues']);

			// Add some more info
			foreach($claim['issues'] as $key => &$issue) {
				if(!is_array($issue)) {
					unset($claim['issues'][$key]);
					continue;
				}
				$issue = array_merge($issue, bo_perp_claim_issue::read($issue['claim_issue_id']));
			}

			$this->read_from_array($claim);
		}

		/**
		*	Initialize a new claim from an array of values
		*/
		protected function read_from_array($values) {
			if($values['claim_id']) {
				$this->claim_id = $values['claim_id'];
			}
			foreach($values as $key => $value) {
				$set_function = "set_$key";
				if(method_exists($this, $set_function)) {
					$this->$set_function($value);
				}
			}
			if($values['issues'] && is_array($values['issues'])) {
				foreach($values['issues'] as $issue_array) {
					$this->add_new_issue($issue_array);
				}
			}
		}

		/**
		*	Write a claim to the database
		*
		*	This will update the claim, all its sub-bits, and any associated tracker entries
		*/
		public function write() {
			$GLOBALS['egw']->db->transaction_begin();
	
			// Write claim
			$GLOBALS['egw']->db->insert(self::DB_TABLE,
				array(
					'claim_id'		=>	$this->claim_id,
					'registration_id'	=>	$this->registration_id,
					'client_id'		=>	$this->client_id,
					'supplier_id'		=>	$this->supplier_id,
					'created_date'		=>	$GLOBALS['egw']->db->to_timestamp($this->created_date),
					'claim_date'		=>	$GLOBALS['egw']->db->to_timestamp($this->claim_date),
					'last_modified_date'	=>	$GLOBALS['egw']->db->to_timestamp(time()),
					'created_by'		=>	$this->created_by,
					'modified_by'		=>	$GLOBALS['egw_info']['user']['account_id'],
					'time_claimed'		=>	$this->time_claimed,
					'travel_claimed'	=>	$this->travel_claimed,
					'description'		=>	$this->description,
					'technician'		=>	$this->technician,
					'replacement_order_id'	=>	$this->replacement_order_id
				),
				array(
					'claim_id'	=>	$this->claim_id
				),
				__LINE__, __FILE__, 'perp_warranty'
			);
			if(!$this->claim_id) {
				$this->claim_id = $GLOBALS['egw']->db->get_last_insert_id(self::DB_TABLE, 'claim_id');
			}

			// Write issues
			foreach($this->issues as $issue) {
				$issue->write();
			}

			$GLOBALS['egw']->db->transaction_commit();

			return $this->claim_id;
		}

		/**
		*	Add a new issue using an array of information
		*/
		public function add_new_issue($issue_array) {
			$issue_array['claim'] =& $this;
			$issue = new ClaimIssue($issue_array);
			$this->add_issue($issue);
		}
		
		/**
		*	Add a new ClaimIssue
		*/
		public function add_issue(ClaimIssue $issue) {
			$this->issues[] = $issue;
		}

		/**
		*	Remove an issue
		*
		*	@param claim_issue_id Identification number of the issue
		*/
		public function remove_issue($claim_issue_id) {
			foreach($this->issues as $key => $issue) {
				if($issue->get_claim_issue_id() == $claim_issue_id) {
					unset($this->issues[$key]);
					$issue->remove();
				}
			}
		}

		/**
		*	Returns all issues on this claim, by reference
		*/
		public function &get_issues() {
			return $this->issues;
		}

		/**
		*	Get a single issue, by claim_issue_id
		*
		*	@param claim_issue_id
		*/
		public function &get_issue($claim_issue_id) {
			foreach($this->issues as $key => &$issue) {
				if($issue->get_claim_issue_id() == $claim_issue_id) {
					return $issue;
				}
			}
		}

		/**
		*	Can this claim be edited?
		*
		*	If any of the issues have been resolved, then no.  Otherwise, yes.
		*	This doesn't mean that the issues themselves cannot be edited, that's
		*	seperate.  Its possible to resolve one issue, and still work on another.
		*/
		public function editable() {
			foreach($this->issues as $issue) {
				if($issue->get_resolution != self::UNRESOLVED) {
					return false;
				}
			}
			return true;
		}

		/**
		*	Accessors
		*	@{
		*/

		public function get_claim_id() {
			return $this->claim_id;
		}


		/**
		*	Get who the product is registered to, if at all
		*
		*	@return int
		*/
		public function get_registration_id() {
			return $this->registration_id;
		}
		/**
		*	Set the registration id for this claim
		*
		*	@param registration ID of the registration info
		*	@throws claim_not_editable_exception if you try to change it when you're not allowed
		*/
		public function set_registration_id($registration) {
			if($this->editable()) {
				$this->registration_id = $registration;
			} else {
				throw new claim_not_editable_exception();
			}
		}
		
		/**
		*	Who the product was sold to 
		*/
		public function get_client_id() {
			return $this->client_id;
		}

		/**
		*	Set which client the product was sold to
		*	@throws claim_not_editable_exception if you try to change it when you're not allowed
		*/
		public function set_client_id($client_id) {
			if($this->editable()) {
				$this->client_id = $client_id;
			} else {
				throw new claim_not_editable_exception();
			}
		}

		/**
		*	Who the product was purchased from.
		*
		*	Actually, who we're making the claim against...
		*/
		public function get_supplier_id() {
			return $this->supplier_id;
		}

		/**
		*	Set who we're making the claim against
		*
		*	@param supplier_id
		*	@throws claim_not_editable_exception if you try to change it when you're not allowed
		*/
		public function set_supplier_id($supplier_id) {
			if($this->editable()) {
				$this->supplier_id = $supplier_id;
			} else {
				throw new claim_not_editable_exception();
			}
		}

		/**
		*	Get the date of the claim.
		*
		*	This is the date that should be used when deciding if it's too late.
		*
		*	@return date, in timestamp format
		*/
		public function get_claim_date() {
			return $this->claim_date;
		}

		/**
		*	Set the date of the claim.  This may not be the same as the creation date, if
		*	people are a little slow entering things...
		*
		*	@param date Date of the claim.
		*	@throws claim_not_editable_exception if you try to change it when you're not allowed
		*/
		public function set_claim_date($date) {
			if($this->editable()) {
				$this->claim_date = $date;
			} else {
				throw new claim_not_editable_exception();
			}
		}

		/**
		*	Get the last time this claim was changed.
		*
		*	@return date, in timestamp format
		*/
		public function get_last_modified_date() {
			return $this->last_modified_date;
		}

		/**
		*	Get the total time claimed
		*
		*	This is the total time across all issues that compensation is being asked for.
		*	We don't break it down by issue, because they're not going to keep accurate track
		*	of which repairs took how long.  
		*
		*	@return integer, time in minutes
		*/
		public function get_time_claimed() {
			return $this->time_claimed;
		}

		/**
		*	Set how much time compensation is requested for
		*
		*	@param time in minutes
		*	@throws claim_not_editable_exception if you try to change it when you're not allowed
		*/
		public function set_time_claimed($time) {
			if($this->editable()) {
				$this->time_claimed = $time;
			} else {
				throw new claim_not_editable_exception();
			}
		}

		/**
		*	Get the travel distance claimed
		*
		*	This is the distance in km compensation is being asked for.
		*
		*	@return integer, travel distance in km
		*/
		public function get_travel_claimed() {
			return $this->travel_claimed;
		}

		/**
		*	Set how much travel compensation is requested for
		*
		*	@param distance in km
		*	@throws claim_not_editable_exception if you try to change it when you're not allowed
		*/
		public function set_travel_claimed($distance) {
			if($this->editable()) {
				$this->travel_claimed = $distance;
			} else {
				throw new claim_not_editable_exception();
			}
		}

		/**
		*	Get the description of the problem
		*
		*	This is an overall problem, and may not be applicable if there are several issues
		*	being dealt with in one claim.
		*
		*	@return string, description of the problem
		*/
		public function get_description() {
			return $this->description;
		}

		/**
		*	Set the description of the problem
		*
		*	This is an overall problem, and may not be applicable if there are several issues
		*	being dealt with in one claim.
		*
		*	@param description A description of the overall problem, if there is one 
		*	@throws claim_not_editable_exception if you try to change it when you're not allowed
		*/
		public function set_description($description) {
			if($this->editable()) {
				$this->description = trim($description);
			} else {
				throw new claim_not_editable_exception();
			}
		}

		/**
		*	Get the name of the technician that dealt with the problem
		*
		*	@return string name of a person
		*/
		public function get_technician() {
			return $this->technician;
		}

		/**
		*	Set the name of the technician that dealt with the problem
		*
		*	@param name The name of the person that dealt with the claim
		*	@throws claim_not_editable_exception if you try to change it when you're not allowed
		*/
		public function set_technician($name) {
			if($this->editable()) {
				$this->technician = $name;
			} else {
				throw new claim_not_editable_exception();
			}
		}

		/**
		*	If this claim is for multiple issues on a single item,
		*	get the ClaimIssue that holds that item.  This ClaimIssue
		*	is usually otherwise hidden.
		*/
		public function get_parent() {
			foreach($this->issues as $issue) {
				if($issue->get_issue_id() == bo_perp_issue::PARENT) {
					return $issue;
				}
			}
			return null;
		}

		/**
		*	@}
		*	End accessors
		*/

		public function get_our_contact_person() {
echo __METHOD__ . ' not implemented at ' . __FILE__ . ':' . __LINE__ . '<br />';
		}
		public function get_their_contact_person() {
echo __METHOD__ . ' not implemented at ' . __FILE__ . ':' . __LINE__ . '<br />';
		}
		public function get_bill_from_address() {
			if($this->supplier_id) {
				$supplier = ExecMethod('perp_ap.bo_perp_supplier.get_supplier', $this->supplier_id);
				return $supplier['supplier_address'];
			} elseif ($this->client_id) {
				$config = config::read('perp_warranty');
				return $config['warranty_address'];
			}
		}
		public function get_bill_to_address() {
			if($this->client_id) {
				$client = ExecMethod('perp_ar.bo_perp_client.get_client', $this->client_id);
				return $client['address'];
			} elseif ($this->supplier_id) {
				$config = config::read('perp_warranty');
				return $config['warranty_address'];
			}
		}
		public function get_invoice_currency() {
echo __METHOD__ . ' not implemented at ' . __FILE__ . ':' . __LINE__ . '<br />';
		}
		public function get_payment_terms() {
echo __METHOD__ . ' not implemented at ' . __FILE__ . ':' . __LINE__ . '<br />';
		}
		
		/**
		*	Get a list of items to go on an invoice
		*/
		public function &get_invoicable_items() {
			$config = config::read('perp_warranty');

			$items = array();

			// Add time
			$items[] = new ClaimInvoicableItem(array(
				'claim_id'		=>	$this->get_claim_id(),
				'claim_issue_id'	=>	0,
				'piece_key'		=>	ClaimInvoicableItem::TIME
			));

			// Add travel
			$items[] = new ClaimInvoicableItem(array(
				'claim_id'		=>	$this->get_claim_id(),
				'claim_issue_id'	=>	0,
				'piece_key'		=>	ClaimInvoicableItem::TRAVEL
			));

			// Get list of issues
			$query = array(
				'claim_id'	=>	$this->get_claim_id(),
				'num_rows'	=>	-1,
			);
			bo_perp_claim_issue::get_claim_issues($query, $issues);
			foreach($issues as &$issue) {
				if($issue['issue_id'] && $issue['resolution'] == Claim::ACCEPTED) {

					// Flat fee
					$items[] = new ClaimInvoicableItem(array(
						'claim_id'		=>	$this->get_claim_id(),
						'claim_issue_id'	=>	$issue['claim_issue_id'],
						'piece_key'		=>	ClaimInvoicableItem::ISSUE
					));

					$issue['components'] = bo_perp_claim_issue::get_components($issue);

					foreach($issue['components'] as $component) {
						$items[] = new ClaimInvoicableItem(array(
							'claim_id'		=>	$this->get_claim_id(),
							'claim_issue_id'	=>	$issue['claim_issue_id'],
							'piece_key'		=>	$component['stock_id'],
						));
					}
				}
			}

			if($config['exclude_0_from_credit']) {
				foreach($items as $key => &$item) {
					if($item->get_unit_price() == 0) {
						unset($items[$key]);
					}
				}
			}

			return $items;
		}

		/**
		*	Utilities
		*	@{
		*/
	
		/**
		*	Convert this object into string form.
		*
		*	Mostly useful for debugging.
		*
		*	@return String representation of this Claim
		*/
		public function __toString() {
			$string = __CLASS__ . ' ' . $this->get_claim_id() . "\n";
			$string .= sprintf("Client: %-20s Supplier: %-20s\n",
				$this->client_id ? ExecMethod('perp_ar.bo_perp_client.link_title', $this->client_id) : 'Us',
				$this->supplier_id ? ExecMethod('perp_ap.bo_perp_supplier.link_title', $this->supplier_id) : 'Us'
			);
			$string .= $this->description . "\n";
			$string .= sprintf("Tech: %-20.30s Time: %4d mins\tTravel: %4d km\n", 
				$this->technician,
				$this->time_claimed, 
				$this->travel_claimed
			);
 
			$string .= "Issues:\n";
			
			foreach($this->issues as $issue) {
				$string .= $issue->__toString() . "\n";
			}
			return $string;	
		}

		/**
		*	Convert this claim into an array, useful for display
		*/
		public function toArray() {
			$data = get_object_vars($this);
			unset($data['issues']);
			foreach($this->issues as $issue) {
				$data['issues'][] = $issue->toArray();
			}

			return $data;
		}

		/**
		*	@}
		*	End utilities
		*/
	}
?>
