<?php

	/**
	*	Map the parts of the claim that can generate a credit to
	*	an InvoicableItem.  invoicable_id is a unique system generated number,
	*	invoicable_item_id is the claim_id.  invoicable_item_id isn't really
	*	needed, but it's there so we'll stick something in it to avoid any 
	*	null issues.
	*
	*	@author Nathan Gray
	*	@ingroup perp_warranty
	*/

	require_once( EGW_INCLUDE_ROOT . '/perp_api/inc/class.InvoicableItem.inc.php');

	class ClaimInvoicableItem implements InvoicableItem {

		/**
		*	Warranty claims have multiple pieces that can generate a credit. 
		*	Time, Travel, Issue and component can all generate a credit, and
		*	linking to them, so we can mark them as invoiced, can be trouble.
		*	These keys are used so we can map to what is actually referred to.
		*	Components use their stock_id.
		*/
		const TIME	=	-1;
		const TRAVEL	=	-2;
		const ISSUE	=	-3;

		protected $invoicable_id;

		// invoicable_item_id
		protected $claim_id;
	
		protected $claim_issue_id;
		protected $piece_key;

		// Convenience
		protected $claim;

		public function __construct($element_array) {
			if($element_array['invoicable_id']) {
				$this->read($element_array['invoicable_id']);
			} else {
				$this->claim_id = $element_array['claim_id'];
				$this->claim_issue_id = $element_array['claim_issue_id'];
				$this->piece_key = $element_array['piece_key'];
			}
			$this->claim =& new Claim($this->claim_id);
		}

		/**
		*	Read an already created ClaimInvoicableItem from the database
		*/
		protected function read($invoicable_id) {
			$row = $GLOBALS['egw']->db->select(
				array(
					'invoicable_id',
					'claim_id',
					'claim_issue_id',
					'piece_key'
				),
				array(
					'invoicable_id'	=>	$this->invoicable_id
				),
				__LINE__, __FILE__, false,
				'', 'perp_warranty', 1
			)->fetch();
			$this->invoicable_id	=	$row['invoicable_id'];
			$this->claim_id		=	$row['claim_id'];
			$this->claim_issue_id	=	$row['claim_issue_id'];
			$this->piece_key	=	$row['piece_key'];
		}

		/**
		*	Mark this InvoicableItem as invoiced.  The InvoiceLine will take care of
		*	any General Ledger transactions, but this object should take care of
		*	any recordkeeping needed to mark itself as invoiced, and prevent 
		*	changes if needed.
		*
		*	@param invoice_line	InvoiceLine that is part of the Invoice
		*/
		public function invoice(InvoiceLine &$invoice_line) {
	
			$GLOBALS['egw']->db->transaction_begin();
			$GLOBALS['egw']->db->insert(
				'perp_warranty_credit',
				array(
					'invoicable_id'		=>	$this->invoicable_id,
					'claim_id'		=>	$this->claim_id,
					'claim_issue_id'	=>	$this->claim_issue_id,
					'piece_key'		=>	$this->piece_key
				),
				array(
					'invoicable_id'		=>	$this->invoicable_id,
				),
				__LINE__, __FILE__,
				'perp_warranty'
			);
			if(!$this->invoicable_id) {
                                $this->invoicable_id = $GLOBALS['egw']->db->get_last_insert_id('perp_warranty_credit', 'invoicable_id');

				/*
				Because we don't know the invoicable ID on the first save, we
				may have to go back and update the invoice line record.
				*/
				$GLOBALS['egw']->db->update(
					InvoiceLine::invoice_line_table,
					array(
						'invoicable_id'	=>	$this->invoicable_id
					),
					array(
						'invoice_id'	=>	$invoice_line->get_invoice()->get_invoice_id(),
						'line_no'	=>	$invoice_line->get_line_number()
					),
					__LINE__, __FILE__, 'perp_api'
				);
                        }
			$GLOBALS['egw']->db->transaction_commit();
		}

		/**
		*	Return the Invoicable object that this InvoicableItem is part of
		*
		*	@return Invoicable
		*/
		public function get_owner() {
			return $this->claim;
		}

		/**
		*	Return a unique identifier for the owner of this InvoicableItem
		*	This is usually the parent object's database key.
		*
		*	@return int
		*/
		public function get_invoicable_id() {
			return $this->invoicable_id;
		}

		/**
		*	Return a unique (amongst the owning object's InvoicableItems)
		*
		*	@return int identifier for this InvoicableItem 
		*/
		public function get_invoicable_line_id() {
			return $this->claim_id;
		}

		/**
		*	Check that this InvoicableItem can be invoiced.  It can't be
		*	invoiced if it has already been completely invoiced.
		*
		*	@throws Exception with error message if it can't be invoiced.
		*	@return boolean 
		*/
		public function can_invoice() {
			return is_null(bo_perp_claim::get_claim_invoice_id($this->claim_id));
		}

		/**
		*	Get the quantity to be invoiced, in the supplier's units of measure.
		*	
		*	@return float quantity to be invoiced
		*/
		public function get_invoice_quantity() {
			if(!$this->claim_issue_id) {
				switch($this->piece_key) {
					case self::TIME:
						return bcdiv($this->claim->get_time_claimed(), 60, 1);
						break;
					case self::TRAVEL:
						return $this->claim->get_travel_claimed();
						break;
				}
			} else {
				$issue =& new ClaimIssue($this->claim_issue_id);
				if($this->piece_key == self::ISSUE) {
					return $issue->get_claim_amount();
				} else {
					// Stock ID
					return $issue->get_component_by_stock_id($this->piece_key)->get_quantity_claimed();
				}
			}
			return 0;
		}

		/**
		*	Get unit price for one unit of the InvoicableItem being invoiced.
		*	This is in the supplier's units of measure.
		*
		*	@return float the price of one unit 
		*/
		public function get_unit_price() {
			if($this->claim->get_client_id()) {
				$settings = bo_perp_client_warranty::get_client_settings($this->claim->get_client_id());
			}

			switch($this->piece_key) {
				case self::TIME:
					$amount = $settings['claim']['time_rate'];
					break;
				case self::TRAVEL:
					$amount = $settings['claim']['travel_rate'];
					break;
				case self::ISSUE:
					$issue = new ClaimIssue($this->claim_issue_id);
					$amount = $issue->get_claim_amount();
					break;
				default:
					$issue = new ClaimIssue($this->claim_issue_id);
					$component = $issue->get_component_by_stock_id($this->piece_key);
					$amount = $component->get_credit();
					break;
			}

			// Multiply by the overall rate
			if(is_array($settings)) {
				$amount *= bcdiv($settings['claim']['credit_rate'], 100, 2);
			}

			// Negate, because this is a credit
			return -$amount;
		}

		/**
		*	Get the description for the InvoicableItem
		*
		*	@return string description
		*/
		public function get_description() {
			switch($this->piece_key) {
				case self::TIME:
					return lang('Time credit for claim %1', $this->claim_id);
					break;
				case self::TRAVEL:
					return lang('Travel credit for claim %1', $this->claim_id);
					break;
				case self::ISSUE:
					$issue = new ClaimIssue($this->claim_issue_id);
					return lang('Flat credit for %1', bo_perp_issue::link_title($issue->get_issue_id()));
					break;
				default:
					return ExecMethod('perp_inventory.bo_perp_stock.link_title', $this->piece_key);
					break;
			}
		}

		/**
		*	Get the general ledger account to be used on the invoice.
		*	May return null if the gl account is not known, in this case
		*	a note will be added to the invoice's comments section.
		*
		*	Implementing classes really should provide an account number.
		*	Methods of determining the account include a custom field, 
		*	but the best is using the config.
		*
		*	@return int general ledger account number
		*/
		public function get_gl_account() {
			$config = config::read('perp_warranty');
			switch($this->piece_key) {
				case self::TIME:
					return $config['time_account'];
					break;
				case self::TRAVEL:
					return $config['travel_account'];
					break;
				case self::ISSUE:
					return $config['issue_account'];
					break;
				default:
					return $config['part_account'];
					break;
			}

		}

		/**
		*	Return an array that will work for an eGroupWare style link.
		*	This is for allowing cross linking and jumping around.
		*	The array contain a sub-array, with keys app, id:
		*	return array(
		*		array(
		*			'app'	=>	$GLOBALS['egw_info']['flags']['currentapp'],
		*			'id'	=>	'CLAIM-' . $this->get_owner->get__id()
		*		)
		*	);
		*
		*	@return array
		*/
		public function get_link() {
			return array(
				array(
					'app'	=>	'perp_warranty',
					'id'	=>	'CLAIM-' . $this->claim_id
				)
			);
		}

		/**
		*	Return a list of extra details to be displayed on the invoice, below
		*	the line.  This list cannot be edited, and is an array of details in
		*	the form ('name' => 'Detail name', 'value' => '123')
		*
		*	@param filter The filter parameter from the display select (usually true)
		*	@return array
		*/
		public function get_invoice_details($filter) {
		}

		public function toArray() {
			return $this->claim->toArray() + array(
				'link'	=>	$this->get_link()
			);
		}
	}
?>
