<?php

	/**
	*	An issue on a claim.  A single claim can include only one issue (20 broken keyboards),
	*	or several issues (a computer with bad RAM, dead HD and a broken keyboard).
	*
	*	@author Nathan Gray
	*	@package perp_warranty
	*/

	class ClaimIssue {
		
		/**
		*	System-wide unique issue identifier
		*/
		protected $claim_issue_id;

		/**
		*	Claim this issue is part of
		*/
		protected $claim;

		/**
		*	Which issue are we having?
		*/
		protected $issue_id = bo_perp_issue::OTHER;

		/**
		*	Tracker ID, if this issue needs a tracker.
		*
		*	If a claim is for multiple issues on a single product, the 
		*	product with the problems is stored in a ClaimIssue with no
		*	tracker ID or issue ID
		*/
		protected $tracker_id;

		protected $tracker;

		/**
		*	How much money?
		*/
		protected $claim_amount;

		/**
		*	Components that are damaged, and may be returned
		*/
		protected $components = array();

		/**
		*	Components that are being replaced
		*/
		protected $replacements = array();

		/**
		*	Create a new ClaimIssue
		*
		*	@param claim_issue_id Unique ID for this claim to read it, or an array
		*		of information to add a new issue to a claim.
		*/
		public function __construct($claim_issue_id) {
			if($claim_issue_id) {
				return $this->read($claim_issue_id);
			}
		}

		/**
		*	Read a ClaimIssue
		*/
		public function read($read_from) {
			if(is_array($read_from)) {
				$this->read_from_array($read_from);
			} elseif (is_numeric($read_from)) {
				$this->read_from_db($read_from);
			}
		}
		
		/**
		*	Initialized a new ClaimIssue from an array of values
		*/
		protected function read_from_array(Array $values) {
			if($values['claim'] instanceof Claim) {
				$this->claim =& $values['claim'];
			}
			if($values['claim_issue_id']) {
				$this->claim_issue_id = $values['claim_issue_id'];
			}
			if($values['tracker_id']) {
				$this->tracker_id = $values['tracker_id'];
			}


			foreach($values as $key => $value) {
				$set_function = "set_$key";
				if(method_exists($this, $set_function)) {
					$this->$set_function($value);
				}
			}
			if($values['components'] && is_array($values['components'])) {
				foreach($values['components'] as $component) {
					$this->add_new_component($component);
				}
			}
			if($values['replacements'] && is_array($values['replacements'])) {
				foreach($values['replacements'] as $replacement) {
					$this->add_new_replacement($replacement);
				}
			}
		}


		/**
		*	Read an existing ClaimIssue from the database
		*/
		protected function read_from_db($values) {
			$this->read_from_array(bo_perp_claim_issue::read($values));
		}

		/**
		*	Write this issue to the database
		*/
		public function write() {
			$this->claim_issue_id = bo_perp_claim_issue::write($this);
		}

		/**
		*	Add a new component that has this issue
		*/
		public function add_new_component($values) {
			$values['parent'] =& $this;

			$component =& new Component($values);
			$this->add_component($component);
		}

		/**
		*	Add a replacement component
		*/
		public function add_new_replacement($values) {
			$values['parent'] =& $this;
			$replacement =& new Replacement($values);
			$this->add_replacement($replacement);
		}

		/**
		*	Add a new component
		*
		*	@param c Component to add
		*/
		public function add_component(Component $c) {
			$this->components[] = $c;
		}

		/**
		*	Retrieve a component by its stock_id
		*/
		public function &get_component_by_stock_id($stock_id) {
			$component = null;
			foreach($this->components as $key => &$value) {
				if($value->get_stock_id() == $stock_id) {
					return $value;
				}
			}
			return false;
		}

		/**
		*	Remove a component
		*/
		public function remove_component(Component $c) {
			foreach($this->components as $key => $value) {
				if($value == $c) {
					unset($this->components[$key]);
				}
			}
		}

		/**
		*	Add a new replacement
		*
		*	@param r Replacement to add
		*/
		public function add_replacement(Replacement $c) {
			$this->replacements[] = $r;
		}

		/**
		*	Remove a replacement component
		*/
		public function remove_replacment(Replacement $r) {
			foreach($this->replacements as $key => $value) {
				if($value == $r) {
					unset($this->replacements[$key]);
				}
			}
		}

		/**
		*	Accessors
		*	@{
		*/
		
		/**
		*	Get the ID for this object
		*
		*	@return integer
		*/
		public function get_claim_issue_id() {
			return $this->claim_issue_id;
		}

		/**
		*	Set the Claim that this issue is part of
		*
		*	@param claim_id 
		*/
		public function set_claim_id($claim_id) {
			$this->claim_id = $claim_id;
		}

		/**
		*	Get the ID of the claim that this issue is part of
		*
		*	@return integer
		*/
		public function get_claim_id() {
			if(!$this->claim_id) {
				$this->set_claim_id($this->get_claim()->get_claim_id());
			}
			return $this->claim_id;
		}

		/**
		*	Get the Claim that this issue is part of
		*
		*	@return Claim
		*/
		public function &get_claim() {
			if(!($this->claim) || !($this->claim instanceof Claim)) {
				$this->claim =& new Claim($this->claim_id);
				if($this->claim->get_client_id()) {
					$this->claim =& new ClientClaim($this->claim_id);
				}
			}
			return $this->claim;
		}

		/**
		*	Set the issue that this ClaimIssue deals with
		*
		*	@param issue_id
		*/
		public function set_issue_id($issue_id) {
			$this->issue_id = $issue_id;
		}

		/**
		*	Get the ID of the issue
		*
		*	@return integer
		*/
		public function get_issue_id() {
			return $this->issue_id;
		}
		
		/**
		*	Get the ID of the tracker that is used for this claim issue
		*
		*	@return int
		*/
		public function get_tracker_id() {
			if(!$this->tracker_id) {
				$links = egw_link::get_links(
					'perp_warranty', 'CLAIM_ISSUE-' . $this->get_claim_issue_id(),
					'tracker'
				);
				if(is_array($links)) {
					$this->tracker_id = current($links);
				}
			}
			return $this->tracker_id;
		}

		/**
		*	Get the amount of money (in GL currency) that will be credited
		*	if this part of the claim is accepted
		*
		*	@return double
		*/
		public function get_claim_amount() {
			return $this->claim_amount;
		}

		/**
		*	Set the amount of money (in GL currency) that will be credited
		*	if this part of the claim is accepted
		*
		*	@param amount to claim
		*/
		public function set_claim_amount($amount) {
			$this->claim_amount = $amount;
		}

		/**
		*	Get the components
		*/
		public function get_components() {
			return $this->components;
		}

		/**
		*	Get the replacements
		*/
		public function get_replacements() {
			return $this->replacements;
		}

		/**
		*	@}
		*	End accessors
		*/
		
		/**
		*	Utilities
		*	@{
		*/
		
		/**
		*	Get a string reperesentation of this object.
		*	Useful for debugging.
		*
		*	@return String
		*/
		public function __toString() {
			$string = sprintf("%4d. %-30s\n", 
				$this->claim_issue_id,
				bo_perp_issue::link_title($this->issue_id)
			);
			$components = array('Components');
			foreach($this->components as $component) {
				$components[] = $component->__toString();
			}
			$replacements = array('Replacements');
			foreach($this->replacements as $replacement) {
				$replacements[] = $replacement->__toString();
			}
			for($i = 0; $i < max(count($components), count($replacements)); $i++) {
				$string .= sprintf("\t%-35.35s\t%-35.35s\n", $components[$i], $replacements[$i]);
			}

			return $string;
		}

		/**
		*	Convert this claim into an array, useful for display
		*/
		public function toArray() {
			$data = get_object_vars($this);
			foreach(array('components', 'replacements') as $parts) {
				unset($data[$parts]);
				$data[$parts] = array();
				foreach($this->$parts as &$part) {
					$data[$parts][] = $part->toArray();
				}
			}
			return $data;
		}

		/**
		*	@}
		*	End utilities
		*/
	}
?>
