<?php
	/**
	*	Container for damaged components that caused a warranty claim
	*
	*	Components are tied to a ClaimIssue, and may have a Replacement.
	*/
	class Component {

		protected $parent;

		protected $stock_id;	// ID of the damaged component
	
		// How many are damaged
		protected $quantity_claimed;	

		protected $serial_numbers = array();

		// Do we want the parts back?
		protected $return_parts = false;
	
		// Will we send replacement parts?
		protected $replace_parts = false;

		// Are retrofits allowed? (?)
		protected $retrofits_allowed = false;

		// Do we need pictures?
		protected $image_required = false;

		// How much will we give for this component?
		protected $credit = 0;

		// Cache of default settings
		public static $issue_settings = array();

		public function __construct($values) {
			$this->parent =& $values['parent'];

			// Set defaults from the associated issue
			$issue = bo_perp_issue::read($this->parent->get_issue_id());

			if(!$values['stock_category']) {
				$stock_item = ExecMethod('perp_inventory.bo_perp_stock.get_stock_item', $values['stock_id']);
				$values['stock_category'] = $stock_item['stock_category'];
			}

			// Get settings for this stock category, index by category and cache
			if(!self::$issue_settings[$values['stock_category']]) {
				$category_settings = bo_perp_issue::get_stock_categories($this->parent->get_issue_id(), true);
				foreach($category_settings as $category) {
					self::$issue_settings[$category['stock_category_id']] = $category;
				}
			}
			self::set_default('return_parts', $values, self::$issue_settings[$values['stock_category']]);
			self::set_default('replace_parts', $values, self::$issue_settings[$values['stock_category']]);
			self::set_default('retrofits_allowed', $values, self::$issue_settings[$values['stock_category']]);
			self::set_default('image_required', $values, self::$issue_settings[$values['stock_category']]);

			// Set values
			$this->stock_id			=	$values['stock_id'];
			$this->quantity_claimed		=	$values['quantity_claimed'];
			$this->serial_numbers		=	$values['serial_numbers'];

			$this->return_parts 		=	$values['return_parts'];
			$this->replace_parts		=	$values['replace_parts'];
			$this->retrofits_allowed	=	$values['retrofits_allowed'];
			$this->image_required		=	$values['image_required'];
			$this->credit			=	$values['credit'];
		}

		/**
		*	Default parameters to what is set in the issue for the stock category for new components
		*/
		private static function set_default($param_name, &$value, &$stock_category) {
			if(isset($value[$param_name])) {
				return;
			}
			if(in_array($stock_category[$param_name], array(bo_perp_issue::ALWAYS, bo_perp_issue::DEFAULT_YES))) {
				$value[$param_name] = true;
			} else {
				$value[$param_name] = false;
			}
		}

		/**
		*	Accessors
		*	@{
		*/
		public function &get_parent() {
			return $this->parent;
		}
		public function get_stock_id() {
			return $this->stock_id;
		}
		public function get_quantity_claimed() {
			return $this->quantity_claimed;
		}

		/**
		*	Set the quantity of a component that was damaged by the issue
		*
		*	@param quantity
		*/
		public function set_quantity_claimed($quantity) {
			if($this->check_status()) {
				$this->quantity_claimed = $quantity;
			} else {
				return false;
			}
		}

		public function set_return_parts($return) {
			$this->return_parts = $return;
		}

		public function set_replace_parts($replace) {
			$this->replace_parts = $replace;
		}

		/**
		*	Set the amount of credit for this component
		*/
		public function set_credit($credit) {
			if($this->check_status()) {
				$this->credit = $credit;
			} else {
				return false;
			}
		}
		public function get_credit() {
			return $this->credit;
		}

		public function get_serial_numbers() {
			return $this->serial_numbers;
		}
		
		/**
		*	Set the serial numbers for this component.
		*
		*	This is where we check to see if this is a valid serial number.
		*
		*	@param serial_numbers Array of (SN, quantity, replacement)
		*
		*	@param boolean success
		*	@throws egw_exception with a descriptive error message, including which serial
		*		number was bad.
		*/
		public function set_serial_numbers(Array $serial_numbers) {
			$this->serial_numbers = array();
			$count = 0;
			foreach($serial_numbers as $key => $values) {
				if(!is_array($values) || !isset($values['serial_number'])) {
					// Extra data that shouldn't have been passed
					continue;
				}
				$result = bo_perp_stock::check_serial($this->stock_id, $values['serial_number']);
				switch($result) {
					case bo_perp_stock::NOT_CONTROLLED:
						// Vacuously true
						return true;
					case bo_perp_stock::SERIAL_OK:
						$this->serial_numbers[] = $values;
						$count += $values['quantity'];
						break;
					case bo_perp_stock::FORMAT_OK:
						// Pattern matches, that's OK if we're sending it back to a supplier
						$this->serial_numbers[] = $values;
						$count += $values['quantity'];
						break;
					default:
						// Bad serial, or no record of it
						throw new egw_exception_not_found(lang(
							'Unable to find a record of %1 with serial number %2', 
							ExecMethod('perp_inventory.bo_perp_stock.link_title', $this->stock_id),
							$values['serial_number']
						));
				}
			}

			// Check quantity
			$stock_item = ExecMethod('perp_inventory.bo_perp_stock.get_stock_item', $this->stock_id);
			if($count != $this->quantity_claimed && $stock_item['controlled']) {
				throw new egw_exception_wrong_parameter(lang(
					'Incorrect number of serial numbers for %3.  Claimed %1, provided %2', 
					$this->quantity_claimed, 
					$count,
					ExecMethod('perp_inventory.bo_perp_stock.link_title', $this->stock_id)
				));
			}
			return true;
		}

		public function get_sent_or_received() {
echo 'Not implemented: ' . __METHOD__ . ' ' . __FILE__ . ':' . __LINE__ . '<br />';
			return false;
		}

		/**
		*	@}
		*	End accessors
		*/
		/**
		*	Utilities
		*	@{
		*/

		/**
		*	Check the status of the parent claim issue to see if editing is allowed
		*
		*/
		protected function check_status() {
			return true;
		}

		public function __toString() {
			return sprintf("%2d %-25s", $this->quantity_claimed, ExecMethod('perp_inventory.bo_perp_stock.link_title', $this->stock_id));
		}

		public function toArray() {
			$data = get_object_vars($this);
			unset($data['parent']);
			return $data;
		}
	}
?>
