<?php
	/**
	*	Data holder (model) for Return Materials Authorization, a list of goods
	*	or components [on a warranty claim] that have been authorized for return by a supplier.  
	*	
	*	This includes returns from clients to us, as well as from us to suppliers.
	*
	*	@package perp_warranty
	*	@author Nathan Gray
	*/

	require_once ( EGW_INCLUDE_ROOT . '/perp_api/inc/class.Order.inc.php' );
	require_once ( EGW_INCLUDE_ROOT . '/perp_warranty/inc/class.RMALine.inc.php');

	class RMA extends Order {

		// Claim this RMA is from
		protected $claim_id;

		public function __construct($rma_id = null) {
			if(is_array($rma_id)) {
				parent::__construct();
				$this->read_from_array($rma_id);
			} else {
				parent::__construct($rma_id);
			}
		}

		public function read($read_from) {
			if(is_array($read_from)) {
				$this->read_from_array($read_from);
			} elseif (is_numeric($read_from)) {
				$this->read_from_db($read_from);
			}
		}

		protected function read_from_db($rma_id) {
			parent::read($rma_id);
			$rma = $GLOBALS['egw']->db->select(
                                'perp_rma',
                                array(
                                        '*',
                                ),
                                array(
                                        'rma_id'        =>      $rma_id
                                ),
                                __LINE__, __FILE__, false,
                                '', 'perp_warranty'
                        )->fetch();

			$this->read_from_array($rma);
		}

		protected function read_from_array(Array $values) {
			if($values['claim_id']) {
				$this->claim_id = $values['claim_id'];
			}

			foreach($values as $key => $value) {
				$set_function = "set_$key";
				if(method_exists($this, $set_function)) {
					$this->$set_function($value);
				}
			}
		}

		/**
		*	Write this RMA to the DB
		*/
		public function write() {
			$GLOBALS['egw']->db->transaction_begin();
			parent::write();

			// Insert RMA
                        $GLOBALS['egw']->db->insert(
                                'perp_rma',
                                array(
                                        'rma_id'        =>      $this->get_rma_id(),
                                        'claim_id'      =>      $this->claim_id
                                ),
                                array(
                                        'rma_id'        =>      $this->get_rma_id()
                                ),
                                __LINE__, __FILE__, 'perp_warranty'
                        );

			$GLOBALS['egw']->db->transaction_commit();

			return $this->get_rma_id();
		}

		/**
		*	Add a new item to this RMA
		*
		*	@param values
		*/
		public function add_new_line(Array $values) {
			$line =& new RMALine($this, $values);
			$this->add_line($line);
		}

		/**
		*	Get a line based on stock_id
		*/
		public function &get_line_by_stock_id($stock_id) {
			foreach($this->lines as &$line) {
				if($line->get_stock_id() == $stock_id) {
					return $line;
				}
			}
			return null;
		}

		/**
		*	Get the RMA ID
		*/
		public function get_rma_id() {
			return $this->order_id;
		}

		/**
		*	Get the claim ID
		*/
		public function get_claim_id() {
			return $this->claim_id;
		}

		/**
		*	Access and permissions
		*	@{
		*/

		/**
		*	Get a list of users allowed to administer this Order
		*/
		public function get_admin_list() {
			$users = array();
			if(!$this->requires_approval) {
				$users[$this->get_originating_user()] = $GLOBALS['egw']->accounts->id2name($this->get_originating_user());
			}
			$approvers = $GLOBALS['egw']->acl->get_ids_for_location('rma', EGW_ACL_CUSTOM_1);
			if(!is_array($approvers)) {
				return $users;
			}
			foreach($approvers as $id) {
				if($id != 0) {
					$users[$id] = $GLOBALS['egw']->accounts->id2name($id);
				}
				if($id < 0) {
					$users += $GLOBALS['egw']->accounts->members($id);
				}
			}
			return $users;
		}

		/**
		*	Can this RMA be shipped?  Only if it's leaving us
		*/
		public function can_ship() {
			if($this->get_client_id() == 0) {
				return parent::can_ship();
			} else {
				return false;
			}
		}

		/**
		*	Can this RMA be received?  Only if it's coming to us, and hasn't arrived
		*	Override from parent to skip ACL check
		*/
		public function can_receive() {
			if($this->get_supplier_id() == 0) {
				$test = (count($this->order_lines) > 0);
				foreach($this->order_lines as $line) {
					$test = ($test && $line->can_receive());
				}

				// Check for approval
				$test = $test && (($this->requires_approval && !is_null($this->get_approval_date())) || !$this->requires_approval);

				// Can't receive if it already arrived
				$test = $test && !($this->actual_delivery_date);
				return $test;
			} else {
				return false;
			}

		}

		/**
		*	Is this RMA completed (closed)?
		*
		*	@return boolean
		*/
		public function is_complete() {
			$test = (count($this->order_lines) > 0) && ($this->closed_date != false);
			return $test;
		}

		/**
		*	Mark this RMA as received.
		*/
		public function receive($box_count = 1) {
			if($this->can_receive() && $GLOBALS['egw']->acl->check('rma', EGW_ACL_EDIT, 'perp_warranty')) {
				$this->set_actual_delivery_date(time());

				$this->set_comments(lang('Received in %1 box(es)', $box_count) . ($this->get_comments() ? "\n" + $this->get_comments() : ''));

				// Add in ActionDetails
				foreach($this->order_lines as &$line) {
					if(!$line->get_serial_numbers()) {
						$action_total = 0;
						$action_details =& $line->get_details('ActionDetail');
						foreach($action_details as &$detail) {
							$action_total += $detail->get_quantity();
						}
						if($action_total < $line->get_quantity_ordered()) {
							$new_detail =& new ActionDetail($line, array(
								'detail_name'	=>	$line->get_quantity_ordered() - $action_total,
							));
							$line->add_detail($new_detail);
						}
					} else {
						$serial_details =& $line->get_details('SerialNumberDetail');
						foreach($serial_details as &$detail) {
							if(!$detail->get_action()) {
								$new_detail =& new ActionDetail($line, array(
									'detail_name'	=>	'SN-'.$detail->get_number()
								));
								$line->add_detail($new_detail);
							}
						}
					}
				}

				$this->write();
				return true;
			} else {
				throw new egw_exception_no_permission(lang('Unable to receive Return Materials Authorization %1.', $this->get_our_reference_number()));
			}
			return false;
		}

		/**
		*	Set the close date, and then take the appropriate actions (put back, return to supplier, discard, etc)
		*	on the stock items
		*/
		public function set_closed_date($date) {
			if(!$this->closed_date) {
				parent::set_closed_date($date);
				
				// Go through the lines and deal with it
				foreach($this->order_lines as &$line) {
					$actions =& $line->get_details('ActionDetail');
					foreach($actions as &$detail) {
						$detail->perform_action();
					}
				}
			}
		}

		/**
		*	@}
		*/

		/**
		*       Utilities
		*       @{
		*/

		/**
		*	Get our location involved in this RMA.
		*	This will be either the source or destination,
		*	depending on the RMA.  It will be backwards from a regular Order
		*
		*	@return int Inventory location ID or false if it can't be found.
		*/
		public function get_location_id() {
			if($this->supplier_id != 0) {
				$address = $this->ship_from_address;
			} else {
				$address = $this->ship_to_address;
			}
			$links = egw_link::get_links('addressbook', $address, 'perp_inventory');
			$key = current($links);
			if($key) {
				list($location, $location_id) = explode('-', $key);
				return $location_id;
			}
			return false;
		}

		/**
		*	Cancel RMA, if possible
		*/
		public function cancel($user_id) {
			$GLOBALS['egw']->db->transaction_begin();

			parent::cancel($user_id);

			$GLOBALS['egw']->db->delete(
				'perp_rma',
				array(
					'rma_id'	=>	$this->get_rma_id()
				),
				__LINE__, __FILE__, 'perp_warranty'
			);

			$GLOBALS['egw']->db->transaction_commit();
		}

		/**
		*	Get a human reference number
		*/
		public function suggest_order_number() {
			if($this->get_supplier_id()) {
				// Supplier should give us a number
				return '';
			}

			$next_id = $GLOBALS['egw']->db->select(
				self::order_table,
				array('max(order_id) + 1 AS next_po_id'),
				array(),
				__LINE__, __FILE__,
				false, '', 'perp_api'
			)->fetchSingle();

			return sprintf('◄%05d', $next_id);
		}

		/**
		*       Get a string reperesentation of this object.
		*       Useful for debugging.
		*
		*       @return String
		*/
		public function __toString() {
			$string = sprintf("RMA #%4d %10s\n",
				$this->get_rma_id(),
				($this->get_claim_id() ? 'for ' . bo_perp_claim::link_title($this->get_claim_id()) : '')
			);
			$string .= parent::__toString();

			return $string;
		}

		/**
		*       Convert this claim into an array, useful for display
		*/
		public function toArray() {
			$data = parent::toArray() + array(
				'rma_id'	=>	$this->get_order_id(),
				'claim_id'	=>	$this->get_claim_id()
			);

			return $data;
		}

		/**
		*	Prepare this object for serialization
		*
		*	@return array Properties to preserve across serialization.
		*/
		public function __sleep() {
			// For some reason just adding the extras as an array doesn't work.
			$properties = parent::__sleep();
			$properties[] = 'rma_id';
			$properties[] = 'claim_id';

			return $properties;
		}

		/**
		*       @}
		*       End utilities
		*/

	}
?>
