<?php

	/**
	*	Business / Logic layer object for dealing with warranty issues
	*
	*	@author Nathan Gray
	*/

	require_once( EGW_INCLUDE_ROOT . '/perp_warranty/inc/class.so_perp_issue.inc.php');

	class bo_perp_issue {

		/**
		*	Coverage constants
		*/
		const INCLUDED		=	0;
		const DISCRETIONARY	=	1;
		const EXCLUDED		=	2;
		const VOIDS_WARRANTY	=	3;

		/**
		*	Settings for things required
		*/
		const ALWAYS		=	3;
		const DEFAULT_YES	=	2;
		const DEFAULT_NO	=	1;
		const NEVER		=	0;

		/**
		*	Special issues that get special handling
		*	@{
		*/
		/**
		*	Other is primarily a place holder used until someone can find out 
		*	what the actual problem is, rather than a real issue.  Either an
		*	existing issue should be assigned, or a new issue created.
		*/
		const OTHER		=	1;
		
		/**
		*	Parent is used to distinguish an assembly that has problems from
		*	the components that have the problem.  The parent assembly is 
		*	recorded, but all the issues are with the components.
		*/
		const PARENT		=	0;

		/**
		*	@}
		*	End special issues
		*/

		/**
		*	Get a list of known issues
		*
		*	@nextmatch
		*
		*	Extra query keys:
		*	- include_exceptions - Includes exceptions for a particular stock id.
		*/
		public static function get_issue_list(&$query = array(), &$rows = array(), &$readonlys = array()) {
			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_READ, 'perp_warranty')) {
				$rows = array();
				return 0;
			}

			if(!$query['order']) {
				$query['order'] = 'issue_id';
			}
			if(!$query['template']) {
				// Get a list of issues for a selectbox
				$query['selectcols'] = 'issue_id,name';
				$query['num_rows'] = -1;
			}

			$rows = so_perp_issue::get_issue_list($query);
			
			$list = array();
			foreach($rows as $row) {
				$list[$row['issue_id']] = $row['name'];
			}
			$can_delete = self::can_delete(array_keys($list));

			if(!$query['template']) {
				return $list;
			}
			
			foreach($rows as $row) {
				$readonlys['delete[' . $row['issue_id'] . ']'] = !$can_delete[$row['issue_id']];
			}
			return $query['total'];
		}

		/**
		*	Get a list of coverage options
		*
		*	Used to decide if the claim should be paid or not
		*/
		public static function get_coverage_list() {
			return array(
				self::INCLUDED		=>	lang('Included'),
				self::DISCRETIONARY	=>	lang('Discretionary'),
				self::EXCLUDED		=>	lang('Excluded'),
				self::VOIDS_WARRANTY	=>	lang('Voids warranty')
			);
		}

		/**
		*	Get a list of options for things that are required, such as
		*	replace parts.
		*
		*	@return array of options
		*/
		public static function get_default_option_list() {
			return array(
				self::ALWAYS		=>	lang('Always'),
				self::DEFAULT_YES	=>	lang('Default Yes'),
				self::DEFAULT_NO	=>	lang('Default No'),
				self::NEVER		=>	lang('Never')
			);
		}

		public static function get_coverage_limit_list() {
			return array(
				7	=>	lang('7 days'),
				14	=>	lang('14 days'),
				30	=>	lang('30 days'),
				45	=>	lang('45 days'),
				60	=>	lang('60 days'),
				90	=>	lang('90 days'),
				183	=>	lang('6 months'),
				365	=>	lang('1 year'),
				365*2	=>	lang('2 years'),
				365*3	=>	lang('3 years'),
				365*4	=>	lang('4 years'),
				365*5	=>	lang('5 years'),
				365*10	=>	lang('10 years'),
				0	=>	lang('Forever')
			);
		}

		/**
		*	Read an issue
		*/
		public static function read($issue_id) {
			if($GLOBALS['egw']->acl->check('issue', EGW_ACL_READ, 'perp_warranty')) {
				return so_perp_issue::read((int)$issue_id);
			} else {
				return array();
			}
		}

		/**
		*	Get a list of stock categories, and their settings for the given issue
		*
		*	@param issue_id The issue to get the settings for
		*	@param include_all Include categories that are excluded from the issue
		*
		*	@return array of settings, one sub-array per stock category
		*/
		public static function get_stock_categories($issue_id, $include_all = false) {
			if($GLOBALS['egw']->acl->check('issue', EGW_ACL_READ, 'perp_warranty')) {
				return so_perp_issue::get_stock_categories((int)$issue_id, $include_all);
			} else {
				return array();
			}
		}

		/**
		*	Get a list of stock items that are exceptions to the given issue,
		*	and their settings.
		*
		*	@param issue_id
		*
		*	@return array of settings, each stock item in a sub-array
		*/
		public static function get_stock_exception_list($issue_id) {
			if($GLOBALS['egw']->acl->check('issue', EGW_ACL_READ, 'perp_warranty')) {
				return so_perp_issue::get_stock_exception_list((int)$issue_id);
			} else {
				return array();
			}
		}

		/**
		*	Get the default settings for a certain stock item
		*	
		*	@param issue_id
		*	@param stock_id
		*
		*	@return array of settings for a stock item
		*/
		public static function get_stock_settings($issue_id, $stock_id) {
			return so_perp_issue::get_stock_settings($issue_id, $stock_id);
		}

		/**
		*	Save the basics of a warranty issue
		*/
		public static function save($issue_id, $issue) {
			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_EDIT, 'perp_warranty')) {
				throw new egw_exception_no_permission(lang('You do not have permission to edit issues'));
			}

			$issue['name'] = trim($issue['name']);
			if(!$issue['name']) {
				throw new egw_exception_wrong_userinput(lang('Missing name'));
			}
	
			$issue['description'] = trim($issue['description']);
			if(!$issue['description']) {
				throw new egw_exception_wrong_userinput(lang('Missing description'));
			}
			

			// Limit is only for when coverage is possible
			if($issue['coverage'] == EXCLUDED || $issue['coverage'] == VOIDS_WARRANTY) {
				$issue['coverage_limit'] = 0;
			}
			
			return so_perp_issue::save($issue_id, $issue);
		}

		/**
		*	Delete an issue from the database
		*
		*	@param issue_id
		*/
		public static function delete($issue_id) {
			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_DELETE, 'perp_warranty')) {
				throw new egw_exception_no_permission(lang('You do not have permission to delete issues'));
			}
			return so_perp_issue::delete($issue_id);
		}

		/**
		*	Check to see if an issue, or list of issues, can be safely deleted.
		*
		*	@param issue_id ID or array of IDs
		*
		*	@return Array of ID => boolean
		*/
		public static function can_delete($issue_id) {
			if(!is_array($issue_id)) {
				$issue_id = array($issue_id);
			}

			$result = array();

			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_DELETE, 'perp_warranty')) {
				foreach($issue_id as $id) {
					$result[$id] = false;
				}
			} else {
				$result = so_perp_issue::can_delete($issue_id);
			}
			return $result;
		}

		/**
		*	Save the stock category settings of a warranty issue
		*/
		public static function set_stock_categories($issue_id, $stock_categories) {
			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_EDIT, 'perp_warranty')) {
				throw new egw_exception_no_permission(lang('You do not have permission to edit issues'));
			}
			foreach($stock_categories as $category) {
				$category['image_description'] = trim($category['image_description']);
				if($category['include'] && $category['image_required'] != self::NEVER && $category['image_description'] == '') {
					throw new egw_exception_wrong_userinput(lang('Missing image description'));
				}
			}
			return so_perp_issue::set_stock_categories($issue_id, $stock_categories);
		}

		/**
		*	Add an exception for a given stock item for a warranty issue
		*
		*	@param issue_id
		*	@param stock_id
		*
		*/
		public static function add_stock_exception($issue_id, $stock_id) {
			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_EDIT, 'perp_warranty')) {
				throw new egw_exception_no_permission(lang('You do not have permission to edit issues'));
			}

			// Check to see if it already exists
			if(self::get_stock_exception($issue_id, $stock_id)) {
				return false;
			}

			$issue = self::read($issue_id);
			$stock_item = ExecMethod('perp_inventory.bo_perp_stock.get_stock_item', $stock_id);

			unset($issue['issue_id']);
			$issue['name'] .= ' - ' . $stock_item['short_description'];

			$exception_id = self::save(null, $issue);
			
			$exception = array(
				'base_issue_id'		=>	$issue_id,
				'stock_id'		=>	(int)$stock_id,
				'exception_issue_id'	=>	$exception_id
			);
			so_perp_issue::set_stock_exception($exception);

			return $exception_id;
		}

		/**
		*	Get the exception settings for a particular stock item on a specified issue
		*/
		public static function get_stock_exception($issue_id, $stock_id) {
			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_READ, 'perp_warranty')) {
				throw new egw_exception_no_permission(lang('You do not have permission to read issues'));
			}
			return so_perp_issue::get_stock_exception_list($issue_id, $stock_id);
		}

		/**
		*	Set the exception settings for a particular stock item on a specified issue
		*/
		public static function set_stock_exception(Array $exception) {
			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_EDIT, 'perp_warranty')) {
				throw new egw_exception_no_permission(lang('You do not have permission to edit issues'));
			}
			return so_perp_issue::set_stock_exception($exception);
		}

		/**
		*	Remove an exception for a stock item, and let it be handled according to its category
		*
		*	@param issue_id
		*	@param stock_id
		*/
		public static function delete_stock_exception($issue_id, $stock_id) {
			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_EDIT, 'perp_warranty')) {
				throw new egw_exception_no_permission(lang('You do not have permission to edit issues'));
			}
			return so_perp_issue::delete_stock_exception($issue_id, $stock_id);
		}

		/**
		*	Normally damaged parts are replaced with the same thing.  Add an exception to that rule.
		*
		*	@param issue_id
		*	@param stock_id
		*	@param replacement_id
		*	@param quantity
		*/
		public static function add_replacement_part($issue_id, $stock_id, $replacement_id, $quantity) {
			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_EDIT, 'perp_warranty')) {
				throw new egw_exception_no_permission(lang('You do not have permission to edit issues'));
			}
			if(!(int)$stock_id) {
				throw new egw_exception_wrong_userinput(lang('Missing the original part number'));
			}
			if(!(int)$replacement_id) {
				throw new egw_exception_wrong_userinput(lang('Missing the replacement part number'));
			}
			return so_perp_issue::add_replacement_part($issue_id, (int)$stock_id, (int)$replacement_id, $quantity);
		}

		/**
		*	Remove an exception for a replacement part, send the same thing as was damaged.
		*
		*	@param issue_id
		*	@param stock_id
		*	@param replacement_id
		*/
		public static function delete_replacement_part($issue_id, $stock_id, $replacement_id) {
			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_EDIT, 'perp_warranty')) {
				throw new egw_exception_no_permission(lang('You do not have permission to edit issues'));
			}
			return so_perp_issue::delete_replacement_part($issue_id, (int)$stock_id, (int)$replacement_id);
		}

		/**
		*	Get a list of replacement parts for a given issue.
		*
		*	If the stock_id parameter is set, it will return the replacement parts for that
		*	stock item, which may be the same stock item.
		*
		*	If stock_id parameter is not set, it will return only replacement parts where
		*	the stock item is replaced with (a) different item(s).
		*
		*	@nextmatch
		*/
		public static function get_replacement_parts(&$query = array(), &$rows = array(), &$readonlys = array()) {
			if(!$GLOBALS['egw']->acl->check('issue', EGW_ACL_READ, 'perp_warranty')) {
				throw new egw_exception_no_permission(lang('You do not have permission to read issues'));
			}

			if(!$query['order']) {
				$query['order'] = 'perp_warranty_replacement_parts.stock_id';
			}
			if(!$query['num_rows']) {
				$query['num_rows'] = -1;
			}

			$rows = so_perp_issue::get_replacement_parts($query);
			foreach($rows as $key => $row) {
				if(!$row['issue_id']) {
					$readonlys['delete[' . $row['issue_id'] . '-' . $row['stock_id'] . '-' . $row['replacement_id'] . ']'] = true;
				}
			}

			return $query['total'];
		}

		/**
		*	Get the name of an issue, for linking or handy reference
		*
		*	@param ID
		*
		*	@return String Name of the issue
		*/
		public static function link_title($id) {
			$issue = self::read($id);
			return $issue['name'];
		}
	}
?>
