<?php


	/**
	*	Business / Logic layer for dealing with the registration of warranties
	*
	*	@author Nathan Gray
	*	@package perp_warranty
	*/

	class bo_perp_registration {
		

		/**
		*	Get the registration information for an item.
		*
		*	This only works with serialized items.
		*/
		public static function get_registration($stock_id, $serial_number) {
			if(!$GLOBALS['egw']->acl->check('registration', EGW_ACL_READ, 'perp_warranty')) {
				throw new egw_exception_no_permission();
			}
			if(!$stock_id || !$serial_number) {
				throw new egw_exception_wrong_parameter("Missing stock_id or serial_number: get_registration($stock_id, $serial_number)");
			}

			$query = array(
				'stock_id'	=>	$stock_id,
				'serial_number'	=>	$serial_number
			);
			$list = so_perp_registration::get_registration_list($query);

			if(count($list) >= 1) {
				return $list[$query['total'] - 1]['registration_id'];
			} else {
				return false;
			}
		}

		/**
		*	Get a list of registrations based on parameters
		*
		*	@nextmatch
		*
		*	Extra parameters:
		*	- stock_id (filter2)
		*	- client_id (filter)
		*/
		public static function get_registration_list(&$query = array(), &$rows = array(), &$readonly = array()) {
			if(!$GLOBALS['egw']->acl->check('registration', EGW_ACL_READ, 'perp_warranty')) {
				$rows = array();
				return 0;
			}

			if(!$query['order']) {
				$query['order'] = 'registration_id';
			}
			if($query['filter']) {
				$query['client_id'] = $query['filter'];
			}
			if($query['filter2']) {
				$query['stock_id'] = $query['filter2'];
			}

			$rows = so_perp_registration::get_registration_list($query);

			return $query['total'];
		}

		/**
		*	Generate a registration key to prevent registration by trouble makers
		*
		*	This key is to be distributed with the product to allow end customers
		*	to register over the internet, on a public web site.  It is reproducible, 
		*	based only on the stock_id and serial number.  It can use more information
		*	to generate a more complicated key if needed, but there is no real need for
		*	these to be guaranteed unique, just prevent typos in serial number and make
		*	sure people who register actually have a product.
		*/
		public static function get_key($stock_id, $serial_number) {
			return strtoupper(substr(md5($stock_id.$serial_number), 0, 8));
		}

		/**
		*	Read registration information
		*/
		public static function read($registration_id) {
			if(!$GLOBALS['egw']->acl->check('registration', EGW_ACL_READ, 'perp_warranty')) {
				throw new egw_exception_no_permission();
			}

			if($registration_id) {
				$registration = so_perp_registration::read($registration_id);

				// Check if the current user is a client
				$client_ids = ExecMethod('perp_ar.bo_perp_client.get_client_by_user_account');
				if(is_array($client_ids) && count($client_ids) > 0 && !in_array($registration['client_id'], $client_ids)) {
					throw new egw_exception_no_permission(lang('Item is registered to another client'));
				}
			} else {
				// Defaults
				$registration = array(
					'registration_date'	=>	time(),
					'registered_by'		=>	$GLOBALS['egw_info']['user']['account_id'],
				);
			}

			return $registration;
		}

		/**
		*	Store registration information
		*/
		public static function save($registration) {
			if($GLOBALS['egw']->acl->check('registration', EGW_ACL_EDIT, 'perp_warranty') || (
				!$registration['registration_id'] && $GLOBALS['egw']->acl->check('registration', EGW_ACL_ADD, 'perp_warranty')
			)) {
				// Checks
				if(!$registration['client_id']) {
					throw new egw_exception(lang('Missing client'));
				}
				if(!$registration['stock_id']) {
					throw new egw_exception(lang('Missing stock item'));
				}
				if(!$registration['serial_number']) {
					throw new egw_exception(lang('Missing serial number'));
				}
				if(!$registration['owner']) {
					throw new egw_exception(lang('Missing owner information'));
				}
				if(!$registration['registration_date']) {
					$registration['registration_date'] = date();
				}
				if(!$registration['start_date']) {
					$registration['start_date'] = $registration['registration_date'];
				}
				if(!$registration['registered_by']) {
					$registration['registered_by'] = $GLOBALS['egw_info']['user']['account_id'];
				}
				if(!$registration['registration_id']) {
					$first_time = true;
				}

				// Check for already registered
				if($registration['registration_id'] != self::get_registration($registration['stock_id'], $registration['serial_number'])) {
					$config = config::read('perp_warranty');
					if(!$GLOBALS['egw']->acl->check('registration', EGW_ACL_EDIT, 'perp_warranty') || !$config['warranty_transfer']) {
						throw new egw_exception(lang('%1 serial number %2 is already registered', 
							ExecMethod('perp_inventory.bo_perp_stock.link_title', $registration['stock_id']),
							$registration['serial_number']
						));
					}
				}
				$registration_id = so_perp_registration::save($registration);

				if($registration_id && $first_time) {
					// Make a record in the item's history
					bo_perp_stock_history::add_history(
						$registration['stock_id'], 
						$registration['serial_number'], 
						lang('Warranty registered'),
						array('app' => 'perp_warranty', 'id' => 'REGISTRATION-' . $registration_id),
						__LINE__, __FILE__
					);
				}
				return $registration_id;
			} else {
				throw new egw_exception_no_permission();
			}
		}

		/**
		*	Handy link title
		*/
		public static function link_title($id) {
			$registration = self::read($id);
			return $registration['serial_number'] . ': ' . ExecMethod('addressbook.addressbook_bo.link_title', $registration['owner']);
		}
	}
?>
